// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#ifndef BX_BOCHS_H
#  define BX_BOCHS_H 1

extern "C" {
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#  include <unistd.h>
#else
#  include <io.h>
#endif
#include <time.h>
#include <ctype.h>
#include <string.h>
}



#include "config.h"
#include "state_file.h"

#if BX_USE_SMF == 0
   // normal member functions, use this pointer
#  define BX_CPU_THIS_PTR  this->
#  define BX_SMF
#  define BX_CPU_C_PREFIX  bx_cpu_c::
#else
   // static member functions, use direct access
#  define BX_CPU_THIS_PTR  bx_cpu.
#  define BX_SMF           static
#  define BX_CPU_C_PREFIX
#endif

#ifndef UNUSED
#  define UNUSED(x) ((void)x)
#endif

#if USE_INLINE
#  define INLINE inline
#else
#  define INLINE 
#endif /* USE_INLINE */

#define uint8   Bit8u
#define uint16  Bit16u
#define uint32  Bit32u

#include "debug/debug.h"

#if BX_PROVIDE_CPU_MEMORY==1
#  include "cpu/cpu.h"
#endif

#if BX_DISASM
#  include "disasm/disasm.h"
#endif


typedef struct {
  Boolean floppy;
  Boolean keyboard;
  Boolean video;
  Boolean disk;
  Boolean pit;
  Boolean pic;
  Boolean bios;
  Boolean cmos;
  Boolean _protected;
  Boolean a20;
  Boolean interrupts;
  Boolean exceptions;
  Boolean unsupported;
  Boolean temp;
  Boolean reset;
  Boolean debugger;
  Boolean mouse;
  Boolean io;
  Boolean xms;
  Boolean v8086;
  Boolean paging;
  Boolean creg;
  Boolean dreg;
  Boolean dma;
  Boolean unsupported_io;
  } bx_debug_t;

void bx_printf(char *fmt, ...);
void bx_panic(char *fmt, ...);
extern bx_debug_t bx_dbg;



#define BX_FLOPPY_NONE   10 // floppy not present
#define BX_FLOPPY_1_2    11 // 1.2M  5.25"
#define BX_FLOPPY_1_44   12 // 1.44M 3.5"
#define BX_FLOPPY_2_88   13 // 2.88M 3.5"
#define BX_FLOPPY_720K   14 // 720K  3.5"


#define BX_READ    10
#define BX_WRITE   11
#define BX_RW      12





#include "memory/memory.h"


#if BX_PROVIDE_CPU_MEMORY==1
extern bx_cpu_c       bx_cpu;
#endif

enum PCS_OP { PCS_CLEAR, PCS_SET, PCS_TOGGLE };

#include "pc_system.h"
#include "gui/gui.h"
extern bx_gui_c   bx_gui;
#include "iodev/iodev.h"


 


#if BX_SUPPORT_A20
#  define A20ADDR(x) ( (x) & bx_pc_system.a20_mask )
#else
#  define A20ADDR(x) (x)
#endif




/* --- EXTERNS --- */

extern FILE *bx_logfd;




#if ( BX_PROVIDE_DEVICE_MODELS==1 )
extern bx_devices_c   bx_devices;
#endif

#define BX_EJECTED   10
#define BX_INSERTED  11


  // device was reset from twiddling register bits
#define BX_RESET_SOFTWARE 10
  // device was reset from hardware RESET
#define BX_RESET_HARDWARE 11


typedef struct {
  char path[512];
  unsigned type;
  unsigned initial_status;
  } bx_floppy_options;

typedef struct {
  char path[512];
  unsigned int cylinders;
  unsigned int heads;
  unsigned int spt;
  } bx_disk_options;

typedef struct {
  char *path;
  } bx_rom_options;

typedef struct {
  char *path;
  } bx_vgarom_options;

typedef struct {
  size_t megs;
  } bx_mem_options;

typedef struct {
  bx_floppy_options floppya;
  bx_floppy_options floppyb;
  bx_disk_options   diskc;
  bx_rom_options    rom;
  bx_vgarom_options vgarom;
  bx_mem_options    memory;
  char              bootdrive[2];
  unsigned long     vga_update_interval;
  unsigned long     keyboard_serial_delay;
  unsigned long     floppy_command_delay;
  unsigned long     ips;
  Boolean           mouse_enabled;
  Boolean           private_colormap;
  } bx_options_t;

extern bx_options_t bx_options;




#if BX_DEBUGGER
#  define BX_CPU_ICOUNT() { \
          Boolean paddr_valid; \
          Bit32u  laddr, paddr; \
          laddr = BX_CPU_THIS_PTR cs.cache.u.segment.base + BX_CPU_THIS_PTR eip; \
          dbg_xlate_linear2phy(laddr, &paddr, &paddr_valid); \
          if (bx_dbg_cpu_icount(BX_CPU_THIS_PTR cs.selector.value, \
                                BX_CPU_THIS_PTR eip, \
                                laddr, paddr, paddr_valid, \
                                BX_CPU_THIS_PTR INTR, \
                                BX_CPU_THIS_PTR cs.cache.u.segment.d_b)) return; \
          }
#endif

#if BX_DEBUGGER
#  define BX_DBG_TAKE_INTR bx_dbg_take_irq
#  define BX_DBG_TAKE_DMA  bx_dbg_take_dma
#else
#  define BX_DBG_TAKE_INTR 1
#  define BX_DBG_TAKE_DMA  1
#endif



#if BX_PROVIDE_CPU_MEMORY==1
#else
// #  include "external_interface.h"
#endif

#define BX_USE_PS2_MOUSE 1

int     bx_bochs_init(int argc, char *argv[]);

// Hacks for win32: always return regular file.  
#ifdef WIN32
#  define S_ISREG(st_mode) 1
#  define S_ISCHR(st_mode) 0
#endif


#include "instrument.h"

#endif  /* BX_BOCHS_H */
