// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com




#ifndef BX_CPU_H
#  define BX_CPU_H 1

#include <setjmp.h>

#include "cpu/lazy_flags.h"



/* segment register encoding */
#define BX_SEG_REG_ES    0
#define BX_SEG_REG_CS    1
#define BX_SEG_REG_SS    2
#define BX_SEG_REG_DS    3
#define BX_SEG_REG_FS    4
#define BX_SEG_REG_GS    5



#ifdef BX_LITTLE_ENDIAN
#define BX_REG8L_OFFSET 0
#define BX_REG8H_OFFSET 1
#define BX_REG16_OFFSET 0
#else // BX_BIG_ENDIAN
#define BX_REG8L_OFFSET 3
#define BX_REG8H_OFFSET 2
#define BX_REG16_OFFSET 2
#endif // ifdef BX_LITTLE_ENDIAN

#define BX_REGISTER_REF 10
#define BX_MEMORY_REF   11

#define BX_8BIT_REG_AL  0
#define BX_8BIT_REG_CL  1
#define BX_8BIT_REG_DL  2
#define BX_8BIT_REG_BL  3
#define BX_8BIT_REG_AH  4
#define BX_8BIT_REG_CH  5
#define BX_8BIT_REG_DH  6
#define BX_8BIT_REG_BH  7

#define BX_16BIT_REG_AX 0
#define BX_16BIT_REG_CX 1
#define BX_16BIT_REG_DX 2
#define BX_16BIT_REG_BX 3
#define BX_16BIT_REG_SP 4
#define BX_16BIT_REG_BP 5
#define BX_16BIT_REG_SI 6
#define BX_16BIT_REG_DI 7

#define BX_32BIT_REG_EAX 0
#define BX_32BIT_REG_ECX 1
#define BX_32BIT_REG_EDX 2
#define BX_32BIT_REG_EBX 3
#define BX_32BIT_REG_ESP 4
#define BX_32BIT_REG_EBP 5
#define BX_32BIT_REG_ESI 6
#define BX_32BIT_REG_EDI 7


// access to 8 bit general registers
#define AL (BX_CPU_THIS_PTR gen_reg[0].word.byte.rl)
#define CL (BX_CPU_THIS_PTR gen_reg[1].word.byte.rl)
#define DL (BX_CPU_THIS_PTR gen_reg[2].word.byte.rl)
#define BL (BX_CPU_THIS_PTR gen_reg[3].word.byte.rl)
#define AH (BX_CPU_THIS_PTR gen_reg[0].word.byte.rh)
#define CH (BX_CPU_THIS_PTR gen_reg[1].word.byte.rh)
#define DH (BX_CPU_THIS_PTR gen_reg[2].word.byte.rh)
#define BH (BX_CPU_THIS_PTR gen_reg[3].word.byte.rh)


// access to 16 bit general registers
#define AX (BX_CPU_THIS_PTR gen_reg[0].word.rx)
#define CX (BX_CPU_THIS_PTR gen_reg[1].word.rx)
#define DX (BX_CPU_THIS_PTR gen_reg[2].word.rx)
#define BX (BX_CPU_THIS_PTR gen_reg[3].word.rx)
#define SP (BX_CPU_THIS_PTR gen_reg[4].word.rx)
#define BP (BX_CPU_THIS_PTR gen_reg[5].word.rx)
#define SI (BX_CPU_THIS_PTR gen_reg[6].word.rx)
#define DI (BX_CPU_THIS_PTR gen_reg[7].word.rx)

// access to 16 bit instruction pointer
#define IP (* (Bit16u *) (((Bit8u *) &BX_CPU_THIS_PTR eip) + BX_REG16_OFFSET))


// accesss to 32 bit general registers
#define EAX BX_CPU_THIS_PTR gen_reg[0].erx
#define ECX BX_CPU_THIS_PTR gen_reg[1].erx
#define EDX BX_CPU_THIS_PTR gen_reg[2].erx
#define EBX BX_CPU_THIS_PTR gen_reg[3].erx
#define ESP BX_CPU_THIS_PTR gen_reg[4].erx
#define EBP BX_CPU_THIS_PTR gen_reg[5].erx
#define ESI BX_CPU_THIS_PTR gen_reg[6].erx
#define EDI BX_CPU_THIS_PTR gen_reg[7].erx

// access to 32 bit instruction pointer
#define EIP BX_CPU_THIS_PTR eip

#define BX_READ_8BIT_REG(index)  (((index) < 4) ? \
  (BX_CPU_THIS_PTR gen_reg[index].word.byte.rl) : \
  (BX_CPU_THIS_PTR gen_reg[(index)-4].word.byte.rh))
#define BX_READ_16BIT_REG(index) (BX_CPU_THIS_PTR gen_reg[index].word.rx)
#define BX_READ_32BIT_REG(index) (BX_CPU_THIS_PTR gen_reg[index].erx)

#define BX_READ_16BIT_BASE_REG(var, index) {\
  var = *BX_CPU_THIS_PTR _16bit_base_reg[index];\
  }

#define BX_READ_16BIT_INDEX_REG(var, index) {\
  var = *BX_CPU_THIS_PTR _16bit_index_reg[index];\
  }

#define BX_WRITE_8BIT_REG(index, val) {\
  if ((index) < 4) \
    BX_CPU_THIS_PTR gen_reg[index].word.byte.rl = val; \
  else \
    BX_CPU_THIS_PTR gen_reg[(index)-4].word.byte.rh = val; \
  }
#define BX_WRITE_16BIT_REG(index, val) {\
  BX_CPU_THIS_PTR gen_reg[index].word.rx = val; \
  }
#define BX_WRITE_32BIT_REG(index, val) {\
  BX_CPU_THIS_PTR gen_reg[index].erx = val; \
  }




#define TF BX_CPU_THIS_PTR eflags.tf
#define IF BX_CPU_THIS_PTR eflags.if_
#define DF BX_CPU_THIS_PTR eflags.df

#define IOPL BX_CPU_THIS_PTR eflags.iopl
#define CPL  BX_CPU_THIS_PTR cs.selector.rpl

#define BX_DE_EXCEPTION   0     /* Divide Error exception */
#define BX_UD_EXCEPTION   6
#define BX_NM_EXCEPTION   7
#define BX_DF_EXCEPTION   8
#define BX_TS_EXCEPTION  10
#define BX_NP_EXCEPTION  11
#define BX_SS_EXCEPTION  12
#define BX_GP_EXCEPTION  13
#define BX_PF_EXCEPTION  14
#define BX_MF_EXCEPTION  16
#define BX_AC_EXCEPTION  17


typedef struct {
  /* 31|30|29|28|27|26|25|24|23|22|21|20|19|18|17|16
   * ==|==|=====|==|==|==|==|==|==|==|==|==|==|==|==
   *  0| 0| 0| 0| 0| 0| 0| 0| 0| 0|ID|VP|VF|AC|VM|RF
   *
   * 15|14|13|12|11|10| 9| 8| 7| 6| 5| 4| 3| 2| 1| 0
   * ==|==|=====|==|==|==|==|==|==|==|==|==|==|==|==
   *  0|NT| IOPL|OF|DF|IF|TF|SF|ZF| 0|AF| 0|PF| 1|CF
   */
  
  Boolean cf;
  Boolean bit1;
  Bit8u   pf_byte;  /* PF derived from last result byte when needed */
  Boolean bit3;
  Boolean af;
  Boolean bit5;
  Boolean zf;
  Boolean sf;
  Boolean tf;
  Boolean if_;
  Boolean df;
  Boolean of;
#if BX_CPU >= 2
  Bit8u   iopl;
  Boolean nt;
#endif
  Boolean bit15;
#if BX_CPU >= 3
  Boolean rf;
  Boolean vm;
#endif
#if BX_CPU >= 4
  Boolean ac;
#endif
  } bx_flags_reg_t;


#if BX_CPU >= 2
typedef struct {
#if BX_CPU >= 3
  Boolean pg; /* paging */
#endif

#if BX_CPU >= 4
  /* uncomment initialization in main.c */
  Boolean cd; /* cache disable */
  Boolean nw; /* no write-through */
  Boolean am; /* alignment mask */
  Boolean wp; /* write-protect */
  Boolean ne; /* numerics exception */
#endif
  /*Boolean et;???*/ /* extension type */

  /* reserved bit = Extension type */
  Boolean ts; /* task switched */
  Boolean em; /* emulate math coprocessor */
  Boolean mp; /* monitor coprocessor */
  Boolean pe; /* protected mode enable */
  } bx_cr0_t;
#endif


typedef struct { /* bx_selector_t */
  Bit16u value;   /* the 16bit value of the selector */
#if BX_CPU >= 2
    /* the following fields are extracted from the value field in protected
       mode only.  They're used for sake of efficiency */
  Bit16u index;   /* 13bit index extracted from value in protected mode */
  Bit8u  ti;      /* table indicator bit extracted from value */
  Bit8u  rpl;     /* RPL extracted from value */
#endif
  } bx_selector_t;



typedef struct {
  Boolean valid;         /* 0 = invalid, 1 = valid */
  Boolean p;             /* present */
  Bit8u   dpl;           /* descriptor privilege level 0..3 */
  Boolean segment;       /* 0 = system/gate, 1 = data/code segment */
  Bit8u   type;          /* For system & gate descriptors, only
                          *  0 = invalid descriptor (reserved)
                          *  1 = 286 available Task State Segment (TSS)
                          *  2 = LDT descriptor
                          *  3 = 286 busy Task State Segment (TSS)
                          *  4 = 286 call gate
                          *  5 = task gate
                          *  6 = 286 interrupt gate
                          *  7 = 286 trap gate
                          *  8 = (reserved)
                          *  9 = 386 available TSS
                          * 10 = (reserved)
                          * 11 = 386 busy TSS
                          * 12 = 386 call gate
                          * 13 = (reserved)
                          * 14 = 386 interrupt gate
                          * 15 = 386 trap gate */
  union {
  struct {
    Boolean executable;    /* 1=code, 0=data or stack segment */
    Boolean c_ed;          /* for code: 1=conforming,
                              for data/stack: 1=expand down */
    Boolean r_w;           /* for code: readable?, for data/stack: writeable? */
    Boolean a;             /* accessed? */
    Bit32u  base;          /* base address: 286=24bits, 386=32bits */
    Bit32u  limit;         /* limit: 286=16bits, 386=20bits */
    Bit32u  limit_scaled;  /* for efficiency, this contrived field is set to
                            * limit for byte granular, and
                            * (limit << 12) | 0xfff for page granular seg's
                            */
#if BX_CPU >= 3
    Boolean g;             /* granularity: 0=byte, 1=4K (page) */
    Boolean d_b;           /* default size: 0=16bit, 1=32bit */
    Boolean avl;           /* available for use by system */
#endif
    } segment;
  struct {
    Bit8u   word_count;    /* 5bits (0..31) #words to copy from caller's stack
                            * to called procedure's stack.  (call gates only)*/
    Bit16u  dest_selector;
    Bit16u  dest_offset;
    } gate286;
  struct {                 // type 5: Task Gate Descriptor
    Bit16u  tss_selector;  // TSS segment selector
    } taskgate;
#if BX_CPU >= 3
  struct {
    Bit8u   dword_count;   /* 5bits (0..31) #dwords to copy from caller's stack
                            * to called procedure's stack.  (call gates only)*/
    Bit16u  dest_selector;
    Bit32u  dest_offset;
    } gate386;
#endif
  struct {
    Bit32u  base;          /* 24 bit 286 TSS base  */
    Bit16u  limit;         /* 16 bit 286 TSS limit */
    } tss286;
#if BX_CPU >= 3
  struct {
    Bit32u  base;          /* 32 bit 386 TSS base  */
    Bit32u  limit;         /* 20 bit 386 TSS limit */
    Boolean g;             /* granularity: 0=byte, 1=4K (page) */
    Boolean avl;           /* available for use by system */
    } tss386;
#endif
  struct {
    Bit32u  base;  /* 286=24 386+ =32 bit LDT base */
    Bit16u  limit; /* 286+ =16 bit LDT limit */
    } ldt;
    } u;

  } bx_descriptor_t;

typedef struct {
  bx_selector_t          selector;
  bx_descriptor_t  cache;
  } bx_segment_reg_t;


#if BX_CPU < 2
  /* no GDTR or IDTR register in an 8086 */
#else
typedef struct {
  Bit32u                 base;      /* base address: 24bits=286,32bits=386 */
  Bit16u                 limit;     /* limit, 16bits */
  } bx_global_segment_reg_t;
#endif



//#define BX_PREFETCH_QSIZE  8
#define BX_PREFETCH_QSIZE  16
//#define BX_PREFETCH_QSIZE  32

#define BX_INVALID_FETCH_INDEX (BX_PREFETCH_QSIZE + 10)


#if BX_USE_TLB
 
#  define BX_TLB_SIZE 4096
 
  typedef struct {
    Bit32u LPF; // linear page frame
    Bit32u PPF; // physical page frame
    Bit32u access;
    } bx_TLB_entry;
#endif  // #if BX_USE_TLB



#ifdef BX_BIG_ENDIAN
typedef struct {
  union {
    Bit32u erx;
    struct {
      Bit16u word_filler;
      union {
        Bit16u rx;
        struct {
          Bit8u rh;
          Bit8u rl;
          } byte;
        };
      } word;
    };
  } bx_gen_reg_t;
#else
typedef struct {
  union {
    Bit32u erx;
    struct {
      union {
        Bit16u rx;
        struct {
          Bit8u rl;
          Bit8u rh;
          } byte;
        };
      Bit16u word_filler;
      } word;
    };
  } bx_gen_reg_t;
#endif


class bx_cpu_c {

public: // for now...

  // General register set
  // eax: accumulator
  // ebx: base
  // ecx: count
  // edx: data
  // ebp: base pointer
  // esi: source index
  // edi: destination index
  // esp: stack pointer
  bx_gen_reg_t  gen_reg[8];

  /* status and control flags register set */
  bx_flags_reg_t eflags; /* status and control flags */
  Boolean inhibit_interrupts;

  Bit32u eip;    /* instruction pointer */
#if BX_CPU > 0
  /* so that we can back up when handling faults, exceptions, etc.
     we need to store the value of the instruction pointer, before
     each fetch/execute cycle.
   */
  Bit32u prev_eip;
  Bit32u prev_esp;
#endif

  bx_segment_reg_t *segment_override;

  /* user segment register set */
  bx_segment_reg_t  cs; /* code segment register */
  bx_segment_reg_t  ss; /* stack segment register */
  bx_segment_reg_t  ds; /* data segment register */
  bx_segment_reg_t  es; /* extra segment register */
  bx_segment_reg_t  fs; /* extra segment register */
  bx_segment_reg_t  gs; /* extra segment register */

  /* system segment registers */
#if BX_CPU >= 2
  bx_global_segment_reg_t gdtr; /* global descriptor table register */
  bx_global_segment_reg_t idtr; /* interrupt descriptor table register */
#endif
  bx_segment_reg_t        ldtr; /* interrupt descriptor table register */
  bx_segment_reg_t        tr;   /* task register */


  /* debug registers 0-7 (unimplemented) */
#if BX_CPU >= 3
  Bit32u dr0;
  Bit32u dr1;
  Bit32u dr2;
  Bit32u dr3;
  Bit32u dr4;
  Bit32u dr5;
  Bit32u dr6;
  Bit32u dr7;
#endif

  /* TR3 - TR7 (Test Register 3-7), unimplemented */
#if 0
  Bit32u tr3;
  Bit32u tr4;
  Bit32u tr5;
  Bit32u tr6;
  Bit32u tr7;
#endif

  /* Control registers */
#if BX_CPU >= 2
  bx_cr0_t  cr0;
  Bit32u    cr1;
  Bit32u    cr2;
  Bit32u    cr3;
#endif
#if BX_CPU >= 5
  Bit32u    cr4;
#endif

  Boolean EXT; /* 1 if processing external interrupt or exception
                * or if not related to current instruction,
                * 0 if current CS:IP caused exception */
  unsigned errorno;   /* signal exception during instruction emulation */
  Boolean INTR;

#if BX_CPU >= 3
  Boolean is_32bit_opsize;
  Boolean is_32bit_addrsize;
#endif
  Boolean  single_step_event;
  volatile Boolean async_event;
  Bit32u   lf_flags_status;
  Boolean  lf_pf;
  bx_lf_flags_entry oszapc;
  bx_lf_flags_entry oszap;


  // for accessing registers by index number
  Bit16u *_16bit_base_reg[8];
  Bit16u *_16bit_index_reg[8];
  Bit32u empty_register;

  // for decoding instructions; accessing seg reg's by index
  bx_segment_reg_t *sreg_mod00_rm16[8];
  bx_segment_reg_t *sreg_mod01_rm16[8];
  bx_segment_reg_t *sreg_mod10_rm16[8];
  bx_segment_reg_t *sreg_mod00_rm32[8];
  bx_segment_reg_t *sreg_mod01_rm32[8];
  bx_segment_reg_t *sreg_mod10_rm32[8];
  bx_segment_reg_t *sreg_mod00_base32[8];
  bx_segment_reg_t *sreg_mod01_base32[8];
  bx_segment_reg_t *sreg_mod10_base32[8];

  // for exceptions
  jmp_buf jmp_buf_env;
  Bit8u curr_exception[2]; 

  static const Boolean is_exception_OK[3][3];
 
  bx_segment_reg_t save_cs;
  bx_segment_reg_t save_ss;
  Bit32u           save_eip;
  Bit32u           save_esp;

  // Page cache
  struct {
    /* cache def here */
    Boolean valid;
    } TLB;

  Bit8u  prefetch_queue[BX_PREFETCH_QSIZE];
  Bit32u linear_block_addr;
  Bit32u fetch_index;

  // for paging
#if BX_USE_TLB
  bx_TLB_entry local_TLB[BX_TLB_SIZE];
#endif

  struct {
    Bit32u  paddress1;  // physical address after translation of 1st len1 bytes of data
    Bit32u  paddress2;  // physical address after translation of 2nd len2 bytes of data
    Bit32u  len1;       // number of bytes in page 1
    Bit32u  len2;       // number of bytes in page 2
    unsigned pages;     // number of pages access spans (1 or 2)
    } address_xlation;

  // for lazy flags processing
  BX_SMF Boolean get_OF(void);
  BX_SMF Boolean get_SF(void);
  BX_SMF Boolean get_ZF(void);
  BX_SMF Boolean get_AF(void);
  BX_SMF Boolean get_PF(void);
  BX_SMF Boolean get_CF(void);

  // constructors & destructors...
  bx_cpu_c(void);
  ~bx_cpu_c(void);

  // For decode_exgx()...
  unsigned          reg_index;
  Bit32u            rm_addr;
  unsigned          rm_type;
  bx_segment_reg_t *rm_seg_reg;
  unsigned          modrm_ttt; // from |mod|ttt|rm|  byte

  // For decode_ESC_modrm()
  unsigned ESC_modrm;
  unsigned ESC_rm;
  unsigned ESC_nnn;
  unsigned ESC_mod;

  // prototypes for CPU instructions...
  BX_SMF INLINE void ADD_EbGb(void);
  BX_SMF INLINE void ADD_EvGv(void);
  BX_SMF INLINE void ADD_GbEb(void);
  BX_SMF INLINE void ADD_GvEv(void);
  BX_SMF INLINE void ADD_ALIb(void);
  BX_SMF INLINE void ADD_eAXIv(void);
  BX_SMF INLINE void PUSH_ES(void);
  BX_SMF INLINE void POP_ES(void);
  BX_SMF INLINE void OR_EbGb(void);
  BX_SMF INLINE void OR_EvGv(void);
  BX_SMF INLINE void OR_GbEb(void);
  BX_SMF INLINE void OR_GvEv(void);
  BX_SMF INLINE void OR_ALIb(void);
  BX_SMF INLINE void OR_eAXIv(void);
  BX_SMF INLINE void PUSH_CS(void);
  
  BX_SMF INLINE void ADC_EbGb(void);
  BX_SMF INLINE void ADC_EvGv(void);
  BX_SMF INLINE void ADC_GbEb(void);
  BX_SMF INLINE void ADC_GvEv(void);
  BX_SMF INLINE void ADC_ALIb(void);
  BX_SMF INLINE void ADC_eAXIv(void);
  BX_SMF INLINE void PUSH_SS(void);
  BX_SMF INLINE void POP_SS(void);
  BX_SMF INLINE void SBB_EbGb(void);
  BX_SMF INLINE void SBB_EvGv(void);
  BX_SMF INLINE void SBB_GbEb(void);
  BX_SMF INLINE void SBB_GvEv(void);
  BX_SMF INLINE void SBB_ALIb(void);
  BX_SMF INLINE void SBB_eAXIv(void);
  BX_SMF INLINE void PUSH_DS(void);
  BX_SMF INLINE void POP_DS(void);
  
  BX_SMF INLINE void AND_EbGb(void);
  BX_SMF INLINE void AND_EvGv(void);
  BX_SMF INLINE void AND_GbEb(void);
  BX_SMF INLINE void AND_GvEv(void);
  BX_SMF INLINE void AND_ALIb(void);
  BX_SMF INLINE void AND_eAXIv(void);
  BX_SMF INLINE void DAA(void);
  BX_SMF INLINE void SUB_EbGb(void);
  BX_SMF INLINE void SUB_EvGv(void);
  BX_SMF INLINE void SUB_GbEb(void);
  BX_SMF INLINE void SUB_GvEv(void);
  BX_SMF INLINE void SUB_ALIb(void);
  BX_SMF INLINE void SUB_eAXIv(void);
  BX_SMF INLINE void DAS(void);
  
  BX_SMF INLINE void XOR_EbGb(void);
  BX_SMF INLINE void XOR_EvGv(void);
  BX_SMF INLINE void XOR_GbEb(void);
  BX_SMF INLINE void XOR_GvEv(void);
  BX_SMF INLINE void XOR_ALIb(void);
  BX_SMF INLINE void XOR_eAXIv(void);
  BX_SMF INLINE void AAA(void);
  BX_SMF INLINE void CMP_EbGb(void);
  BX_SMF INLINE void CMP_EvGv(void);
  BX_SMF INLINE void CMP_GbEb(void);
  BX_SMF INLINE void CMP_GvEv(void);
  BX_SMF INLINE void CMP_ALIb(void);
  BX_SMF INLINE void CMP_eAXIv(void);
  BX_SMF INLINE void AAS(void);
  
  BX_SMF INLINE void PUSHAD(void);
  BX_SMF INLINE void POPAD(void);
  BX_SMF INLINE void BOUND_GvMa(void);
  BX_SMF INLINE void ARPL_EwGw(void);
  BX_SMF INLINE void PUSH_Iv(void);
  BX_SMF INLINE void IMUL_GvEvIv(void);
  BX_SMF INLINE void PUSH_Ib(void);
  BX_SMF INLINE void IMUL_GvEvIb(void);
  BX_SMF INLINE void INSB_YbDX(void);
  BX_SMF INLINE void INSW_YvDX(void);
  BX_SMF INLINE void OUTSB_DXXb(void);
  BX_SMF INLINE void OUTSW_DXXv(void);
  
  BX_SMF INLINE void TEST_EbGb(void);
  BX_SMF INLINE void TEST_EvGv(void);
  BX_SMF INLINE void XCHG_EbGb(void);
  BX_SMF INLINE void XCHG_EvGv(void);
  BX_SMF INLINE void MOV_EbGb(void);
  BX_SMF INLINE void MOV_EvGv(void);
  BX_SMF INLINE void MOV_GbEb(void);
  BX_SMF INLINE void MOV_GvEv(void);
  BX_SMF INLINE void MOV_EwSw(void);
  BX_SMF INLINE void LEA_GvM(void);
  BX_SMF INLINE void MOV_SwEw(void);
  BX_SMF INLINE void POP_Ev(void);
  
  BX_SMF INLINE void CBW(void);
  BX_SMF INLINE void CWD(void);
  BX_SMF INLINE void CALL_Ap(void);
  BX_SMF INLINE void WAIT(void);
  BX_SMF INLINE void PUSHF_Fv(void);
  BX_SMF INLINE void POPF_Fv(void);
  BX_SMF INLINE void SAHF(void);
  BX_SMF INLINE void LAHF(void);
  
  BX_SMF INLINE void MOV_ALOb(void);
  BX_SMF INLINE void MOV_eAXOv(void);
  BX_SMF INLINE void MOV_ObAL(void);
  BX_SMF INLINE void MOV_OveAX(void);
  BX_SMF INLINE void MOVSB_XbYb(void);
  BX_SMF INLINE void MOVSW_XvYv(void);
  BX_SMF INLINE void CMPSB_XbYb(void);
  BX_SMF INLINE void CMPSW_XvYv(void);
  BX_SMF INLINE void TEST_ALIb(void);
  BX_SMF INLINE void TEST_eAXIv(void);
  BX_SMF INLINE void STOSB_YbAL(void);
  BX_SMF INLINE void STOSW_YveAX(void);
  BX_SMF INLINE void LODSB_ALXb(void);
  BX_SMF INLINE void LODSW_eAXXv(void);
  BX_SMF INLINE void SCASB_ALXb(void);
  BX_SMF INLINE void SCASW_eAXXv(void);
  
  BX_SMF INLINE void RETnear(void);
  BX_SMF INLINE void LES_GvMp(void);
  BX_SMF INLINE void LDS_GvMp(void);
  BX_SMF INLINE void MOV_EbIb(void);
  BX_SMF INLINE void MOV_EvIv(void);
  BX_SMF INLINE void ENTER_IwIb(void);
  BX_SMF INLINE void LEAVE(void);
  BX_SMF INLINE void RETfar(void);

  BX_SMF INLINE void INT3(void);
  BX_SMF INLINE void INT_Ib(void);
  BX_SMF INLINE void INTO(void);
  BX_SMF INLINE void IRET(void);
  
  BX_SMF INLINE void AAM(void);
  BX_SMF INLINE void AAD(void);
  BX_SMF INLINE void SALC(void);
  BX_SMF INLINE void XLAT(void);
  
  BX_SMF INLINE void LOOPNE_Jb(void);
  BX_SMF INLINE void LOOPE_Jb(void);
  BX_SMF INLINE void LOOP_Jb(void);
  BX_SMF INLINE void JCXZ_Jb(void);
  BX_SMF INLINE void IN_ALIb(void);
  BX_SMF INLINE void IN_eAXIb(void);
  BX_SMF INLINE void OUT_IbAL(void);
  BX_SMF INLINE void OUT_IbeAX(void);
  BX_SMF INLINE void CALL_Av(void);
  BX_SMF INLINE void JMP_Jv(void);
  BX_SMF INLINE void JMP_Ap(void);
  BX_SMF INLINE void JMP_Jb(void);
  BX_SMF INLINE void IN_ALDX(void);
  BX_SMF INLINE void IN_eAXDX(void);
  BX_SMF INLINE void OUT_DXAL(void);
  BX_SMF INLINE void OUT_DXeAX(void);
  
  BX_SMF INLINE void HLT(void);
  BX_SMF INLINE void CMC(void);
  BX_SMF INLINE void CLC(void);
  BX_SMF INLINE void STC(void);
  BX_SMF INLINE void CLI(void);
  BX_SMF INLINE void STI(void);
  BX_SMF INLINE void CLD(void);
  BX_SMF INLINE void STD(void);
  
  
  BX_SMF INLINE void LAR_GvEw(void);
  BX_SMF INLINE void LSL_GvEw(void);
  BX_SMF INLINE void CLTS(void);
  BX_SMF INLINE void INVD(void);
  BX_SMF INLINE void WBINVD(void);
  
  BX_SMF INLINE void INVLPG_Ea(void);
  
  BX_SMF INLINE void MOV_CdRd(void);
  BX_SMF INLINE void MOV_DdRd(void);
  BX_SMF INLINE void MOV_RdCd(void);
  BX_SMF INLINE void MOV_RdDd(void);
  BX_SMF INLINE void MOV_TdRd(void);
  BX_SMF INLINE void MOV_RdTd(void);

  BX_SMF INLINE void JO_Jv(void);
  BX_SMF INLINE void JNO_Jv(void);
  BX_SMF INLINE void JB_Jv(void);
  BX_SMF INLINE void JNB_Jv(void);
  BX_SMF INLINE void JZ_Jv(void);
  BX_SMF INLINE void JNZ_Jv(void);
  BX_SMF INLINE void JBE_Jv(void);
  BX_SMF INLINE void JNBE_Jv(void);
  BX_SMF INLINE void JS_Jv(void);
  BX_SMF INLINE void JNS_Jv(void);
  BX_SMF INLINE void JP_Jv(void);
  BX_SMF INLINE void JNP_Jv(void);
  BX_SMF INLINE void JL_Jv(void);
  BX_SMF INLINE void JNL_Jv(void);
  BX_SMF INLINE void JLE_Jv(void);
  BX_SMF INLINE void JNLE_Jv(void);
  
  BX_SMF INLINE void SETO_Eb(void);
  BX_SMF INLINE void SETNO_Eb(void);
  BX_SMF INLINE void SETB_Eb(void);
  BX_SMF INLINE void SETNB_Eb(void);
  BX_SMF INLINE void SETZ_Eb(void);
  BX_SMF INLINE void SETNZ_Eb(void);
  BX_SMF INLINE void SETBE_Eb(void);
  BX_SMF INLINE void SETNBE_Eb(void);
  BX_SMF INLINE void SETS_Eb(void);
  BX_SMF INLINE void SETNS_Eb(void);
  BX_SMF INLINE void SETP_Eb(void);
  BX_SMF INLINE void SETNP_Eb(void);
  BX_SMF INLINE void SETL_Eb(void);
  BX_SMF INLINE void SETNL_Eb(void);
  BX_SMF INLINE void SETLE_Eb(void);
  BX_SMF INLINE void SETNLE_Eb(void);
  
  BX_SMF INLINE void PUSH_FS(void);
  BX_SMF INLINE void POP_FS(void);
  BX_SMF INLINE void CPUID(void);
  BX_SMF INLINE void BT_EvGv(void);
  BX_SMF INLINE void SHLD_EvGv(unsigned count);
  
  
  BX_SMF INLINE void PUSH_GS(void);
  BX_SMF INLINE void POP_GS(void);
  BX_SMF INLINE void BTS_EvGv(void);
  BX_SMF INLINE void SHRD_EvGv(unsigned count);
  BX_SMF INLINE void IMUL_GvEv(void);
  
  BX_SMF INLINE void LSS_GvMp(void);
  BX_SMF INLINE void BTR_EvGv(void);
  BX_SMF INLINE void LFS_GvMp(void);
  BX_SMF INLINE void LGS_GvMp(void);
  BX_SMF INLINE void MOVZX_GvEb(void);
  BX_SMF INLINE void MOVZX_GvEw(void);
  BX_SMF INLINE void BTC_EvGv(void);
  BX_SMF INLINE void BSF_GvEv(void);
  BX_SMF INLINE void BSR_GvEv(void);
  BX_SMF INLINE void MOVSX_GvEb(void);
  BX_SMF INLINE void MOVSX_GvEw(void);
  
  BX_SMF INLINE void BSWAP_EAX(void);
  BX_SMF INLINE void BSWAP_ECX(void);
  BX_SMF INLINE void BSWAP_EDX(void);
  BX_SMF INLINE void BSWAP_EBX(void);
  BX_SMF INLINE void BSWAP_ESP(void);
  BX_SMF INLINE void BSWAP_EBP(void);
  BX_SMF INLINE void BSWAP_ESI(void);
  BX_SMF INLINE void BSWAP_EDI(void);
  
  BX_SMF INLINE void ADD_EbIb(void);
  BX_SMF INLINE void OR_EbIb(void);
  BX_SMF INLINE void ADC_EbIb(void);
  BX_SMF INLINE void SBB_EbIb(void);
  BX_SMF INLINE void AND_EbIb(void);
  BX_SMF INLINE void SUB_EbIb(void);
  BX_SMF INLINE void XOR_EbIb(void);
  BX_SMF INLINE void CMP_EbIb(void);
  
  BX_SMF INLINE void ADD_EvIv(void);
  BX_SMF INLINE void OR_EvIv(void);
  BX_SMF INLINE void ADC_EvIv(void);
  BX_SMF INLINE void SBB_EvIv(void);
  BX_SMF INLINE void AND_EvIv(void);
  BX_SMF INLINE void SUB_EvIv(void);
  BX_SMF INLINE void XOR_EvIv(void);
  BX_SMF INLINE void CMP_EvIv(void);
  
  BX_SMF INLINE void ADD_EvIb(void);
  BX_SMF INLINE void OR_EvIb(void);
  BX_SMF INLINE void ADC_EvIb(void);
  BX_SMF INLINE void SBB_EvIb(void);
  BX_SMF INLINE void AND_EvIb(void);
  BX_SMF INLINE void SUB_EvIb(void);
  BX_SMF INLINE void XOR_EvIb(void);
  BX_SMF INLINE void CMP_EvIb(void);
  
  BX_SMF INLINE void ROL_Eb(unsigned count);
  BX_SMF INLINE void ROR_Eb(unsigned count);
  BX_SMF INLINE void RCL_Eb(unsigned count);
  BX_SMF INLINE void RCR_Eb(unsigned count);
  BX_SMF INLINE void SHL_Eb(unsigned count);
  BX_SMF INLINE void SHR_Eb(unsigned count);
  BX_SMF INLINE void SAR_Eb(unsigned count);
  
  BX_SMF INLINE void ROL_Ev(unsigned count);
  BX_SMF INLINE void ROR_Ev(unsigned count);
  BX_SMF INLINE void RCL_Ev(unsigned count);
  BX_SMF INLINE void RCR_Ev(unsigned count);
  BX_SMF INLINE void SHL_Ev(unsigned count);
  BX_SMF INLINE void SHR_Ev(unsigned count);
  BX_SMF INLINE void SAR_Ev(unsigned count);
  
  BX_SMF INLINE void TEST_EbIb(void);
  BX_SMF INLINE void NOT_Eb(void);
  BX_SMF INLINE void NEG_Eb(void);
  BX_SMF INLINE void MUL_ALEb(void);
  BX_SMF INLINE void IMUL_ALEb(void);
  BX_SMF INLINE void DIV_ALEb(void);
  BX_SMF INLINE void IDIV_ALEb(void);
  
  BX_SMF INLINE void TEST_EvIv(void);
  BX_SMF INLINE void NOT_Ev(void);
  BX_SMF INLINE void NEG_Ev(void);
  BX_SMF INLINE void MUL_eAXEv(void);
  BX_SMF INLINE void IMUL_eAXEv(void);
  BX_SMF INLINE void DIV_eAXEv(void);
  BX_SMF INLINE void IDIV_eAXEv(void);
  
  BX_SMF INLINE void INC_Eb(void);
  BX_SMF INLINE void DEC_Eb(void);
  
  BX_SMF INLINE void INC_Ev(void);
  BX_SMF INLINE void DEC_Ev(void);
  BX_SMF INLINE void CALL_Ev(void);
  BX_SMF INLINE void CALL_Ep(void);
  BX_SMF INLINE void JMP_Ev(void);
  BX_SMF INLINE void JMP_Ep(void);
  BX_SMF INLINE void PUSH_Ev(void);
  
  BX_SMF INLINE void SLDT_Ew(void);
  BX_SMF INLINE void STR_Ew(void);
  BX_SMF INLINE void LLDT_Ew(void);
  BX_SMF INLINE void LTR_Ew(void);
  BX_SMF INLINE void VERR_Ew(void);
  BX_SMF INLINE void VERW_Ew(void);
  
  BX_SMF INLINE void SGDT_Ms(void);
  BX_SMF INLINE void SIDT_Ms(void);
  BX_SMF INLINE void LGDT_Ms(void);
  BX_SMF INLINE void LIDT_Ms(void);
  BX_SMF INLINE void SMSW_Ew(void);
  BX_SMF INLINE void LMSW_Ew(void);
  
  
  BX_SMF INLINE void BT_EvIb(void);
  BX_SMF INLINE void BTS_EvIb(void);
  BX_SMF INLINE void BTR_EvIb(void);
  BX_SMF INLINE void BTC_EvIb(void);
  
  BX_SMF INLINE void FADD_Es(void);
  BX_SMF INLINE void FMUL_Es(void);
  BX_SMF INLINE void FCOM_Es(void);
  BX_SMF INLINE void FCOMP_Es(void);
  BX_SMF INLINE void FSUB_Es(void);
  BX_SMF INLINE void FSUBR_Es(void);
  BX_SMF INLINE void FDIV_Es(void);
  BX_SMF INLINE void FDIVR_Es(void);
  
  BX_SMF INLINE void FADD_ST_STi(void);
  BX_SMF INLINE void FMUL_ST_STi(void);
  BX_SMF INLINE void FCOM_ST_STi(void);
  BX_SMF INLINE void FCOMP_ST_STi(void);
  BX_SMF INLINE void FSUB_ST_STi(void);
  BX_SMF INLINE void FSUBR_ST_STi(void);
  BX_SMF INLINE void FDIV_ST_STi(void);
  BX_SMF INLINE void FDIVR_ST_STi(void);
  
  BX_SMF INLINE void FLD_Es(void);
  BX_SMF INLINE void FST_Es(void);
  BX_SMF INLINE void FSTP_Es(void);
  BX_SMF INLINE void FLDENV_Ea(void);
  BX_SMF INLINE void FLDCW_Ew(void);
  BX_SMF INLINE void FSTENV_Ea(void);
  BX_SMF INLINE void FSTCW_Ew(void);
  
  BX_SMF INLINE void FNOP(void);
  BX_SMF INLINE void FCHS(void);
  BX_SMF INLINE void FLD1(void);
  BX_SMF INLINE void F2XM1(void);
  BX_SMF INLINE void FPREM(void);
  
  BX_SMF INLINE void FABS(void);
  BX_SMF INLINE void FLDL2T(void);
  BX_SMF INLINE void FYL2X(void);
  BX_SMF INLINE void FYL2XP1(void);
  
  BX_SMF INLINE void FLDL2E(void);
  BX_SMF INLINE void FPTAN(void);
  BX_SMF INLINE void FSQRT(void);
  
  BX_SMF INLINE void FXCH_3(void);
  BX_SMF INLINE void FLDPI(void);
  BX_SMF INLINE void FPATAN(void);
  BX_SMF INLINE void FSINCOS(void);
  
  BX_SMF INLINE void FTST(void);
  BX_SMF INLINE void FLDLG2(void);
  BX_SMF INLINE void FXTRACT(void);
  BX_SMF INLINE void FRNDINT(void);
  
  BX_SMF INLINE void FXAM(void);
  BX_SMF INLINE void FSCALE(void);
  
  BX_SMF INLINE void FLDZ(void);
  BX_SMF INLINE void FDECSTP(void);
  BX_SMF INLINE void FSIN(void);
  BX_SMF INLINE void FINCSTP(void);
  BX_SMF INLINE void FCOS(void);
  
  BX_SMF INLINE void FIADD_Ew(void);
  BX_SMF INLINE void FIMUL_Ew(void);
  BX_SMF INLINE void FICOM_Ew(void);
  BX_SMF INLINE void FICOMP_Ew(void);
  BX_SMF INLINE void FISUB_Ew(void);
  BX_SMF INLINE void FISUBR_Ew(void);
  BX_SMF INLINE void FIDIV_Ew(void);
  BX_SMF INLINE void FIDIVR_Ew(void);
  
  BX_SMF INLINE void FUCOMPP(void);
  
  BX_SMF INLINE void FILD_Ew(void);
  BX_SMF INLINE void FIST_Ew(void);
  BX_SMF INLINE void FISTP_Ew(void);
  BX_SMF INLINE void FLD_Et(void);
  BX_SMF INLINE void FSTP_Et(void);
  
  BX_SMF INLINE void FENI(void);
  BX_SMF INLINE void FDISI(void);
  BX_SMF INLINE void FCLEX(void);
  BX_SMF INLINE void FINIT(void);
  BX_SMF INLINE void FSETPM(void);
  
  BX_SMF INLINE void FADD_El(void);
  BX_SMF INLINE void FMUL_El(void);
  BX_SMF INLINE void FCOM_El(void);
  BX_SMF INLINE void FCOMP_El(void);
  BX_SMF INLINE void FSUB_El(void);
  BX_SMF INLINE void FSUBR_El(void);
  BX_SMF INLINE void FDIV_El(void);
  BX_SMF INLINE void FDIVR_El(void);
  
  BX_SMF INLINE void FADD_STi_ST(void);
  BX_SMF INLINE void FMUL_STi_ST(void);
  BX_SMF INLINE void FCOM_STi_ST(void);
  BX_SMF INLINE void FCOMP_STi_ST(void);
  BX_SMF INLINE void FSUB_STi_ST(void);
  BX_SMF INLINE void FSUBR_STi_ST(void);
  BX_SMF INLINE void FDIV_STi_ST(void);
  BX_SMF INLINE void FDIVR_STi_ST(void);
  
  BX_SMF INLINE void FLD_El(void);
  BX_SMF INLINE void FST_El(void);
  BX_SMF INLINE void FSTP_El(void);
  BX_SMF INLINE void FRSTOR_Ea(void);
  BX_SMF INLINE void FSAVE_Ea(void);
  BX_SMF INLINE void FSTSW_Ew(void);
  
  BX_SMF INLINE void FFREE_STi(void);
  BX_SMF INLINE void FST_STi(void);
  BX_SMF INLINE void FSTP_STi(void);
  BX_SMF INLINE void FUCOMP_STi(void);
  BX_SMF INLINE void FIADD_Ed(void);
  BX_SMF INLINE void FIMUL_Ed(void);
  BX_SMF INLINE void FICOM_Ed(void);
  BX_SMF INLINE void FICOMP_Ed(void);
  BX_SMF INLINE void FISUB_Ed(void);
  BX_SMF INLINE void FISUBR_Ed(void);
  BX_SMF INLINE void FIDIV_Ed(void);
  BX_SMF INLINE void FIDIVR_Ed(void);
  
  BX_SMF INLINE void FADDP_STi_ST(void);
  BX_SMF INLINE void FMULP_STi_ST(void);
  BX_SMF INLINE void FCOMPP(void);
  BX_SMF INLINE void FSUBP_STi_ST(void);
  BX_SMF INLINE void FSUBRP_STi_ST(void);
  BX_SMF INLINE void FDIVP_STi_ST(void);
  BX_SMF INLINE void FDIVRP_STi_ST(void);
  
  BX_SMF INLINE void FILD_Ed(void);
  BX_SMF INLINE void FIST_Ed(void);
  BX_SMF INLINE void FISTP_Ed(void);
  BX_SMF INLINE void FBLD_Eb(void);
  BX_SMF INLINE void FILD_Eq(void);
  BX_SMF INLINE void FBSTP_Eb(void);
  BX_SMF INLINE void FISTP_Eq(void);
  
  BX_SMF INLINE void FSTSW_AX(void);
  
  BX_SMF INLINE void CMPXCHG_XBTS(void);
  BX_SMF INLINE void CMPXCHG_IBTS(void);
  BX_SMF INLINE void CMPXCHG_EbGb(void);
  BX_SMF INLINE void CMPXCHG_EvGv(void);
  BX_SMF INLINE void XADD_EbGb(void);
  BX_SMF INLINE void XADD_EvGv(void);
  BX_SMF INLINE void RETnear_Iw(void);
  BX_SMF INLINE void RETfar_Iw(void);
  BX_SMF INLINE void FLD_STi(void);
  BX_SMF INLINE void FXCH_STi(void);
  BX_SMF INLINE void FLDLN2(void);
  BX_SMF INLINE void FPREM1(void);
  BX_SMF INLINE void FUCOM_STi_ST(void);
  BX_SMF INLINE void FCOMP_STi(void);
  
  BX_SMF INLINE void LOADALL(void);
  BX_SMF INLINE void CMOV(Boolean condition);

  // now for some ancillary functions...
  BX_SMF void cpu_loop(void);
  BX_SMF void decode_exgx(void);
  BX_SMF void decode_ESC_modrm(void);

  BX_SMF void prefetch(void);
  BX_SMF void revalidate_prefetch_q(void);
  BX_SMF INLINE void invalidate_prefetch_q(void);

  BX_SMF INLINE Bit8u fetch_next_byte(void);
  BX_SMF INLINE Bit16u fetch_next_word(void);
  BX_SMF INLINE Bit32u fetch_next_dword(void);

  BX_SMF inline void write_virtual_checks(bx_segment_reg_t *seg, Bit32u offset, unsigned length);
  BX_SMF inline void read_virtual_checks(bx_segment_reg_t *seg, Bit32u offset, unsigned length);
  BX_SMF void write_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data);
  BX_SMF void write_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data);
  BX_SMF void write_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data);
  BX_SMF void read_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data);
  BX_SMF void read_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data);
  BX_SMF void read_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data);

  BX_SMF void read_RMW_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data);
  BX_SMF void read_RMW_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data);
  BX_SMF void read_RMW_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data);
  BX_SMF void write_RMW_virtual_byte(Bit8u val8);
  BX_SMF void write_RMW_virtual_word(Bit16u val16);
  BX_SMF void write_RMW_virtual_dword(Bit32u val32);

  BX_SMF void access_linear(Bit32u address, unsigned length, unsigned pl,
                     unsigned rw, void *data);
  BX_SMF Bit32u itranslate_linear(Bit32u laddress, unsigned pl);
  BX_SMF Bit32u dtranslate_linear(Bit32u laddress, unsigned pl, unsigned rw);
  BX_SMF void TLB_flush(void);
  BX_SMF void set_INTR(Boolean value);
  BX_SMF char *strseg(bx_segment_reg_t *seg);
  BX_SMF void interrupt(Bit8u vector, Boolean is_INT, Boolean is_error_code,
                 Bit16u error_code);
#if BX_CPU >= 2
  BX_SMF void exception(unsigned vector, Bit16u error_code, Boolean is_INT);
#endif
  BX_SMF int  int_number(bx_segment_reg_t *seg);
  BX_SMF void reset_cpu(void);
  BX_SMF void shutdown_cpu(void);
  BX_SMF void enable_paging(void);
  BX_SMF void disable_paging(void);
  BX_SMF void CR3_change(Bit32u value32);
  BX_SMF void init_cpu(void);

  BX_SMF void jump_protected(Bit16u cs, Bit32u disp32);
  BX_SMF void call_protected(Bit16u cs, Bit32u disp32);
  BX_SMF void return_protected(Bit16u pop_bytes);
  BX_SMF void iret_protected(void);
  BX_SMF void validate_seg_regs(void);
  BX_SMF void stack_return_to_v86(Bit32u new_eip, Bit32u raw_cs_selector,
                                     Bit32u flags32);
  BX_SMF void stack_return_from_v86(void);
  BX_SMF void init_v8086_mode(void);
  BX_SMF void v8086_message(void);
  BX_SMF void task_switch(bx_selector_t *selector,
                     bx_descriptor_t *descriptor,
                     unsigned source,
                     Bit32u dword1, Bit32u dword2);
  BX_SMF void write_flags(Bit16u flags, Boolean change_IOPL, Boolean change_IF);
  BX_SMF void write_eflags(Bit32u eflags, Boolean change_IOPL, Boolean change_IF,
                    Boolean change_VM, Boolean change_RF);
  BX_SMF Bit16u read_flags(void);
  BX_SMF Bit32u read_eflags(void);

  BX_SMF Bit8u   inp8(Bit16u addr);
  BX_SMF void    outp8(Bit16u addr, Bit8u value);
  BX_SMF Bit16u  inp16(Bit16u addr);
  BX_SMF void    outp16(Bit16u addr, Bit16u value);
  BX_SMF Bit32u  inp32(Bit16u addr);
  BX_SMF void    outp32(Bit16u addr, Bit32u value);
  BX_SMF Boolean allow_io(Bit16u addr, unsigned len);
  BX_SMF void    enter_protected_mode(void);
  BX_SMF void    enter_real_mode(void);
  BX_SMF Bit16u  read_seg_reg(Bit32u index);
  BX_SMF void    write_seg_reg(Bit32u index, Bit16u value);
  BX_SMF void    parse_selector(Bit16u raw_selector, bx_selector_t *selector);
  BX_SMF void    parse_descriptor(Bit32u dword1, Bit32u dword2, bx_descriptor_t *temp);
  BX_SMF void    load_ldtr(bx_selector_t *selector, bx_descriptor_t *descriptor);
  BX_SMF void    load_cs(bx_selector_t *selector, bx_descriptor_t *descriptor, Bit8u cpl);
  BX_SMF void    load_ss(bx_selector_t *selector, bx_descriptor_t *descriptor, Bit8u cpl);
  BX_SMF void    fetch_raw_descriptor(bx_selector_t *selector,
                               Bit32u *dword1, Bit32u *dword2, Bit8u exception);
  BX_SMF void    load_seg_reg(bx_segment_reg_t *seg, Bit16u new_value);
  BX_SMF Boolean fetch_raw_descriptor2(bx_selector_t *selector,
                                Bit32u *dword1, Bit32u *dword2);
  BX_SMF void    push_16(Bit16u value16);
  BX_SMF void    push_32(Bit32u value32);
  BX_SMF void    pop_16(Bit16u *value16_ptr);
  BX_SMF void    pop_32(Bit32u *value32_ptr);
  BX_SMF Boolean can_push(bx_descriptor_t *descriptor, Bit32u esp, Bit32u bytes);
  BX_SMF Boolean can_pop(Bit32u bytes);
  BX_SMF void    sanity_checks(void);

  BX_SMF void    debug(Bit32u offset);

  BX_SMF inline void set_CF(Boolean val);
  BX_SMF inline void set_AF(Boolean val);
  BX_SMF inline void set_ZF(Boolean val);
  BX_SMF inline void set_SF(Boolean val);
  BX_SMF inline void set_OF(Boolean val);
  BX_SMF inline void set_PF(Boolean val);
  BX_SMF inline void set_PF_base(Bit8u val);


  BX_SMF inline void set_AX(Bit16u ax);
  BX_SMF inline void set_BX(Bit16u bx);
  BX_SMF inline void set_CX(Bit16u cx);
  BX_SMF inline void set_DX(Bit16u dx);
  BX_SMF inline void set_AL(Bit8u  al);
  BX_SMF inline void set_AH(Bit8u  ah);
  BX_SMF inline void set_BL(Bit8u  bl);
  BX_SMF inline void set_BH(Bit8u  bh);
  BX_SMF inline void set_CL(Bit8u  cl);
  BX_SMF inline void set_CH(Bit8u  ch);
  BX_SMF inline void set_DL(Bit8u  dl);
  BX_SMF inline void set_DH(Bit8u  dh);

  BX_SMF inline Bit8u get_AL(void);
  BX_SMF inline Bit8u get_AH(void);
  BX_SMF inline Bit8u get_BL(void);
  BX_SMF inline Bit8u get_BH(void);
  BX_SMF inline Bit8u get_CL(void);
  BX_SMF inline Bit8u get_CH(void);
  BX_SMF inline Bit8u get_DL(void);
  BX_SMF inline Bit8u get_DH(void);

  BX_SMF inline Bit16u get_AX(void);
  BX_SMF inline Bit16u get_BX(void);
  BX_SMF inline Bit16u get_CX(void);
  BX_SMF inline Bit16u get_DX(void);

#if BX_CPU >= 2
  BX_SMF inline Boolean real_mode(void);
#endif
#if BX_CPU >= 3
  BX_SMF inline Boolean protected_mode(void);
  BX_SMF inline Boolean v8086_mode(void);
#endif
  };



extern bx_cpu_c       bx_cpu;


BX_SMF inline void BX_CPU_C_PREFIX set_AX(Bit16u ax) { AX = ax; };
BX_SMF inline void BX_CPU_C_PREFIX set_BX(Bit16u bx) { BX = bx; };
BX_SMF inline void BX_CPU_C_PREFIX set_CX(Bit16u cx) { CX = cx; };
BX_SMF inline void BX_CPU_C_PREFIX set_DX(Bit16u dx) { DX = dx; };
BX_SMF inline void BX_CPU_C_PREFIX set_AL(Bit8u  al) { AL = al; };
BX_SMF inline void BX_CPU_C_PREFIX set_AH(Bit8u  ah) { AH = ah; };
BX_SMF inline void BX_CPU_C_PREFIX set_BL(Bit8u  bl) { BL = bl; };
BX_SMF inline void BX_CPU_C_PREFIX set_BH(Bit8u  bh) { BH = bh; };
BX_SMF inline void BX_CPU_C_PREFIX set_CL(Bit8u  cl) { CL = cl; };
BX_SMF inline void BX_CPU_C_PREFIX set_CH(Bit8u  ch) { CH = ch; };
BX_SMF inline void BX_CPU_C_PREFIX set_DL(Bit8u  dl) { DL = dl; };
BX_SMF inline void BX_CPU_C_PREFIX set_DH(Bit8u  dh) { DH = dh; };

BX_SMF inline Bit8u BX_CPU_C_PREFIX get_AL(void) { return(AL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_AH(void) { return(AH); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_BL(void) { return(BL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_BH(void) { return(BH); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_CL(void) { return(CL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_CH(void) { return(CH); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_DL(void) { return(DL); };
BX_SMF inline Bit8u BX_CPU_C_PREFIX get_DH(void) { return(DH); };

BX_SMF inline Bit16u BX_CPU_C_PREFIX get_AX(void) { return(AX); };
BX_SMF inline Bit16u BX_CPU_C_PREFIX get_BX(void) { return(BX); };
BX_SMF inline Bit16u BX_CPU_C_PREFIX get_CX(void) { return(CX); };
BX_SMF inline Bit16u BX_CPU_C_PREFIX get_DX(void) { return(DX); };


#if BX_CPU >= 2
  inline Boolean bx_cpu_c::real_mode(void) { return( !BX_CPU_THIS_PTR cr0.pe ); };
#endif

#if BX_CPU == 2
  inline Boolean bx_cpu_c::protected_mode(void) { return( BX_CPU_THIS_PTR cr0.pe ); };
#endif


#if BX_CPU >= 3
#  if BX_SUPPORT_V8086_MODE
  inline Boolean
bx_cpu_c::v8086_mode(void) {
  return(BX_CPU_THIS_PTR eflags.vm);
  }

  inline Boolean
bx_cpu_c::protected_mode(void) {
  return(BX_CPU_THIS_PTR cr0.pe && !BX_CPU_THIS_PTR eflags.vm);
  }
#  else
  inline Boolean
bx_cpu_c::v8086_mode(void) {
  return(0);
  }

  inline Boolean
bx_cpu_c::protected_mode(void) {
  return(BX_CPU_THIS_PTR cr0.pe);
  }
#  endif
#endif

    inline void
bx_cpu_c::set_CF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xfffff0;
    BX_CPU_THIS_PTR eflags.cf = val;
    }

    inline void
bx_cpu_c::set_AF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xfff0ff;
    BX_CPU_THIS_PTR eflags.af = val;
    }

    inline void
bx_cpu_c::set_ZF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xff0fff;
    BX_CPU_THIS_PTR eflags.zf = val;
    }

    inline void
bx_cpu_c::set_SF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xf0ffff;
    BX_CPU_THIS_PTR eflags.sf = val;
    }


    inline void
bx_cpu_c::set_OF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0x0fffff;
    BX_CPU_THIS_PTR eflags.of = val;
    }

    inline void
bx_cpu_c::set_PF(Boolean val) {
    BX_CPU_THIS_PTR lf_flags_status &= 0xffff0f;
    BX_CPU_THIS_PTR lf_pf = val;
    }

    inline void
bx_cpu_c::set_PF_base(Bit8u val) {
    BX_CPU_THIS_PTR eflags.pf_byte = val;
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0xffff0f) | BX_LF_MASK_P;
    }


#define SET_FLAGS_OSZAPC_8(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszapc.op1_8 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_8 = op2; \
    BX_CPU_THIS_PTR oszapc.result_8 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_8_CF(op1, op2, result, ins, last_CF) { \
    BX_CPU_THIS_PTR oszapc.op1_8 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_8 = op2; \
    BX_CPU_THIS_PTR oszapc.result_8 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR oszapc.prev_CF = last_CF; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_16(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszapc.op1_16 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_16 = op2; \
    BX_CPU_THIS_PTR oszapc.result_16 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_16_CF(op1, op2, result, ins, last_CF) { \
    BX_CPU_THIS_PTR oszapc.op1_16 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_16 = op2; \
    BX_CPU_THIS_PTR oszapc.result_16 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR oszapc.prev_CF = last_CF; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_32(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszapc.op1_32 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_32 = op2; \
    BX_CPU_THIS_PTR oszapc.result_32 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }

#define SET_FLAGS_OSZAPC_32_CF(op1, op2, result, ins, last_CF) { \
    BX_CPU_THIS_PTR oszapc.op1_32 = op1; \
    BX_CPU_THIS_PTR oszapc.op2_32 = op2; \
    BX_CPU_THIS_PTR oszapc.result_32 = result; \
    BX_CPU_THIS_PTR oszapc.instr = ins; \
    BX_CPU_THIS_PTR oszapc.prev_CF = last_CF; \
    BX_CPU_THIS_PTR lf_flags_status = BX_LF_MASK_OSZAPC; \
    }


#define SET_FLAGS_OSZAP_8(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszap.op1_8 = op1; \
    BX_CPU_THIS_PTR oszap.op2_8 = op2; \
    BX_CPU_THIS_PTR oszap.result_8 = result; \
    BX_CPU_THIS_PTR oszap.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP; \
    }

#define SET_FLAGS_OSZAP_16(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszap.op1_16 = op1; \
    BX_CPU_THIS_PTR oszap.op2_16 = op2; \
    BX_CPU_THIS_PTR oszap.result_16 = result; \
    BX_CPU_THIS_PTR oszap.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP; \
    }

#define SET_FLAGS_OSZAP_32(op1, op2, result, ins) { \
    BX_CPU_THIS_PTR oszap.op1_32 = op1; \
    BX_CPU_THIS_PTR oszap.op2_32 = op2; \
    BX_CPU_THIS_PTR oszap.result_32 = result; \
    BX_CPU_THIS_PTR oszap.instr = ins; \
    BX_CPU_THIS_PTR lf_flags_status = (BX_CPU_THIS_PTR lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP; \
    }

#define SET_FLAGS_OxxxxC(new_of, new_cf) { \
    BX_CPU_THIS_PTR eflags.of = (Boolean) new_of; \
    BX_CPU_THIS_PTR eflags.cf = (Boolean) new_cf; \
    BX_CPU_THIS_PTR lf_flags_status &= 0x0ffff0; \
    /* ??? could also mark other bits undefined here */ \
    }



#if 0
    inline void
  SET_FLAGS_OSZAPC_8(Bit8u op1, Bit8u op2, Bit8u result, unsigned instr) {
    oszapc.op1_8 = op1;
    oszapc.op2_8 = op2;
    oszapc.result_8 = result;
    oszapc.instr = instr;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_8_CF(Bit8u op1, Bit8u op2, Bit8u result, unsigned instr,
                         Boolean prev_CF) {
    oszapc.op1_8 = op1;
    oszapc.op2_8 = op2;
    oszapc.result_8 = result;
    oszapc.instr = instr;
    oszapc.prev_CF = prev_CF;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_16(Bit16u op1, Bit16u op2, Bit16u result, unsigned instr) {
    oszapc.op1_16 = op1;
    oszapc.op2_16 = op2;
    oszapc.result_16 = result;
    oszapc.instr = instr;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_16_CF(Bit16u op1, Bit16u op2, Bit16u result, unsigned instr,
                          Boolean prev_CF) {
    oszapc.op1_16 = op1;
    oszapc.op2_16 = op2;
    oszapc.result_16 = result;
    oszapc.instr = instr;
    oszapc.prev_CF = prev_CF;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_32(Bit32u op1, Bit32u op2, Bit32u result, unsigned instr) {
    oszapc.op1_32 = op1;
    oszapc.op2_32 = op2;
    oszapc.result_32 = result;
    oszapc.instr = instr;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }

    inline void
  SET_FLAGS_OSZAPC_32_CF(Bit32u op1, Bit32u op2, Bit32u result, unsigned instr,
                          Boolean prev_CF) {
    oszapc.op1_32 = op1;
    oszapc.op2_32 = op2;
    oszapc.result_32 = result;
    oszapc.instr = instr;
    oszapc.prev_CF = prev_CF;
    lf_flags_status = BX_LF_MASK_OSZAPC;
    }


    inline void
  SET_FLAGS_OSZAP_8(Bit8u op1, Bit8u op2, Bit8u result, unsigned instr) {
    oszap.op1_8 = op1;
    oszap.op2_8 = op2;
    oszap.result_8 = result;
    oszap.instr = instr;
    lf_flags_status = (lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP;
    }

    inline void
  SET_FLAGS_OSZAP_16(Bit16u op1, Bit16u op2, Bit16u result, unsigned instr) {
    oszap.op1_16 = op1;
    oszap.op2_16 = op2;
    oszap.result_16 = result;
    oszap.instr = instr;
    lf_flags_status = (lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP;
    }

    inline void
  SET_FLAGS_OSZAP_32(Bit32u op1, Bit32u op2, Bit32u result, unsigned instr) {
    oszap.op1_32 = op1;
    oszap.op2_32 = op2;
    oszap.result_32 = result;
    oszap.instr = instr;
    lf_flags_status = (lf_flags_status & 0x00000f) | BX_LF_MASK_OSZAP;
    }

    inline void
  SET_FLAGS_OxxxxC(Boolean new_of, Boolean new_cf) {
    eflags.of = new_of;
    eflags.cf = new_cf;
    lf_flags_status &= 0x0ffff0;
    /* ??? could also mark other bits undefined here */
    }
#endif



extern const Boolean bx_parity_lookup[256];

#define BX_REPE_PREFIX  10
#define BX_REPNE_PREFIX 11



#define BX_TASK_FROM_JUMP         10
#define BX_TASK_FROM_CALL_OR_INT  11
#define BX_TASK_FROM_IRET         12


#if USE_INLINE
#  error "rehash this in cpu.h"
#  include "fetch.cc"
#endif

#endif  // #ifndef BX_CPU_H
