// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"





  INLINE void
bx_cpu_c::JCXZ_Jb()
{
  Bit8s offset;
  Bit32u temp_ECX;

  offset = (Bit8s) fetch_next_byte();

  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    temp_ECX = ECX;
  else
    temp_ECX = CX;

  if ( temp_ECX == 0 ) {
    Bit32u new_EIP;

    new_EIP = EIP + offset;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if (protected_mode()) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("jcxz_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
#endif
    EIP = new_EIP;
    BX_INSTR_CNEAR_BRANCH(new_EIP);
    revalidate_prefetch_q();
    }
}


  INLINE void
bx_cpu_c::RETnear_Iw()
{
  Bit16u imm16;
  Bit32u temp_ESP;

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
    temp_ESP = ESP;
  else
    temp_ESP = SP;

  imm16 = fetch_next_word();

  invalidate_prefetch_q();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u return_EIP;
    
    if (protected_mode()) {
      if ( !can_pop(4) ) {
        bx_panic("retnear_iw: can't pop EIP\n");
        /* ??? #SS(0) -or #GP(0) */
        }
      
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        4, CPL==3, BX_READ, &return_EIP);
  
      if (protected_mode() &&
          (return_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) ) {
        bx_panic("retnear_iw: EIP > limit\n");
        }
  
      /* Pentium book says imm16 is number of words ??? */
      if ( !can_pop(4 + imm16) ) {
        bx_panic("retnear_iw: can't release bytes from stack\n");
        /* #GP(0) -or #SS(0) ??? */
        }
  
      BX_CPU_THIS_PTR eip = return_EIP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 4 + imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += 4 + imm16;
      }
    else {
      pop_32(&return_EIP);
      BX_CPU_THIS_PTR eip = return_EIP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += imm16;
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u return_IP;

    if (protected_mode()) {
      if ( !can_pop(2) ) {
        bx_panic("retnear_iw: can't pop IP\n");
        /* ??? #SS(0) -or #GP(0) */
        }
      
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        2, CPL==3, BX_READ, &return_IP);
  
      if ( return_IP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("retnear_iw: IP > limit\n");
        }
  
      if ( !can_pop(2 + imm16) ) {
        bx_panic("retnear_iw: can't release bytes from stack\n");
        /* #GP(0) -or #SS(0) ??? */
        }
  
      BX_CPU_THIS_PTR eip = return_IP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 2 + imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += 2 + imm16;
      }
    else {
      pop_16(&return_IP);
      BX_CPU_THIS_PTR eip = return_IP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += imm16; /* ??? should it be 2*imm16 ? */
      else
        SP  += imm16;
      }
    }

  BX_INSTR_UCNEAR_BRANCH(BX_CPU_THIS_PTR eip);
}

  INLINE void
bx_cpu_c::RETnear()
{
  Bit32u temp_ESP;

  invalidate_prefetch_q();

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
    temp_ESP = ESP;
  else
    temp_ESP = SP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u return_EIP;
    
    if (protected_mode()) {
      if ( !can_pop(4) ) {
        bx_panic("retnear: can't pop EIP\n");
        /* ??? #SS(0) -or #GP(0) */
        }
    
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        4, CPL==3, BX_READ, &return_EIP);

      if ( return_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("retnear: EIP > limit\n");
        //exception(BX_GP_EXCEPTION, 0, 0);
        }
      BX_CPU_THIS_PTR eip = return_EIP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 4;
      else
        SP  += 4;
      }
    else {
      pop_32(&return_EIP);
      BX_CPU_THIS_PTR eip = return_EIP;
      }
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u return_IP;

    if (protected_mode()) {
      if ( !can_pop(2) ) {
        bx_panic("retnear: can't pop IP\n");
        /* ??? #SS(0) -or #GP(0) */
        }
    
      access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 0,
        2, CPL==3, BX_READ, &return_IP);

      if ( return_IP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("retnear: IP > limit\n");
        }

      BX_CPU_THIS_PTR eip = return_IP;
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) /* 32bit stack */
        ESP += 2;
      else
        SP  += 2;
      }
    else {
      pop_16(&return_IP);
      BX_CPU_THIS_PTR eip = return_IP;
      }
    }

  BX_INSTR_UCNEAR_BRANCH(BX_CPU_THIS_PTR eip);
}

  INLINE void
bx_cpu_c::RETfar_Iw()
{
  Bit16s imm16;

  /* ??? is imm16, number of bytes/words depending on operandsize ? */

  imm16 = fetch_next_word();

  invalidate_prefetch_q();

#if BX_CPU >= 2
  if (protected_mode()) {
    BX_CPU_THIS_PTR return_protected(imm16);
    goto done;
    }
#endif

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eip, ecs_raw;

    pop_32(&eip);
    pop_32(&ecs_raw);
    BX_CPU_THIS_PTR eip = eip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, (Bit16u) ecs_raw);
    if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
      ESP += imm16;
    else
      SP  += imm16;
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u ip, cs_raw;

    pop_16(&ip);
    pop_16(&cs_raw);
    BX_CPU_THIS_PTR eip = (Bit32u) ip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
      ESP += imm16;
    else
      SP  += imm16;
    }

done:
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
  return;
}

  INLINE void
bx_cpu_c::RETfar()
{
  invalidate_prefetch_q();

#if BX_CPU >= 2
  if ( protected_mode() ) {
    BX_CPU_THIS_PTR return_protected(0);
    goto done;
    }
#endif

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eip, ecs_raw;

    pop_32(&eip);
    pop_32(&ecs_raw); /* 32bit pop, MSW discarded */
    BX_CPU_THIS_PTR eip = eip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, (Bit16u) ecs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u ip, cs_raw;

    pop_16(&ip);
    pop_16(&cs_raw);
    BX_CPU_THIS_PTR eip = (Bit32u) ip;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }

done:
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
  return;
}


  INLINE void
bx_cpu_c::LOOPNE_Jb()
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = fetch_next_byte();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    count = ECX;
  else
#endif /* BX_CPU >= 3 */
    count = CX;

  count--;
  if ( (count!=0) && (get_ZF()==0) ) {

    new_EIP = EIP + imm8;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loopne_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    BX_INSTR_CNEAR_BRANCH(new_EIP);
    revalidate_prefetch_q();
    }

  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    ECX--;
  else
    CX--;
}

  INLINE void
bx_cpu_c::LOOPE_Jb()
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = fetch_next_byte();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    count = ECX;
  else
#endif /* BX_CPU >= 3 */
    count = CX;

  count--;
  if ( (count!=0) && get_ZF()) {

    new_EIP = EIP + imm8;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loope_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    BX_INSTR_CNEAR_BRANCH(new_EIP);
    revalidate_prefetch_q();
    }
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    ECX--;
  else
    CX--;
}

  INLINE void
bx_cpu_c::LOOP_Jb()
{
  Bit8s imm8;
  Bit32u count, new_EIP;

  imm8 = fetch_next_byte();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    count = ECX;
  else
#endif /* BX_CPU >= 3 */
    count = CX;

  count--;
  if (count != 0) {

    new_EIP = EIP + imm8;
    if (BX_CPU_THIS_PTR is_32bit_opsize==0)
      new_EIP &= 0x0000ffff;
    if (protected_mode()) {
      if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("loop_jb: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }
    BX_CPU_THIS_PTR eip = new_EIP;
    BX_INSTR_CNEAR_BRANCH(new_EIP);
    revalidate_prefetch_q();
    }
  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    ECX--;
  else
    CX--;
}

  INLINE void
bx_cpu_c::CALL_Av()
{
  Bit32u new_EIP;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32s disp32;

    disp32 = fetch_next_dword();
    invalidate_prefetch_q();

    new_EIP = EIP + disp32;

    if ( protected_mode() ) {
      if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
        bx_panic("call_av: offset outside of CS limits\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      }

    /* push 32 bit EA of next instruction */
    push_32(BX_CPU_THIS_PTR eip);
    BX_CPU_THIS_PTR eip = new_EIP;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit opsize mode */
    Bit16s disp16;


    disp16 = fetch_next_word();
    invalidate_prefetch_q();

    new_EIP = EIP + disp16;
    new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
    if ( protected_mode() &&
         (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) ) {
      bx_panic("call_av: new_IP > BX_CPU_THIS_PTR cs.limit\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      }
#endif

    /* push 16 bit EA of next instruction */
    push_16(IP);

    BX_CPU_THIS_PTR eip = new_EIP;
    }

  BX_INSTR_UCNEAR_BRANCH(BX_CPU_THIS_PTR eip);
}

  INLINE void
bx_cpu_c::CALL_Ap()
{
  Bit16u cs_raw;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u disp32;

    disp32 = fetch_next_dword();
    cs_raw     = fetch_next_word();
    invalidate_prefetch_q();

    if (protected_mode()) {
      BX_CPU_THIS_PTR call_protected(cs_raw, disp32);
      goto done;
      }
    push_32(BX_CPU_THIS_PTR cs.selector.value);
    push_32(BX_CPU_THIS_PTR eip);
    BX_CPU_THIS_PTR eip = disp32;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit opsize mode */
    Bit16u disp16;

    disp16 = fetch_next_word();
    cs_raw     = fetch_next_word();
    invalidate_prefetch_q();

#if BX_CPU >= 2
    if (protected_mode()) {
      BX_CPU_THIS_PTR call_protected(cs_raw, disp16);
      goto done;
      }
#endif
    push_16(BX_CPU_THIS_PTR cs.selector.value);
    push_16((Bit16u) BX_CPU_THIS_PTR eip);
    BX_CPU_THIS_PTR eip = (Bit32u) disp16;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }

done:
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
  return;
}

  INLINE void
bx_cpu_c::CALL_Ev()
{
  Bit32u temp_ESP;

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    temp_ESP = ESP;
  else
    temp_ESP = SP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }
    invalidate_prefetch_q();

    if (protected_mode()) {
      if (op1_32 > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("call_ev: EIP out of CS limits!\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, 4) ) {
        bx_panic("call_ev: can't push EIP\n");
        }
      }

    push_32(BX_CPU_THIS_PTR eip);

    BX_CPU_THIS_PTR eip = op1_32;
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }
    invalidate_prefetch_q();

#if BX_CPU >= 2
    if (protected_mode()) {
      if (op1_16 > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
        bx_panic("call_ev: IP out of CS limits!\n");
        exception(BX_GP_EXCEPTION, 0, 0);
        }
      if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, 2) ) {
        bx_panic("call_ev: can't push IP\n");
        }
      }
#endif

    push_16(IP);

    BX_CPU_THIS_PTR eip = op1_16;
    }

  BX_INSTR_UCNEAR_BRANCH(BX_CPU_THIS_PTR eip);
}

  INLINE void
bx_cpu_c::CALL_Ep()
{
  Bit16u cs_raw;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      bx_panic("CALL_Ep: op1 is a register\n");
      }

    /* pointer, segment address pair */
    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+4, &cs_raw);
    invalidate_prefetch_q();

    if ( protected_mode() ) {
      BX_CPU_THIS_PTR call_protected(cs_raw, op1_32);
      goto done;
      }

    push_32(BX_CPU_THIS_PTR cs.selector.value);
    push_32(BX_CPU_THIS_PTR eip);

    BX_CPU_THIS_PTR eip = op1_32;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      bx_panic("CALL_Ep: op1 is a register");
      }

    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+2, &cs_raw);
    invalidate_prefetch_q();

    if ( protected_mode() ) {
      BX_CPU_THIS_PTR call_protected(cs_raw, op1_16);
      goto done;
      }

    push_16(BX_CPU_THIS_PTR cs.selector.value);
    push_16(IP);

    BX_CPU_THIS_PTR eip = op1_16;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }

done:
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
  return;
}


  INLINE void
bx_cpu_c::JMP_Jv()
{
  Bit32u new_EIP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32s imm32;

    imm32 = fetch_next_dword();
    invalidate_prefetch_q();

    new_EIP = EIP + imm32;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit operand size mode */
    Bit16s imm16;

    imm16 = fetch_next_word();
    invalidate_prefetch_q();

    new_EIP = EIP + imm16;
    new_EIP &= 0x0000ffff;
    }

#if BX_CPU >= 2
  if (protected_mode()) {
    if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
      bx_panic("jmp_jv: offset outside of CS limits\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      }
    }
#endif

  BX_CPU_THIS_PTR eip = new_EIP;
  BX_INSTR_UCNEAR_BRANCH(new_EIP);
}

  INLINE void
bx_cpu_c::JMP_Ap()
{
  Bit16u cs_raw;
  Bit32u disp32;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    disp32 = fetch_next_dword();
    }
  else { /* 16bit operands */
    disp32 = fetch_next_word(); /* clear upper 16 bits */
    }

  cs_raw = fetch_next_word();
  invalidate_prefetch_q();

#if BX_CPU >= 2
  if (protected_mode()) {
    BX_CPU_THIS_PTR jump_protected(cs_raw, disp32);
    goto done;
    }
#endif

  load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
  BX_CPU_THIS_PTR eip = disp32;

done:
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
  return;
}



  INLINE void
bx_cpu_c::JMP_Jb()
{
  Bit8s imm8;
  Bit32u new_EIP;

  imm8 = fetch_next_byte();
  invalidate_prefetch_q();

#if BX_WEIRDISMS
  if (imm8 == -2) {
    if (BX_CPU_THIS_PTR eflags.if_ == 0)
      bx_panic("endless loop detected!\n");
    }
#endif

  new_EIP = EIP + imm8;
  if (BX_CPU_THIS_PTR is_32bit_opsize==0)
    new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
  if (protected_mode()) {
    if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
      bx_panic("jmp_jb: offset outside of CS limits\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      }
    }
#endif

  EIP = new_EIP;
  BX_INSTR_UCNEAR_BRANCH(new_EIP);
}



  INLINE void
bx_cpu_c::JMP_Ev()
{
  Bit32u new_EIP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    invalidate_prefetch_q();
    new_EIP = op1_32;
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    invalidate_prefetch_q();
    new_EIP = op1_16;
    }

#if BX_CPU >= 2
  if (protected_mode()) {
    if (new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
      bx_panic("jmp_ev: IP out of CS limits!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      }
    }
#endif

  BX_CPU_THIS_PTR eip = new_EIP;

  BX_INSTR_UCNEAR_BRANCH(new_EIP);
}

  /* Far indirect jump */

  INLINE void
bx_cpu_c::JMP_Ep()
{
  Bit16u cs_raw;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      /* far indirect must specify a memory address */
      bx_panic("JMP_Ep(): op1 is a register\n");
      }

    /* pointer, segment address pair */
    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+4, &cs_raw);
    invalidate_prefetch_q();

    if ( protected_mode() ) {
      BX_CPU_THIS_PTR jump_protected(cs_raw, op1_32);
      goto done;
      }

    BX_CPU_THIS_PTR eip = op1_32;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      /* far indirect must specify a memory address */
      bx_panic("JMP_Ep(): op1 is a register\n");
      }

    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+2, &cs_raw);
    invalidate_prefetch_q();

#if BX_CPU >= 2
    if ( protected_mode() ) {
      BX_CPU_THIS_PTR jump_protected(cs_raw, op1_16);
      goto done;
      }
#endif

    BX_CPU_THIS_PTR eip = op1_16;
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    }

done:
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
  return;
}

  INLINE void
bx_cpu_c::IRET()
{
  invalidate_prefetch_q();

  if (v8086_mode()) {
    // IOPL check in stack_return_from_v86()
    stack_return_from_v86();
    goto done;
    }

#if BX_CPU >= 2
  if (protected_mode()) {
    iret_protected();
    goto done;
    }
#endif

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eip, ecs_raw, eflags;

    pop_32(&eip);
    pop_32(&ecs_raw);
    pop_32(&eflags);

    load_seg_reg(&BX_CPU_THIS_PTR cs, (Bit16u) ecs_raw);
    BX_CPU_THIS_PTR eip = eip;
    write_eflags(eflags, 1, 1, 0, 1);
    }
  else
#endif /* BX_CPU >= 3 */
    {
    Bit16u ip, cs_raw, flags;

    pop_16(&ip);
    pop_16(&cs_raw);
    pop_16(&flags);

    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    BX_CPU_THIS_PTR eip = (Bit32u) ip;
    write_flags(flags, 1, 1);
    }

done:
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
  return;
}
