// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com






#include "bochs.h"


  INLINE void
bx_cpu_c::MOV_EbGb()
{
  Bit8u op2;

  decode_exgx();

  /* op2 is a register, op2_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* now write op2 to op1 */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, op2);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2);
    }
}

  INLINE void
bx_cpu_c::MOV_EvGv()
{
  decode_exgx();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op2_32;

    /* op2_32 is a register, op2_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_32 is a register or memory reference */
    /* now write op2 to op1 */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, op2_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit operand size mode */
    Bit16u op2_16;

    /* op2_16 is a register, op2_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_16 is a register or memory reference */
    /* now write op2 to op1 */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, op2_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }
    }
}

  INLINE void
bx_cpu_c::MOV_GbEb()
{
  Bit8u op2;

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2);
    }

  BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR reg_index, op2);
}

  INLINE void
bx_cpu_c::MOV_GvEv()
{
  decode_exgx();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op2_32;

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, op2_32);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit operand size mode */
    Bit16u op2_16;

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, op2_16);
    }
}

  INLINE void
bx_cpu_c::MOV_EwSw()
{
  Bit16u seg_reg;

  decode_exgx();

  seg_reg = BX_CPU_THIS_PTR read_seg_reg(BX_CPU_THIS_PTR reg_index);

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    // ??? BX_WRITE_16BIT_REG(mem_addr, seg_reg);
    if ( BX_CPU_THIS_PTR is_32bit_opsize ) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, seg_reg);
      }
    else {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, seg_reg);
      }
    }
  else {
    write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &seg_reg);
    }
}

  INLINE void
bx_cpu_c::MOV_SwEw()
{
  Bit16u op2_16;

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
    }

  BX_CPU_THIS_PTR write_seg_reg(BX_CPU_THIS_PTR reg_index, op2_16);
}

  INLINE void
bx_cpu_c::LEA_GvM()
{
  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    bx_panic("LEA_GvM: op2 is a register");
    exception(6, 0, 0);
    return;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* write effective address of op2 in op1 */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, BX_CPU_THIS_PTR rm_addr);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit operand size mode */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit16u) BX_CPU_THIS_PTR rm_addr);
    }
}


  INLINE void
bx_cpu_c::MOV_ALOb()
{
  Bit32u addr_32;
  Bit8u  temp_8;



#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    addr_32 = fetch_next_dword();
    }
  else
#endif /* BX_CPU >= 3 */
    {
    addr_32 = fetch_next_word();
    }

  /* read from memory address */
  if (BX_CPU_THIS_PTR segment_override) {
    read_virtual_byte(BX_CPU_THIS_PTR segment_override, addr_32, &temp_8);
    }
  else {
    read_virtual_byte(&BX_CPU_THIS_PTR ds, addr_32, &temp_8);
    }
  

  /* write to register */
  AL = temp_8;
}

  INLINE void
bx_cpu_c::MOV_eAXOv()
{
  Bit32u addr_32;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    addr_32 = fetch_next_dword();
    }
  else
#endif /* BX_CPU >= 3 */
    {
    addr_32 = fetch_next_word();
    }

  /* read from memory address */
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u temp_32;

    if (BX_CPU_THIS_PTR segment_override) {
      read_virtual_dword(BX_CPU_THIS_PTR segment_override, addr_32, &temp_32);
      }
    else {
      read_virtual_dword(&BX_CPU_THIS_PTR ds, addr_32, &temp_32);
      }

    /* write to register */
    EAX = temp_32;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u temp_16;

    if (BX_CPU_THIS_PTR segment_override) {
      read_virtual_word(BX_CPU_THIS_PTR segment_override, addr_32, &temp_16);
      }
    else {
      read_virtual_word(&BX_CPU_THIS_PTR ds, addr_32, &temp_16);
      }

    /* write to register */
    AX = temp_16;
    }
}

  INLINE void
bx_cpu_c::MOV_ObAL()
{
  Bit32u addr_32;
  Bit8u  temp_8;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    addr_32 = fetch_next_dword();
    }
  else
#endif /* BX_CPU >= 3 */
    {
    addr_32 = fetch_next_word();
    }

  /* read from register */
  temp_8 = AL;

  /* write to memory address */
  if (BX_CPU_THIS_PTR segment_override) {
    write_virtual_byte(BX_CPU_THIS_PTR segment_override, addr_32, &temp_8);
    }
  else {
    write_virtual_byte(&BX_CPU_THIS_PTR ds, addr_32, &temp_8);
    }
}

  INLINE void
bx_cpu_c::MOV_OveAX()
{
  Bit32u addr_32;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    addr_32 = fetch_next_dword();
    }
  else
#endif /* BX_CPU >= 3 */
    {
    addr_32 = fetch_next_word();
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u temp_32;

    /* read from register */
    temp_32 = EAX;

    /* write to memory address */
    if (BX_CPU_THIS_PTR segment_override) {
      write_virtual_dword(BX_CPU_THIS_PTR segment_override, addr_32, &temp_32);
      }
    else {
      write_virtual_dword(&BX_CPU_THIS_PTR ds, addr_32, &temp_32);
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u temp_16;

    /* read from register */
    temp_16 = AX;

    /* write to memory address */
    if (BX_CPU_THIS_PTR segment_override) {
      write_virtual_word(BX_CPU_THIS_PTR segment_override, addr_32, &temp_16);
      }
    else {
      write_virtual_word(&BX_CPU_THIS_PTR ds, addr_32, &temp_16);
      }
    }
}


  INLINE void
bx_cpu_c::MOV_EbIb()
{
  Bit8u op2;

  /* op1 is mod+r/m, op2 is an 8bit immediate */
  decode_exgx();

  op2 = fetch_next_byte();

  /* now write op2 back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, op2);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2);
    }
}

  INLINE void
bx_cpu_c::MOV_EvIv()
{
  decode_exgx();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op2_32;

    op2_32 = fetch_next_dword();

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, op2_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit operand size mode */
    Bit16u op2_16;

    op2_16 = fetch_next_word();

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, op2_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }
    }
}


  INLINE void
bx_cpu_c::MOVZX_GvEb()
{
#if BX_CPU < 3
  bx_panic("MOVZX_GvEb: not supported on < 386\n");
#else
  Bit8u  op2_8;

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_8);
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* zero extend byte op2 into dword op1 */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit32u) op2_8);
    }
  else {
    /* zero extend byte op2 into word op1 */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit16u) op2_8);
    }
#endif /* BX_CPU < 3 */
}

  INLINE void
bx_cpu_c::MOVZX_GvEw()
{
#if BX_CPU < 3
  bx_panic("MOVZX_GvEw: not supported on < 386\n");
#else
  Bit16u op2_16;

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* zero extend word op2 into dword op1 */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit32u) op2_16);
    }
  else {
    /* normal move */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, op2_16);
    }
#endif /* BX_CPU < 3 */
}

  INLINE void
bx_cpu_c::MOVSX_GvEb()
{
#if BX_CPU < 3
  bx_panic("MOVSX_GvEb: not supported on < 386\n");
#else
  Bit8u op2_8;

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_8);
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* sign extend byte op2 into dword op1 */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit8s) op2_8);
    }
  else {
    /* sign extend byte op2 into word op1 */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit8s) op2_8);
    }
#endif /* BX_CPU < 3 */
}

  INLINE void
bx_cpu_c::MOVSX_GvEw()
{
#if BX_CPU < 3
  bx_panic("MOVSX_GvEw: not supported on < 386\n");
#else
  Bit16u op2_16;

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* sign extend word op2 into dword op1 */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit16s) op2_16);
    }
  else {
    /* normal move */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, op2_16);
    }
#endif /* BX_CPU < 3 */
}

  INLINE void
bx_cpu_c::XLAT()
{
  Bit32u offset_32;
  Bit8u  al;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    offset_32 = EBX + AL;
    }
  else
#endif /* BX_CPU >= 3 */
    {
    offset_32 = BX + AL;
    }

  if (BX_CPU_THIS_PTR segment_override) {
    read_virtual_byte(BX_CPU_THIS_PTR segment_override, offset_32, &al);
    }
  else {
    read_virtual_byte(&BX_CPU_THIS_PTR ds, offset_32, &al);
    }
  AL = al;
}

  INLINE void
bx_cpu_c::XCHG_EbGb()
{
  Bit8u op2, op1;

  decode_exgx();

  /* op2 is a register, op2_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, op2);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2);
    }

  BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR reg_index, op1);
}

  INLINE void
bx_cpu_c::XCHG_EvGv()
{
  decode_exgx();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op2_32, op1_32;

    /* op2_32 is a register, op2_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, op2_32);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, op1_32);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16;

    /* op2_16 is a register, op2_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, op2_16);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, op1_16);
    }
}


#if (BX_CPU >= 6) || (BX_CPU_HACKED >= 6)
  INLINE void
bx_cpu_c::CMOV(Boolean condition)
{
  // Note: CMOV accesses a memory source operand (read), regardless
  //       of whether condition is true or not.  Thus, exceptions may
  //       occur even if the MOV does not take place.
  decode_exgx();

  if (BX_CPU_THIS_PTR is_32bit_opsize) { // 32 bit operand size mode
    Bit32u op2_32;

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    if (condition) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, op2_32);
      }
    }
  else { // 16 bit operand size mode
    Bit16u op2_16;

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    if (condition) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, op2_16);
      }
    }
}
#endif
