// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com




extern "C" {
#include <stdio.h>
#include <string.h>
}

#include "bochs.h"

#if BX_DEBUGGER
static Bit32u bx_dbg_get_descriptor_l(bx_descriptor_t *);
static Bit32u bx_dbg_get_descriptor_h(bx_descriptor_t *);
#endif


#if 0
#define SNAPSHOT_SIZE 50
static bx_cpu_c cpu_snapshot[SNAPSHOT_SIZE];
static unsigned snapshot_index = 0;




  void
bx_save_cpu_snapshot(bx_cpu_c *cpu)
{
  cpu_snapshot[snapshot_index] = *cpu;
  snapshot_index = (snapshot_index + 1) % SNAPSHOT_SIZE;
}

  void
bx_show_last_instructions(void)
{
  unsigned index = snapshot_index;

  bx_printf("=======================================\n");
  bx_printf("LAST %u instructions:\n", SNAPSHOT_SIZE);
  for (unsigned i = 0; i<SNAPSHOT_SIZE; i++) {
    bx_cpu = cpu_snapshot[index];
    bx_cpu.debug(bx_cpu.prev_eip);
    index = (index + 1) % SNAPSHOT_SIZE;
    }
  bx_printf("=======================================\n");
}
#else
  void
bx_save_cpu_snapshot(bx_cpu_c *cpu)
{
  UNUSED(cpu);
}

  void
bx_show_last_instructions(void)
{
}
#endif

  void
bx_cpu_c::debug(Bit32u offset)
{
#if BX_CPU <= 2
  bx_printf("\n| AX=%04x  BX=%04x  CX=%04x  DX=%04x  SP=%04x  BP=%04x  SI=%04x  DI=%04x\n",
    (unsigned) AX, (int) BX, (int) CX, (int) DX, (int) SP, (int) BP, (int) SI,
    (unsigned) DI);
  bx_printf("| DS=%04x  ES=%04x  SS=%04x  CS=%04x  IP=%04x   ",
    (unsigned) bx_cpu.ds.selector.value,
    (unsigned) bx_cpu.es.selector.value,
    (unsigned) bx_cpu.ss.selector.value,
    (unsigned) bx_cpu.cs.selector.value,
    (unsigned) bx_cpu.eip);
  bx_printf("%s %s %s %s %s %s %s %s\n",
    bx_cpu.get_OF() ? "OV" : "NV",
    bx_get_DF() ? "DW" : "UP",
    bx_cpu.eflags.if_ ? "EI" : "DI",
    bx_cpu.get_SF() ? "NG" : "PL",
    bx_cpu.get_ZF() ? "ZR" : "NZ",
    bx_cpu.get_AF() ? "AC" : "NA",
    bx_cpu.get_PF() ? "PE" : "PO",
    bx_cpu.get_CF() ? "CY" : "NC");


  bx_printf(">> [%x:%04x] ", (unsigned) seg->selector.value,
    (unsigned) offset);
#else
  bx_printf("\n| EAX=%08x  EBX=%08x  ECX=%08x  EDX=%08x\n",
          (unsigned) EAX, (unsigned) EBX, (unsigned) ECX, (unsigned) EDX);
  bx_printf("| ESP=%08x  EBP=%08x  ESI=%08x  EDI=%08x\n",
          (unsigned) ESP, (unsigned) EBP, (unsigned) ESI, (unsigned) EDI);
  bx_printf("| IOPL=%1u %s %s %s %s %s %s %s %s\n",
    bx_cpu.eflags.iopl,
    bx_cpu.get_OF()       ? "OV" : "NV",
    bx_cpu.eflags.df  ? "DW" : "UP",
    bx_cpu.eflags.if_ ? "EI" : "DI",
    bx_cpu.get_SF()       ? "NG" : "PL",
    bx_cpu.get_ZF()       ? "ZR" : "NZ",
    bx_cpu.get_AF()       ? "AC" : "NA",
    bx_cpu.get_PF()       ? "PE" : "PO",
    bx_cpu.get_CF()       ? "CY" : "NC");
  bx_printf("| SEG selector     base    limit G D\n");
  bx_printf("| SEG sltr(index|ti|rpl)     base    limit G D\n");
  bx_printf("|  DS:%04x( %04x| %01u|  %1u) %08x %08x %1u %1u\n",
    (unsigned) bx_cpu.ds.selector.value,
    (unsigned) bx_cpu.ds.selector.index,
    (unsigned) bx_cpu.ds.selector.ti,
    (unsigned) bx_cpu.ds.selector.rpl,
    (unsigned) bx_cpu.ds.cache.u.segment.base,
    (unsigned) bx_cpu.ds.cache.u.segment.limit,
    (unsigned) bx_cpu.ds.cache.u.segment.g,
    (unsigned) bx_cpu.ds.cache.u.segment.d_b);
  bx_printf("|  ES:%04x( %04x| %01u|  %1u) %08x %08x %1u %1u\n",
    (unsigned) bx_cpu.es.selector.value,
    (unsigned) bx_cpu.es.selector.index,
    (unsigned) bx_cpu.es.selector.ti,
    (unsigned) bx_cpu.es.selector.rpl,
    (unsigned) bx_cpu.es.cache.u.segment.base,
    (unsigned) bx_cpu.es.cache.u.segment.limit,
    (unsigned) bx_cpu.es.cache.u.segment.g,
    (unsigned) bx_cpu.es.cache.u.segment.d_b);
  bx_printf("|  FS:%04x( %04x| %01u|  %1u) %08x %08x %1u %1u\n",
    (unsigned) bx_cpu.fs.selector.value,
    (unsigned) bx_cpu.fs.selector.index,
    (unsigned) bx_cpu.fs.selector.ti,
    (unsigned) bx_cpu.fs.selector.rpl,
    (unsigned) bx_cpu.fs.cache.u.segment.base,
    (unsigned) bx_cpu.fs.cache.u.segment.limit,
    (unsigned) bx_cpu.fs.cache.u.segment.g,
    (unsigned) bx_cpu.fs.cache.u.segment.d_b);
  bx_printf("|  GS:%04x( %04x| %01u|  %1u) %08x %08x %1u %1u\n",
    (unsigned) bx_cpu.gs.selector.value,
    (unsigned) bx_cpu.gs.selector.index,
    (unsigned) bx_cpu.gs.selector.ti,
    (unsigned) bx_cpu.gs.selector.rpl,
    (unsigned) bx_cpu.gs.cache.u.segment.base,
    (unsigned) bx_cpu.gs.cache.u.segment.limit,
    (unsigned) bx_cpu.gs.cache.u.segment.g,
    (unsigned) bx_cpu.gs.cache.u.segment.d_b);
  bx_printf("|  SS:%04x( %04x| %01u|  %1u) %08x %08x %1u %1u\n",
    (unsigned) bx_cpu.ss.selector.value,
    (unsigned) bx_cpu.ss.selector.index,
    (unsigned) bx_cpu.ss.selector.ti,
    (unsigned) bx_cpu.ss.selector.rpl,
    (unsigned) bx_cpu.ss.cache.u.segment.base,
    (unsigned) bx_cpu.ss.cache.u.segment.limit,
    (unsigned) bx_cpu.ss.cache.u.segment.g,
    (unsigned) bx_cpu.ss.cache.u.segment.d_b);
  bx_printf("|  CS:%04x( %04x| %01u|  %1u) %08x %08x %1u %1u\n",
    (unsigned) bx_cpu.cs.selector.value,
    (unsigned) bx_cpu.cs.selector.index,
    (unsigned) bx_cpu.cs.selector.ti,
    (unsigned) bx_cpu.cs.selector.rpl,
    (unsigned) bx_cpu.cs.cache.u.segment.base,
    (unsigned) bx_cpu.cs.cache.u.segment.limit,
    (unsigned) bx_cpu.cs.cache.u.segment.g,
    (unsigned) bx_cpu.cs.cache.u.segment.d_b);
  bx_printf("| EIP=%08x (%08x)\n", (unsigned) bx_cpu.eip,
    (unsigned) bx_cpu.prev_eip);
  bx_printf(">> ");
#endif


#if BX_DISASM
  Boolean valid;
  Bit32u  phy_addr;
  Bit8u   instr_buf[32];
  char    char_buf[256];
  unsigned isize;

  dbg_xlate_linear2phy(BX_CPU_THIS_PTR cs.cache.u.segment.base + offset,
                       &phy_addr, &valid);
  if (valid) {
    bx_dbg_fetch_mem(phy_addr, 16, instr_buf);
    isize = bx_disassemble.disasm(BX_CPU_THIS_PTR cs.cache.u.segment.d_b,
                        instr_buf, char_buf);
    for (unsigned j=0; j<isize; j++)
      bx_printf("%02x", (unsigned) instr_buf[j]);
    bx_printf(": %s\n", char_buf);
    }
  else {
    bx_printf("(instruction unavailable) page not present\n");
    }
#else
  UNUSED(offset);
#endif  // #if BX_DISASM
}


#if BX_DEBUGGER
  Bit32u
bx_dbg_get_register(unsigned which_reg)
{
  Bit32u return_val32;

  switch (which_reg) {
    case BX_DBG_REG_EAX: return(EAX);
    case BX_DBG_REG_ECX: return(ECX);
    case BX_DBG_REG_EDX: return(EDX);
    case BX_DBG_REG_EBX: return(EBX);
    case BX_DBG_REG_ESP: return(ESP);
    case BX_DBG_REG_EBP: return(EBP);
    case BX_DBG_REG_ESI: return(ESI);
    case BX_DBG_REG_EDI: return(EDI);
    case BX_DBG_REG_EIP: return(EIP);
    case BX_DBG_REG_EFLAGS:
      return_val32 = bx_dbg_get_eflags();
      return(return_val32);
    case BX_DBG_REG_CS: return(bx_cpu.cs.selector.value);
    case BX_DBG_REG_SS: return(bx_cpu.ss.selector.value);
    case BX_DBG_REG_DS: return(bx_cpu.ds.selector.value);
    case BX_DBG_REG_ES: return(bx_cpu.es.selector.value);
    case BX_DBG_REG_FS: return(bx_cpu.fs.selector.value);
    case BX_DBG_REG_GS: return(bx_cpu.gs.selector.value);
    default:
      bx_panic("get_register: request for unknown register\n");
      return(0);
    }
}

  Boolean
bx_dbg_set_reg(unsigned reg, Bit32u val)
{
  // returns 1=OK, 0=can't change
  bx_segment_reg_t *seg;
  Bit32u current_sys_bits;

  switch (reg) {
    case BX_DBG_REG_EAX: EAX = val; return(1);
    case BX_DBG_REG_ECX: ECX = val; return(1);
    case BX_DBG_REG_EDX: EDX = val; return(1);
    case BX_DBG_REG_EBX: EBX = val; return(1);
    case BX_DBG_REG_ESP: ESP = val; return(1);
    case BX_DBG_REG_EBP: EBP = val; return(1);
    case BX_DBG_REG_ESI: ESI = val; return(1);
    case BX_DBG_REG_EDI: EDI = val; return(1);
    case BX_DBG_REG_EIP: EIP = val; return(1);
    case BX_DBG_REG_EFLAGS:
      bx_printf("dbg_set_reg: can not handle eflags yet.\n");
      if ( val & 0xffff0000 ) {
        bx_printf("dbg_set_reg: can not set upper 16 bits of eflags.\n");
        return(0);
        }
      // make sure none of the system bits are being changed
      current_sys_bits = (bx_cpu.eflags.nt << 14) |
                         (bx_cpu.eflags.iopl << 12) |
                         (bx_cpu.eflags.tf << 8);
      if ( current_sys_bits != (val & 0x0000f100) ) {
        bx_printf("dbg_set_reg: can not modify NT, IOPL, or TF.\n");
        return(0);
        }
      bx_cpu.set_CF(val & 0x01); val >>= 2;
      bx_cpu.set_PF(val & 0x01); val >>= 2;
      bx_cpu.set_AF(val & 0x01); val >>= 2;
      bx_cpu.set_ZF(val & 0x01); val >>= 1;
      bx_cpu.set_SF(val & 0x01); val >>= 2;
      bx_cpu.eflags.if_ = val & 0x01; val >>= 1;
      bx_cpu.eflags.df  = val & 0x01; val >>= 1;
      bx_cpu.set_OF(val & 0x01);
      if (bx_cpu.eflags.if_)
        bx_cpu.async_event = 1;
      return(1);
    case BX_DBG_REG_CS:
      seg = &bx_cpu.cs;
      break;
    case BX_DBG_REG_SS:
      seg = &bx_cpu.ss;
      break;
    case BX_DBG_REG_DS:
      seg = &bx_cpu.ds;
      break;
    case BX_DBG_REG_ES:
      seg = &bx_cpu.es;
      break;
    case BX_DBG_REG_FS:
      seg = &bx_cpu.fs;
      break;
    case BX_DBG_REG_GS:
      seg = &bx_cpu.gs;
      break;
    default:
      bx_panic("dbg_set_reg: unrecognized register ID (%u)\n", reg);
      return(0);
    }

  if (bx_cpu.real_mode()) {
    seg->selector.value = val;
    seg->cache.valid = 1;
    seg->cache.p = 1;
    seg->cache.dpl = 0;
    seg->cache.segment = 1; // regular segment
    if (reg == BX_DBG_REG_CS) {
      seg->cache.u.segment.executable = 1; // code segment
      }
    else {
      seg->cache.u.segment.executable = 0; // data segment
      }
    seg->cache.u.segment.c_ed = 0;       // expand up/non-conforming
    seg->cache.u.segment.r_w = 1;        // writeable
    seg->cache.u.segment.a = 1;          // accessed
    seg->cache.u.segment.base = val << 4;
    seg->cache.u.segment.limit        = 0xffff;
    seg->cache.u.segment.limit_scaled = 0xffff;
    seg->cache.u.segment.g     = 0;      // byte granular
    seg->cache.u.segment.d_b   = 0;      // default 16bit size
    seg->cache.u.segment.avl   = 0;
    return(1); // ok
    }

  return(0); // can't change when not in real mode
}

  void
bx_dbg_query_pending(void)
{
  Boolean found = 0;

  if ( bx_pc_system.HRQ ) {  // DMA Hold Request
    fprintf(stderr, "pending DMA\n");
    found = 1;
    }

  if ( BX_CPU_THIS_PTR INTR && BX_CPU_THIS_PTR eflags.if_ ) {
    fprintf(stderr, "pending IRQ\n");
    found = 1;
    }

  if (!found)
    fprintf(stderr, "pending none\n");

  fprintf(stderr, "done\n");
}



  Bit32u
bx_dbg_get_eflags(void)
{
  Bit32u val32;

  val32 =
    (bx_cpu.get_CF()) |
    (bx_cpu.eflags.bit1 << 1) |
    ((bx_cpu.get_PF()) << 2) |
    (bx_cpu.eflags.bit3 << 3) |
    ((bx_cpu.get_AF()>0) << 4) |
    (bx_cpu.eflags.bit5 << 5) |
    ((bx_cpu.get_ZF()>0) << 6) |
    ((bx_cpu.get_SF()>0) << 7) |
    (bx_cpu.eflags.tf << 8) |
    (bx_cpu.eflags.if_ << 9) |
    (bx_cpu.eflags.df << 10) |
    ((bx_cpu.get_OF()>0) << 11) |
    (bx_cpu.eflags.iopl << 12) |
    (bx_cpu.eflags.nt << 14) |
    (bx_cpu.eflags.bit15 << 15) |
    (bx_cpu.eflags.rf << 16) |
    (bx_cpu.eflags.vm << 17);
#if BX_CPU >= 5
#  error "dbg_get_eflags: unfinished for >= 586"
#endif
  return(val32);
}


  Bit32u
bx_dbg_get_descriptor_l(bx_descriptor_t *d)
{
  Bit32u val;

  if (d->valid == 0) {
    return(0);
    }

  if (d->segment) {
    val = ((d->u.segment.base & 0xffff) << 16) |
          (d->u.segment.limit & 0xffff);
    return(val);
    }
  else {
    switch (d->type) {
      case 0: // Reserved (not yet defined)
        fprintf(stderr, "#get_descriptor_l(): type %d not finished\n",
          d->type);
        return(0);

      case 1: // available 16bit TSS
        val = ((d->u.tss286.base & 0xffff) << 16) |
               (d->u.tss286.limit & 0xffff);
        return(val);

      case 2: // LDT
        val = ((d->u.ldt.base & 0xffff) << 16) |
              d->u.ldt.limit;
        return(val);

      case 9: // available 32bit TSS
        val = ((d->u.tss386.base & 0xffff) << 16) |
               (d->u.tss386.limit & 0xffff);
        return(val);

      default:
        fprintf(stderr, "#get_descriptor_l(): type %d not finished\n",
          d->type);
        return(0);
      }
    }
}

  Bit32u
bx_dbg_get_descriptor_h(bx_descriptor_t *d)
{
  Bit32u val;

  if (d->valid == 0) {
    return(0);
    }

  if (d->segment) {
    val = (d->u.segment.base & 0xff000000) |
          ((d->u.segment.base >> 16) & 0x000000ff) |
          (d->u.segment.executable << 11) |
          (d->u.segment.c_ed << 10) |
          (d->u.segment.r_w << 9) |
          (d->u.segment.a << 8) |
          (d->segment << 12) |
          (d->dpl << 13) |
          (d->p << 15) |
          (d->u.segment.limit & 0xf0000) |
          (d->u.segment.avl << 20) |
          (d->u.segment.d_b << 22) |
          (d->u.segment.g << 23);
    return(val);
    }
  else {
    switch (d->type) {
      case 0: // Reserved (not yet defined)
        fprintf(stderr, "#get_descriptor_h(): type %d not finished\n",
          d->type);
        return(0);

      case 1: // available 16bit TSS
        val = ((d->u.tss286.base >> 16) & 0xff) |
              (d->type << 8) |
              (d->dpl << 13) |
              (d->p << 15);
        return(val);

      case 2: // LDT
        val = ((d->u.ldt.base >> 16) & 0xff) |
              (d->type << 8) |
              (d->dpl << 13) |
              (d->p << 15) |
              (d->u.ldt.base & 0xff000000);
        return(val);

      case 9: // available 32bit TSS
        val = ((d->u.tss386.base >> 16) & 0xff) |
              (d->type << 8) |
              (d->dpl << 13) |
              (d->p << 15) |
              (d->u.tss386.limit & 0xf0000) |
              (d->u.tss386.avl << 20) |
              (d->u.tss386.g << 23) |
              (d->u.tss386.base & 0xff000000);
        return(val);

      default:
        fprintf(stderr, "#get_descriptor_h(): type %d not finished\n",
          d->type);
        return(0);
      }
    }
}

  void
bx_dbg_dump_cpu(void)
{
  Bit32u val32;

  fprintf(stderr, "eax:0x%x\n", (unsigned) EAX);
  fprintf(stderr, "ebx:0x%x\n", (unsigned) EBX);
  fprintf(stderr, "ecx:0x%x\n", (unsigned) ECX);
  fprintf(stderr, "edx:0x%x\n", (unsigned) EDX);

  fprintf(stderr, "ebp:0x%x\n", (unsigned) EBP);
  fprintf(stderr, "esi:0x%x\n", (unsigned) ESI);
  fprintf(stderr, "edi:0x%x\n", (unsigned) EDI);
  fprintf(stderr, "esp:0x%x\n", (unsigned) ESP);

  fprintf(stderr, "eflags:0x%x\n", (unsigned) bx_dbg_get_eflags());
  fprintf(stderr, "eip:0x%x\n", (unsigned) bx_cpu.eip);

  fprintf(stderr, "cs:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.cs.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.cs.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.cs.cache),
    bx_cpu.cs.cache.valid);

  fprintf(stderr, "ss:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.ss.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.ss.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.ss.cache),
    bx_cpu.ss.cache.valid);

  fprintf(stderr, "ds:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.ds.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.ds.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.ds.cache),
    bx_cpu.ds.cache.valid);

  fprintf(stderr, "es:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.es.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.es.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.es.cache),
    bx_cpu.es.cache.valid);

  fprintf(stderr, "fs:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.fs.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.fs.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.fs.cache),
    bx_cpu.fs.cache.valid);

  fprintf(stderr, "gs:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.gs.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.gs.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.gs.cache),
    bx_cpu.gs.cache.valid);

  fprintf(stderr, "ldtr:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.ldtr.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.ldtr.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.ldtr.cache),
    bx_cpu.ldtr.cache.valid);

  fprintf(stderr, "tr:s=0x%x, dl=0x%x, dh=0x%x, valid=%u\n",
    (unsigned) bx_cpu.tr.selector.value,
    bx_dbg_get_descriptor_l(&bx_cpu.tr.cache),
    bx_dbg_get_descriptor_h(&bx_cpu.tr.cache),
    bx_cpu.tr.cache.valid);

  fprintf(stderr, "gdtr:base=0x%x, limit=0x%x\n",
    (unsigned) bx_cpu.gdtr.base,
    (unsigned) bx_cpu.gdtr.limit);

  fprintf(stderr, "idtr:base=0x%x, limit=0x%x\n",
    (unsigned) bx_cpu.idtr.base,
    (unsigned) bx_cpu.idtr.limit);

  fprintf(stderr, "dr0:0x%x\n", bx_cpu.dr0);
  fprintf(stderr, "dr1:0x%x\n", bx_cpu.dr1);
  fprintf(stderr, "dr2:0x%x\n", bx_cpu.dr2);
  fprintf(stderr, "dr3:0x%x\n", bx_cpu.dr3);
  fprintf(stderr, "dr4:0x%x\n", bx_cpu.dr4);
  fprintf(stderr, "dr5:0x%x\n", bx_cpu.dr5);
  fprintf(stderr, "dr6:0x%x\n", bx_cpu.dr6);
  fprintf(stderr, "dr7:0x%x\n", bx_cpu.dr7);

  fprintf(stderr, "tr3:0x%x\n", 0);
  fprintf(stderr, "tr4:0x%x\n", 0);
  fprintf(stderr, "tr5:0x%x\n", 0);
  fprintf(stderr, "tr6:0x%x\n", 0);
  fprintf(stderr, "tr7:0x%x\n", 0);

  // cr0:32=pg,cd,nw,am,wp,ne,ts,em,mp,pe
  val32 = (bx_cpu.cr0.pg << 31) |

#if BX_CPU >= 4
          (bx_cpu.cr0.cd << 30) |
          (bx_cpu.cr0.nw << 29) |
          (bx_cpu.cr0.am << 18) |
          (bx_cpu.cr0.wp << 16) |
          (bx_cpu.cr0.ne << 5) |
#endif

#if BX_CPU == 3
          (0 << 4) |
#elif BX_CPU == 4
          (1 << 4) |
#elif BX_CPU >= 5
          (bx_cpu.cr0.et << 4) |
#endif

          (bx_cpu.cr0.ts << 3) |
          (bx_cpu.cr0.em << 2) |
          (bx_cpu.cr0.mp << 1) |
          (bx_cpu.cr0.pe << 0);
  fprintf(stderr, "cr0:0x%x\n", (unsigned) val32);
  fprintf(stderr, "cr1:0x%x\n", (unsigned) 0);
  fprintf(stderr, "cr2:0x%x\n", (unsigned) bx_cpu.cr2);
  fprintf(stderr, "cr3:0x%x\n", (unsigned) bx_cpu.cr3);
  fprintf(stderr, "cr4:0x%x\n", (unsigned) 0);

  fprintf(stderr, "inhibit_int:%u\n", (unsigned) bx_cpu.inhibit_interrupts);

  fprintf(stderr, "done\n");
}

  Boolean
bx_dbg_set_cpu(bx_dbg_cpu_t *cpu)
{
  // returns 1=OK, 0=Error
  Bit32u val;
  Bit32u type;

  // =================================================
  // Do checks first, before setting any CPU registers
  // =================================================

  // CS, SS, DS, ES, FS, GS descriptor checks
  if (!cpu->cs.valid) {
    fprintf(stderr, "Error: CS not valid\n");
    return(0); // error
    }
  if ( (cpu->cs.des_h & 0x1000) == 0 ) {
    fprintf(stderr, "Error: CS not application type\n");
    return(0); // error
    }
  if ( (cpu->cs.des_h & 0x0800) == 0 ) {
    fprintf(stderr, "Error: CS not executable\n");
    return(0); // error
    }

  if (!cpu->ss.valid) {
    fprintf(stderr, "Error: SS not valid\n");
    return(0); // error
    }
  if ( (cpu->ss.des_h & 0x1000) == 0 ) {
    fprintf(stderr, "Error: SS not application type\n");
    return(0); // error
    }

  if (cpu->ds.valid) {
    if ( (cpu->ds.des_h & 0x1000) == 0 ) {
      fprintf(stderr, "Error: DS not application type\n");
      return(0); // error
      }
    }

  if (cpu->es.valid) {
    if ( (cpu->es.des_h & 0x1000) == 0 ) {
      fprintf(stderr, "Error: ES not application type\n");
      return(0); // error
      }
    }

  if (cpu->fs.valid) {
    if ( (cpu->fs.des_h & 0x1000) == 0 ) {
      fprintf(stderr, "Error: FS not application type\n");
      return(0); // error
      }
    }

  if (cpu->gs.valid) {
    if ( (cpu->gs.des_h & 0x1000) == 0 ) {
      fprintf(stderr, "Error: GS not application type\n");
      return(0); // error
      }
    }

  if (cpu->ldtr.valid) {
    if ( cpu->ldtr.des_h & 0x1000 ) {
      fprintf(stderr, "Error: LDTR not system type\n");
      return(0); // error
      }
    if ( ((cpu->ldtr.des_h >> 8) & 0x0f) != 2 ) {
      fprintf(stderr, "Error: LDTR descriptor type not LDT\n");
      return(0); // error
      }
    }

  if (cpu->tr.valid) {
    if ( cpu->tr.des_h & 0x1000 ) {
      fprintf(stderr, "Error: TR not system type\n");
      return(0); // error
      }
    type = (cpu->tr.des_h >> 8) & 0x0f;

    if ( (type != 1) && (type != 9) ) {
      fprintf(stderr, "Error: TR descriptor type not TSS\n");
      return(0); // error
      }
    }

  // =============
  // end of checks
  // =============

  EAX = cpu->eax;
  EBX = cpu->ebx;
  ECX = cpu->ecx;
  EDX = cpu->edx;
  EBP = cpu->ebp;
  ESI = cpu->esi;
  EDI = cpu->edi;
  ESP = cpu->esp;

  // eflags
  val = cpu->eflags;
  bx_cpu.set_CF(val & 0x01); val >>= 2;
  bx_cpu.set_PF(val & 0x01); val >>= 2;
  bx_cpu.set_AF(val & 0x01); val >>= 2;
  bx_cpu.set_ZF(val & 0x01); val >>= 1;
  bx_cpu.set_SF(val & 0x01); val >>= 1;
  bx_cpu.eflags.tf  = val & 0x01; val >>= 1;
  bx_cpu.eflags.if_ = val & 0x01; val >>= 1;
  if (bx_cpu.eflags.if_)
    bx_cpu.async_event = 1;
  bx_cpu.eflags.df  = val & 0x01; val >>= 1;
  bx_cpu.set_OF(val & 0x01); val >>= 1;
  bx_cpu.eflags.iopl = val & 0x03; val >>= 2;
  bx_cpu.eflags.nt   = val & 0x01; val >>= 2;
  bx_cpu.eflags.rf   = val & 0x01; val >>= 1;
  bx_cpu.eflags.vm   = val & 0x01; val >>= 1;
#if BX_CPU >= 4
  bx_cpu.eflags.ac   = val & 0x01; val >>= 1;
#endif
#if BX_CPU >= 5
  bx_cpu.eflags.vif  = val & 0x01; val >>= 1;
  bx_cpu.eflags.vip  = val & 0x01; val >>= 1;
  bx_cpu.eflags.id   = val & 0x01;
#endif

  bx_cpu.eip = cpu->eip;



  // CS:
  bx_cpu.cs.selector.value = cpu->cs.sel;
  bx_cpu.cs.selector.index = cpu->cs.sel >> 3;
  bx_cpu.cs.selector.ti    = (cpu->cs.sel >> 2) & 0x01;
  bx_cpu.cs.selector.rpl   = cpu->cs.sel & 0x03;

  bx_cpu.cs.cache.valid            = cpu->cs.valid;
  bx_cpu.cs.cache.p                = (cpu->cs.des_h >> 15) & 0x01;
  bx_cpu.cs.cache.dpl              = (cpu->cs.des_h >> 13) & 0x03;
  bx_cpu.cs.cache.segment          = (cpu->cs.des_h >> 12) & 0x01;
  bx_cpu.cs.cache.type             = (cpu->cs.des_h >> 8) & 0x0f;
  bx_cpu.cs.cache.u.segment.executable = (cpu->cs.des_h >> 11) & 0x01;
  bx_cpu.cs.cache.u.segment.c_ed   = (cpu->cs.des_h >> 10) & 0x01;
  bx_cpu.cs.cache.u.segment.r_w    = (cpu->cs.des_h >> 9) & 0x01;
  bx_cpu.cs.cache.u.segment.a      = (cpu->cs.des_h >> 8) & 0x01;
  bx_cpu.cs.cache.u.segment.base   = (cpu->cs.des_l >> 16);
  bx_cpu.cs.cache.u.segment.base  |= (cpu->cs.des_h & 0xff) << 16;
  bx_cpu.cs.cache.u.segment.base  |= (cpu->cs.des_h & 0xff000000);
  bx_cpu.cs.cache.u.segment.limit  = (cpu->cs.des_l & 0xffff);
  bx_cpu.cs.cache.u.segment.limit |= (cpu->cs.des_h & 0x000f0000);
  bx_cpu.cs.cache.u.segment.g      = (cpu->cs.des_h >> 23) & 0x01;
  bx_cpu.cs.cache.u.segment.d_b    = (cpu->cs.des_h >> 22) & 0x01;
  bx_cpu.cs.cache.u.segment.avl    = (cpu->cs.des_h >> 20) & 0x01;
  if (bx_cpu.cs.cache.u.segment.g)
    bx_cpu.cs.cache.u.segment.limit_scaled =
      (bx_cpu.cs.cache.u.segment.limit << 12) | 0x0fff;
  else
    bx_cpu.cs.cache.u.segment.limit_scaled =
      bx_cpu.cs.cache.u.segment.limit;


  // SS:
  bx_cpu.ss.selector.value = cpu->ss.sel;
  bx_cpu.ss.selector.index = cpu->ss.sel >> 3;
  bx_cpu.ss.selector.ti    = (cpu->ss.sel >> 2) & 0x01;
  bx_cpu.ss.selector.rpl   = cpu->ss.sel & 0x03;

  bx_cpu.ss.cache.valid            = cpu->ss.valid;
  bx_cpu.ss.cache.p                = (cpu->ss.des_h >> 15) & 0x01;
  bx_cpu.ss.cache.dpl              = (cpu->ss.des_h >> 13) & 0x03;
  bx_cpu.ss.cache.segment          = (cpu->ss.des_h >> 12) & 0x01;
  bx_cpu.ss.cache.type             = (cpu->ss.des_h >> 8) & 0x0f;
  bx_cpu.ss.cache.u.segment.executable = (cpu->ss.des_h >> 11) & 0x01;
  bx_cpu.ss.cache.u.segment.c_ed   = (cpu->ss.des_h >> 10) & 0x01;
  bx_cpu.ss.cache.u.segment.r_w    = (cpu->ss.des_h >> 9) & 0x01;
  bx_cpu.ss.cache.u.segment.a      = (cpu->ss.des_h >> 8) & 0x01;
  bx_cpu.ss.cache.u.segment.base   = (cpu->ss.des_l >> 16);
  bx_cpu.ss.cache.u.segment.base  |= (cpu->ss.des_h & 0xff) << 16;
  bx_cpu.ss.cache.u.segment.base  |= (cpu->ss.des_h & 0xff000000);
  bx_cpu.ss.cache.u.segment.limit  = (cpu->ss.des_l & 0xffff);
  bx_cpu.ss.cache.u.segment.limit |= (cpu->ss.des_h & 0x000f0000);
  bx_cpu.ss.cache.u.segment.g      = (cpu->ss.des_h >> 23) & 0x01;
  bx_cpu.ss.cache.u.segment.d_b    = (cpu->ss.des_h >> 22) & 0x01;
  bx_cpu.ss.cache.u.segment.avl    = (cpu->ss.des_h >> 20) & 0x01;
  if (bx_cpu.ss.cache.u.segment.g)
    bx_cpu.ss.cache.u.segment.limit_scaled =
      (bx_cpu.ss.cache.u.segment.limit << 12) | 0x0fff;
  else
    bx_cpu.ss.cache.u.segment.limit_scaled =
      bx_cpu.ss.cache.u.segment.limit;


  // DS:
  bx_cpu.ds.selector.value = cpu->ds.sel;
  bx_cpu.ds.selector.index = cpu->ds.sel >> 3;
  bx_cpu.ds.selector.ti    = (cpu->ds.sel >> 2) & 0x01;
  bx_cpu.ds.selector.rpl   = cpu->ds.sel & 0x03;

  bx_cpu.ds.cache.valid            = cpu->ds.valid;
  bx_cpu.ds.cache.p                = (cpu->ds.des_h >> 15) & 0x01;
  bx_cpu.ds.cache.dpl              = (cpu->ds.des_h >> 13) & 0x03;
  bx_cpu.ds.cache.segment          = (cpu->ds.des_h >> 12) & 0x01;
  bx_cpu.ds.cache.type             = (cpu->ds.des_h >> 8) & 0x0f;
  bx_cpu.ds.cache.u.segment.executable = (cpu->ds.des_h >> 11) & 0x01;
  bx_cpu.ds.cache.u.segment.c_ed   = (cpu->ds.des_h >> 10) & 0x01;
  bx_cpu.ds.cache.u.segment.r_w    = (cpu->ds.des_h >> 9) & 0x01;
  bx_cpu.ds.cache.u.segment.a      = (cpu->ds.des_h >> 8) & 0x01;
  bx_cpu.ds.cache.u.segment.base   = (cpu->ds.des_l >> 16);
  bx_cpu.ds.cache.u.segment.base  |= (cpu->ds.des_h & 0xff) << 16;
  bx_cpu.ds.cache.u.segment.base  |= (cpu->ds.des_h & 0xff000000);
  bx_cpu.ds.cache.u.segment.limit  = (cpu->ds.des_l & 0xffff);
  bx_cpu.ds.cache.u.segment.limit |= (cpu->ds.des_h & 0x000f0000);
  bx_cpu.ds.cache.u.segment.g      = (cpu->ds.des_h >> 23) & 0x01;
  bx_cpu.ds.cache.u.segment.d_b    = (cpu->ds.des_h >> 22) & 0x01;
  bx_cpu.ds.cache.u.segment.avl    = (cpu->ds.des_h >> 20) & 0x01;
  if (bx_cpu.ds.cache.u.segment.g)
    bx_cpu.ds.cache.u.segment.limit_scaled =
      (bx_cpu.ds.cache.u.segment.limit << 12) | 0x0fff;
  else
    bx_cpu.ds.cache.u.segment.limit_scaled =
      bx_cpu.ds.cache.u.segment.limit;



  // ES:
  bx_cpu.es.selector.value = cpu->es.sel;
  bx_cpu.es.selector.index = cpu->es.sel >> 3;
  bx_cpu.es.selector.ti    = (cpu->es.sel >> 2) & 0x01;
  bx_cpu.es.selector.rpl   = cpu->es.sel & 0x03;

  bx_cpu.es.cache.valid            = cpu->es.valid;
  bx_cpu.es.cache.p                = (cpu->es.des_h >> 15) & 0x01;
  bx_cpu.es.cache.dpl              = (cpu->es.des_h >> 13) & 0x03;
  bx_cpu.es.cache.segment          = (cpu->es.des_h >> 12) & 0x01;
  bx_cpu.es.cache.type             = (cpu->es.des_h >> 8) & 0x0f;
  bx_cpu.es.cache.u.segment.executable = (cpu->es.des_h >> 11) & 0x01;
  bx_cpu.es.cache.u.segment.c_ed   = (cpu->es.des_h >> 10) & 0x01;
  bx_cpu.es.cache.u.segment.r_w    = (cpu->es.des_h >> 9) & 0x01;
  bx_cpu.es.cache.u.segment.a      = (cpu->es.des_h >> 8) & 0x01;
  bx_cpu.es.cache.u.segment.base   = (cpu->es.des_l >> 16);
  bx_cpu.es.cache.u.segment.base  |= (cpu->es.des_h & 0xff) << 16;
  bx_cpu.es.cache.u.segment.base  |= (cpu->es.des_h & 0xff000000);
  bx_cpu.es.cache.u.segment.limit  = (cpu->es.des_l & 0xffff);
  bx_cpu.es.cache.u.segment.limit |= (cpu->es.des_h & 0x000f0000);
  bx_cpu.es.cache.u.segment.g      = (cpu->es.des_h >> 23) & 0x01;
  bx_cpu.es.cache.u.segment.d_b    = (cpu->es.des_h >> 22) & 0x01;
  bx_cpu.es.cache.u.segment.avl    = (cpu->es.des_h >> 20) & 0x01;
  if (bx_cpu.es.cache.u.segment.g)
    bx_cpu.es.cache.u.segment.limit_scaled =
      (bx_cpu.es.cache.u.segment.limit << 12) | 0x0fff;
  else
    bx_cpu.es.cache.u.segment.limit_scaled =
      bx_cpu.es.cache.u.segment.limit;


  // FS:
  bx_cpu.fs.selector.value = cpu->fs.sel;
  bx_cpu.fs.selector.index = cpu->fs.sel >> 3;
  bx_cpu.fs.selector.ti    = (cpu->fs.sel >> 2) & 0x01;
  bx_cpu.fs.selector.rpl   = cpu->fs.sel & 0x03;

  bx_cpu.fs.cache.valid            = cpu->fs.valid;
  bx_cpu.fs.cache.p                = (cpu->fs.des_h >> 15) & 0x01;
  bx_cpu.fs.cache.dpl              = (cpu->fs.des_h >> 13) & 0x03;
  bx_cpu.fs.cache.segment          = (cpu->fs.des_h >> 12) & 0x01;
  bx_cpu.fs.cache.type             = (cpu->fs.des_h >> 8) & 0x0f;
  bx_cpu.fs.cache.u.segment.executable = (cpu->fs.des_h >> 11) & 0x01;
  bx_cpu.fs.cache.u.segment.c_ed   = (cpu->fs.des_h >> 10) & 0x01;
  bx_cpu.fs.cache.u.segment.r_w    = (cpu->fs.des_h >> 9) & 0x01;
  bx_cpu.fs.cache.u.segment.a      = (cpu->fs.des_h >> 8) & 0x01;
  bx_cpu.fs.cache.u.segment.base   = (cpu->fs.des_l >> 16);
  bx_cpu.fs.cache.u.segment.base  |= (cpu->fs.des_h & 0xff) << 16;
  bx_cpu.fs.cache.u.segment.base  |= (cpu->fs.des_h & 0xff000000);
  bx_cpu.fs.cache.u.segment.limit  = (cpu->fs.des_l & 0xffff);
  bx_cpu.fs.cache.u.segment.limit |= (cpu->fs.des_h & 0x000f0000);
  bx_cpu.fs.cache.u.segment.g      = (cpu->fs.des_h >> 23) & 0x01;
  bx_cpu.fs.cache.u.segment.d_b    = (cpu->fs.des_h >> 22) & 0x01;
  bx_cpu.fs.cache.u.segment.avl    = (cpu->fs.des_h >> 20) & 0x01;
  if (bx_cpu.fs.cache.u.segment.g)
    bx_cpu.fs.cache.u.segment.limit_scaled =
      (bx_cpu.fs.cache.u.segment.limit << 12) | 0x0fff;
  else
    bx_cpu.fs.cache.u.segment.limit_scaled =
      bx_cpu.fs.cache.u.segment.limit;


  // GS:
  bx_cpu.gs.selector.value = cpu->gs.sel;
  bx_cpu.gs.selector.index = cpu->gs.sel >> 3;
  bx_cpu.gs.selector.ti    = (cpu->gs.sel >> 2) & 0x01;
  bx_cpu.gs.selector.rpl   = cpu->gs.sel & 0x03;

  bx_cpu.gs.cache.valid            = cpu->gs.valid;
  bx_cpu.gs.cache.p                = (cpu->gs.des_h >> 15) & 0x01;
  bx_cpu.gs.cache.dpl              = (cpu->gs.des_h >> 13) & 0x03;
  bx_cpu.gs.cache.segment          = (cpu->gs.des_h >> 12) & 0x01;
  bx_cpu.gs.cache.type             = (cpu->gs.des_h >> 8) & 0x0f;
  bx_cpu.gs.cache.u.segment.executable = (cpu->gs.des_h >> 11) & 0x01;
  bx_cpu.gs.cache.u.segment.c_ed   = (cpu->gs.des_h >> 10) & 0x01;
  bx_cpu.gs.cache.u.segment.r_w    = (cpu->gs.des_h >> 9) & 0x01;
  bx_cpu.gs.cache.u.segment.a      = (cpu->gs.des_h >> 8) & 0x01;
  bx_cpu.gs.cache.u.segment.base   = (cpu->gs.des_l >> 16);
  bx_cpu.gs.cache.u.segment.base  |= (cpu->gs.des_h & 0xff) << 16;
  bx_cpu.gs.cache.u.segment.base  |= (cpu->gs.des_h & 0xff000000);
  bx_cpu.gs.cache.u.segment.limit  = (cpu->gs.des_l & 0xffff);
  bx_cpu.gs.cache.u.segment.limit |= (cpu->gs.des_h & 0x000f0000);
  bx_cpu.gs.cache.u.segment.g      = (cpu->gs.des_h >> 23) & 0x01;
  bx_cpu.gs.cache.u.segment.d_b    = (cpu->gs.des_h >> 22) & 0x01;
  bx_cpu.gs.cache.u.segment.avl    = (cpu->gs.des_h >> 20) & 0x01;
  if (bx_cpu.gs.cache.u.segment.g)
    bx_cpu.gs.cache.u.segment.limit_scaled =
      (bx_cpu.gs.cache.u.segment.limit << 12) | 0x0fff;
  else
    bx_cpu.gs.cache.u.segment.limit_scaled =
      bx_cpu.gs.cache.u.segment.limit;

  // LDTR:
  bx_cpu.ldtr.selector.value = cpu->ldtr.sel;
  bx_cpu.ldtr.selector.index = cpu->ldtr.sel >> 3;
  bx_cpu.ldtr.selector.ti    = (cpu->ldtr.sel >> 2) & 0x01;
  bx_cpu.ldtr.selector.rpl   = cpu->ldtr.sel & 0x03;

  bx_cpu.ldtr.cache.valid            = cpu->ldtr.valid;
  bx_cpu.ldtr.cache.p                = (cpu->ldtr.des_h >> 15) & 0x01;
  bx_cpu.ldtr.cache.dpl              = (cpu->ldtr.des_h >> 13) & 0x03;
  bx_cpu.ldtr.cache.segment          = (cpu->ldtr.des_h >> 12) & 0x01;
  bx_cpu.ldtr.cache.type             = (cpu->ldtr.des_h >> 8) & 0x0f;
  bx_cpu.ldtr.cache.u.ldt.base       = (cpu->ldtr.des_l >> 16);
  bx_cpu.ldtr.cache.u.ldt.base      |= (cpu->ldtr.des_h & 0xff) << 16;
  bx_cpu.ldtr.cache.u.ldt.base      |= (cpu->ldtr.des_h & 0xff000000);
  bx_cpu.ldtr.cache.u.ldt.limit      = (cpu->ldtr.des_l & 0xffff);

  // TR
  type = (cpu->tr.des_h >> 8) & 0x0f;
  bx_cpu.tr.selector.value = cpu->tr.sel;
  bx_cpu.tr.selector.index = cpu->tr.sel >> 3;
  bx_cpu.tr.selector.ti    = (cpu->tr.sel >> 2) & 0x01;
  bx_cpu.tr.selector.rpl   = cpu->tr.sel & 0x03;

  bx_cpu.tr.cache.valid            = cpu->tr.valid;
  bx_cpu.tr.cache.p                = (cpu->tr.des_h >> 15) & 0x01;
  bx_cpu.tr.cache.dpl              = (cpu->tr.des_h >> 13) & 0x03;
  bx_cpu.tr.cache.segment          = (cpu->tr.des_h >> 12) & 0x01;
  bx_cpu.tr.cache.type             = type;
  if (type == 1) { // 286 TSS
    bx_cpu.tr.cache.u.tss286.base   = (cpu->tr.des_l >> 16);
    bx_cpu.tr.cache.u.tss286.base  |= (cpu->tr.des_h & 0xff) << 16;
    bx_cpu.tr.cache.u.tss286.limit  = (cpu->tr.des_l & 0xffff);
    }
  else { // type == 9, 386 TSS
    bx_cpu.tr.cache.u.tss386.base   = (cpu->tr.des_l >> 16);
    bx_cpu.tr.cache.u.tss386.base  |= (cpu->tr.des_h & 0xff) << 16;
    bx_cpu.tr.cache.u.tss386.base  |= (cpu->tr.des_h & 0xff000000);
    bx_cpu.tr.cache.u.tss386.limit  = (cpu->tr.des_l & 0xffff);
    bx_cpu.tr.cache.u.tss386.limit |= (cpu->tr.des_h & 0x000f0000);
    bx_cpu.tr.cache.u.tss386.g      = (cpu->tr.des_h >> 23) & 0x01;
    bx_cpu.tr.cache.u.tss386.avl    = (cpu->tr.des_h >> 20) & 0x01;
    }


  // gdtr
  bx_cpu.gdtr.base  = cpu->gdtr.base;
  bx_cpu.gdtr.limit = cpu->gdtr.limit;

  // idtr
  bx_cpu.idtr.base  = cpu->idtr.base;
  bx_cpu.idtr.limit = cpu->idtr.limit;


  bx_cpu.dr0 = cpu->dr0;
  bx_cpu.dr1 = cpu->dr1;
  bx_cpu.dr2 = cpu->dr2;
  bx_cpu.dr3 = cpu->dr3;
  bx_cpu.dr4 = cpu->dr4;
  bx_cpu.dr5 = cpu->dr5;
  bx_cpu.dr6 = cpu->dr6;
  bx_cpu.dr7 = cpu->dr7;

  // bx_cpu.tr3 = cpu->tr3;
  // bx_cpu.tr4 = cpu->tr4;
  // bx_cpu.tr5 = cpu->tr5;
  // bx_cpu.tr6 = cpu->tr6;
  // bx_cpu.tr7 = cpu->tr7;


  // cr0, cr1, cr2, cr3, cr4
  val = cpu->cr0;
  bx_cpu.cr0.pe = val & 0x01; val >>= 1;
  bx_cpu.cr0.mp = val & 0x01; val >>= 1;
  bx_cpu.cr0.em = val & 0x01; val >>= 1;
  bx_cpu.cr0.ts = val & 0x01; val >>= 1;

#if BX_CPU >= 5
  bx_cpu.cr0.et = val & 0x01;
#endif
  val >>= 1;

#if BX_CPU >= 4
  bx_cpu.cr0.ne = val & 0x01;
#endif
  val >>= 11;

#if BX_CPU >= 4
  bx_cpu.cr0.wp = val & 0x01; val >>= 2;
  bx_cpu.cr0.am = val & 0x01; val >>= 11;
  bx_cpu.cr0.nw = val & 0x01; val >>= 1;
  bx_cpu.cr0.cd = val & 0x01; val >>= 1;
#else
  val >>= 15;
#endif

#if BX_CPU >= 3
  bx_cpu.cr0.pg = val & 0x01;
#endif


  bx_cpu.cr1 = cpu->cr1;
  bx_cpu.cr2 = cpu->cr2;
  bx_cpu.cr3 = cpu->cr3;
#if BX_CPU >= 5
  bx_cpu.cr4 = cpu->cr4;
#endif

  bx_cpu.inhibit_interrupts = cpu->inhibit_int;

  //
  // flush cached items, prefetch, paging, etc
  //
  bx_cpu.CR3_change(cpu->cr3);
  bx_cpu.invalidate_prefetch_q();

  return(1);
}

#endif  // #if BX_DEBUGGER
