// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com





#include "bochs.h"



#if BX_SHOW_IPS
unsigned long ips_count=0;
#endif





/* This array defines a look-up table for the even parity-ness
   of an 8bit quantity, for optimal assignment of the parity bit
   in the EFLAGS register
  */
const Boolean bx_parity_lookup[256] = {
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1
  };






#if USE_INLINE
#include "arith.c"
#include "arith2.c"
#include "bit.c"
#include "ctrl_xfer.c"
#include "data_xfer.c"
#include "flag_ctrl.c"
#include "io.c"
#include "logical.c"
#include "shift.c"
#include "ndp.c"
#include "proc_ctrl.c"
#include "protect_ctrl.c"
#include "segment_ctrl.c"
#include "soft_int.c"
#include "stack.c"
#include "string.c"
#include "bcd.c"
#endif



  void
bx_cpu_c::cpu_loop(void)
{

  Bit8u next_byte;
  Boolean jb_condition, jv_condition;
  Bit8s  offset8s;
  Bit32s offset32s;
  Boolean rep_prefix;


  BX_CPU_THIS_PTR errorno = 0;

  (void) setjmp( BX_CPU_THIS_PTR jmp_buf_env );

  //BX_CPU_LOOP_ENTRY();

  while (1) {

#if BX_CPU >= 2
    BX_CPU_THIS_PTR EXT = 0;
    BX_CPU_THIS_PTR prev_eip = EIP;   // save instruction pointer
    BX_CPU_THIS_PTR prev_esp = ESP;   // save stack pointer
#endif

    BX_INSTR_OPCODE_BEGIN();

    if (BX_CPU_THIS_PTR async_event) {
      /* INPUT PROCESSING ORDER:
       * 1) instruction exception
       * 2) single step
       * 3) NMI
       * 4) processor extension segment overrun
       * 5) INTR
       */
      /* ??? look at priorities for various cpu levels */

      /* single step recognized here */
      if ( BX_CPU_THIS_PTR single_step_event ) {
        bx_panic("decode: single step recognized\n");
        /* should BX_CPU_THIS_PTR EXT be set to 1 here ??? */
        exception(1, 0, 0); /* single step interrupt */
        BX_CPU_THIS_PTR single_step_event = 0;
        /* fall thru to other interrupt processing */
        }

      /* single step generated here */
      if ( BX_CPU_THIS_PTR eflags.tf ) {
        BX_CPU_THIS_PTR single_step_event = 1;
        bx_panic("decode: single step generated\n");
        }

      /* if last instruction caused processing of interrupts to pend,
       * such as STI or load of SS, then skip for one instruction */
      if (BX_CPU_THIS_PTR inhibit_interrupts) {
        BX_CPU_THIS_PTR inhibit_interrupts = 0;
        }

      else if ( BX_CPU_THIS_PTR INTR && BX_CPU_THIS_PTR eflags.if_ && BX_DBG_TAKE_INTR) {
        Bit8u vector;

        // NOTE: similar code in debugstuff.cc
        vector = bx_pc_system.IAC(); /* may set INTR with next interrupt */
        //if (bx_dbg.interrupts) bx_printf("decode: interrupt %u\n",
        //                                   (unsigned) vector);
        BX_CPU_THIS_PTR errorno = 0;
        BX_CPU_THIS_PTR EXT   = 1; /* external event */
        interrupt(vector, 0, 0, 0);
        BX_INSTR_HWINTERRUPT(vector, BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
        }

      // DMA Hold Request
      else if ( bx_pc_system.HRQ && BX_DBG_TAKE_DMA ) {
        // NOTE: similar code in debugstuff.cc
        // assert Hold Acknowledge (HLDA) and go into a bus hold state
        bx_pc_system.raise_HLDA();
        }

      if ( !(BX_CPU_THIS_PTR INTR ||
             BX_CPU_THIS_PTR single_step_event ||
             bx_pc_system.HRQ ||
             BX_CPU_THIS_PTR eflags.tf) )
        BX_CPU_THIS_PTR async_event = 0;

      }

#if BX_CPU >= 3
    BX_CPU_THIS_PTR is_32bit_opsize   = BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
    BX_CPU_THIS_PTR is_32bit_addrsize = BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
#endif

    BX_CPU_THIS_PTR segment_override = NULL;
    rep_prefix = 0;

#if BX_CPU >= 2
    BX_CPU_THIS_PTR errorno = 0;
#endif



#if 0
  {
  void bx_save_cpu_snapshot(bx_cpu_c *cpu);

  bx_save_cpu_snapshot(&bx_cpu);
  }
#endif

//if (bx_dbg.debugger) debug(BX_CPU_THIS_PTR eip);


    BX_CPU_ICOUNT();

#if BX_SHOW_IPS
    ips_count++;
#endif


    /* On the 386 and above, instructions must be a maximum of 15 bytes long.
     * this means redundant prefix codes can put the byte count over 15 and
     * cause an illegal instruction.
     */
    // ?? should check for >15 byte instructions
    while (1) {
      next_byte = fetch_next_byte(); 

      BX_INSTR_OPCODE_BYTE1(next_byte);

      switch (next_byte) {

        case 0x00: ADD_EbGb(); break;
        case 0x01: ADD_EvGv(); break;
        case 0x02: ADD_GbEb(); break;
        case 0x03: ADD_GvEv(); break;
        case 0x04: ADD_ALIb(); break;
        case 0x05: ADD_eAXIv(); break;
        case 0x06: PUSH_ES(); break;
        case 0x07: POP_ES(); break;
        case 0x08: OR_EbGb(); break;
        case 0x09: OR_EvGv(); break;
        case 0x0A: OR_GbEb(); break;
        case 0x0B: OR_GvEv(); break;
        case 0x0C: OR_ALIb(); break;
        case 0x0D: OR_eAXIv(); break;
        case 0x0E: PUSH_CS(); break;
        case 0x0F: /* 2-byte escape */
          next_byte = fetch_next_byte(); 
          BX_INSTR_OPCODE_BYTE2(next_byte);
          switch (next_byte) {
            case 0x00: // Group 6
              decode_exgx();
              BX_INSTR_OPCODE_G6(BX_CPU_THIS_PTR modrm_ttt);
              switch (BX_CPU_THIS_PTR modrm_ttt) {
                case 0x00: SLDT_Ew(); break;
                case 0x01: STR_Ew(); break;
                case 0x02: LLDT_Ew(); break;
                case 0x03: LTR_Ew(); break;
                case 0x04: VERR_Ew(); break;
                case 0x05: VERW_Ew(); break;
                case 0x06: bx_panic("decode: : Grp6:\n"); break;
                case 0x07: bx_panic("decode: : Grp6:\n"); break;
                default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
                }
              break;
            case 0x01: // Group 7
              decode_exgx();
              BX_INSTR_OPCODE_G7(BX_CPU_THIS_PTR modrm_ttt);
              switch (BX_CPU_THIS_PTR modrm_ttt) {
                case 0x00: SGDT_Ms(); break;
                case 0x01: SIDT_Ms(); break;
                case 0x02: LGDT_Ms(); break;
                case 0x03: LIDT_Ms(); break;
                case 0x04: SMSW_Ew(); break;
                case 0x05: bx_panic("decode: : Grp7:\n"); break;
                case 0x06: LMSW_Ew(); break;
                case 0x07: bx_panic("decode: : Grp7:\n"); break;
                default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
                }
              break;

            case 0x02: LAR_GvEw(); break;
            case 0x03: LSL_GvEw(); break;
            case 0x04: bx_panic("decode: : 0F 04:\n"); break;
            case 0x05: LOADALL(); break;
            case 0x06: CLTS(); break;
            case 0x07: bx_panic("decode: : 0F 07:\n"); break;
            case 0x08: INVD(); break;
            case 0x09: WBINVD(); break;
            case 0x0A:
            case 0x0B:
            case 0x0C:
            case 0x0D:
            case 0x0E:
            case 0x0F: bx_panic("decode: : 0F %02x:\n", (unsigned) next_byte); break;

            case 0x10:
            case 0x11:
            case 0x12:
            case 0x13:
            case 0x14:
            case 0x15:
            case 0x16:
            case 0x17:
            case 0x18:
            case 0x19:
            case 0x1A:
            case 0x1B:
            case 0x1C:
            case 0x1D:
            case 0x1E:
            case 0x1F:
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte); break;
              break;

            case 0x20: MOV_RdCd(); break;
            case 0x21: MOV_RdDd(); break;
            case 0x22: MOV_CdRd(); break;
            case 0x23: MOV_DdRd(); break;
            case 0x24: MOV_RdTd(); break;
            case 0x25: bx_panic("decode: : 0F 25:\n"); break;
            case 0x26: MOV_TdRd(); break;
            case 0x27:
            case 0x28:
            case 0x29:
            case 0x2A:
            case 0x2B:
            case 0x2C:
            case 0x2D:
            case 0x2E:
            case 0x2F: bx_panic("decode: 0F %02x:\n", (unsigned) next_byte); break;

            case 0x30:
            case 0x31:
            case 0x32:
            case 0x33:
            case 0x34:
            case 0x35:
            case 0x36:
            case 0x37:
            case 0x38:
            case 0x39:
            case 0x3A:
            case 0x3B:
            case 0x3C:
            case 0x3D:
            case 0x3E:
            case 0x3F:
              bx_printf("decode: 0F %02x:\n", (unsigned) next_byte);
              exception(6, 0, 0);
              break;

#if (BX_CPU >= 6) || (BX_CPU_HACKED >= 6)
            // CMOV opcodes:
            case 0x40: CMOV(get_OF()); break;
            case 0x41: CMOV(!get_OF()); break;
            case 0x42: CMOV(get_CF()); break;
            case 0x43: CMOV(!get_CF()); break;
            case 0x44: CMOV(get_ZF()); break;
            case 0x45: CMOV(!get_ZF()); break;
            case 0x46: CMOV(get_CF() || get_ZF()); break;
            case 0x47: CMOV(!get_CF() && !get_ZF()); break;
            case 0x48: CMOV(get_SF()); break;
            case 0x49: CMOV(!get_SF()); break;
            case 0x4A: CMOV(get_PF()); break;
            case 0x4B: CMOV(!get_PF()); break;
            case 0x4C: CMOV(get_SF() != get_OF()); break;
            case 0x4D: CMOV(get_SF() == get_OF()); break;
            case 0x4E: CMOV(get_ZF() || (get_SF() != get_OF())); break;
            case 0x4F: CMOV(!get_ZF() && (get_SF() == get_OF())); break;
#else
            case 0x40:
            case 0x41:
            case 0x42:
            case 0x43:
            case 0x44:
            case 0x45:
            case 0x46:
            case 0x47:
            case 0x48:
            case 0x49:
            case 0x4A:
            case 0x4B:
            case 0x4C:
            case 0x4D:
            case 0x4E:
            case 0x4F: 
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte); break;
#endif

            case 0x50:
            case 0x51:
            case 0x52:
            case 0x53:
            case 0x54:
            case 0x55:
            case 0x56:
            case 0x57:
            case 0x58:
            case 0x59:
            case 0x5A:
            case 0x5B:
            case 0x5C:
            case 0x5D:
            case 0x5E:
            case 0x5F:

            case 0x60:
            case 0x61:
            case 0x62:
            case 0x63:
            case 0x64:
            case 0x65:
            case 0x66:
            case 0x67:
            case 0x68:
            case 0x69:
            case 0x6A:
            case 0x6B:
            case 0x6C:
            case 0x6D:
            case 0x6E:
            case 0x6F:

            case 0x70:
            case 0x71:
            case 0x72:
            case 0x73:
            case 0x74:
            case 0x75:
            case 0x76:
            case 0x77:
            case 0x78:
            case 0x79:
            case 0x7A:
            case 0x7B:
            case 0x7C:
            case 0x7D:
            case 0x7E:
            case 0x7F: bx_panic("decode: : 0F %02x:\n", (unsigned) next_byte); break;

            case 0x80: /* JO_Jv(); */ jv_condition = get_OF(); goto jv_routine;
            case 0x81: /* JNO_Jv(); */ jv_condition = !get_OF(); goto jv_routine;
            case 0x82: /* JB_Jv(); */ jv_condition = get_CF(); goto jv_routine;
            case 0x83: /* JNB_Jv(); */ jv_condition = !get_CF(); goto jv_routine;
            case 0x84: /* JZ_Jv(); */ jv_condition = get_ZF(); goto jv_routine;
            case 0x85: /* JNZ_Jv(); */ jv_condition = !get_ZF(); goto jv_routine;
            case 0x86: /* JBE_Jv(); */ jv_condition = get_CF() || get_ZF(); goto jv_routine;
            case 0x87: /* JNBE_Jv(); */ jv_condition = !get_CF() && !get_ZF(); goto jv_routine;
            case 0x88: /* JS_Jv(); */ jv_condition = get_SF(); goto jv_routine;
            case 0x89: /* JNS_Jv(); */ jv_condition = !get_SF(); goto jv_routine;
            case 0x8A: /* JP_Jv(); */ jv_condition = get_PF(); goto jv_routine;
            case 0x8B: /* JNP_Jv(); */ jv_condition = !get_PF(); goto jv_routine;
            case 0x8C: /* JL_Jv(); */ jv_condition = get_SF() != get_OF(); goto jv_routine;
            case 0x8D: /* JNL_Jv(); */ jv_condition = get_SF() == get_OF(); goto jv_routine;
            case 0x8E: /* JLE_Jv(); */ jv_condition = get_ZF() || (get_SF() != get_OF()); goto jv_routine;
            case 0x8F: /* JNLE_Jv(); */ jv_condition = (get_SF() == get_OF())  &&  !get_ZF(); goto jv_routine;

jv_routine:
#if BX_CPU < 3
              bx_panic("jv_routine: not supported on < 386\n");
              break;
#else
 
              if (BX_CPU_THIS_PTR is_32bit_opsize) {
                offset32s = (Bit32s) fetch_next_dword();
                }
              else {
                offset32s = (Bit16s) fetch_next_word();
                }
 
              if (jv_condition) {
                Bit32u new_EIP;
 
                invalidate_prefetch_q();
                new_EIP = EIP + offset32s;
                if (BX_CPU_THIS_PTR is_32bit_opsize==0)
                  new_EIP &= 0x0000ffff;
                if (protected_mode()) {
                  if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
                    bx_panic("jv_routine: offset outside of CS limits\n");
                    exception(BX_GP_EXCEPTION, 0, 0);
                    }
                  }
                EIP = new_EIP;
                BX_INSTR_CNEAR_BRANCH(new_EIP);
                }
              break;
#endif

            case 0x90: SETO_Eb(); break;
            case 0x91: SETNO_Eb(); break;
            case 0x92: SETB_Eb(); break;
            case 0x93: SETNB_Eb(); break;
            case 0x94: SETZ_Eb(); break;
            case 0x95: SETNZ_Eb(); break;
            case 0x96: SETBE_Eb(); break;
            case 0x97: SETNBE_Eb(); break;
            case 0x98: SETS_Eb(); break;
            case 0x99: SETNS_Eb(); break;
            case 0x9A: SETP_Eb(); break;
            case 0x9B: SETNP_Eb(); break;
            case 0x9C: SETL_Eb(); break;
            case 0x9D: SETNL_Eb(); break;
            case 0x9E: SETLE_Eb(); break;
            case 0x9F: SETNLE_Eb(); break;

            case 0xA0: PUSH_FS(); break;
            case 0xA1: POP_FS(); break;
            case 0xA2: CPUID(); break;
            case 0xA3: BT_EvGv(); break;

            case 0xA4: // SHLD_EvGvIb
              decode_exgx();
              SHLD_EvGv(fetch_next_byte());
              break;
            case 0xA5: // SHLD_EvGvCL
              decode_exgx();
              SHLD_EvGv(CL);
              break;

            case 0xA6: CMPXCHG_XBTS(); break;
            case 0xA7: CMPXCHG_IBTS(); break;
            case 0xA8: PUSH_GS(); break;
            case 0xA9: POP_GS(); break;
            case 0xAA: bx_panic("decode: : 0F AA:\n"); break;
            case 0xAB: BTS_EvGv(); break;

            case 0xAC: // SHRD_EvGvIb
              decode_exgx();
              SHRD_EvGv(fetch_next_byte());
              break;
            case 0xAD: // SHRD_EvGvCL
              decode_exgx();
              SHRD_EvGv(CL);
              break;

            case 0xAE: bx_panic("decode: 0F AE:\n"); break;
            case 0xAF: IMUL_GvEv(); break;

            case 0xB0: CMPXCHG_EbGb(); break;
            case 0xB1: CMPXCHG_EvGv(); break;
            case 0xB2: LSS_GvMp(); break;
            case 0xB3: BTR_EvGv(); break;
            case 0xB4: LFS_GvMp(); break;
            case 0xB5: LGS_GvMp(); break;
            case 0xB6: MOVZX_GvEb(); break;
            case 0xB7: MOVZX_GvEw(); break;
            case 0xB8: bx_panic("decode: 0F B8:\n"); break;
            case 0xB9: bx_panic("decode: 0F B9:\n"); break;
            case 0xBA: // Group 8 Ev,Ib
              decode_exgx();
              BX_INSTR_OPCODE_G8EvIb(BX_CPU_THIS_PTR modrm_ttt);
              switch (BX_CPU_THIS_PTR modrm_ttt) {
                case 0x00:
                case 0x01:
                case 0x02:
                case 0x03: bx_panic("decode: Grp8:\n"); break;
                case 0x04: BT_EvIb(); break;
                case 0x05: BTS_EvIb(); break;
                case 0x06: BTR_EvIb(); break;
                case 0x07: BTC_EvIb(); break;
                default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
                }
              break;

            case 0xBB: BTC_EvGv(); break;
            case 0xBC: BSF_GvEv(); break;
            case 0xBD: BSR_GvEv(); break;
            case 0xBE: MOVSX_GvEb(); break;
            case 0xBF: MOVSX_GvEw(); break;

            case 0xC0: XADD_EbGb(); break;
            case 0xC1: XADD_EvGv(); break;
            case 0xC2: 
            case 0xC3: 
            case 0xC4: 
            case 0xC5: 
            case 0xC6: 
            case 0xC7: bx_panic("decode: 0F C7:\n"); break;
            case 0xC8: BSWAP_EAX(); break;
            case 0xC9: BSWAP_ECX(); break;
            case 0xCA: BSWAP_EDX(); break;
            case 0xCB: BSWAP_EBX(); break;
            case 0xCC: BSWAP_ESP(); break;
            case 0xCD: BSWAP_EBP(); break;
            case 0xCE: BSWAP_ESI(); break;
            case 0xCF: BSWAP_EDI(); break;

            case 0xD0:
            case 0xD1:
            case 0xD2:
            case 0xD3:
            case 0xD4:
            case 0xD5:
            case 0xD6:
            case 0xD7:
            case 0xD8:
            case 0xD9:
            case 0xDA:
            case 0xDB:
            case 0xDC:
            case 0xDD:
            case 0xDE:
            case 0xDF:
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte);
              break;

            case 0xE0:
            case 0xE1:
            case 0xE2:
            case 0xE3:
            case 0xE4:
            case 0xE5:
            case 0xE6:
            case 0xE7:
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte);
              break;
            case 0xE8:
              bx_panic("decode: 0F %02x: caused undefined opcode exception\n",
                       (unsigned) next_byte);
              exception(6, 0, 0);
              break;
            case 0xE9:
            case 0xEA:
            case 0xEB:
            case 0xEC:
            case 0xED:
            case 0xEE:
            case 0xEF:
              bx_panic("decode: 0F %02x:\n", (unsigned) next_byte);
              exception(6, 0, 0);
              break;

            case 0xF0:
            case 0xF1:
            case 0xF2:
            case 0xF3:
            case 0xF4:
            case 0xF5:
            case 0xF6:
            case 0xF7:
            case 0xF8:
            case 0xF9:
            case 0xFA:
            case 0xFB:
            case 0xFC:
            case 0xFD:
            case 0xFE:
            case 0xFF:
              bx_printf("decode: instruction 0F %02x causes exception 6\n",
                        (unsigned) next_byte);
              exception(6, 0, 0);
              break;

            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0x10: ADC_EbGb(); break;
        case 0x11: ADC_EvGv(); break;
        case 0x12: ADC_GbEb(); break;
        case 0x13: ADC_GvEv(); break;
        case 0x14: ADC_ALIb(); break;
        case 0x15: ADC_eAXIv(); break;
        case 0x16: PUSH_SS(); break;
        case 0x17: POP_SS(); break;
        case 0x18: SBB_EbGb(); break;
        case 0x19: SBB_EvGv(); break;
        case 0x1A: SBB_GbEb(); break;
        case 0x1B: SBB_GvEv(); break;
        case 0x1C: SBB_ALIb(); break;
        case 0x1D: SBB_eAXIv(); break;
        case 0x1E: PUSH_DS(); break;
        case 0x1F: POP_DS(); break;


        case 0x20: AND_EbGb(); break;
        case 0x21: AND_EvGv(); break;
        case 0x22: AND_GbEb(); break;
        case 0x23: AND_GvEv(); break;
        case 0x24: AND_ALIb(); break;
        case 0x25: AND_eAXIv(); break;
        case 0x26: 
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR es;
          continue;
          break;
        case 0x27: DAA(); break;
        case 0x28: SUB_EbGb(); break;
        case 0x29: SUB_EvGv(); break;
        case 0x2A: SUB_GbEb(); break;
        case 0x2B: SUB_GvEv(); break;
        case 0x2C: SUB_ALIb(); break;
        case 0x2D: SUB_eAXIv(); break;
        case 0x2E:
#if BX_CPU <= 2
          bx_panic("decode: CS prefix used\n");
#else
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR cs;
#endif
          continue;
          break;
        case 0x2F: DAS(); break;

        case 0x30: XOR_EbGb(); break;
        case 0x31: XOR_EvGv(); break;
        case 0x32: XOR_GbEb(); break;
        case 0x33: XOR_GvEv(); break;
        case 0x34: XOR_ALIb(); break;
        case 0x35: XOR_eAXIv(); break;
        case 0x36: 
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR ss;
          continue;
          break;
        case 0x37: AAA(); break;
        case 0x38: CMP_EbGb(); break;
        case 0x39: CMP_EvGv(); break;
        case 0x3A: CMP_GbEb(); break;
        case 0x3B: CMP_GvEv(); break;
        case 0x3C: CMP_ALIb(); break;
        case 0x3D: CMP_eAXIv(); break;
        case 0x3E: 
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR ds;
          continue;
          break;
        case 0x3F: AAS(); break;

        case 0x40: // INC_eAX
        case 0x41: // INC_eCX
        case 0x42: // INC_eDX
        case 0x43: // INC_eBX
        case 0x44: // INC_eSP
        case 0x45: // INC_eBP
        case 0x46: // INC_eSI
        case 0x47: // INC_eDI

#if BX_CPU >= 3
          if (BX_CPU_THIS_PTR is_32bit_opsize) {
            Bit32u erx;

            erx = ++ BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].erx;
            SET_FLAGS_OSZAP_32(0, 0, erx, BX_INSTR_INC32);
            }
          else
#endif
            {
            Bit16u rx;

            rx = ++ BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].word.rx;
            SET_FLAGS_OSZAP_16(0, 0, rx, BX_INSTR_INC16);
            }
          break;




        case 0x48: // DEC_eAX
        case 0x49: // DEC_eCX
        case 0x4A: // DEC_eDX
        case 0x4B: // DEC_eBX
        case 0x4C: // DEC_eSP
        case 0x4D: // DEC_eBP
        case 0x4E: // DEC_eSI
        case 0x4F: // DEC_eDI

#if BX_CPU >= 3
          if (BX_CPU_THIS_PTR is_32bit_opsize) {
            Bit32u erx;

            erx = -- BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].erx;
            SET_FLAGS_OSZAP_32(0, 0, erx, BX_INSTR_DEC32);
            }
          else
#endif
            {
            Bit16u rx;

            rx = -- BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].word.rx;
            SET_FLAGS_OSZAP_16(0, 0, rx, BX_INSTR_DEC16);
            }
          break;


        case 0x50: // PUSH_eAX
        case 0x51: // PUSH_eCX
        case 0x52: // PUSH_eDX
        case 0x53: // PUSH_eBX
        case 0x54: // PUSH_eSP
        case 0x55: // PUSH_eBP
        case 0x56: // PUSH_eSI
        case 0x57: // PUSH_eDI

#if BX_CPU >= 3
          if (BX_CPU_THIS_PTR is_32bit_opsize) {
            push_32( BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].erx );
            }
          else
#endif
            {
            push_16( BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].word.rx );
            }
          break;

        case 0x58: // POP_eAX
        case 0x59: // POP_eCX
        case 0x5A: // POP_eDX
        case 0x5B: // POP_eBX
        case 0x5C: // POP_eSP
        case 0x5D: // POP_eBP
        case 0x5E: // POP_eSI
        case 0x5F: // POP_eDI

#if BX_CPU >= 3
          if (BX_CPU_THIS_PTR is_32bit_opsize) {
            Bit32u erx;

            pop_32(&erx);
            BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].erx = erx;
            }
          else
#endif
            {
            Bit16u rx;

            pop_16(&rx);
            BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].word.rx = rx;
            }
          break;


        case 0x60: PUSHAD(); break;
        case 0x61: POPAD(); break;
        case 0x62: BOUND_GvMa(); break;
        case 0x63: ARPL_EwGw(); break;
        case 0x64: 
#if BX_CPU <= 2
          bx_panic("decode: FS segment override used\n");
#else
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR fs;
#endif
          continue;
          break;
        case 0x65: 
#if BX_CPU <= 2
          bx_panic("decode: GS segment override used\n");
#else
          BX_CPU_THIS_PTR segment_override = &BX_CPU_THIS_PTR gs;
#endif
          continue;
          break;
        case 0x66:
#if BX_CPU < 3
          bx_panic("decode: 32bit_opsize prefix not supported on < 386\n");
          exception(6, 0, 0);
#else /* 386+ */
          BX_CPU_THIS_PTR is_32bit_opsize   = !BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
#endif
          continue;
          break;
        case 0x67:
#if BX_CPU < 3
          bx_panic("decode: 32bit_addrsize prefix not supported < 386!\n");
          exception(6, 0, 0);
#else /* 386+ */
          BX_CPU_THIS_PTR is_32bit_addrsize = !BX_CPU_THIS_PTR cs.cache.u.segment.d_b;
#endif
          continue;
          break;
        case 0x68: PUSH_Iv(); break;
        case 0x69: IMUL_GvEvIv(); break;
        case 0x6A: PUSH_Ib(); break;
        case 0x6B: IMUL_GvEvIb(); break;
        case 0x6C:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                INSB_YbDX();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                INSB_YbDX();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          INSB_YbDX();
          break;


        case 0x6D:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                INSW_YvDX();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                INSW_YvDX();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          INSW_YvDX();
          break;

        case 0x6E:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                OUTSB_DXXb();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                OUTSB_DXXb();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          OUTSB_DXXb();
          break;


        case 0x6F:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                OUTSW_DXXv();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                OUTSW_DXXv();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          OUTSW_DXXv();
          break;


        case 0x70: /* JO_Jb(); */ jb_condition = get_OF(); goto jb_routine;
        case 0x71: /* JNO_Jb(); */ jb_condition = !get_OF(); goto jb_routine;
        case 0x72: /* JB_Jb(); */ jb_condition = get_CF(); goto jb_routine;
        case 0x73: /* JNB_Jb(); */ jb_condition = !get_CF(); goto jb_routine;
        case 0x74: /* JZ_Jb(); */ jb_condition = get_ZF(); goto jb_routine;
        case 0x75: /* JNZ_Jb(); */ jb_condition = !get_ZF(); goto jb_routine;
        case 0x76: /* JBE_Jb(); */ jb_condition = get_CF() || get_ZF(); goto jb_routine;
        case 0x77: /* JNBE_Jb(); */ jb_condition = !get_CF() && !get_ZF(); goto jb_routine;
        case 0x78: /* JS_Jb(); */ jb_condition = get_SF(); goto jb_routine;
        case 0x79: /* JNS_Jb(); */ jb_condition = !get_SF(); goto jb_routine;
        case 0x7A: /* JP_Jb(); */ jb_condition = get_PF(); goto jb_routine;
        case 0x7B: /* JNP_Jb(); */ jb_condition = !get_PF(); goto jb_routine;
        case 0x7C: /* JL_Jb(); */ jb_condition = get_SF() != get_OF(); goto jb_routine;
        case 0x7D: /* JNL_Jb(); */ jb_condition = get_SF() == get_OF(); goto jb_routine;
        case 0x7E: /* JLE_Jb(); */ jb_condition = get_ZF() || (get_SF() != get_OF()); goto jb_routine;
        case 0x7F: /* JNLE_Jb(); */ jb_condition = (get_SF() == get_OF())  &&  !get_ZF(); goto jb_routine;

jb_routine:
          offset8s = (Bit8s) fetch_next_byte();
 
          if (jb_condition) {
            Bit32u new_EIP;
 
            new_EIP = EIP + offset8s;
            if (BX_CPU_THIS_PTR is_32bit_opsize==0)
              new_EIP &= 0x0000ffff;
#if BX_CPU >= 2
            if (protected_mode()) {
              if ( new_EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled ) {
                bx_panic("jo_routine: offset outside of CS limits\n");
                exception(BX_GP_EXCEPTION, 0, 0);
                }
              }
#endif
            EIP = new_EIP;
            BX_INSTR_CNEAR_BRANCH(new_EIP);
            revalidate_prefetch_q();
            }
          break;

        case 0x80: /* Immdediate Group 1 EbIb */
        case 0x82: // same behaviour as 0x80
          decode_exgx();
          BX_INSTR_OPCODE_G1EbIb(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ADD_EbIb(); break;
            case 0x01: OR_EbIb();  break;
            case 0x02: ADC_EbIb(); break;
            case 0x03: SBB_EbIb(); break;
            case 0x04: AND_EbIb(); break;
            case 0x05: SUB_EbIb(); break;
            case 0x06: XOR_EbIb(); break;
            case 0x07: CMP_EbIb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0x81: // Immdediate Group 1 EvIv
          decode_exgx();
          BX_INSTR_OPCODE_G1EvIv(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ADD_EvIv(); break;
            case 0x01: OR_EvIv(); break;
            case 0x02: ADC_EvIv(); break;
            case 0x03: SBB_EvIv(); break;
            case 0x04: AND_EvIv(); break;
            case 0x05: SUB_EvIv(); break;
            case 0x06: XOR_EvIv(); break;
            case 0x07: CMP_EvIv(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;


        case 0x83: /* Immdediate Group 1 EvIb */
          decode_exgx();
          BX_INSTR_OPCODE_G1EvIb(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ADD_EvIb(); break;
            case 0x01: OR_EvIb(); break;
            case 0x02: ADC_EvIb(); break;
            case 0x03: SBB_EvIb(); break;
            case 0x04: AND_EvIb(); break;
            case 0x05: SUB_EvIb(); break;
            case 0x06: XOR_EvIb(); break;
            case 0x07: CMP_EvIb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0x84: TEST_EbGb(); break;
        case 0x85: TEST_EvGv(); break;
        case 0x86: XCHG_EbGb(); break;
        case 0x87: XCHG_EvGv(); break;
        case 0x88: MOV_EbGb(); break;
        case 0x89: MOV_EvGv(); break;
        case 0x8A: MOV_GbEb(); break;
        case 0x8B: MOV_GvEv(); break;
        case 0x8C: MOV_EwSw(); break;
        case 0x8D: LEA_GvM(); break;
        case 0x8E: MOV_SwEw(); break;
        case 0x8F: POP_Ev(); break;


        case 0x90: /* NOP */ break;

        case 0x91: // XCHG_eCXeAX
        case 0x92: // XCHG_eDXeAX
        case 0x93: // XCHG_eBXeAX
        case 0x94: // XCHG_eSPeAX
        case 0x95: // XCHG_eBPeAX
        case 0x96: // XCHG_eSIeAX
        case 0x97: // XCHG_eDIeAX

#if BX_CPU >= 3
          if (BX_CPU_THIS_PTR is_32bit_opsize) {
            Bit32u temp32;
 
            temp32 = EAX;
            EAX = BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].erx;
            BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].erx = temp32;
            }
          else
#endif
            {
            Bit16u temp16;
 
            temp16 = AX;
            AX = BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].word.rx;
            BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].word.rx = temp16;
            }
          break;

        case 0x98: CBW(); break;
        case 0x99: CWD(); break;
        case 0x9A: CALL_Ap(); break;
        case 0x9B: WAIT(); break;
        case 0x9C: PUSHF_Fv(); break;
        case 0x9D: POPF_Fv(); break;
        case 0x9E: SAHF(); break;
        case 0x9F: LAHF(); break;


        case 0xA0: MOV_ALOb(); break;
        case 0xA1: MOV_eAXOv(); break;
        case 0xA2: MOV_ObAL(); break;
        case 0xA3: MOV_OveAX(); break;
        case 0xA4:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                MOVSB_XbYb();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                MOVSB_XbYb();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          MOVSB_XbYb();
          break;

        case 0xA5:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                MOVSW_XvYv();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                MOVSW_XvYv();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          MOVSW_XvYv();
          break;

        case 0xA6:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                CMPSB_XbYb();
                ECX = ECX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                CMPSB_XbYb();
                CX = CX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          CMPSB_XbYb();
          break;

        case 0xA7:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                CMPSW_XvYv();
                ECX = ECX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                CMPSW_XvYv();
                CX = CX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          CMPSW_XvYv();
          break;

        case 0xA8: TEST_ALIb(); break;
        case 0xA9: TEST_eAXIv(); break;
        case 0xAA:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                STOSB_YbAL();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                STOSB_YbAL();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          STOSB_YbAL();
          break;

        case 0xAB:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                STOSW_YveAX();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                STOSW_YveAX();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          STOSW_YveAX();
          break;

        case 0xAC:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                LODSB_ALXb();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                LODSB_ALXb();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          LODSB_ALXb();
          break;

        case 0xAD:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                LODSW_eAXXv();
                ECX = ECX - 1;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                LODSW_eAXXv();
                CX = CX - 1;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          LODSW_eAXXv();
          break;

        case 0xAE:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                SCASB_ALXb();
                ECX = ECX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                SCASB_ALXb();
                CX = CX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          SCASB_ALXb();
          break;

        case 0xAF:
          if (rep_prefix) {
#if BX_CPU > 2
            if (BX_CPU_THIS_PTR is_32bit_addrsize) {
              if (ECX != 0) {
                SCASW_eAXXv();
                ECX = ECX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (ECX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }
            else {
#endif
              if (CX != 0) {
                SCASW_eAXXv();
                CX = CX - 1;
                if (rep_prefix==BX_REPE_PREFIX && get_ZF()==0) break;
                if (rep_prefix==BX_REPNE_PREFIX && get_ZF()!=0) break;
                if (CX != 0) {
                  BX_CPU_THIS_PTR eip = BX_CPU_THIS_PTR prev_eip; // backup IP and repeat
                  BX_INSTR_OPCODE_REPEATING();
                  revalidate_prefetch_q();
                  }
                }
              }

            break;
            }

          /* no repeat prefix used */
          SCASW_eAXXv();
          break;


        case 0xB0:
        case 0xB1:
        case 0xB2:
        case 0xB3:
          BX_CPU_THIS_PTR gen_reg[next_byte & 0x03].word.byte.rl = fetch_next_byte();
          break;

        case 0xB4:
        case 0xB5:
        case 0xB6:
        case 0xB7:
          BX_CPU_THIS_PTR gen_reg[next_byte & 0x03].word.byte.rh = fetch_next_byte();
          break;

        case 0xB8:
        case 0xB9:
        case 0xBA:
        case 0xBB:
        case 0xBC:
        case 0xBD:
        case 0xBE:
        case 0xBF:
#if BX_CPU >= 3
          if (BX_CPU_THIS_PTR is_32bit_opsize) {
            BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].erx = fetch_next_dword();
            }
          else
#endif
            {
            BX_CPU_THIS_PTR gen_reg[next_byte & 0x07].word.rx = fetch_next_word();
            }
          break;

        case 0xC0: // Group 2 Eb,Ib
          decode_exgx();
          BX_INSTR_OPCODE_G2EbIb(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ROL_Eb(fetch_next_byte()); break;
            case 0x01: ROR_Eb(fetch_next_byte()); break;
            case 0x02: RCL_Eb(fetch_next_byte()); break;
            case 0x03: RCR_Eb(fetch_next_byte()); break;
            case 0x04: case 0x06: SHL_Eb(fetch_next_byte()); break;
            case 0x05: SHR_Eb(fetch_next_byte()); break;
            case 0x07: SAR_Eb(fetch_next_byte()); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xC1: // Group 2 Ev,Ib
          decode_exgx();
          BX_INSTR_OPCODE_G2EvIb(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ROL_Ev(fetch_next_byte()); break;
            case 0x01: ROR_Ev(fetch_next_byte()); break;
            case 0x02: RCL_Ev(fetch_next_byte()); break;
            case 0x03: RCR_Ev(fetch_next_byte()); break;
            case 0x04: case 0x06: SHL_Ev(fetch_next_byte()); break;
            case 0x05: SHR_Ev(fetch_next_byte()); break;
            case 0x07: SAR_Ev(fetch_next_byte()); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xC2: RETnear_Iw(); break;
        case 0xC3: RETnear(); break;
        case 0xC4: LES_GvMp(); break;
        case 0xC5: LDS_GvMp(); break;
        case 0xC6: MOV_EbIb(); break;
        case 0xC7: MOV_EvIv(); break;
        case 0xC8: ENTER_IwIb(); break;
        case 0xC9: LEAVE(); break;
        case 0xCA: RETfar_Iw(); break;
        case 0xCB: RETfar(); break;
        case 0xCC: INT3(); break;
        case 0xCD: INT_Ib(); break;
        case 0xCE: INTO(); break;
        case 0xCF: IRET(); break;


        case 0xD0: // group 2 Eb,1
          decode_exgx();
          BX_INSTR_OPCODE_G2Eb1(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ROL_Eb(1); break;
            case 0x01: ROR_Eb(1); break;
            case 0x02: RCL_Eb(1); break;
            case 0x03: RCR_Eb(1); break;
            case 0x04: case 0x06: SHL_Eb(1); break;
            case 0x05: SHR_Eb(1); break;
            case 0x07: SAR_Eb(1); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD1: // group2 Ev,1
          decode_exgx();
          BX_INSTR_OPCODE_G2Ev1(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ROL_Ev(1); break;
            case 0x01: ROR_Ev(1); break;
            case 0x02: RCL_Ev(1); break;
            case 0x03: RCR_Ev(1); break;
            case 0x04: case 0x06: SHL_Ev(1); break;
            case 0x05: SHR_Ev(1); break;
            case 0x07: SAR_Ev(1); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD2: // group2 Eb,CL
          decode_exgx();
          BX_INSTR_OPCODE_G2EbCL(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ROL_Eb(CL); break;
            case 0x01: ROR_Eb(CL); break;
            case 0x02: RCL_Eb(CL); break;
            case 0x03: RCR_Eb(CL); break;
            case 0x04: case 0x06: SHL_Eb(CL); break;
            case 0x05: SHR_Eb(CL); break;
            case 0x07: SAR_Eb(CL); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD3: // group2 Ev,CL
          decode_exgx();
          BX_INSTR_OPCODE_G2EvCL(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: ROL_Ev(CL); break;
            case 0x01: ROR_Ev(CL); break;
            case 0x02: RCL_Ev(CL); break;
            case 0x03: RCR_Ev(CL); break;
            case 0x04: case 0x06: SHL_Ev(CL); break;
            case 0x05: SHR_Ev(CL); break;
            case 0x07: SAR_Ev(CL); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xD4: AAM(); break;
        case 0xD5: AAD(); break;
        case 0xD6: SALC(); break;
        case 0xD7: XLAT(); break;

        case 0xD8: // ESC0
          //bx_printf("ESC0:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC0:\n");
          decode_ESC_modrm();
          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FADD_Es(); break;
              case 0x01: FMUL_Es(); break;
              case 0x02: FCOM_Es(); break;
              case 0x03: FCOMP_Es(); break;
              case 0x04: FSUB_Es(); break;
              case 0x05: FSUBR_Es(); break;
              case 0x06: FDIV_Es(); break;
              case 0x07: FDIVR_Es(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FADD_ST_STi(); break;
              case 0x01: FMUL_ST_STi(); break;
              case 0x02: FCOM_ST_STi(); break;
              case 0x03: FCOMP_ST_STi(); break;
              case 0x04: FSUB_ST_STi(); break;
              case 0x05: FSUBR_ST_STi(); break;
              case 0x06: FDIV_ST_STi(); break;
              case 0x07: FDIVR_ST_STi(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          break;

        case 0xD9: // ESC1
          //bx_printf("ESC1:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          decode_ESC_modrm();

          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FLD_Es(); break;
              case 0x01: bx_panic("decode: ESC1:\n"); break;
              case 0x02: FST_Es(); break;
              case 0x03: FSTP_Es(); break;
              case 0x04: FLDENV_Ea(); break;
              case 0x05: FLDCW_Ew(); break;
              case 0x06: FSTENV_Ea(); break;
              case 0x07: FSTCW_Ew(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00:
                FLD_STi(); break;
              case 0x01:
                FXCH_STi(); break;
              case 0x02:
                if (BX_CPU_THIS_PTR ESC_rm == 0) {
                  FNOP(); break;
                  }
                else {
                  bx_panic("decode: ESC9:\n"); break;
                  }
              case 0x03:
                FSTP_STi(); break;

              case 0x04:
                switch (BX_CPU_THIS_PTR ESC_rm) {
                  case 0x00: FCHS(); break;
                  case 0x01: FABS(); break;
                  case 0x02:
                  case 0x03: bx_panic("decode: ESC9:\n"); break;
                  case 0x04: FTST(); break;
                  case 0x05: FXAM(); break;
                  case 0x06:
                  case 0x07: bx_panic("decode: ESC9:\n"); break;
                  }
                break;
              case 0x05:
                switch (BX_CPU_THIS_PTR ESC_rm) {
                  case 0x00: FLD1(); break;
                  case 0x01: FLDL2T(); break;
                  case 0x02: FLDL2E(); break;
                  case 0x03: FLDPI(); break;
                  case 0x04: FLDLG2(); break;
                  case 0x05: FLDLN2(); break;
                  case 0x06: FLDZ(); break;
                  case 0x07: bx_panic("decode: ESC9:\n"); break;
                  }
                break;
              case 0x06:
                switch (BX_CPU_THIS_PTR ESC_rm) {
                  case 0x00: F2XM1(); break;
                  case 0x01: FYL2X(); break;
                  case 0x02: FPTAN(); break;
                  case 0x03: FPATAN(); break;
                  case 0x04: FXTRACT(); break;
                  case 0x05: FPREM1(); break;
                  case 0x06: FDECSTP(); break;
                  case 0x07: FINCSTP(); break;
                  }
                break;
              case 0x07:
                switch (BX_CPU_THIS_PTR ESC_rm) {
                  case 0x00: FPREM(); break;
                  case 0x01: FYL2XP1(); break;
                  case 0x02: FSQRT(); break;
                  case 0x03: FSINCOS(); break;
                  case 0x04: FRNDINT(); break;
                  case 0x05: FSCALE(); break;
                  case 0x06: FSIN(); break;
                  case 0x07: FCOS(); break;
                  }
                break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          break;

        case 0xDA: // ESC2
          //bx_printf("ESC2:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC2:\n");
          decode_ESC_modrm();
          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FIADD_Ed(); break;
              case 0x01: FIMUL_Ed(); break;
              case 0x02: FICOM_Ed(); break;
              case 0x03: FICOMP_Ed(); break;
              case 0x04: FISUB_Ed(); break;
              case 0x05: FISUBR_Ed(); break;
              case 0x06: FIDIV_Ed(); break;
              case 0x07: FIDIVR_Ed(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x05:
                if (BX_CPU_THIS_PTR ESC_rm == 1) {
                  FUCOMPP(); break;
                  }
                else {
                  bx_panic("decode: ESC2:\n"); break;
                  }
              default: bx_panic("decode: ESC2:\n"); break;
              }
            break;
            }
          break;

        case 0xDB: // ESC3
          //bx_printf("ESC3:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          decode_ESC_modrm();
          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FILD_Ed(); break;
              case 0x01: bx_panic("decode: ESC3:\n"); break;
              case 0x02: FIST_Ed(); break;
              case 0x03: FISTP_Ed(); break;
              case 0x04: bx_panic("decode: ESC3:\n"); break;
              case 0x05: FLD_Et(); break;
              case 0x06: bx_panic("decode: ESC3:\n"); break;
              case 0x07: FSTP_Et(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            switch (BX_CPU_THIS_PTR ESC_modrm) {
              case 0xe0: FENI(); break;
              case 0xe1: FDISI(); break;
              case 0xe2: FCLEX(); break;
              case 0xe3: FINIT(); break;
              case 0xe4: FSETPM(); break;
              default:
                bx_panic("decode: ESC3: reserved opcode\n"); break;
              }
            break;
            }
          break;

        case 0xDC: // ESC4
          //bx_printf("ESC4:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC4:\n");
          decode_ESC_modrm();
          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FADD_El(); break;
              case 0x01: FMUL_El(); break;
              case 0x02: FCOM_El(); break;
              case 0x03: FCOMP_El(); break;
              case 0x04: FSUB_El(); break;
              case 0x05: FSUBR_El(); break;
              case 0x06: FDIV_El(); break;
              case 0x07: FDIVR_El(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FADD_STi_ST(); break;
              case 0x01: FMUL_STi_ST(); break;
              case 0x02: FCOM_STi_ST(); break;
              case 0x03: FCOMP_STi_ST(); break;
              case 0x04: FSUBR_STi_ST(); break;
              case 0x05: FSUB_STi_ST(); break;
              case 0x06: FDIVR_STi_ST(); break;
              case 0x07: FDIV_STi_ST(); break;
              default: bx_panic("decode: ESC4:\n"); break;
              }
            break;
            }
          break;


        case 0xDD: // ESC5
          //bx_printf("ESC5:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          decode_ESC_modrm();
          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FLD_El(); break;
              case 0x01: bx_panic("decode: ESC5:\n"); break;
              case 0x02: FST_El(); break;
              case 0x03: FSTP_El(); break;
              case 0x04: FRSTOR_Ea(); break;
              case 0x05: bx_panic("decode: ESC5:\n"); break;
              case 0x06: FSAVE_Ea(); break;
              case 0x07: FSTSW_Ew(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
bx_panic("ESC5: recode\n");
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FFREE_STi(); break;
              case 0x01: FXCH_STi(); break;
              case 0x02: FST_STi(); break;
              case 0x03: FSTP_STi(); break;
              case 0x04: FUCOM_STi_ST(); break;
              case 0x05: FUCOMP_STi(); break;
              case 0x06: bx_panic("decode: ESC5:\n"); break;
              case 0x07: bx_panic("decode: ESC5:\n"); break;
              default: bx_panic("decode: ESC5:\n"); break;
              }
            break;
            }
          break;

        case 0xDE: // ESC6
          //bx_printf("ESC6:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          bx_panic("ESC6:\n");
          decode_ESC_modrm();
          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FIADD_Ew(); break;
              case 0x01: FIMUL_Ew(); break;
              case 0x02: FICOM_Ew(); break;
              case 0x03: FICOMP_Ew(); break;
              case 0x04: FISUB_Ew(); break;
              case 0x05: FISUBR_Ew(); break;
              case 0x06: FIDIV_Ew(); break;
              case 0x07: FIDIVR_Ew(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FADDP_STi_ST(); break;
              case 0x01: FMULP_STi_ST(); break;
              case 0x02: FCOMP_STi(); break;
              case 0x03:
                switch (BX_CPU_THIS_PTR ESC_rm) {
                  case 0x01: FCOMPP(); break;
                  default: bx_panic("decode: ESC6:\n"); break;
                  }
                break;
              case 0x04: FSUBRP_STi_ST(); break;
              case 0x05: FSUBP_STi_ST(); break;
              case 0x06: FDIVRP_STi_ST(); break;
              case 0x07: FDIVP_STi_ST(); break;
              default: bx_panic("decode: ESC6:\n"); break;
              }
            break;
            }
          break;

        case 0xDF: // ESC7
          //bx_printf("ESC7:\n");
          if ( BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts ) {
            exception(BX_NM_EXCEPTION, 0, 0);
            break;
            }

          decode_ESC_modrm();
          if (BX_CPU_THIS_PTR ESC_mod != 3) {
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FILD_Ew(); break;
              case 0x01: bx_panic("decode: ESC7: mod!=3 case 0x01:\n"); break;
              case 0x02: FIST_Ew(); break;
              case 0x03: FISTP_Ew(); break;
              case 0x04: FBLD_Eb(); break;
              case 0x05: FILD_Eq(); break;
              case 0x06: FBSTP_Eb(); break;
              case 0x07: FISTP_Eq(); break;
              default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
              }
            break;
            }
          else { /* mod == 3 */
            switch (BX_CPU_THIS_PTR ESC_nnn) {
              case 0x00: FFREE_STi(); break;
              case 0x01: FXCH_STi(); break;
              case 0x02: FST_STi(); break;
              case 0x03: FSTP_STi(); break;
              case 0x04:
                switch (BX_CPU_THIS_PTR ESC_rm) {
                  case 0x00: FSTSW_AX(); break;
                  default: bx_panic("decode: ESC7: mod==3 case 0x04: default:\n"); break;
                  }
                break;
              default: bx_panic("decode: ESC7: mod==3: default\n"); break;
              }
            break;
            }
          break;


        case 0xE0: LOOPNE_Jb(); break;
        case 0xE1: LOOPE_Jb(); break;
        case 0xE2: LOOP_Jb(); break;
        case 0xE3: JCXZ_Jb(); break;
        case 0xE4: IN_ALIb(); break;
        case 0xE5: IN_eAXIb(); break;
        case 0xE6: OUT_IbAL(); break;
        case 0xE7: OUT_IbeAX(); break;
        case 0xE8: CALL_Av(); break;
        case 0xE9: JMP_Jv(); break;
        case 0xEA: JMP_Ap(); break;
        case 0xEB: JMP_Jb(); break;
        case 0xEC: IN_ALDX(); break;
        case 0xED: IN_eAXDX(); break;
        case 0xEE: OUT_DXAL(); break;
        case 0xEF: OUT_DXeAX(); break;

        case 0xF0: /* LOCK */
          bx_printf("decode: LOCK prefix encountered\n");
          if (protected_mode() && CPL>IOPL) {
            bx_panic("decode: LOCK prefix found: CPL>IOPL\n");
            /* #GP(0) */
            break;
            }
          continue;
          break;
        case 0xF1: bx_panic("decode: F1: Intel reserved opcode\n");
                   break;
        case 0xF2: /* REPNE/REPNZ */
          rep_prefix = BX_REPNE_PREFIX;
          continue;
          break;
        case 0xF3: /* REP/REPE/REPZ */
          rep_prefix = BX_REPE_PREFIX;
          continue;
          break;
        case 0xF4:
          HLT();
          break;
        case 0xF5: CMC(); break;
        case 0xF6: /* Group 3 Eb */
          decode_exgx();
          BX_INSTR_OPCODE_G3Eb(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: case 0x01: TEST_EbIb(); break;
            case 0x02: NOT_Eb(); break;
            case 0x03: NEG_Eb(); break;
            case 0x04: MUL_ALEb(); break;
            case 0x05: IMUL_ALEb(); break;
            case 0x06: DIV_ALEb(); break;
            case 0x07: IDIV_ALEb(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;

        case 0xF7: /* GROUP3 Ev */
          decode_exgx();
          BX_INSTR_OPCODE_G3Ev(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: case 0x01: TEST_EvIv(); break;
            case 0x02: NOT_Ev(); break;
            case 0x03: NEG_Ev(); break;
            case 0x04: MUL_eAXEv(); break;
            case 0x05: IMUL_eAXEv(); break;
            case 0x06: DIV_eAXEv(); break;
            case 0x07: IDIV_eAXEv(); break;
            default: bx_panic("decode: error in bx_cpu_loop()!\n"); break;
            }
          break;
        case 0xF8: CLC(); break;
        case 0xF9: STC(); break;
        case 0xFA: CLI(); break;
        case 0xFB: STI(); break;
        case 0xFC: CLD(); break;
        case 0xFD: STD(); break;
        case 0xFE: // GROUP4
          decode_exgx();
          BX_INSTR_OPCODE_G4(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: INC_Eb(); break;
            case 0x01: DEC_Eb(); break;
            default: bx_panic("decode: FE default:\n"); break;
            }
          break;

        case 0xFF: // GROUP 5
          decode_exgx();
          BX_INSTR_OPCODE_G5(BX_CPU_THIS_PTR modrm_ttt);
          switch (BX_CPU_THIS_PTR modrm_ttt) {
            case 0x00: INC_Ev(); break;
            case 0x01: DEC_Ev(); break;
            case 0x02: CALL_Ev(); break;
            case 0x03: CALL_Ep(); break;
            case 0x04: JMP_Ev(); break;
            case 0x05: JMP_Ep(); break;
            case 0x06: PUSH_Ev(); break;
            default: bx_panic("decode: FF default:\n"); break;
            }
          break;

        default: /* only invalid instructions left */
          bx_panic("decode: error in bx_cpu_loop()!\n");
          break;
        } /* switch (next_byte) */
      break;
      } // while (1)...

    // advance instruction tick clock...
    bx_pc_system.TICK();
    BX_INSTR_OPCODE_END();
    } /* while (1) */

  //BX_CPU_LOOP_EXIT();
}
