// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"




/* Exception classes.  These are used as indexes into the 'is_exception_OK'
 * array below, and are stored in the 'exception' array also
 */
#define BX_ET_BENIGN       0
#define BX_ET_CONTRIBUTORY 1
#define BX_ET_PAGE_FAULT   2

#define BX_ET_DOUBLE_FAULT 10


const Boolean bx_cpu_c::is_exception_OK[3][3] = {
    { 1, 1, 1 }, /* 1st exception is BENIGN */
    { 1, 0, 1 }, /* 1st exception is CONTRIBUTORY */
    { 1, 0, 0 }  /* 1st exception is PAGE_FAULT */
    };


  void
bx_cpu_c::interrupt(Bit8u vector, Boolean is_INT, Boolean is_error_code,
                    Bit16u error_code)
{
  invalidate_prefetch_q();

#if BX_CPU >= 2
//  unsigned prev_errno;

  if (bx_dbg.interrupts)
    bx_printf("interrupt(): vector = %u, INT = %u\n",
      (unsigned) vector, (unsigned) is_INT);

BX_CPU_THIS_PTR save_cs  = BX_CPU_THIS_PTR cs;
BX_CPU_THIS_PTR save_ss  = BX_CPU_THIS_PTR ss;
BX_CPU_THIS_PTR save_eip = EIP;
BX_CPU_THIS_PTR save_esp = ESP;

//  prev_errno = BX_CPU_THIS_PTR errorno;

  if(!real_mode()) {
    Bit32u  dword1, dword2;
    bx_descriptor_t gate_descriptor, cs_descriptor;
    bx_selector_t cs_selector;

    Bit16u raw_tss_selector;
    bx_selector_t   tss_selector;
    bx_descriptor_t tss_descriptor;


    /* Comments closely copied from Intel's "80286 and 80287 Programmers's
     * Reference Manual", page B-49 */

    /* interrupt vector must be within IDT table limits,
     * else #GP(vector number*8 + 2 + EXT) */
    if ( (vector*8 + 7) > BX_CPU_THIS_PTR idtr.limit) {
      if (bx_dbg.interrupts) {
        bx_printf("IDT.limit = %04x\n", (unsigned) BX_CPU_THIS_PTR idtr.limit);
        bx_printf("IDT.base  = %06x\n", (unsigned) BX_CPU_THIS_PTR idtr.base);
        bx_printf("interrupt vector must be within IDT table limits\n");
        bx_printf("bailing\n");
        }
      bx_printf("interrupt(): vector > idtr.limit\n");
      //reset_cpu(); /* ??? */
      //return;

      exception(BX_GP_EXCEPTION, vector*8 + 2, 0);
      return;
      }

    /* descriptor AR byte must indicate interrupt gate, trap gate,
     * or task gate, else #GP(vector*8 + 2 + EXT) */
    access_linear(BX_CPU_THIS_PTR idtr.base + vector*8,     4, 0,
      BX_READ, &dword1);
    access_linear(BX_CPU_THIS_PTR idtr.base + vector*8 + 4, 4, 0,
      BX_READ, &dword2);

    parse_descriptor(dword1, dword2, &gate_descriptor);

    if ( (gate_descriptor.valid==0) || gate_descriptor.segment) {
      bx_panic("interrupt(): gate descriptor is not valid sys seg\n");
      exception(BX_GP_EXCEPTION, vector*8 + 2, 0);
      return;
      }

    switch (gate_descriptor.type) {
      case 5: /* task gate */
      case 6: /* 286 interrupt gate */
      case 7: /* 286 trap gate */
      case 14: /* 386 interrupt gate */
      case 15: /* 386 trap gate */
        break;
      default:
        bx_panic("interrupt(): gate.type(%u) != {5,6,7,14,15}\n",
          (unsigned) gate_descriptor.type);
        exception(BX_GP_EXCEPTION, vector*8 + 2, 0);
        return;
      }

    /* if software interrupt, then gate descripor DPL must be >= CPL,
     * else #GP(vector * 8 + 2 + EXT) */
    if (is_INT  &&  (gate_descriptor.dpl < CPL)) {
/* ??? */
      bx_printf("interrupt(): is_INT && (dpl < CPL)\n");
      exception(BX_GP_EXCEPTION, vector*8 + 2, 0);
      return;
      }

    /* Gate must be present, else #NP(vector * 8 + 2 + EXT) */
    if (gate_descriptor.p == 0) { /* not present */
      bx_panic("interrupt(): p == 0\n");
      exception(BX_NP_EXCEPTION, vector*8 + 2, 0);
      }

    switch (gate_descriptor.type) {
      case 5: // 286/386 task gate
        bx_printf("interrupt(): task gate\n");

        // examine selector to TSS, given in task gate descriptor
        raw_tss_selector = gate_descriptor.u.taskgate.tss_selector;
        parse_selector(raw_tss_selector, &tss_selector);

        // must specify global in the local/global bit,
        //      else #TS(TSS selector)
        if (tss_selector.ti) {
          bx_panic("interrupt: tss_selector.ti=1\n");
          exception(BX_TS_EXCEPTION, raw_tss_selector & 0xfffc, 0);
          return;
          }

        // index must be within GDT limits, else #TS(TSS selector)
        fetch_raw_descriptor(&tss_selector, &dword1, &dword2,
          BX_TS_EXCEPTION);

        // AR byte must specify available TSS,
        //   else #TS(TSS selector)
        parse_descriptor(dword1, dword2, &tss_descriptor);
        if (tss_descriptor.valid==0 || tss_descriptor.segment) {
          bx_panic("exception: TSS selector points to bad TSS\n");
          exception(BX_TS_EXCEPTION, raw_tss_selector & 0xfffc, 0);
          return;
          }
        if (tss_descriptor.type!=9 && tss_descriptor.type!=1) {
          bx_panic("exception: TSS selector points to bad TSS\n");
          exception(BX_TS_EXCEPTION, raw_tss_selector & 0xfffc, 0);
          return;
          }


        // TSS must be present, else #NP(TSS selector)
        // done in task_switch()

bx_printf("task gate: ready for task_switch()\n");
        // switch tasks with nesting to TSS
        task_switch(&tss_selector, &tss_descriptor,
                    BX_TASK_FROM_CALL_OR_INT, dword1, dword2);

bx_printf("task gate: ready to push error code\n");
        // if interrupt was caused by fault with error code
        //   stack limits must allow push of 2 more bytes, else #SS(0)
        // push error code onto stack

bx_printf("tss.type=%u, CS.size=%u\n",
  (unsigned) tss_descriptor.type,
  (unsigned) BX_CPU_THIS_PTR cs.cache.u.segment.d_b);

        //??? push_16 vs push_32
        if ( is_error_code ) {
          //if (tss_descriptor.type==9)
          if (BX_CPU_THIS_PTR cs.cache.u.segment.d_b)
            push_32(error_code);
          else
            push_16(error_code);
          }

        // instruction pointer must be in CS limit, else #GP(0)
        //if (EIP > cs_descriptor.u.segment.limit_scaled) {}
        if (EIP > BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled) {
          bx_panic("exception(): eIP > CS.limit\n");
          exception(BX_GP_EXCEPTION, 0x0000, 0);
          }
bx_printf("task gate: all checks OK\n");
        return;
        break;

      case 6: /* 286 interrupt gate */
      case 7: /* 286 trap gate */
        if (v8086_mode()) {
          bx_panic("interrupt: 286-int/trap gate from v8086 mode\n");
          // Note: processor leaves v8086 mode as the result of an exception
          //   or interrupt which calls a trap or interrupt gate
          }

        if (bx_dbg.interrupts) {
          if (gate_descriptor.type==6)
            bx_printf("interrupt(): int gate 16\n");
          else
            bx_printf("interrupt(): trap gate 16\n");
          }

        /* TRAP-OR-INT-GATE */

        /* examine CS selector and descriptor given in gate descriptor */
        /* selector must be non-null else #GP(EXT) */
        if ( (gate_descriptor.u.gate286.dest_selector & 0xfffc) == 0 ) {
          bx_panic("int_trap_gate286(): selector null\n");
          /* GP(ext) */
          return;
          }

        parse_selector(gate_descriptor.u.gate286.dest_selector,
                          &cs_selector);

        /* fetch 2 dwords of descriptor; call handles out of limits checks */
        fetch_raw_descriptor(&cs_selector, &dword1, &dword2,
                                BX_GP_EXCEPTION);
        parse_descriptor(dword1, dword2, &cs_descriptor);

        /* descriptor AR byte must indicate code seg, else #GP(selector+EXT) */
        if ( cs_descriptor.valid==0 ||
             cs_descriptor.segment==0 ||
             cs_descriptor.u.segment.executable==0 ) {
          bx_panic("interrupt(): not code segment\n");
          /* GP(selector + ext) */
          return;
          }

        /* segment must be present, else #NP(selector + EXT) */
        if ( cs_descriptor.p==0 ) {
          bx_panic("interrupt(): segment not present\n");
          /* NP(selector + EXT) */
          return;
          }

        /* if code segment is non-conforming and DPL < CPL then
         * INTERRUPT TO INNER PRIVILEGE: */
        if ( cs_descriptor.u.segment.c_ed==0 && cs_descriptor.dpl<CPL ) {
          Bit16u old_SP, old_SS, old_IP, old_CS;
          Bit16u SS_for_cpl_x, SP_for_cpl_x;
          bx_descriptor_t ss_descriptor;
          bx_selector_t   ss_selector;
          int bytes;

          if (bx_dbg.interrupts)
            bx_printf("interrupt(): INTERRUPT TO INNER PRIVILEGE\n");

          /* check selector and descriptor for new stack in current
           * Task State Segment */
          
          if (BX_CPU_THIS_PTR tr.cache.valid==0)
            bx_panic("interrupt(): TR.cache invalid\n");
          if (BX_CPU_THIS_PTR tr.cache.type!=1)
            bx_panic("interrupt(): gate286: TR.type(%u)!=9\n",
              (unsigned) BX_CPU_THIS_PTR tr.cache.type);

          access_linear(BX_CPU_THIS_PTR tr.cache.u.tss286.base + 4 +
            4*cs_descriptor.dpl, 2, 0, BX_READ, &SS_for_cpl_x);
          access_linear(BX_CPU_THIS_PTR tr.cache.u.tss286.base + 2 +
            4*cs_descriptor.dpl, 2, 0, BX_READ, &SP_for_cpl_x);

          /* Selector must be non-null else #TS(EXT) */
          if ( (SS_for_cpl_x & 0xfffc) == 0 ) {
            bx_panic("interrupt(): SS selector null\n");
            /* TS(ext) */
            return;
            }

          /* selector index must be within its descriptor table limits
           * else #TS(SS selector + EXT) */
          parse_selector(SS_for_cpl_x, &ss_selector);
          /* fetch 2 dwords of descriptor; call handles out of limits checks */
          fetch_raw_descriptor(&ss_selector, &dword1, &dword2,
                                  BX_TS_EXCEPTION);
          parse_descriptor(dword1, dword2, &ss_descriptor);

          /* selector rpl must = dpl of code segment,
           * else #TS(SS selector + ext)*/
          if (ss_selector.rpl != cs_descriptor.dpl) {
            bx_panic("interrupt(): SS.rpl != CS.dpl\n");
            /* TS(SS selector) */
            return;
            }

          /* stack seg DPL must = DPL of code segment,
           * else #TS(SS selector + ext) */
          if (ss_descriptor.dpl != cs_descriptor.dpl) {
            bx_panic("interrupt(): SS.dpl != CS.dpl\n");
            /* TS(SS selector) */
            return;
            }

          /* descriptor must indicate writable data segment,
           * else #TS(SS selector + EXT) */
          if (ss_descriptor.valid==0 ||
              ss_descriptor.segment==0  ||
              ss_descriptor.u.segment.executable==1  ||
              ss_descriptor.u.segment.r_w==0) {
            bx_panic("interrupt(): SS not writable data segment\n");
            /* TS(SS selector) */
            return;
            }

          /* seg must be present, else #SS(SS selector + ext) */
          if (ss_descriptor.p==0) {
            bx_panic("interrupt(): SS not present\n");
            /* SS(SS selector) */
            return;
            }

          if ( is_error_code )
            bytes = 12;
          else
            bytes = 10;

          /* new stack must have room for 10 bytes, else #SS(0) */
          if ( !can_push(&ss_descriptor, SP_for_cpl_x, bytes) ) {
            bx_panic("interrupt(): new stack doesn't have room for 10\n");
            /* SS(0) */
            return;
            }

          /* IP must be within CS segment boundaries, else #GP(0) */
          if (gate_descriptor.u.gate286.dest_offset >
              cs_descriptor.u.segment.limit_scaled) {
            bx_panic("interrupt(): IP > CS.limit\n");
            /* GP(0) */
            return;
            }

          old_SP = SP;
          old_SS = BX_CPU_THIS_PTR ss.selector.value;
          old_IP = IP;
          old_CS = BX_CPU_THIS_PTR cs.selector.value;

          /* load new SS:SP values from TSS */
          /* load SS descriptor */
          load_ss(&ss_selector, &ss_descriptor, cs_descriptor.dpl);
          ESP = SP_for_cpl_x; /* ??? */

          /* load new CS:IP values from gate */
          /* load CS descriptor */
          /* set CPL to new code segment DPL */
          /* set RPL of CS to CPL */
          load_cs(&cs_selector, &cs_descriptor, cs_descriptor.dpl);
          BX_CPU_THIS_PTR eip = gate_descriptor.u.gate286.dest_offset;

          /* push long pointer to old stack onto new stack */
          push_16(old_SS);
          push_16(old_SP);

          /* ??? push flags ??? */
          push_16(read_flags());

          /* push return address onto new stack */
          push_16(old_CS);
          push_16(old_IP);
          if ( is_error_code )
            push_16(error_code);

          /* if INTERRUPT GATE set IF to 0 */
          /* set TF to 0 */
          /* set NT to 0 */
          if ( gate_descriptor.type==6 )
            BX_CPU_THIS_PTR eflags.if_ = 0;
          BX_CPU_THIS_PTR eflags.tf = 0;
          BX_CPU_THIS_PTR eflags.nt = 0;

          if (bx_dbg.interrupts)
            bx_printf("interrupt(): INTERRUPT TO INNER PRIVILEGE succeeding\n");
          return;
          }

        /* if code segment is conforming OR code segment DPL = CPL then
         * INTERRUPT TO SAME PRIVILEGE LEVEL: */
        if ( cs_descriptor.u.segment.c_ed==1 || cs_descriptor.dpl==CPL ) {
          int bytes;

          if (bx_dbg.interrupts)
            bx_printf("int_trap_gate286(): INTERRUPT TO SAME PRIVILEGE\n");
      
          /* Current stack limits must allow pushing 6 bytes, else #SS(0)
           * If interrupt was caused by fault with error code the
           * stack limits must allow push of 2 more bytes, else #SS(0) */
          if ( is_error_code )
            bytes = 8;
          else
            bytes = 6;

          if ( !can_push(&BX_CPU_THIS_PTR ss.cache, SP, bytes) ) {
            bx_panic("interrupt(): stack doesn't have room\n");
            exception(BX_SS_EXCEPTION, 0, 0);
            return;
            }

          /* IP must be in CS limit else #GP(0) */
          if (gate_descriptor.u.gate286.dest_offset >
              cs_descriptor.u.segment.limit_scaled) {
            bx_panic("interrupt(): IP > cs descriptor limit\n");
            /* #GP(0) */
            return;
            }

          /* push flags onto stack */
          /* push current CS selector onto stack */
          /* push return offset onto stack */
          push_16(read_flags());
          push_16(BX_CPU_THIS_PTR cs.selector.value);
          push_16(IP);
          if ( is_error_code )
            push_16(error_code);

          /* load CS:IP from gate */
          /* load CS descriptor */
          /* set the RPL field of CS to CPL */
          load_cs(&cs_selector, &cs_descriptor, CPL);
          BX_CPU_THIS_PTR eip = gate_descriptor.u.gate286.dest_offset;

          /* push error code (if any) onto stack */
          /* (do this in segment_exception() ) */

          /* if interrupt gate then set IF to 0 */
          /* set the trap flag to 0 */
          /* set the nested task flag to 0 */
          if ( gate_descriptor.type==6 )
            BX_CPU_THIS_PTR eflags.if_ = 0;
          BX_CPU_THIS_PTR eflags.tf = 0;
          BX_CPU_THIS_PTR eflags.nt = 0;

          if (bx_dbg.interrupts)
            bx_printf("int_trap_gate286(): INTERRUPT TO SAME PRIVILEGE succeeded\n");
          return;
          }

        bx_printf("interrupt: bad descriptor\n");
        /* else #GP(CS selector + ext) */
        bx_printf("c_ed=%u, descriptor.dpl=%u, CPL=%u\n",
          (unsigned) cs_descriptor.u.segment.c_ed,
          (unsigned) cs_descriptor.dpl,
          (unsigned) CPL);
        bx_panic("cs.segment = %u\n", (unsigned) cs_descriptor.segment);
        exception(BX_GP_EXCEPTION, cs_selector.value & 0xfffc, 0);
        return;

        break; 

      case 14: // 386 interrupt gate
      case 15: // 386 trap gate
        // Note: processor leaves v8086 mode as the result of an exception
        //   or interrupt which calls a trap or interrupt gate

        if (bx_dbg.interrupts) {
          if (gate_descriptor.type==14)
            bx_printf("interrupt(): int gate 32\n");
          else
            bx_printf("interrupt(): trap gate 32\n");
          }

        /* TRAP-OR-INT-GATE */

        /* examine CS selector and descriptor given in gate descriptor */
        /* selector must be non-null else #GP(EXT) */
        if ( (gate_descriptor.u.gate386.dest_selector & 0xfffc) == 0 ) {
          bx_panic("int_trap_gate386(): selector null\n");
          /* GP(ext) */
          return;
          }

        parse_selector(gate_descriptor.u.gate386.dest_selector,
                          &cs_selector);

        /* selector must be within its descriptor table limits
         * else #GP(selector+EXT) */
        fetch_raw_descriptor(&cs_selector, &dword1, &dword2,
                                BX_GP_EXCEPTION);
        parse_descriptor(dword1, dword2, &cs_descriptor);

        /* descriptor AR byte must indicate code seg, else #GP(selector+EXT) */
        if ( cs_descriptor.valid==0 ||
             cs_descriptor.segment==0 ||
             cs_descriptor.u.segment.executable==0 ) {
          bx_panic("interrupt(): not code segment\n");
          /* GP(selector + ext) */
          return;
          }

        /* segment must be present, else #NP(selector + EXT) */
        if ( cs_descriptor.p==0 ) {
          bx_panic("interrupt(): segment not present\n");
          /* NP(selector + EXT) */
          return;
          }

        /* if code segment is non-conforming AND DPL < CPL then
         * INTERRUPT TO INNER PRIVILEGE: */
/* ??? use cs_descriptor.dpl or gate.dpl */
        if ( (cs_descriptor.u.segment.c_ed==0) && (cs_descriptor.dpl<CPL) ) {
          Bit16u old_SS, old_CS, SS_for_cpl_x;
          Bit32u old_ESP, old_EIP, ESP_for_cpl_x;
          bx_descriptor_t ss_descriptor;
          bx_selector_t   ss_selector;
          int bytes;

          if (bx_dbg.interrupts)
            bx_printf("interrupt(): INTERRUPT TO INNER PRIVILEGE\n");

          // check selector and descriptor for new stack in current TSS
          if (BX_CPU_THIS_PTR tr.cache.valid==0) {
            bx_panic("interrupt: TR.cache invalid\n");
            return;
            }
          else if (BX_CPU_THIS_PTR tr.cache.type == 9) {
            access_linear(BX_CPU_THIS_PTR tr.cache.u.tss386.base + 8 +
              8*cs_descriptor.dpl, 2, 0, BX_READ, &SS_for_cpl_x);
            access_linear(BX_CPU_THIS_PTR tr.cache.u.tss386.base + 4 +
              8*cs_descriptor.dpl, 4, 0, BX_READ, &ESP_for_cpl_x);
            }
          else { // tr.cache.type == 1
            Bit16u temp16;

            access_linear(BX_CPU_THIS_PTR tr.cache.u.tss286.base + 4 +
              4*cs_descriptor.dpl, 2, 0, BX_READ, &SS_for_cpl_x);
            access_linear(BX_CPU_THIS_PTR tr.cache.u.tss286.base + 2 +
              4*cs_descriptor.dpl, 2, 0, BX_READ, &temp16);
            ESP_for_cpl_x = temp16;
            }

          // Selector must be non-null else #TS(EXT)
          if ( (SS_for_cpl_x & 0xfffc) == 0 ) {
            bx_panic("interrupt(): SS selector null\n");
            /* TS(ext) */
            return;
            }

          // selector index must be within its descriptor table limits
          // else #TS(SS selector + EXT)
          parse_selector(SS_for_cpl_x, &ss_selector);
          /* fetch 2 dwords of descriptor; call handles out of limits checks */
          fetch_raw_descriptor(&ss_selector, &dword1, &dword2,
                                  BX_TS_EXCEPTION);
          parse_descriptor(dword1, dword2, &ss_descriptor);

          // selector rpl must = dpl of code segment,
          // else #TS(SS selector + ext)
          if (ss_selector.rpl != cs_descriptor.dpl) {
            bx_panic("interrupt(): SS.rpl != CS.dpl\n");
            /* TS(SS selector) */
            return;
            }

          /* stack seg DPL must = DPL of code segment,
           * else #TS(SS selector + ext) */
          if (ss_descriptor.dpl != cs_descriptor.dpl) {
            bx_panic("interrupt(): SS.dpl != CS.dpl\n");
            /* TS(SS selector) */
            return;
            }

          /* descriptor must indicate writable data segment,
           * else #TS(SS selector + EXT) */
          if (ss_descriptor.valid==0 ||
              ss_descriptor.segment==0  ||
              ss_descriptor.u.segment.executable==1  ||
              ss_descriptor.u.segment.r_w==0) {
            bx_panic("interrupt(): SS not writable data segment\n");
            /* TS(SS selector) */
            return;
            }

          /* seg must be present, else #SS(SS selector + ext) */
          if (ss_descriptor.p==0) {
            bx_panic("interrupt(): SS not present\n");
            /* SS(SS selector) */
            return;
            }

          /* new stack must have room for 20 bytes, else #SS(0) */
          if ( is_error_code )
            bytes = 24;
          else
            bytes = 20;
          if ( !can_push(&ss_descriptor, ESP_for_cpl_x, bytes) ) {
            bx_panic("interrupt(): new stack doesn't have room for 10\n");
            /* SS(0) */
            return;
            }

          /* EIP must be within CS segment boundaries, else #GP(0) */
          if (gate_descriptor.u.gate386.dest_offset >
              cs_descriptor.u.segment.limit_scaled) {
            bx_panic("interrupt(): EIP > CS.limit\n");
            /* GP(0) */
            return;
            }

          old_ESP = ESP;
          old_SS  = BX_CPU_THIS_PTR ss.selector.value;
          old_EIP = EIP;
          old_CS  = BX_CPU_THIS_PTR cs.selector.value;

          /* load new SS:ESP values from TSS */
          /* load SS descriptor */
          load_ss(&ss_selector, &ss_descriptor, cs_descriptor.dpl);
          ESP = ESP_for_cpl_x;

          /* load new CS:EIP values from gate */
          /* load CS descriptor */
          /* set CPL to new code segment DPL */
          /* set RPL of CS to CPL */
          load_cs(&cs_selector, &cs_descriptor, cs_descriptor.dpl);
          BX_CPU_THIS_PTR eip = gate_descriptor.u.gate386.dest_offset;

          // check for extra stack push-ability???
          if (v8086_mode()) {
            push_32(BX_CPU_THIS_PTR gs.selector.value);
            push_32(BX_CPU_THIS_PTR fs.selector.value);
            push_32(BX_CPU_THIS_PTR ds.selector.value);
            push_32(BX_CPU_THIS_PTR es.selector.value);
            BX_CPU_THIS_PTR gs.cache.valid = 0;
            BX_CPU_THIS_PTR fs.cache.valid = 0;
            BX_CPU_THIS_PTR ds.cache.valid = 0;
            BX_CPU_THIS_PTR es.cache.valid = 0;
            }

          /* push long pointer to old stack onto new stack */
          push_32((Bit32u) old_SS);
          push_32(old_ESP);

          /* ??? push eflags */
          push_32(read_eflags());

          /* push long pointer to return address onto new stack */
          push_32((Bit32u) old_CS);
          push_32(old_EIP);
          if ( is_error_code )
            push_32(error_code);

          /* if INTERRUPT GATE set IF to 0 */
          /* set TF to 0 */
          /* set NT to 0 */
          if ( gate_descriptor.type==14 )
            BX_CPU_THIS_PTR eflags.if_ = 0;
          BX_CPU_THIS_PTR eflags.tf = 0;
          BX_CPU_THIS_PTR eflags.nt = 0;
          BX_CPU_THIS_PTR eflags.vm = 0;
          return;
          }

        /* if code segment is conforming OR code segment DPL = CPL then
         * INTERRUPT TO SAME PRIVILEGE LEVEL: */
        if ( cs_descriptor.u.segment.c_ed==1 || cs_descriptor.dpl==CPL ) {
          int bytes;

          if (bx_dbg.interrupts)
            bx_printf("int_trap_gate386(): INTERRUPT TO SAME PRIVILEGE\n");
      
          /* Current stack limits must allow pushing 10 bytes, else #SS(0)
           * If interrupt was caused by fault with error code the
           * stack limits must allow push of 2 more bytes, else #SS(0) */
          if ( is_error_code )
            bytes = 16;
          else
            bytes = 12;

          if ( !can_push(&BX_CPU_THIS_PTR ss.cache, ESP, bytes) ) {
            bx_panic("interrupt(): stack doesn't have room\n");
            exception(BX_SS_EXCEPTION, 0, 0);
            return;
            }

          /* EIP must be in CS limit else #GP(0) */
          if (gate_descriptor.u.gate386.dest_offset >
              cs_descriptor.u.segment.limit_scaled) {
            bx_panic("interrupt(): EIP > cs descriptor limit\n");
            /* #GP(0) */
            return;
            }

          /* push flags onto stack */
          /* push current CS selector onto stack */
          /* push return offset onto stack */
          push_32(read_eflags());
          push_32((Bit32u) BX_CPU_THIS_PTR cs.selector.value);
          push_32(EIP);
          if ( is_error_code )
            push_32(error_code);

          /* load CS:EIP from gate */
          /* load CS descriptor */
          /* set the RPL field of CS to CPL */
          load_cs(&cs_selector, &cs_descriptor, CPL);
          BX_CPU_THIS_PTR eip = gate_descriptor.u.gate386.dest_offset;

          /* push error code (if any) onto stack */
          /* (do this in segment_exception() ) */

          /* if interrupt gate then set IF to 0 */
          /* set the trap flag to 0 */
          /* set the nested task flag to 0 */
          if ( gate_descriptor.type==14 )
            BX_CPU_THIS_PTR eflags.if_ = 0;
          BX_CPU_THIS_PTR eflags.tf = 0;
          BX_CPU_THIS_PTR eflags.nt = 0;
          BX_CPU_THIS_PTR eflags.vm = 0;
          return;
          }

        /* else #GP(CS selector + ext) */
        bx_printf("interrupt: bad descriptor\n");
        bx_printf("c_ed=%u, descriptor.dpl=%u, CPL=%u\n",
          (unsigned) cs_descriptor.u.segment.c_ed,
          (unsigned) cs_descriptor.dpl,
          (unsigned) CPL);
        bx_panic("cs.segment = %u\n", (unsigned) cs_descriptor.segment);
// should I kick out of vm8086 mode first ???
        exception(BX_GP_EXCEPTION, cs_selector.value & 0xfffc, 0);
        return;
        break; 

      default:
        bx_panic("bad descriptor type in interrupt()!\n");
        break;
      }
    }
  else
#endif
    { /* real mode */
    Bit16u cs_selector, ip;

/* ??? */
if (BX_CPU_THIS_PTR idtr.base != 0) bx_panic("interrupt(real mode) nonzero base\n");
    if ( (vector*4+3) > BX_CPU_THIS_PTR idtr.limit )
      bx_panic("interrupt(real mode) vector > limit\n");

    push_16(read_flags());

    cs_selector = BX_CPU_THIS_PTR cs.selector.value;
    push_16(cs_selector);
    ip = BX_CPU_THIS_PTR eip;
    push_16(ip);

    access_linear(BX_CPU_THIS_PTR idtr.base + 4 * vector,     2, 0, BX_READ, &ip);
    IP = ip;
    access_linear(BX_CPU_THIS_PTR idtr.base + 4 * vector + 2, 2, 0, BX_READ, &cs_selector);
    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_selector);

    /* INT affects the following flags: I,T */
    BX_CPU_THIS_PTR eflags.if_ = 0;
    BX_CPU_THIS_PTR eflags.tf  = 0;
#if BX_CPU >= 4
    BX_CPU_THIS_PTR eflags.ac  = 0;
#endif
    }
}




  void
bx_cpu_c::exception(unsigned vector, Bit16u error_code, Boolean is_INT)
  /* vector:     0..255: vector in IDT
   * error_code: if exception generates and error, push this error code
   */
{
  Boolean  push_error;
  Bit8u    exception_type;
  unsigned prev_errno;

  invalidate_prefetch_q();

  UNUSED(is_INT);
  BX_INSTR_EXCEPTION(vector);

  if (bx_dbg.exceptions)
    bx_printf("exception(%02x h)\n", (unsigned) vector);

  // if not initial error, restore previous register values from
  // previous attempt to handle exception
  if (BX_CPU_THIS_PTR errorno) {
    BX_CPU_THIS_PTR cs  = BX_CPU_THIS_PTR save_cs;
    BX_CPU_THIS_PTR ss  = BX_CPU_THIS_PTR save_ss;
    EIP = BX_CPU_THIS_PTR save_eip;
    ESP = BX_CPU_THIS_PTR save_esp;
    }

  BX_CPU_THIS_PTR errorno++;
  if (BX_CPU_THIS_PTR errorno >= 3) {
    bx_panic("exception(): 3rd exception with no resolution\n");
    }

  /* careful not to get here with curr_exception[1]==DOUBLE_FAULT */
  /* ...index on DOUBLE_FAULT below, will be out of bounds */

  /* if 1st was a double fault (software INT?), then shutdown */
  if ( (BX_CPU_THIS_PTR errorno==2) && (BX_CPU_THIS_PTR curr_exception[0]==BX_ET_DOUBLE_FAULT) ) {
    bx_panic("exception(): tripple fault encountered\n");
    }

  /* ??? this is not totally correct, should be done depending on
   * vector */
  /* backup IP to value before error occurred */
  EIP = BX_CPU_THIS_PTR prev_eip;
  ESP = BX_CPU_THIS_PTR prev_esp;


  switch (vector) {
    case  0: /* DIV by 0 */
      push_error = 0;
      exception_type = BX_ET_CONTRIBUTORY;
      break;
    case  1: /* debug exceptions */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
    case  2: /* NMI */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
    case  3: /* breakpoint */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
    case  4: /* overflow */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
    case  5: /* bounds check */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
    case  6: /* invalid opcode */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
    case  7: /* device not available */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
    case  8: /* double fault */
      push_error = 1;
      exception_type = BX_ET_DOUBLE_FAULT;
      break;
    case  9: /* coprocessor segment overrun (286,386 only) */
      push_error = 0;
      exception_type = BX_ET_CONTRIBUTORY;
      bx_panic("exception(9): unfinished\n");
      break;
    case 10: /* invalid TSS */
      push_error = 1;
      exception_type = BX_ET_CONTRIBUTORY;
      error_code = (error_code & 0xfffe) | BX_CPU_THIS_PTR EXT;
      break;
    case 11: /* segment not present */
      push_error = 1;
      exception_type = BX_ET_CONTRIBUTORY;
      error_code = (error_code & 0xfffe) | BX_CPU_THIS_PTR EXT;
      break;
    case 12: /* stack fault */
      push_error = 1;
      exception_type = BX_ET_CONTRIBUTORY;
      error_code = (error_code & 0xfffe) | BX_CPU_THIS_PTR EXT;
      break;
    case 13: /* general protection */
      push_error = 1;
      exception_type = BX_ET_CONTRIBUTORY;
      error_code = (error_code & 0xfffe) | BX_CPU_THIS_PTR EXT;
//bx_printf("GP fault CS:EIP=%04x:%08x\n",
//  (unsigned) bx_cpu.cs.selector.value,
//  (unsigned) bx_cpu.eip
//  );
      break;
    case 14: /* page fault */
      push_error = 1;
      exception_type = BX_ET_PAGE_FAULT;
      /* ??? special format error returned */
      break;
    case 15: /* reserved */
      bx_panic("exception(15): reserved\n");
      push_error = 0;     /* keep compiler happy for now */
      exception_type = 0; /* keep compiler happy for now */
      break;
    case 16: /* floating-point error */
      push_error = 0;
      exception_type = BX_ET_BENIGN;
      break;
#if BX_CPU >= 4
    case 17: /* alignment check */
      bx_panic("exception(): alignment-check, vector 17 unimplemented\n");
      push_error = 0;     /* keep compiler happy for now */
      exception_type = 0; /* keep compiler happy for now */
      break;
#endif
#if BX_CPU >= 5
    case 18: /* machine check */
      bx_panic("exception(): machine-check, vector 18 unimplemented\n");
      push_error = 0;     /* keep compiler happy for now */
      exception_type = 0; /* keep compiler happy for now */
      break;
#endif
    default:
      bx_panic("exception(%u): bad vector\n", (unsigned) vector);
      push_error = 0;     /* keep compiler happy for now */
      exception_type = 0; /* keep compiler happy for now */
      break;
    }

  if (exception_type != BX_ET_PAGE_FAULT) {
    // Page faults have different format
    error_code = (error_code & 0xfffe) | BX_CPU_THIS_PTR EXT;
    }
  BX_CPU_THIS_PTR EXT = 1;

  /* if we've already had 1st exception, see if 2nd causes a
   * Double Fault instead.  Otherwise, just record 1st exception
   */
  if (BX_CPU_THIS_PTR errorno >= 2) {
    if (is_exception_OK[BX_CPU_THIS_PTR curr_exception[0]][exception_type])
      BX_CPU_THIS_PTR curr_exception[1] = exception_type;
    else
      BX_CPU_THIS_PTR curr_exception[1] = BX_ET_DOUBLE_FAULT;
    }
  else {
    BX_CPU_THIS_PTR curr_exception[0] = exception_type;
    }


#if BX_CPU >= 2
  if (!real_mode()) {
    prev_errno = BX_CPU_THIS_PTR errorno;
    BX_CPU_THIS_PTR interrupt(vector, 0, push_error, error_code);
//    if (BX_CPU_THIS_PTR errorno > prev_errno) {
//      bx_printf("segment_exception(): errorno changed\n");
//      longjmp(jmp_buf_env, 1); // go back to main decode loop
//      return;
//      }

//    if (push_error) {
//      /* push error code on stack, after handling interrupt */
//      /* pushed as a word or dword depending upon default size ??? */
//      if (ss.cache.u.segment.d_b)
//        push_32((Bit32u) error_code); /* upper bits reserved */
//      else
//        push_16(error_code);
//      if (BX_CPU_THIS_PTR errorno > prev_errno) {
//        bx_panic("segment_exception(): errorno changed\n");
//        return;
//        }
//      }
    BX_CPU_THIS_PTR errorno = 0; // error resolved
    longjmp(BX_CPU_THIS_PTR jmp_buf_env, 1); // go back to main decode loop
    }
  else /* real mode */
#endif
    {
/* ??? */
    if (vector!=5 && vector!=6 && vector!=7 && vector!=13)
      bx_panic("exception %u occurred in REAL mode\n", (unsigned) vector);
    BX_CPU_THIS_PTR interrupt(vector, 0, 0, 0); /* not INT, no error code pushed */
    BX_CPU_THIS_PTR errorno = 0; // error resolved
    longjmp(BX_CPU_THIS_PTR jmp_buf_env, 1); // go back to main decode loop
    }
}


  int
bx_cpu_c::int_number(bx_segment_reg_t *seg)
{
  if (seg == &BX_CPU_THIS_PTR ss)
    return(BX_SS_EXCEPTION);
  else
    return(BX_GP_EXCEPTION);
}

  void
bx_cpu_c::shutdown_cpu(void)
{
  Bit32u  dword1, dword2;
  bx_descriptor_t gate_descriptor;
  /*bx_selector_t cs_selector;*/
  int vector;
  Bit16u ip, cs;
  
#if BX_CPU > 2
  bx_panic("shutdown_cpu(): not implemented for 386\n");
#endif

  invalidate_prefetch_q();
  bx_panic("shutdown_cpu(): not finished\n");

  if (real_mode()) {
    bx_panic("shutdown_cpu(): don't know how to handle real mode\n");
    }

    vector = 2;
    bx_mem.read_physical(4 * vector,     2, &ip);
    bx_mem.read_physical(4 * vector + 2, 2, &cs);
    bx_printf("real mode vector 2 to CS:IP = %04x:%04x\n",
      (unsigned) cs, (unsigned) ip);

    vector = 8;
    bx_mem.read_physical(4 * vector,     2, &ip);
    bx_mem.read_physical(4 * vector + 2, 2, &cs);
    bx_printf("real mode vector 8 to CS:IP = %04x:%04x\n",
      (unsigned) cs, (unsigned) ip);

    vector = 2;
    /* descriptor AR byte must indicate interrupt gate, trap gate,
       or task gate, else #GP(vector*8 + 2 + EXT)*/
    bx_mem.read_physical(BX_CPU_THIS_PTR idtr.base + vector*8,     4, &dword1);
    bx_mem.read_physical(BX_CPU_THIS_PTR idtr.base + vector*8 + 4, 4, &dword2);

    parse_descriptor(dword1, dword2, &gate_descriptor);

bx_printf("gate.segment = %u\n", (unsigned) gate_descriptor.segment);
bx_printf("gate.type    = %u\n", (unsigned) gate_descriptor.type);
  bx_panic("shutting down CPU due to unrecoverable errors\n");
}
