// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com






#include "bochs.h"





  INLINE void
bx_cpu_c::SAHF()
{
  set_SF((AH & 0x80) >> 7);
  set_ZF((AH & 0x40) >> 6);
  set_AF((AH & 0x10) >> 4);
  set_CF(AH & 0x01);
  set_PF((AH & 0x04) >> 2);
}

  INLINE void
bx_cpu_c::LAHF()
{
  AH = (get_SF() ? 0x80 : 0) |
       (get_ZF() ? 0x40 : 0) |
       (get_AF() ? 0x10 : 0) |
       (get_PF() ? 0x04 : 0) |
       (0x02) |
       (get_CF() ? 0x01 : 0);
}

  INLINE void
bx_cpu_c::CLC()
{
  set_CF(0);
}

  INLINE void
bx_cpu_c::STC()
{
  set_CF(1);
}

  INLINE void
bx_cpu_c::CLI()
{
#if BX_CPU >= 2
  if (protected_mode()) {
    if (CPL > IOPL) {
      //bx_printf("CLI: CPL > IOPL\n"); /* ??? */
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
#if BX_CPU >= 3
  else if (v8086_mode()) {
    if (IOPL != 3) {
      //bx_printf("CLI: IOPL != 3\n"); /* ??? */
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
#endif
#endif

  BX_CPU_THIS_PTR eflags.if_ = 0; 
}

  INLINE void
bx_cpu_c::STI()
{
#if BX_CPU >= 2
  if (protected_mode()) {
    if (CPL > IOPL) {
      //bx_printf("STI: CPL > IOPL\n"); /* ??? */
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
#if BX_CPU >= 3
  else if (v8086_mode()) {
    if (IOPL != 3) {
      //bx_printf("STI: IOPL != 3\n"); /* ??? */
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
#endif
#endif

  if (!BX_CPU_THIS_PTR eflags.if_) {
    BX_CPU_THIS_PTR eflags.if_ = 1;
    BX_CPU_THIS_PTR inhibit_interrupts = 1;
    BX_CPU_THIS_PTR async_event = 1;
    }
}

  INLINE void
bx_cpu_c::CLD()
{
  BX_CPU_THIS_PTR eflags.df = 0;
}

  INLINE void
bx_cpu_c::STD()
{
  BX_CPU_THIS_PTR eflags.df = 1;
}

  INLINE void
bx_cpu_c::CMC()
{
  set_CF( !get_CF() );
}

  INLINE void
bx_cpu_c::PUSHF_Fv()
{
  if (v8086_mode() && (IOPL<3)) {
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    push_32(read_eflags() & 0x00fcffff);
    /* will return after error anyway */
    }
  else
#endif
    {
    push_16(read_flags());
    /* will return after error anyway */
    }
}


  INLINE void
bx_cpu_c::POPF_Fv()
{

#if BX_CPU >= 3
  if (v8086_mode()) {
    if (IOPL < 3) {
      //bx_printf("popf_fv: IOPL < 3\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      bx_panic("POPFD(): not supported in virtual mode\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eflags;

    pop_32(&eflags);

    eflags &= 0x00277fd7;
    if (!real_mode()) {
      write_eflags(eflags, /*IOPL*/ CPL==0, /*IF*/ CPL<=IOPL, 0, 0);
      }
    else { /* real mode */
      write_eflags(eflags, 1, 1, 0, 0);
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u flags;

    pop_16(&flags);

    if (!real_mode()) {
      write_flags(flags, /*IOPL*/ CPL==0, /*IF*/ CPL<=IOPL);
      }
    else { /* real mode */
      write_flags(flags, 1, 1);
      }
    }
}


  INLINE void
bx_cpu_c::SALC()
{
  if ( get_CF() ) {
    AL = 0xff;
    }
  else {
    AL = 0x00;
    }
}
