// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com






#include "bochs.h"






  void
bx_cpu_c::write_flags(Bit16u flags, Boolean change_IOPL, Boolean change_IF)
{
  BX_CPU_THIS_PTR set_CF(flags & 0x01);
  BX_CPU_THIS_PTR set_PF((flags >> 2) & 0x01);
  BX_CPU_THIS_PTR set_AF((flags >> 4) & 0x01);
  BX_CPU_THIS_PTR set_ZF((flags >> 6) & 0x01);
  BX_CPU_THIS_PTR set_SF((flags >> 7) & 0x01);
  BX_CPU_THIS_PTR eflags.tf = (flags >> 8) & 0x01;
  if (BX_CPU_THIS_PTR eflags.tf) {
    BX_CPU_THIS_PTR async_event = 1;
    bx_panic("write_flags(): TF set\n");
    }

  if (change_IF)
    BX_CPU_THIS_PTR eflags.if_ = (flags >> 9) & 0x01;

  BX_CPU_THIS_PTR eflags.df = (flags >> 10) & 0x01;
  BX_CPU_THIS_PTR set_OF((flags >> 11) & 0x01);

  /* real-mode doesn't change IOPL & NT flags ??? */
  if (protected_mode()) {
    if (change_IOPL)
      BX_CPU_THIS_PTR eflags.iopl = (flags >> 12) & 0x03;
    BX_CPU_THIS_PTR eflags.nt = (flags >> 14) & 0x01;
    }
  else { /* real mode */
    /* 286: bits 12-15 always clear
     * 386: bit 15 always clear, bits 14-12 maintain last value loaded
     */
#if BX_CPU == 2
    BX_CPU_THIS_PTR eflags.iopl = 0;
    BX_CPU_THIS_PTR eflags.nt = 0;
#elif BX_CPU >= 3
    /* values shouldn't be used in real mode, but are loaded into eflags */
    BX_CPU_THIS_PTR eflags.iopl = (flags >> 12) & 0x03;
    BX_CPU_THIS_PTR eflags.nt = (flags >> 14) & 0x01;
#endif
    }
}


#if BX_CPU >= 3
  void
bx_cpu_c::write_eflags(Bit32u eflags_raw, Boolean change_IOPL, Boolean change_IF,
                Boolean change_VM, Boolean change_RF)
{
  BX_CPU_THIS_PTR set_CF(eflags_raw & 0x01);
  BX_CPU_THIS_PTR set_PF((eflags_raw >> 2) & 0x01);
  BX_CPU_THIS_PTR set_AF((eflags_raw >> 4) & 0x01);
  BX_CPU_THIS_PTR set_ZF((eflags_raw >> 6) & 0x01);
  BX_CPU_THIS_PTR set_SF((eflags_raw >> 7) & 0x01);
  BX_CPU_THIS_PTR eflags.tf = (eflags_raw >> 8) & 0x01;
  if (BX_CPU_THIS_PTR eflags.tf) {
    BX_CPU_THIS_PTR async_event = 1;
    bx_panic("write_eflags(): TF set\n");
    }

  if (change_IF)
    BX_CPU_THIS_PTR eflags.if_ = (eflags_raw >> 9) & 0x01;

  BX_CPU_THIS_PTR eflags.df = (eflags_raw >> 10) & 0x01;
  BX_CPU_THIS_PTR set_OF((eflags_raw >> 11) & 0x01);

  /* real-mode doesn't change IOPL & NT flags ??? */
  if (protected_mode()) {
    if (change_IOPL)
      BX_CPU_THIS_PTR eflags.iopl = (eflags_raw >> 12) & 0x03;
    BX_CPU_THIS_PTR eflags.nt = (eflags_raw >> 14) & 0x01;
    }
  else { /* real mode */
    /* 386+: bit 15 always clear, bits 14-12 maintain last value loaded
     */
    /* values shouldn't be used in real mode, but are loaded into eflags */
    BX_CPU_THIS_PTR eflags.iopl = (eflags_raw >> 12) & 0x03;
    BX_CPU_THIS_PTR eflags.nt = (eflags_raw >> 14) & 0x01;
    }

  if (change_VM) {
    BX_CPU_THIS_PTR eflags.vm = (eflags_raw >> 17) & 0x01;
#if BX_SUPPORT_V8086_MODE == 0
    if (BX_CPU_THIS_PTR eflags.vm)
      bx_panic("write_eflags: VM bit set: BX_SUPPORT_V8086_MODE==0\n");
#endif
    }
  if (change_RF) {
    BX_CPU_THIS_PTR eflags.rf = (eflags_raw >> 16) & 0x01;
    }
}
#endif /* BX_CPU >= 3 */


  Bit16u
bx_cpu_c::read_flags(void)
{
  Bit16u flags;

  flags = (get_CF()) |
          (BX_CPU_THIS_PTR eflags.bit1 << 1) |
          ((get_PF()) << 2) |
          (BX_CPU_THIS_PTR eflags.bit3 << 3) |
          ((get_AF()>0) << 4) |
          (BX_CPU_THIS_PTR eflags.bit5 << 5) |
          ((get_ZF()>0) << 6) |
          ((get_SF()>0) << 7) |
          (BX_CPU_THIS_PTR eflags.tf << 8) |
          (BX_CPU_THIS_PTR eflags.if_ << 9) |
          (BX_CPU_THIS_PTR eflags.df << 10) |
          ((get_OF()>0) << 11) |
          (BX_CPU_THIS_PTR eflags.iopl << 12) |
          (BX_CPU_THIS_PTR eflags.nt << 14) |
          (BX_CPU_THIS_PTR eflags.bit15 << 15);

  /* 8086: bits 12-15 always set to 1.
   * 286: in real mode, bits 12-15 always cleared.
   * 386+: real-mode: bit15 cleared, bits 14..12 are last loaded value
   *       protected-mode: bit 15 clear, bit 14 = last loaded, IOPL?
   */
#if BX_CPU < 2
  flags |= 0xF000;  /* 8086 nature */
#elif BX_CPU == 2
  if (real_mode()) {
    flags &= 0x0FFF;  /* 80286 in real mode nature */
    }
#else /* 386+ */
  if (real_mode()) {
    flags &= 0x7FFF;  /* 80386 in real mode nature */
    }
#endif

  return(flags);
}


#if BX_CPU >= 3
  Bit32u
bx_cpu_c::read_eflags(void)
{
  Bit32u eflags_raw;

#if BX_CPU >= 4
  bx_panic("read_eflags(): not implemented for 80486+\n");
#endif

  eflags_raw =
          (get_CF()) |
          (BX_CPU_THIS_PTR eflags.bit1 << 1) |
          ((get_PF()) << 2) |
          (BX_CPU_THIS_PTR eflags.bit3 << 3) |
          ((get_AF()>0) << 4) |
          (BX_CPU_THIS_PTR eflags.bit5 << 5) |
          ((get_ZF()>0) << 6) |
          ((get_SF()>0) << 7) |
          (BX_CPU_THIS_PTR eflags.tf << 8) |
          (BX_CPU_THIS_PTR eflags.if_ << 9) |
          (BX_CPU_THIS_PTR eflags.df << 10) |
          ((get_OF()>0) << 11) |
          (BX_CPU_THIS_PTR eflags.iopl << 12) |
          (BX_CPU_THIS_PTR eflags.nt << 14) |
          (BX_CPU_THIS_PTR eflags.bit15 << 15) |
          (BX_CPU_THIS_PTR eflags.rf << 16) |
          (BX_CPU_THIS_PTR eflags.vm << 17);

  /*
   * 386+: real-mode: bit15 cleared, bits 14..12 are last loaded value
   *       protected-mode: bit 15 clear, bit 14 = last loaded, IOPL?
   */
  if (real_mode()) {
    eflags_raw &= 0x7FFF;  /* 80386 in real mode nature */
    }

  return(eflags_raw);
}
#endif /* BX_CPU >= 3 */
