// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
//
//     EMail:    bochs@world.std.com


#include "bochs.h"


/* the device id and stepping id are loaded into DH & DL upon processor
   startup.  for device id: 3 = 80386, 4 = 80486.  just make up a
   number for the stepping (revision) id. */
#define BX_DEVICE_ID     3
#define BX_STEPPING_ID   0



bx_cpu_c::bx_cpu_c(void)
{
  // bx_cpu_c constructor

  fprintf(stderr, "#bx_cpu_c::bx_cpu_c(void) called\n");

  /* hack for the following fields.  Its easier to decode mod-rm bytes if
     you can assume there's always a base & index register used.  For
     modes which don't really use them, point to an empty (zeroed) register.
   */
  empty_register = 0;

  // 16bit address mode base register, used for mod-rm decoding
#if 0
  _16bit_base_reg[0] = (Bit16u *) (((Bit8u *) &ebx) + BX_REG16_OFFSET);
  _16bit_base_reg[1] = (Bit16u *) (((Bit8u *) &ebx) + BX_REG16_OFFSET);
  _16bit_base_reg[2] = (Bit16u *) (((Bit8u *) &ebp) + BX_REG16_OFFSET);
  _16bit_base_reg[3] = (Bit16u *) (((Bit8u *) &ebp) + BX_REG16_OFFSET);
  _16bit_base_reg[4] = (Bit16u *) (((Bit8u *) &empty_register) + BX_REG16_OFFSET);
  _16bit_base_reg[5] = (Bit16u *) (((Bit8u *) &empty_register) + BX_REG16_OFFSET);
  _16bit_base_reg[6] = (Bit16u *) (((Bit8u *) &ebp) + BX_REG16_OFFSET);
  _16bit_base_reg[7] = (Bit16u *) (((Bit8u *) &ebx) + BX_REG16_OFFSET);
#endif

  _16bit_base_reg[0] = &gen_reg[BX_16BIT_REG_BX].word.rx;
  _16bit_base_reg[1] = &gen_reg[BX_16BIT_REG_BX].word.rx;
  _16bit_base_reg[2] = &gen_reg[BX_16BIT_REG_BP].word.rx;
  _16bit_base_reg[3] = &gen_reg[BX_16BIT_REG_BP].word.rx;
  _16bit_base_reg[4] = (Bit16u*) &empty_register;
  _16bit_base_reg[5] = (Bit16u*) &empty_register;
  _16bit_base_reg[6] = &gen_reg[BX_16BIT_REG_BP].word.rx;
  _16bit_base_reg[7] = &gen_reg[BX_16BIT_REG_BX].word.rx;

  // 16bit address mode index register, used for mod-rm decoding
#if 0
  _16bit_index_reg[0] = (Bit16u *) (((Bit8u *) &esi) + BX_REG16_OFFSET);
  _16bit_index_reg[1] = (Bit16u *) (((Bit8u *) &edi) + BX_REG16_OFFSET);
  _16bit_index_reg[2] = (Bit16u *) (((Bit8u *) &esi) + BX_REG16_OFFSET);
  _16bit_index_reg[3] = (Bit16u *) (((Bit8u *) &edi) + BX_REG16_OFFSET);
  _16bit_index_reg[4] = (Bit16u *) (((Bit8u *) &esi) + BX_REG16_OFFSET);
  _16bit_index_reg[5] = (Bit16u *) (((Bit8u *) &edi) + BX_REG16_OFFSET);
  _16bit_index_reg[6] = (Bit16u *) (((Bit8u *) &empty_register) + BX_REG16_OFFSET);
  _16bit_index_reg[7] = (Bit16u *) (((Bit8u *) &empty_register) + BX_REG16_OFFSET);
#endif
  _16bit_index_reg[0] = &gen_reg[BX_16BIT_REG_SI].word.rx;
  _16bit_index_reg[1] = &gen_reg[BX_16BIT_REG_DI].word.rx;
  _16bit_index_reg[2] = &gen_reg[BX_16BIT_REG_SI].word.rx;
  _16bit_index_reg[3] = &gen_reg[BX_16BIT_REG_DI].word.rx;
  _16bit_index_reg[4] = &gen_reg[BX_16BIT_REG_SI].word.rx;
  _16bit_index_reg[5] = &gen_reg[BX_16BIT_REG_DI].word.rx;
  _16bit_index_reg[6] = (Bit16u*) &empty_register;
  _16bit_index_reg[7] = (Bit16u*) &empty_register;

  // for decoding instructions: access to seg reg's via index number
  sreg_mod00_rm16[0] = &ds;
  sreg_mod00_rm16[1] = &ds;
  sreg_mod00_rm16[2] = &ss;
  sreg_mod00_rm16[3] = &ss;
  sreg_mod00_rm16[4] = &ds;
  sreg_mod00_rm16[5] = &ds;
  sreg_mod00_rm16[6] = &ds;
  sreg_mod00_rm16[7] = &ds;

  sreg_mod01_rm16[0] = &ds;
  sreg_mod01_rm16[1] = &ds;
  sreg_mod01_rm16[2] = &ss;
  sreg_mod01_rm16[3] = &ss;
  sreg_mod01_rm16[4] = &ds;
  sreg_mod01_rm16[5] = &ds;
  sreg_mod01_rm16[6] = &ss;
  sreg_mod01_rm16[7] = &ds;

  sreg_mod10_rm16[0] = &ds;
  sreg_mod10_rm16[1] = &ds;
  sreg_mod10_rm16[2] = &ss;
  sreg_mod10_rm16[3] = &ss;
  sreg_mod10_rm16[4] = &ds;
  sreg_mod10_rm16[5] = &ds;
  sreg_mod10_rm16[6] = &ss;
  sreg_mod10_rm16[7] = &ds;

  sreg_mod00_rm32[0] = &ds;
  sreg_mod00_rm32[1] = &ds;
  sreg_mod00_rm32[2] = &ds;
  sreg_mod00_rm32[3] = &ds;
  sreg_mod00_rm32[4] = NULL; // this entry should never be accessed
                           // (escape to 2-byte)
  sreg_mod00_rm32[5] = &ds;
  sreg_mod00_rm32[6] = &ds;
  sreg_mod00_rm32[7] = &ds;

  // the default segment to use for a one-byte modrm with mod==01b
  // and rm==i
  //
  sreg_mod01_rm32[0] = &ds;
  sreg_mod01_rm32[1] = &ds;
  sreg_mod01_rm32[2] = &ds;
  sreg_mod01_rm32[3] = &ds;
  sreg_mod01_rm32[4] = NULL; // this entry should never be accessed
                           // (escape to 2-byte)
  sreg_mod01_rm32[5] = &ss;
  sreg_mod01_rm32[6] = &ds;
  sreg_mod01_rm32[7] = &ds;

  // the default segment to use for a one-byte modrm with mod==10b
  // and rm==i
  //
  sreg_mod10_rm32[0] = &ds;
  sreg_mod10_rm32[1] = &ds;
  sreg_mod10_rm32[2] = &ds;
  sreg_mod10_rm32[3] = &ds;
  sreg_mod10_rm32[4] = NULL;  // this entry should never be accessed
                            // (escape to 2-byte)
  sreg_mod10_rm32[5] = &ss;
  sreg_mod10_rm32[6] = &ds;
  sreg_mod10_rm32[7] = &ds;


  // the default segment to use for a two-byte modrm with mod==00b
  // and base==i
  //
  sreg_mod00_base32[0] = &ds;
  sreg_mod00_base32[1] = &ds;
  sreg_mod00_base32[2] = &ds;
  sreg_mod00_base32[3] = &ds;
  sreg_mod00_base32[4] = &ss;
  sreg_mod00_base32[5] = &ds;
  sreg_mod00_base32[6] = &ds;
  sreg_mod00_base32[7] = &ds;

  // the default segment to use for a two-byte modrm with mod==01b
  // and base==i
  //
  sreg_mod01_base32[0] = &ds;
  sreg_mod01_base32[1] = &ds;
  sreg_mod01_base32[2] = &ds;
  sreg_mod01_base32[3] = &ds;
  sreg_mod01_base32[4] = &ss;
  sreg_mod01_base32[5] = &ss;
  sreg_mod01_base32[6] = &ds;
  sreg_mod01_base32[7] = &ds;

  // the default segment to use for a two-byte modrm with mod==10b
  // and base==i
  //
  sreg_mod10_base32[0] = &ds;
  sreg_mod10_base32[1] = &ds;
  sreg_mod10_base32[2] = &ds;
  sreg_mod10_base32[3] = &ds;
  sreg_mod10_base32[4] = &ss;
  sreg_mod10_base32[5] = &ss;
  sreg_mod10_base32[6] = &ds;
  sreg_mod10_base32[7] = &ds;
  BX_INSTR_INIT();
}


bx_cpu_c::~bx_cpu_c(void)
{
  fprintf(stderr, "#bx_cpu_c::~bx_cpu_c(void) called\n");
  BX_INSTR_SHUTDOWN();
}


  void
bx_cpu_c::init_cpu(void)
{
  /* general registers */
  EAX = 0; /* processor passed test :-) */
  EBX = 0; /* undefined */
  ECX = 0; /* undefined */
  EDX = (BX_DEVICE_ID << 8) | BX_STEPPING_ID; // ???
  EBP = 0; /* undefined */
  ESI = 0; /* undefined */
  EDI = 0; /* undefined */
  ESP = 0; /* undefined */

  /* all status flags at known values, use BX_CPU_THIS_PTR eflags structure */
  BX_CPU_THIS_PTR lf_flags_status = 0x000000;
  BX_CPU_THIS_PTR lf_pf = 0;

  /* status and control flags register set */
  BX_CPU_THIS_PTR set_CF(0);
  BX_CPU_THIS_PTR eflags.bit1 = 1;
  BX_CPU_THIS_PTR set_PF(0);
  BX_CPU_THIS_PTR eflags.bit3 = 0;
  BX_CPU_THIS_PTR set_AF(0);
  BX_CPU_THIS_PTR eflags.bit5 = 0;
  BX_CPU_THIS_PTR set_ZF(0);
  BX_CPU_THIS_PTR set_SF(0);
  BX_CPU_THIS_PTR eflags.tf = 0;
  BX_CPU_THIS_PTR eflags.if_ = 0;
  BX_CPU_THIS_PTR eflags.df = 0;
  BX_CPU_THIS_PTR set_OF(0);
#if BX_CPU >= 2
  BX_CPU_THIS_PTR eflags.iopl = 0;
  BX_CPU_THIS_PTR eflags.nt = 0;
#endif
  BX_CPU_THIS_PTR eflags.bit15 = 0;
#if BX_CPU >= 3
  BX_CPU_THIS_PTR eflags.rf = 0;
  BX_CPU_THIS_PTR eflags.vm = 0;
#endif
#if BX_CPU >= 4
  BX_CPU_THIS_PTR eflags.ac = 0;
#endif

  BX_CPU_THIS_PTR inhibit_interrupts = 0;
  BX_CPU_THIS_PTR single_step_event = 0;

  /* instruction pointer */
#if BX_CPU < 2
  BX_CPU_THIS_PTR eip = 0x00000000;
#else /* from 286 up */
  BX_CPU_THIS_PTR eip = 0x0000FFF0;
#endif


  /* CS (Code Segment) and descriptor cache */
  /* Note: on a real cpu, CS initially points to upper memory.  After
   * the 1st jump, the descriptor base is zero'd out.  Since I'm just
   * going to jump to my BIOS, I don't need to do this.
   * For future reference:
   *   processor  cs.selector   cs.base    cs.limit    EIP
   *        8086    FFFF          FFFF0        FFFF   0000
   *        286     F000         FF0000        FFFF   FFF0
   *        386+    F000       FFFF0000        FFFF   FFF0
   */
  BX_CPU_THIS_PTR cs.selector.value =     0xf000;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR cs.selector.index =     0x0000;
  BX_CPU_THIS_PTR cs.selector.ti = 0;
  BX_CPU_THIS_PTR cs.selector.rpl = 0;

  BX_CPU_THIS_PTR cs.cache.valid =     1;
  BX_CPU_THIS_PTR cs.cache.p = 1;
  BX_CPU_THIS_PTR cs.cache.dpl = 0;
  BX_CPU_THIS_PTR cs.cache.segment = 1; /* data/code segment */
  BX_CPU_THIS_PTR cs.cache.type = 3; /* read/write access */

  BX_CPU_THIS_PTR cs.cache.u.segment.executable   = 1; /* data/stack segment */
  BX_CPU_THIS_PTR cs.cache.u.segment.c_ed         = 0; /* normal expand up */
  BX_CPU_THIS_PTR cs.cache.u.segment.r_w          = 1; /* writeable */
  BX_CPU_THIS_PTR cs.cache.u.segment.a            = 1; /* accessed */
  BX_CPU_THIS_PTR cs.cache.u.segment.base         = 0x000F0000;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit        =     0xFFFF;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled =     0xFFFF;
#endif
#if BX_CPU >= 3
  BX_CPU_THIS_PTR cs.cache.u.segment.g   = 0; /* byte granular */
  BX_CPU_THIS_PTR cs.cache.u.segment.d_b = 0; /* 16bit default size */
  BX_CPU_THIS_PTR cs.cache.u.segment.avl = 0;
#endif


  /* SS (Stack Segment) and descriptor cache */
  BX_CPU_THIS_PTR ss.selector.value =     0x0000;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR ss.selector.index =     0x0000;
  BX_CPU_THIS_PTR ss.selector.ti = 0;
  BX_CPU_THIS_PTR ss.selector.rpl = 0;

  BX_CPU_THIS_PTR ss.cache.valid =     1;
  BX_CPU_THIS_PTR ss.cache.p = 1;
  BX_CPU_THIS_PTR ss.cache.dpl = 0;
  BX_CPU_THIS_PTR ss.cache.segment = 1; /* data/code segment */
  BX_CPU_THIS_PTR ss.cache.type = 3; /* read/write access */

  BX_CPU_THIS_PTR ss.cache.u.segment.executable   = 0; /* data/stack segment */
  BX_CPU_THIS_PTR ss.cache.u.segment.c_ed         = 0; /* normal expand up */
  BX_CPU_THIS_PTR ss.cache.u.segment.r_w          = 1; /* writeable */
  BX_CPU_THIS_PTR ss.cache.u.segment.a            = 1; /* accessed */
  BX_CPU_THIS_PTR ss.cache.u.segment.base         = 0x00000000;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit        =     0xFFFF;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled =     0xFFFF;
#endif
#if BX_CPU >= 3
  BX_CPU_THIS_PTR ss.cache.u.segment.g   = 0; /* byte granular */
  BX_CPU_THIS_PTR ss.cache.u.segment.d_b = 0; /* 16bit default size */
  BX_CPU_THIS_PTR ss.cache.u.segment.avl = 0;
#endif


  /* DS (Data Segment) and descriptor cache */
  BX_CPU_THIS_PTR ds.selector.value =     0x0000;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR ds.selector.index =     0x0000;
  BX_CPU_THIS_PTR ds.selector.ti = 0;
  BX_CPU_THIS_PTR ds.selector.rpl = 0;

  BX_CPU_THIS_PTR ds.cache.valid =     1;
  BX_CPU_THIS_PTR ds.cache.p = 1;
  BX_CPU_THIS_PTR ds.cache.dpl = 0;
  BX_CPU_THIS_PTR ds.cache.segment = 1; /* data/code segment */
  BX_CPU_THIS_PTR ds.cache.type = 3; /* read/write access */

  BX_CPU_THIS_PTR ds.cache.u.segment.executable   = 0; /* data/stack segment */
  BX_CPU_THIS_PTR ds.cache.u.segment.c_ed         = 0; /* normal expand up */
  BX_CPU_THIS_PTR ds.cache.u.segment.r_w          = 1; /* writeable */
  BX_CPU_THIS_PTR ds.cache.u.segment.a            = 1; /* accessed */
  BX_CPU_THIS_PTR ds.cache.u.segment.base         = 0x00000000;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit        =     0xFFFF;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit_scaled =     0xFFFF;
#endif
#if BX_CPU >= 3
  BX_CPU_THIS_PTR ds.cache.u.segment.g   = 0; /* byte granular */
  BX_CPU_THIS_PTR ds.cache.u.segment.d_b = 0; /* 16bit default size */
  BX_CPU_THIS_PTR ds.cache.u.segment.avl = 0;
#endif


  /* ES (Extra Segment) and descriptor cache */
  BX_CPU_THIS_PTR es.selector.value =     0x0000;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR es.selector.index =     0x0000;
  BX_CPU_THIS_PTR es.selector.ti = 0;
  BX_CPU_THIS_PTR es.selector.rpl = 0;

  BX_CPU_THIS_PTR es.cache.valid =     1;
  BX_CPU_THIS_PTR es.cache.p = 1;
  BX_CPU_THIS_PTR es.cache.dpl = 0;
  BX_CPU_THIS_PTR es.cache.segment = 1; /* data/code segment */
  BX_CPU_THIS_PTR es.cache.type = 3; /* read/write access */

  BX_CPU_THIS_PTR es.cache.u.segment.executable   = 0; /* data/stack segment */
  BX_CPU_THIS_PTR es.cache.u.segment.c_ed         = 0; /* normal expand up */
  BX_CPU_THIS_PTR es.cache.u.segment.r_w          = 1; /* writeable */
  BX_CPU_THIS_PTR es.cache.u.segment.a            = 1; /* accessed */
  BX_CPU_THIS_PTR es.cache.u.segment.base         = 0x00000000;
  BX_CPU_THIS_PTR es.cache.u.segment.limit        =     0xFFFF;
  BX_CPU_THIS_PTR es.cache.u.segment.limit_scaled =     0xFFFF;
#endif
#if BX_CPU >= 3
  BX_CPU_THIS_PTR es.cache.u.segment.g   = 0; /* byte granular */
  BX_CPU_THIS_PTR es.cache.u.segment.d_b = 0; /* 16bit default size */
  BX_CPU_THIS_PTR es.cache.u.segment.avl = 0;
#endif


  /* FS and descriptor cache */
#if BX_CPU >= 3
  BX_CPU_THIS_PTR fs.selector.value =     0x0000;
  BX_CPU_THIS_PTR fs.selector.index =     0x0000;
  BX_CPU_THIS_PTR fs.selector.ti = 0;
  BX_CPU_THIS_PTR fs.selector.rpl = 0;

  BX_CPU_THIS_PTR fs.cache.valid =     1;
  BX_CPU_THIS_PTR fs.cache.p = 1;
  BX_CPU_THIS_PTR fs.cache.dpl = 0;
  BX_CPU_THIS_PTR fs.cache.segment = 1; /* data/code segment */
  BX_CPU_THIS_PTR fs.cache.type = 3; /* read/write access */

  BX_CPU_THIS_PTR fs.cache.u.segment.executable   = 0; /* data/stack segment */
  BX_CPU_THIS_PTR fs.cache.u.segment.c_ed         = 0; /* normal expand up */
  BX_CPU_THIS_PTR fs.cache.u.segment.r_w          = 1; /* writeable */
  BX_CPU_THIS_PTR fs.cache.u.segment.a            = 1; /* accessed */
  BX_CPU_THIS_PTR fs.cache.u.segment.base         = 0x00000000;
  BX_CPU_THIS_PTR fs.cache.u.segment.limit        =     0xFFFF;
  BX_CPU_THIS_PTR fs.cache.u.segment.limit_scaled =     0xFFFF;
  BX_CPU_THIS_PTR fs.cache.u.segment.g   = 0; /* byte granular */
  BX_CPU_THIS_PTR fs.cache.u.segment.d_b = 0; /* 16bit default size */
  BX_CPU_THIS_PTR fs.cache.u.segment.avl = 0;
#endif


  /* GS and descriptor cache */
#if BX_CPU >= 3
  BX_CPU_THIS_PTR gs.selector.value =     0x0000;
  BX_CPU_THIS_PTR gs.selector.index =     0x0000;
  BX_CPU_THIS_PTR gs.selector.ti = 0;
  BX_CPU_THIS_PTR gs.selector.rpl = 0;

  BX_CPU_THIS_PTR gs.cache.valid =     1;
  BX_CPU_THIS_PTR gs.cache.p = 1;
  BX_CPU_THIS_PTR gs.cache.dpl = 0;
  BX_CPU_THIS_PTR gs.cache.segment = 1; /* data/code segment */
  BX_CPU_THIS_PTR gs.cache.type = 3; /* read/write access */

  BX_CPU_THIS_PTR gs.cache.u.segment.executable   = 0; /* data/stack segment */
  BX_CPU_THIS_PTR gs.cache.u.segment.c_ed         = 0; /* normal expand up */
  BX_CPU_THIS_PTR gs.cache.u.segment.r_w          = 1; /* writeable */
  BX_CPU_THIS_PTR gs.cache.u.segment.a            = 1; /* accessed */
  BX_CPU_THIS_PTR gs.cache.u.segment.base         = 0x00000000;
  BX_CPU_THIS_PTR gs.cache.u.segment.limit        =     0xFFFF;
  BX_CPU_THIS_PTR gs.cache.u.segment.limit_scaled =     0xFFFF;
  BX_CPU_THIS_PTR gs.cache.u.segment.g   = 0; /* byte granular */
  BX_CPU_THIS_PTR gs.cache.u.segment.d_b = 0; /* 16bit default size */
  BX_CPU_THIS_PTR gs.cache.u.segment.avl = 0;
#endif


  /* GDTR (Global Descriptor Table Register) */
#if BX_CPU >= 2
  BX_CPU_THIS_PTR gdtr.base         = 0x00000000;  /* undefined */
  BX_CPU_THIS_PTR gdtr.limit        =     0x0000;  /* undefined */
  /* ??? AR=Present, Read/Write */
#endif

  /* IDTR (Interrupt Descriptor Table Register) */
#if BX_CPU >= 2
  BX_CPU_THIS_PTR idtr.base         = 0x00000000;
  BX_CPU_THIS_PTR idtr.limit        =     0x03FF; /* always byte granular */ /* ??? */
  /* ??? AR=Present, Read/Write */
#endif

  /* LDTR (Local Descriptor Table Register) */
#if BX_CPU >= 2
  BX_CPU_THIS_PTR ldtr.selector.value =     0x0000;
  BX_CPU_THIS_PTR ldtr.selector.index =     0x0000;
  BX_CPU_THIS_PTR ldtr.selector.ti = 0;
  BX_CPU_THIS_PTR ldtr.selector.rpl = 0;

  BX_CPU_THIS_PTR ldtr.cache.valid   = 0; /* not valid */
  BX_CPU_THIS_PTR ldtr.cache.p       = 0; /* not present */
  BX_CPU_THIS_PTR ldtr.cache.dpl     = 0; /* field not used */
  BX_CPU_THIS_PTR ldtr.cache.segment = 0; /* system segment */
  BX_CPU_THIS_PTR ldtr.cache.type    = 2; /* LDT descriptor */

  BX_CPU_THIS_PTR ldtr.cache.u.ldt.base      = 0x00000000;
  BX_CPU_THIS_PTR ldtr.cache.u.ldt.limit     =     0xFFFF;
#endif

  /* TR (Task Register) */
#if BX_CPU >= 2
  /* ??? I don't know what state the TR comes up in */
  BX_CPU_THIS_PTR tr.selector.value =     0x0000;
  BX_CPU_THIS_PTR tr.selector.index =     0x0000; /* undefined */
  BX_CPU_THIS_PTR tr.selector.ti    =     0;
  BX_CPU_THIS_PTR tr.selector.rpl   =     0;

  BX_CPU_THIS_PTR tr.cache.valid    = 0;
  BX_CPU_THIS_PTR tr.cache.p        = 0;
  BX_CPU_THIS_PTR tr.cache.dpl      = 0; /* field not used */
  BX_CPU_THIS_PTR tr.cache.segment  = 0;
  BX_CPU_THIS_PTR tr.cache.type     = 0; /* invalid */
  BX_CPU_THIS_PTR tr.cache.u.tss286.base             = 0x00000000; /* undefined */
  BX_CPU_THIS_PTR tr.cache.u.tss286.limit            =     0x0000; /* undefined */
#endif

  /* DR0 - DR7 (Debug Register 0 to 7), unimplemented */
#if BX_CPU >= 3
  BX_CPU_THIS_PTR dr0 = 0;   /* undefined */
  BX_CPU_THIS_PTR dr1 = 0;   /* undefined */
  BX_CPU_THIS_PTR dr2 = 0;   /* undefined */
  BX_CPU_THIS_PTR dr3 = 0;   /* undefined */
#endif
#if BX_CPU >= 3
  BX_CPU_THIS_PTR dr4 = 0;   /* undefined */
  BX_CPU_THIS_PTR dr5 = 0;   /* undefined */
#endif
#if   BX_CPU == 3
  BX_CPU_THIS_PTR dr6 = 0xFFFF0FF0;   /* ??? unimplemented */
  BX_CPU_THIS_PTR dr7 = 0x00000000;   /* unimplemented */
#elif BX_CPU == 4
  BX_CPU_THIS_PTR dr6 = 0xFFFF0FF0;   /* unimplemented */
  BX_CPU_THIS_PTR dr7 = 0x00000000;   /* unimplemented */
#elif BX_CPU == 5
  BX_CPU_THIS_PTR dr6 = 0xFFFF0FF0;   /* unimplemented */
  BX_CPU_THIS_PTR dr7 = 0x00000400;   /* unimplemented */
#endif

#if 0
  /* test registers 3-7 (unimplemented) */
  BX_CPU_THIS_PTR tr3 = 0;   /* undefined */
  BX_CPU_THIS_PTR tr4 = 0;   /* undefined */
  BX_CPU_THIS_PTR tr5 = 0;   /* undefined */
  BX_CPU_THIS_PTR tr6 = 0;   /* undefined */
  BX_CPU_THIS_PTR tr7 = 0;   /* undefined */
#endif

  /* MSW (Machine Status Word), so called on 286 */
  /* CR0 (Control Register 0), so called on 386+ */
#if BX_CPU >= 3
  BX_CPU_THIS_PTR cr0.pg = 0; /* paging disabled */
#endif
#if BX_CPU >= 4
  BX_CPU_THIS_PTR cr0.cd = 1; /* caching disabled */
  BX_CPU_THIS_PTR cr0.nw = 1; /* not write-through */
  BX_CPU_THIS_PTR cr0.am = 0; /* disable alignment check */
  BX_CPU_THIS_PTR cr0.wp = 0; /* disable write-protect */
  BX_CPU_THIS_PTR cr0.ne = 0; /* ndp exceptions through int 13H, DOS compat */
#endif
#if BX_CPU >= 2
  BX_CPU_THIS_PTR cr0.ts = 0; /* no task switch */
  BX_CPU_THIS_PTR cr0.em = 0; /* emulate math coprocessor */
  BX_CPU_THIS_PTR cr0.mp = 0; /* wait instructions not trapped */
  BX_CPU_THIS_PTR cr0.pe = 0; /* real mode */
#endif

#if BX_CPU >= 3
  BX_CPU_THIS_PTR cr2 = 0;
  BX_CPU_THIS_PTR cr3 = 0;
#endif


  BX_CPU_THIS_PTR EXT = 0;
  BX_CPU_THIS_PTR INTR = 0;

  BX_CPU_THIS_PTR TLB.valid = 0;

  BX_CPU_THIS_PTR linear_block_addr = 0;
  BX_CPU_THIS_PTR fetch_index = BX_INVALID_FETCH_INDEX;
}


  void
bx_cpu_c::sanity_checks(void)
{
  Bit8u al, cl, dl, bl, ah, ch, dh, bh;
  Bit16u ax, cx, dx, bx, sp, bp, si, di;
  Bit32u eax, ecx, edx, ebx, esp, ebp, esi, edi;

  EAX = 0xFFEEDDCC;
  ECX = 0xBBAA9988;
  EDX = 0x77665544;
  EBX = 0x332211FF;
  ESP = 0xEEDDCCBB;
  EBP = 0xAA998877;
  ESI = 0x66554433;
  EDI = 0x2211FFEE;

  al = AL;
  cl = CL;
  dl = DL;
  bl = BL;
  ah = AH;
  ch = CH;
  dh = DH;
  bh = BH;

  if ( al != (EAX & 0xFF) ||
       cl != (ECX & 0xFF) ||
       dl != (EDX & 0xFF) ||
       bl != (EBX & 0xFF) ||
       ah != ((EAX >> 8) & 0xFF) ||
       ch != ((ECX >> 8) & 0xFF) ||
       dh != ((EDX >> 8) & 0xFF) ||
       bh != ((EBX >> 8) & 0xFF) ) {
    bx_panic("problems using BX_READ_8BIT_REG()!\n");
    }

  ax = AX;
  cx = CX;
  dx = DX;
  bx = BX;
  sp = SP;
  bp = BP;
  si = SI;
  di = DI;

  if ( ax != (EAX & 0xFFFF) ||
       cx != (ECX & 0xFFFF) ||
       dx != (EDX & 0xFFFF) ||
       bx != (EBX & 0xFFFF) ||
       sp != (ESP & 0xFFFF) ||
       bp != (EBP & 0xFFFF) ||
       si != (ESI & 0xFFFF) ||
       di != (EDI & 0xFFFF) ) {
    bx_panic("problems using BX_READ_16BIT_REG()!\n");
    }


  eax = EAX;
  ecx = ECX;
  edx = EDX;
  ebx = EBX;
  esp = ESP;
  ebp = EBP;
  esi = ESI;
  edi = EDI;
  

  if (sizeof(Bit8u)  != 1  ||  sizeof(Bit8s)  != 1)
    bx_panic("data type Bit8u or Bit8s is not of length 1 byte!\n");
  if (sizeof(Bit16u) != 2  ||  sizeof(Bit16s) != 2)
    bx_panic("data type Bit16u or Bit16s is not of length 2 bytes!\n");
  if (sizeof(Bit32u) != 4  ||  sizeof(Bit32s) != 4)
    bx_panic("data type Bit32u or Bit32s is not of length 4 bytes!\n");
  
  fprintf(stderr, "#all sanity checks passed!\n");
}


  void
bx_cpu_c::reset_cpu(void)
{
#if BX_CPU > 2
  bx_panic("reset_cpu: not implemented for 386\n");
#endif

  if (v8086_mode()) bx_panic("main: v8086 mode unsupported\n");

  if (bx_dbg.reset)
    bx_printf("CPU RESET encountered.\n");

  /* all status flags at known values, use eflags structure */
  BX_CPU_THIS_PTR lf_flags_status = 0x000000;
  BX_CPU_THIS_PTR lf_pf = 0;

  /* status and control flags register set */
  BX_CPU_THIS_PTR set_CF(0);
  BX_CPU_THIS_PTR eflags.bit1 = 1;
  BX_CPU_THIS_PTR set_PF(0);
  BX_CPU_THIS_PTR eflags.bit3 = 0;
  BX_CPU_THIS_PTR set_AF(0);
  BX_CPU_THIS_PTR eflags.bit5 = 0;
  BX_CPU_THIS_PTR set_ZF(0);
  BX_CPU_THIS_PTR set_SF(0);
  BX_CPU_THIS_PTR eflags.tf = 0;
  BX_CPU_THIS_PTR eflags.if_ = 0;
  BX_CPU_THIS_PTR eflags.df = 0;
  BX_CPU_THIS_PTR set_OF(0);
#if BX_CPU >= 2
  BX_CPU_THIS_PTR eflags.iopl = 0;
  BX_CPU_THIS_PTR eflags.nt = 0;
#endif
  BX_CPU_THIS_PTR eflags.bit15 = 0;

  BX_CPU_THIS_PTR inhibit_interrupts = 0;
  BX_CPU_THIS_PTR single_step_event = 0;
#if 0
  /* ??? clear out old pending interrupts */
  BX_CPU_THIS_PTR INTR = 0;
#endif

  /* MSW */
#if BX_CPU >= 2
  BX_CPU_THIS_PTR cr0.ts = 0; /* no task switch */
  BX_CPU_THIS_PTR cr0.em = 0; /* emulate math coprocessor */
  BX_CPU_THIS_PTR cr0.mp = 0; /* wait instructions not trapped */
  BX_CPU_THIS_PTR cr0.pe = 0; /* real mode */
#endif
#if BX_CPU >= 3
  BX_CPU_THIS_PTR cr0.pg = 0; /* paging disabled */
#endif

#if BX_CPU >= 3
  BX_CPU_THIS_PTR cr2 = 0;
  BX_CPU_THIS_PTR cr3 = 0;
#endif

  /* ??? 0 for 8086 */
  BX_CPU_THIS_PTR eip    = 0x0000FFF0; /* instruction pointer */

  /* CS selector */
  BX_CPU_THIS_PTR cs.selector.value =     0xf000;
  BX_CPU_THIS_PTR cs.cache.valid =     1;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR cs.selector.index =     0x0000;
  BX_CPU_THIS_PTR cs.cache.u.segment.base           = 0x000F0000;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit          =     0xFFFF;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled   =     0xFFFF;
  BX_CPU_THIS_PTR cs.cache.dpl =          0; /* ??? */
#endif

  BX_CPU_THIS_PTR ds.selector.value =     0x0000;
  BX_CPU_THIS_PTR ds.cache.valid =     1;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR ds.selector.index =     0x0000;
  BX_CPU_THIS_PTR ds.cache.u.segment.base           = 0x00000000;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit          =     0xFFFF;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit_scaled   =     0xFFFF;
  BX_CPU_THIS_PTR ds.cache.dpl =          0; /* ??? */
#endif

  BX_CPU_THIS_PTR ss.selector.value =     0x0000;
  BX_CPU_THIS_PTR ss.cache.valid =     1;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR ss.selector.index =     0x0000;
  BX_CPU_THIS_PTR ss.cache.u.segment.base           = 0x00000000;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit          =     0xFFFF;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled   =     0xFFFF;
  BX_CPU_THIS_PTR ss.cache.dpl =          0; /* ??? */
#endif

  BX_CPU_THIS_PTR es.selector.value =     0x0000;
  BX_CPU_THIS_PTR es.cache.valid =     1;
#if BX_CPU >= 2
  BX_CPU_THIS_PTR es.selector.index =     0x0000;
  BX_CPU_THIS_PTR es.cache.u.segment.base           = 0x00000000;
  BX_CPU_THIS_PTR es.cache.u.segment.limit          =     0xFFFF;
  BX_CPU_THIS_PTR es.cache.u.segment.limit_scaled   =     0xFFFF;
  BX_CPU_THIS_PTR es.cache.dpl =          0; /* ??? */
#endif

#if BX_CPU >= 3
  /* FS */
  BX_CPU_THIS_PTR fs.selector.value =     0x0000;
  BX_CPU_THIS_PTR fs.selector.index =     0x0000;
  BX_CPU_THIS_PTR fs.cache.valid    =     1;
  BX_CPU_THIS_PTR fs.cache.u.segment.base           = 0x00000000;
  BX_CPU_THIS_PTR fs.cache.u.segment.limit          =     0xFFFF;
  BX_CPU_THIS_PTR fs.cache.u.segment.limit_scaled   =     0xFFFF;
  BX_CPU_THIS_PTR fs.cache.dpl =          0; /* ??? */

  /* GS */
  BX_CPU_THIS_PTR gs.selector.value =     0x0000;
  BX_CPU_THIS_PTR gs.selector.index =     0x0000;
  BX_CPU_THIS_PTR gs.cache.valid    =     1;
  BX_CPU_THIS_PTR gs.cache.u.segment.base           = 0x00000000;
  BX_CPU_THIS_PTR gs.cache.u.segment.limit          =     0xFFFF;
  BX_CPU_THIS_PTR gs.cache.u.segment.limit_scaled   =     0xFFFF;
  BX_CPU_THIS_PTR gs.cache.dpl =          0; /* ??? */
#endif

#if BX_CPU >= 2
  BX_CPU_THIS_PTR idtr.base         = 0x00000000;
  BX_CPU_THIS_PTR idtr.limit        =     0x03FF;
#endif

  BX_CPU_THIS_PTR EXT = 0;

  BX_CPU_THIS_PTR TLB.valid = 0;

  BX_CPU_THIS_PTR linear_block_addr = 0;
  BX_CPU_THIS_PTR fetch_index = BX_INVALID_FETCH_INDEX;
}


  void
bx_cpu_c::set_INTR(Boolean value)
{
  BX_CPU_THIS_PTR INTR = value;
  BX_CPU_THIS_PTR async_event = 1;
}
