/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/






#include "bochs.h"





  INLINE void
bx_cpu_c::INSB_YbDX()
{
  Bit8u value8;

  /* output value16 to port DX */
  value8 = BX_CPU_THIS_PTR inp8(DX);

  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    /* no seg override possible */
    write_virtual_byte(&BX_CPU_THIS_PTR es, EDI, &value8);

    if (BX_CPU_THIS_PTR eflags.df) {
      EDI = EDI - 1;
      }
    else {
      EDI = EDI + 1;
      }
    }
  else {
    /* no seg override possible */
    write_virtual_byte(&BX_CPU_THIS_PTR es, DI, &value8);

    if (BX_CPU_THIS_PTR eflags.df) {
      DI = DI - 1;
      }
    else {
      DI = DI + 1;
      }
    }
}

  INLINE void
bx_cpu_c::INSW_YvDX()
  // input word/doubleword from port to string
{
  Bit32u edi;
  unsigned int incr;
  // addrsize determines use of eDI
  // opsize   determines whether word/dword used, and increment of eDI

  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    edi = EDI;
  else
    edi = DI;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u value32;

    /* input value16 from port DX */
    value32 = BX_CPU_THIS_PTR inp32(DX);

    /* no seg override allowed */
    write_virtual_dword(&BX_CPU_THIS_PTR es, edi, &value32);
    incr = 4;
    }
  else {
    Bit16u value16;

    /* input value16 from port DX */
    value16 = BX_CPU_THIS_PTR inp16(DX);

    /* no seg override allowed */
    write_virtual_word(&BX_CPU_THIS_PTR es, edi, &value16);
    incr = 2;
    }

  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    if (BX_CPU_THIS_PTR eflags.df)
      EDI = EDI - incr;
    else
      EDI = EDI + incr;
    }
  else {
    if (BX_CPU_THIS_PTR eflags.df)
      DI = DI - incr;
    else
      DI = DI + incr;
    }
}

  INLINE void
bx_cpu_c::OUTSB_DXXb()
{
  bx_segment_reg_t *seg;
  Bit8u value8;


  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    bx_panic("outsb(): 32bit addrsize unhandled\n");
    }

  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

  read_virtual_byte(seg, SI, &value8);

  /* output value16 to port DX */
  BX_CPU_THIS_PTR outp8(DX, value8);

  if (BX_CPU_THIS_PTR eflags.df) {
    SI = SI - 1;
    }
  else {
    SI = SI + 1;
    }
}

  INLINE void
bx_cpu_c::OUTSW_DXXv()
  // output word/doubleword string to port
{
  bx_segment_reg_t *seg;
  Bit32u esi;
  unsigned int incr;
  // addrsize determines use of eSI
  // opsize   determines wheter word/dword used, and increment of eSI


  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

  if (BX_CPU_THIS_PTR is_32bit_addrsize)
    esi = ESI;
  else
    esi = SI;

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u value32;
    read_virtual_dword(seg, esi, &value32);

    /* output value32 to port DX */
    BX_CPU_THIS_PTR outp32(DX, value32);
    incr = 4;
    }
  else {
    Bit16u value16;
    read_virtual_word(seg, esi, &value16);

    /* output value16 to port DX */
    BX_CPU_THIS_PTR outp16(DX, value16);
    incr = 2;
    }

  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    if (BX_CPU_THIS_PTR eflags.df)
      ESI = ESI - incr;
    else
      ESI = ESI + incr;
    }
  else {
    if (BX_CPU_THIS_PTR eflags.df)
      SI = SI - incr;
    else
      SI = SI + incr;
    }
}


  INLINE void
bx_cpu_c::IN_ALIb()
{
  Bit8u al, imm8;

  imm8 = fetch_next_byte();

  al = BX_CPU_THIS_PTR inp8(imm8);

  AL = al;
}

  INLINE void
bx_cpu_c::IN_eAXIb()
{
  Bit8u imm8;


  imm8 = fetch_next_byte();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eax;

    eax = BX_CPU_THIS_PTR inp32(imm8);
    EAX = eax;
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u ax;

    ax = BX_CPU_THIS_PTR inp16(imm8);
    AX = ax;
    }
}

  INLINE void
bx_cpu_c::OUT_IbAL()
{
  Bit8u al, imm8;

  imm8 = fetch_next_byte();

  al = AL;

  if (protected_mode() && bx_dbg._protected)
    bx_printf("out_ibal(%02x, %02x):\n", (unsigned) imm8, (unsigned) al);
  BX_CPU_THIS_PTR outp8(imm8, al);
}

  INLINE void
bx_cpu_c::OUT_IbeAX()
{
  Bit8u imm8;

  imm8 = fetch_next_byte();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    BX_CPU_THIS_PTR outp32(imm8, EAX);
    }
  else
#endif /* BX_CPU > 2 */
    {
    BX_CPU_THIS_PTR outp16(imm8, AX);
    }
}

  INLINE void
bx_cpu_c::IN_ALDX()
{
  Bit8u al;

  al = BX_CPU_THIS_PTR inp8(DX);

  AL = al;
}

  INLINE void
bx_cpu_c::IN_eAXDX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eax;

    eax = BX_CPU_THIS_PTR inp32(DX);
    EAX = eax;
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u ax;

    ax = BX_CPU_THIS_PTR inp16(DX);
    AX = ax;
    }
}

  INLINE void
bx_cpu_c::OUT_DXAL()
{
  Bit16u dx;
  Bit8u al;

  dx = DX;
  al = AL;

  BX_CPU_THIS_PTR outp8(dx, al);
}

  INLINE void
bx_cpu_c::OUT_DXeAX()
{
  Bit16u dx;

  dx = DX;

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    BX_CPU_THIS_PTR outp32(dx, EAX);
    }
  else
#endif /* BX_CPU > 2 */
    {
    BX_CPU_THIS_PTR outp16(dx, AX);
    }
}
