// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com







#include "bochs.h"




  Bit16u
bx_cpu_c::inp16(Bit16u addr)
{
  if (BX_CPU_THIS_PTR cr0.pe && (BX_CPU_THIS_PTR eflags.vm || (CPL>IOPL))) {
    if ( !BX_CPU_THIS_PTR allow_io(addr, 2) ) {
      // bx_printf("cpu_inp16: GP0()!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return(0);
      }
    }

  return( bx_pc_system.inp(addr, 2) );
}

  void
bx_cpu_c::outp16(Bit16u addr, Bit16u value)
{
  /* If CPL <= IOPL, then all IO addresses are accessible.
   * Otherwise, must check the IO permission map on >286.
   * On the 286, there is no IO permissions map */

  if (BX_CPU_THIS_PTR cr0.pe && (BX_CPU_THIS_PTR eflags.vm || (CPL>IOPL))) {
    if ( !BX_CPU_THIS_PTR allow_io(addr, 2) ) {
      // bx_printf("cpu_outp16: GP0()!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }

  bx_pc_system.outp(addr, value, 2);
}

  Bit32u
bx_cpu_c::inp32(Bit16u addr)
{
  if (BX_CPU_THIS_PTR cr0.pe && (BX_CPU_THIS_PTR eflags.vm || (CPL>IOPL))) {
    if ( !BX_CPU_THIS_PTR allow_io(addr, 4) ) {
      // bx_printf("cpu_inp32: GP0()!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return(0);
      }
    }

  return( bx_pc_system.inp(addr, 4) );
}

  void
bx_cpu_c::outp32(Bit16u addr, Bit32u value)
{
  /* If CPL <= IOPL, then all IO addresses are accessible.
   * Otherwise, must check the IO permission map on >286.
   * On the 286, there is no IO permissions map */

  if (BX_CPU_THIS_PTR cr0.pe && (BX_CPU_THIS_PTR eflags.vm || (CPL>IOPL))) {
    if ( !BX_CPU_THIS_PTR allow_io(addr, 4) ) {
      // bx_printf("cpu_outp32: GP0()!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }

  bx_pc_system.outp(addr, value, 4);
}

  Bit8u
bx_cpu_c::inp8(Bit16u addr)
{
  if (BX_CPU_THIS_PTR cr0.pe && (BX_CPU_THIS_PTR eflags.vm || (CPL>IOPL))) {
    if ( !BX_CPU_THIS_PTR allow_io(addr, 1) ) {
      // bx_printf("cpu_inp8: GP0()!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return(0);
      }
    }

  return( bx_pc_system.inp(addr, 1) );
}


  void
bx_cpu_c::outp8(Bit16u addr, Bit8u value)
{
  /* If CPL <= IOPL, then all IO addresses are accessible.
   * Otherwise, must check the IO permission map on >286.
   * On the 286, there is no IO permissions map */

  if (BX_CPU_THIS_PTR cr0.pe && (BX_CPU_THIS_PTR eflags.vm || (CPL>IOPL))) {
    if ( !BX_CPU_THIS_PTR allow_io(addr, 1) ) {
      // bx_printf("cpu_outp8: GP0()!\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }

  bx_pc_system.outp(addr, value, 1);
}


  Boolean
bx_cpu_c::allow_io(Bit16u addr, unsigned len)
{
  Bit16u io_base, permission16;
  unsigned bit_index, i;

  if (BX_CPU_THIS_PTR tr.cache.valid==0 || BX_CPU_THIS_PTR tr.cache.type!=9) {
    bx_printf("allow_io(): TR doesn't point to a valid 32bit TSS\n");
    return(0);
    }

#if 0
  bx_printf("allow_io(): TR.limit = %u\n",
    (unsigned) BX_CPU_THIS_PTR tr.cache.u.tss386.limit);
#endif
  if (BX_CPU_THIS_PTR tr.cache.u.tss386.limit < 103) {
    bx_panic("allow_io(): TR.limit < 103\n");
    }

  access_linear(BX_CPU_THIS_PTR tr.cache.u.tss386.base + 102, 2, 0, BX_READ,
                         &io_base);
  if (io_base <= 103) {
    bx_panic("allow_io(): TR:io_base <= 103\n");
    }

  if (io_base > BX_CPU_THIS_PTR tr.cache.u.tss386.limit) {
    bx_printf("allow_io(): CPL > IOPL: no IO bitmap defined #GP(0)\n");
    return(0);
    }

  access_linear(BX_CPU_THIS_PTR tr.cache.u.tss386.base + io_base + addr/8,
                   2, 0, BX_READ, &permission16);

  bit_index = addr & 0x07;
  permission16 >>= bit_index;
  for (i=0; i<len; i++) {
    if (permission16 & 0x01)
      return(0);
    permission16 >>= 1;
    }

  return(1);
}
