//
// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com
//





#include "bochs.h"



#if BX_SUPPORT_PAGING

  void
bx_cpu_c::enable_paging(void)
{
  TLB_flush();
  if (bx_dbg.paging) bx_printf("enable_paging():\n");
//fprintf(stderr, "#enable_paging():-------------------------\n");
}

  void
bx_cpu_c::disable_paging(void)
{
  TLB_flush();
  if (bx_dbg.paging) bx_printf("disable_paging():\n");
}

  void
bx_cpu_c::CR3_change(Bit32u value32)
{
  if (bx_dbg.paging) {
    bx_printf("CR3_change(): flush TLB cache\n");
    bx_printf("Page Directory Base %08x\n", (unsigned) value32);
    }

  // only flush page cache if a new value is supplied
  //if (value32 != BX_CPU_THIS_PTR cr3) {
    TLB_flush();
    BX_CPU_THIS_PTR cr3 = value32;
    //}
}

  void
bx_cpu_c::TLB_flush(void)
{
  BX_CPU_THIS_PTR TLB.valid = 0;

#if BX_USE_TLB
  for (unsigned i=0; i<BX_TLB_SIZE; i++) {
    BX_CPU_THIS_PTR local_TLB[i].access = 0; // invalid
    }
#endif  // #if BX_USE_TLB
}


  Bit32u
bx_cpu_c::dtranslate_linear(Bit32u laddress, unsigned pl, unsigned rw)
{
  Bit32u dir_entry_addr, page_table_entry_addr;
  Bit32u dir_entry, page_table_entry;
  Bit32u new_dir_entry, new_page_table_entry;
  Bit32u page_frame_address, paddress, poffset;
  Bit32u local_TLB_index;
  Bit32u need;
  Bit32u lpf;
  Bit16u errorcode;

  // pl: {0,1}
  //     0: access from CPL=0/1/2 or from internal CPU architecture
  //     1: access from CPL=3

// take out ???
if ( pl > 1) bx_panic("dtranslate_linear: pl > 1\n");

  /* Protection:
   *   When the processor is running in supervisor level, all pages are both
   *   readable and writable (write-protect ignored).  When running at user
   *   level, only pages which belong to the user level are accessible;
   *   read/write & read-only are readable, read/write are writable.
   *
   * Accessed bits:
   *   Processor sets the Accessed bits in both levels of page tables before
   *   a read/write operation to a page.
   *
   * Dirty bits:
   *   Processor sets the Dirty bit in the 2nd-level page table before a
   *   write operation to an address mapped by that page table entry.
   *   Dirty bit in directory entries is undefined.
   *   
   */

  lpf = (laddress & 0xfffff000); // linear page frame
  local_TLB_index = (lpf & 0x003ff000) >> 12;
  poffset = laddress & 0x00000fff;


#if BX_USE_TLB
  if ( BX_CPU_THIS_PTR local_TLB[local_TLB_index].LPF == lpf ) {
    need = 0x80; // valid
    if (pl) {
      need |= ((rw==BX_WRITE) ? 0x42 : 0x00) | 0x04;
      }
    if ( BX_CPU_THIS_PTR local_TLB[local_TLB_index].access >= need )
      return( BX_CPU_THIS_PTR local_TLB[local_TLB_index].PPF | poffset );
    }
#endif // #if BX_USE_TLB

  dir_entry_addr = (BX_CPU_THIS_PTR cr3 & 0xfffff000) |
                   ((laddress & 0xffc00000) >> 20);
  bx_mem.read_physical(dir_entry_addr, 4, &dir_entry);
  /*bx_printf("paging: DIR entry: %08x\n", (unsigned) dir_entry);*/
  if ( (dir_entry & 0x01) == 0) {
    bx_printf("dtranslate_linear: dir entry not present\n");
    BX_CPU_THIS_PTR cr2 = laddress;
    errorcode = (pl << 2) | ((rw==BX_WRITE) << 1);
    exception(BX_PF_EXCEPTION, errorcode, 0);
    return(0);
    }

  if ( pl ) {
    if ( (dir_entry & 0x04) == 0 ) {
      bx_printf("paging: no permission: directory entry\n");

      BX_CPU_THIS_PTR cr2 = laddress;
      errorcode = 0x05 | ((rw==BX_WRITE) << 1);
      exception(BX_PF_EXCEPTION, errorcode, 0);
      return(0);
      }
    if ( (rw == BX_WRITE) && ((dir_entry & 0x02)==0) ) {
      bx_panic("paging: no write permission: directory entry\n");
      }
    }

  page_table_entry_addr = (dir_entry & 0xfffff000) |
                          ((laddress & 0x003ff000) >> 10);
  bx_mem.read_physical(page_table_entry_addr, 4, &page_table_entry);
  /*bx_printf("paging: page table entry: %08x\n", (unsigned) page_table_entry);*/

  if ( (page_table_entry & 0x01) == 0 ) {
    //bx_printf("dtranslate_linear: page table entry not present\n");
    BX_CPU_THIS_PTR cr2 = laddress;
    errorcode = (pl << 2) | ((rw==BX_WRITE) << 1);
    exception(BX_PF_EXCEPTION, errorcode, 0);
    return(0);
    }

  if ( pl ) {
    if ( (page_table_entry & 0x04) == 0 ) {
      bx_printf("paging: no permission: page table entry\n");
      BX_CPU_THIS_PTR cr2 = laddress;
      errorcode = 0x05 | ((rw==BX_WRITE) << 1);
      exception(BX_PF_EXCEPTION, errorcode, 0);
      return(0);
      }
    if ( (rw == BX_WRITE) && ((page_table_entry & 0x02)==0) ) {
      //bx_printf("paging: no write permission: page table entry\n");
//debug(BX_CPU_THIS_PTR prev_eip);
//bx_dbg.debugger = 1;
      BX_CPU_THIS_PTR cr2 = laddress;
      errorcode = 0x07; // (pl << 2) | ((rw==BX_WRITE) << 1) | 0x01;
      exception(BX_PF_EXCEPTION, errorcode, 0);
      return(0);
      }
    }

  page_frame_address = (page_table_entry & 0xfffff000);
  paddress = page_frame_address | poffset;

  //dir_entry        |= 0x20;
  //page_table_entry |= 0x20;
  //if (rw==BX_WRITE)
  //  page_table_entry |= 0x40;

  new_dir_entry = dir_entry | 0x20;
  new_page_table_entry = page_table_entry | 0x20;
  if (rw==BX_WRITE)
    new_page_table_entry |= 0x40;

  if (new_dir_entry != dir_entry)
    bx_mem.write_physical(dir_entry_addr, 4, &new_dir_entry);
  if (new_page_table_entry != page_table_entry)
    bx_mem.write_physical(page_table_entry_addr, 4, &new_page_table_entry);
  
#if BX_USE_TLB
  BX_CPU_THIS_PTR local_TLB[ local_TLB_index ].LPF = lpf;
  BX_CPU_THIS_PTR local_TLB[ local_TLB_index ].PPF = page_frame_address;
  BX_CPU_THIS_PTR local_TLB[ local_TLB_index ].access =
    0x80 |                                             // valid
    ( new_page_table_entry & 0x40 ) |                  // Dirty bit
    ( (new_dir_entry & new_page_table_entry) & 0x06 ); // U/S & R/W bits
#endif  // #if BX_USE_TLB

  return(paddress);
}



  Bit32u
bx_cpu_c::itranslate_linear(Bit32u laddress, unsigned pl)
{
  Bit32u dir_entry_addr, page_table_entry_addr;
  Bit32u dir_entry, page_table_entry;
  Bit32u new_dir_entry, new_page_table_entry;
  Bit32u page_frame_address, paddress, poffset;
  Bit32u local_TLB_index;
  Bit32u need;
  Bit32u lpf;
  Bit16u errorcode;


  /* Protection:
   *   When the processor is running in supervisor level, all pages are both
   *   readable and writable (write-protect ignored).  When running at user
   *   level, only pages which belong to the user level are accessible;
   *   read/write & read-only are readable, read/write are writable.
   *
   * Accessed bits:
   *   Processor sets the Accessed bits in both levels of page tables before
   *   a read/write operation to a page.
   *
   * Dirty bits:
   *   Processor sets the Dirty bit in the 2nd-level page table before a
   *   write operation to an address mapped by that page table entry.
   *   Dirty bit in directory entries is undefined.
   *   
   */

  lpf = (laddress & 0xfffff000); // linear page frame
  local_TLB_index = (lpf & 0x003ff000) >> 12;
  poffset = laddress & 0x00000fff;


#if BX_USE_TLB
  if ( BX_CPU_THIS_PTR local_TLB[local_TLB_index].LPF == lpf ) {
    need = 0x80; // valid
    if (pl) {
      need |= 0x04;
      }
    if ( BX_CPU_THIS_PTR local_TLB[local_TLB_index].access >= need )
      return( BX_CPU_THIS_PTR local_TLB[local_TLB_index].PPF | poffset );
    }
#endif // #if BX_USE_TLB

  dir_entry_addr = (BX_CPU_THIS_PTR cr3 & 0xfffff000) |
                   ((laddress & 0xffc00000) >> 20);
  bx_mem.read_physical(dir_entry_addr, 4, &dir_entry);
  /*bx_printf("paging: DIR entry: %08x\n", (unsigned) dir_entry);*/
  if ( (dir_entry & 0x01) == 0) {
    bx_printf("itranslate_linear: dir entry not present\n");
    BX_CPU_THIS_PTR cr2 = laddress;
    errorcode = (pl << 2);
    exception(BX_PF_EXCEPTION, errorcode, 0);
    return(0);
    }

  if ( pl ) {
    if ( (dir_entry & 0x04) == 0 ) {
      bx_printf("paging: no permission: directory entry\n");

      BX_CPU_THIS_PTR cr2 = laddress;
      errorcode = (pl << 2) | 0x01;
      exception(BX_PF_EXCEPTION, errorcode, 0);
      return(0);
      }
    }

  page_table_entry_addr = (dir_entry & 0xfffff000) |
                          ((laddress & 0x003ff000) >> 10);
  bx_mem.read_physical(page_table_entry_addr, 4, &page_table_entry);
  /*bx_printf("paging: page table entry: %08x\n", (unsigned) page_table_entry);*/

  if ( (page_table_entry & 0x01) == 0 ) {
    //bx_printf("itranslate_linear: page table entry not present\n");
    BX_CPU_THIS_PTR cr2 = laddress;
    errorcode = (pl << 2);
    exception(BX_PF_EXCEPTION, errorcode, 0);
    return(0);
    }

  if ( pl ) {
    if ( (page_table_entry & 0x04) == 0 ) {
      bx_printf("paging: no permission: page table entry\n");
      BX_CPU_THIS_PTR cr2 = laddress;
      errorcode = (pl << 2) | 0x01;
      exception(BX_PF_EXCEPTION, errorcode, 0);
      return(0);
      }
    }

  page_frame_address = (page_table_entry & 0xfffff000);
  paddress = page_frame_address | poffset;

  //dir_entry        |= 0x20;
  //page_table_entry |= 0x20;

  new_dir_entry = dir_entry | 0x20;
  new_page_table_entry = page_table_entry | 0x20;

  if (new_dir_entry != dir_entry)
    bx_mem.write_physical(dir_entry_addr, 4, &new_dir_entry);
  if (new_page_table_entry != page_table_entry)
    bx_mem.write_physical(page_table_entry_addr, 4, &new_page_table_entry);
  
#if BX_USE_TLB
  BX_CPU_THIS_PTR local_TLB[ local_TLB_index ].LPF = lpf;
  BX_CPU_THIS_PTR local_TLB[ local_TLB_index ].PPF = page_frame_address;
  BX_CPU_THIS_PTR local_TLB[ local_TLB_index ].access =
    0x80 |                                             // valid
    ( new_page_table_entry & 0x40 ) |                  // Dirty bit
    ( (new_dir_entry & new_page_table_entry) & 0x06 ); // U/S & R/W bits
#endif  // #if BX_USE_TLB

  return(paddress);
}


#if BX_DEBUGGER || BX_DISASM
  void
dbg_xlate_linear2phy(Bit32u laddress, Bit32u *phy, Boolean *valid)
{
  Bit32u dir_entry_addr, page_table_entry_addr;
  Bit32u dir_entry, page_table_entry;
  Bit32u page_frame_address, paddress, poffset;
  Bit32u local_TLB_index;
  Bit32u need;
  Bit32u lpf;

  if (BX_CPU_THIS_PTR cr0.pg == 0) {
    *phy = laddress;
    *valid = 1;
    return;
    }

  lpf = (laddress & 0xfffff000); // linear page frame
  local_TLB_index = (lpf & 0x003ff000) >> 12;
  poffset = laddress & 0x00000fff;


#if BX_USE_TLB
  if ( BX_CPU_THIS_PTR local_TLB[local_TLB_index].LPF == lpf ) {
    need = 0x80; // valid
    if ( BX_CPU_THIS_PTR local_TLB[local_TLB_index].access >= need ) {
      *phy = ( BX_CPU_THIS_PTR local_TLB[local_TLB_index].PPF | poffset );
      *valid = 1;
      return;
      }
    }
#endif // #if BX_USE_TLB

  dir_entry_addr = (BX_CPU_THIS_PTR cr3 & 0xfffff000) |
                   ((laddress & 0xffc00000) >> 20);
  bx_mem.read_physical(dir_entry_addr, 4, &dir_entry);
  if ( (dir_entry & 0x01) == 0) {
    *phy = 0;
    *valid = 0;
    return;
    }

  page_table_entry_addr = (dir_entry & 0xfffff000) |
                          ((laddress & 0x003ff000) >> 10);
  bx_mem.read_physical(page_table_entry_addr, 4, &page_table_entry);

  if ( (page_table_entry & 0x01) == 0 ) {
    *phy = 0;
    *valid = 0;
    return;
    }

  page_frame_address = (page_table_entry & 0xfffff000);
  paddress = page_frame_address | poffset;

  *phy = paddress;
  *valid = 1;
  return;
}
#endif



  void
bx_cpu_c::access_linear(Bit32u laddress, unsigned length, unsigned pl,
    unsigned rw, void *data)
{
  Bit32u mod4096;
  unsigned xlate_rw;

  if (rw==BX_RW) {
    xlate_rw = BX_WRITE;
    rw = BX_READ;
    }
  else {
    xlate_rw = rw;
    }

  /* perhaps put this check before all code which calls this function,
   * so we don't have to here
   */
  if (BX_CPU_THIS_PTR cr0.pg == 0) {
    if (rw == BX_READ)
      bx_mem.read_physical(laddress, length, data);
    else
      bx_mem.write_physical(laddress, length, data);
    return;
    }

  /* check for reference across multiple pages */
  mod4096 = laddress & 0x00000fff;
  if ( (mod4096 + length) > 4096 ) {
    //Bit32u paddress1, paddress2, len1;

    /* access across 2 pages */
    BX_CPU_THIS_PTR address_xlation.paddress1 = dtranslate_linear(laddress, pl, xlate_rw);
    BX_CPU_THIS_PTR address_xlation.len1      = 4096 - mod4096;
    BX_CPU_THIS_PTR address_xlation.len2      = length - BX_CPU_THIS_PTR address_xlation.len1;
    BX_CPU_THIS_PTR address_xlation.pages     = 2;

    BX_CPU_THIS_PTR address_xlation.paddress2 = dtranslate_linear(laddress + BX_CPU_THIS_PTR address_xlation.len1, pl, xlate_rw);

#ifdef BX_LITTLE_ENDIAN
    if (rw == BX_READ) {
      bx_mem.read_physical(BX_CPU_THIS_PTR address_xlation.paddress1, BX_CPU_THIS_PTR address_xlation.len1, data);
      bx_mem.read_physical(BX_CPU_THIS_PTR address_xlation.paddress2, BX_CPU_THIS_PTR address_xlation.len2, ((Bit8u*)data) + BX_CPU_THIS_PTR address_xlation.len1);
      }
    else {
      bx_mem.write_physical(BX_CPU_THIS_PTR address_xlation.paddress1, BX_CPU_THIS_PTR address_xlation.len1, data);
      bx_mem.write_physical(BX_CPU_THIS_PTR address_xlation.paddress2, BX_CPU_THIS_PTR address_xlation.len2, ((Bit8u*)data) + BX_CPU_THIS_PTR address_xlation.len1);
      }
#else // BX_BIG_ENDIAN
    if (rw == BX_READ) {
      bx_mem.read_physical(BX_CPU_THIS_PTR address_xlation.paddress1, BX_CPU_THIS_PTR address_xlation.len1, ((Bit8u*)data) + (length -BX_CPU_THIS_PTR address_xlation.len1));
      bx_mem.read_physical(BX_CPU_THIS_PTR address_xlation.paddress2, BX_CPU_THIS_PTR address_xlation.len2, data);
      }
    else {
      bx_mem.write_physical(BX_CPU_THIS_PTR address_xlation.paddress1, BX_CPU_THIS_PTR address_xlation.len1, ((Bit8u*)data) + (length -BX_CPU_THIS_PTR address_xlation.len1));
      bx_mem.write_physical(BX_CPU_THIS_PTR address_xlation.paddress2, BX_CPU_THIS_PTR address_xlation.len2, data);
      }
#endif
    }
  else {
    // Bit32u paddress1;

    /* access within single page */
    BX_CPU_THIS_PTR address_xlation.paddress1 = dtranslate_linear(laddress, pl, xlate_rw);
    BX_CPU_THIS_PTR address_xlation.pages     = 1;

    if (rw == BX_READ)
      bx_mem.read_physical(BX_CPU_THIS_PTR address_xlation.paddress1, length, data);
    else
      bx_mem.write_physical(BX_CPU_THIS_PTR address_xlation.paddress1, length, data);
    }
}








#else   // BX_SUPPORT_PAGING

// stub functions for non-support of paging
  void
bx_cpu_c::enable_paging(void)
{
  bx_panic("enable_paging(): not implemented\n");
}

  void
bx_cpu_c::disable_paging(void)
{
  bx_panic("disable_paging() called\n");
}

  void
bx_cpu_c::CR3_change(Bit32u value32)
{
  bx_printf("CR3_change(): flush TLB cache\n");
  bx_printf("Page Directory Base %08x\n", (unsigned) value32);
}


  void
bx_cpu_c::access_linear(Bit32u laddress, unsigned length, unsigned pl,
    unsigned rw, void *data)
{
  /* perhaps put this check before all code which calls this function,
   * so we don't have to here
   */
  if (BX_CPU_THIS_PTR cr0.pg == 0) {
    if (rw == BX_READ)
      bx_mem.read_physical(laddress, length, data);
    else
      bx_mem.write_physical(laddress, length, data);
    return;
    }

  bx_panic("access_linear: paging not supported\n");
}



#endif  // BX_SUPPORT_PAGING
