/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/






#include "bochs.h"





  INLINE void
bx_cpu_c::HLT()
{
  // hack to panic if HLT comes from BIOS
  if ( BX_CPU_THIS_PTR cs.selector.value == 0xf000 )
    bx_panic("HALT instruction encountered\n");

  if (CPL!=0) {
    bx_printf("HLT(): CPL!=0\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  // stops instruction execution and places the processor in a
  // HALT state.  An enabled interrupt, NMI, or reset will resume
  // execution.  If interrupt is used to resume execution after HLT,
  // the saved CS:eIP points to instruction following HLT.
  while ( !(BX_CPU_THIS_PTR INTR) ) {
    bx_pc_system.TICK();
    }
}

  INLINE void
bx_cpu_c::CLTS()
{
#if BX_CPU < 2
  bx_panic("CLTS: not implemented for < 286\n");
#else

  if (v8086_mode()) bx_panic("clts: v8086 mode unsupported\n");

  /* read errata file */
  // does CLTS also clear NT flag???

  // #GP(0) if CPL is not 0
  if (CPL!=0) {
    bx_printf("CLTS(): CPL!=0\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  BX_CPU_THIS_PTR cr0.ts = 0;
  bx_printf("CLTS called\n");
#endif
}

  INLINE void
bx_cpu_c::INVD()
{
  bx_printf("---------------\n");
  bx_printf("- INVD called -\n");
  bx_printf("---------------\n");
  invalidate_prefetch_q();
  BX_INSTR_CACHE_CNTRL(BX_INSTR_INVD);
}

  INLINE void
bx_cpu_c::WBINVD()
{
  bx_panic("WBINVD:\n");
  invalidate_prefetch_q();
  BX_INSTR_CACHE_CNTRL(BX_INSTR_WBINVD);
}

  INLINE void
bx_cpu_c::INVLPG_Ea()
{
  bx_panic("INVLPG_Ea:\n");
  invalidate_prefetch_q();
  BX_INSTR_TLB_CNTRL(BX_INSTR_INVLPG, 0);
}

  INLINE void
bx_cpu_c::MOV_DdRd()
{
#if BX_CPU < 3
  bx_panic("MOV_DdRd: not supported on < 386\n");
#else
  Bit32u val_32;

  if (v8086_mode()) bx_panic("MOV_DdRd: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   *   Protected mode: #GP(0) if attempt to write a 1 to any reserved bit
   *     of CR4
   */

  decode_exgx();
  if (BX_CPU_THIS_PTR rm_type != BX_REGISTER_REF) {
    bx_panic("MOV_DdRd(): rm field not a register!\n");
    }

  invalidate_prefetch_q();

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_DdRd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    }

  val_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
  if (bx_dbg.dreg)
    bx_printf("MOV_DdRd: DR[%u]=%08xh unhandled\n",
      (unsigned) BX_CPU_THIS_PTR reg_index, (unsigned) val_32);

  switch (BX_CPU_THIS_PTR reg_index) {
    case 0: // DR0
      BX_CPU_THIS_PTR dr0 = val_32;
      break;
    case 1: // DR1
      BX_CPU_THIS_PTR dr1 = val_32;
      break;
    case 2: // DR2
      BX_CPU_THIS_PTR dr2 = val_32;
      break;
    case 3: // DR3
      BX_CPU_THIS_PTR dr3 = val_32;
      break;
    case 4: // DR4
      BX_CPU_THIS_PTR dr4 = val_32;
      break;
    case 5: // DR5
      BX_CPU_THIS_PTR dr5 = val_32;
      break;
    case 6: // DR6
      BX_CPU_THIS_PTR dr6 = val_32;
      break;
    case 7: // DR7
      BX_CPU_THIS_PTR dr7 = val_32;
      // if any breakpointing is enabled, bail
      // looking for: GD, G3,L3,G2,L2,G1,L1,G0,L0
      if ( val_32 & 0x000020ff ) {
        bx_panic("MOV_DdRd: DR7=%08xh: enabling breakpointing\n", (unsigned) val_32);
        }
      break;
    default:
      bx_panic("MOV_DdRd: control register index out of range\n");
      break;
    }
#endif
}

  INLINE void
bx_cpu_c::MOV_RdDd()
{
#if BX_CPU < 3
  bx_panic("MOV_RdDd: not supported on < 386\n");
#else
  Bit32u val_32;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   */

  decode_exgx();
  if (BX_CPU_THIS_PTR rm_type != BX_REGISTER_REF) {
    bx_panic("MOV_RdDd(): rm field not a register!\n");
    }

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_RdDd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  if (bx_dbg.dreg)
    bx_printf("MOV_RdDd: DR%u not implemented yet\n", BX_CPU_THIS_PTR reg_index);

  switch (BX_CPU_THIS_PTR reg_index) {
    case 0: // DR0
      val_32 = BX_CPU_THIS_PTR dr0;
      break;
    case 1: // DR1
      val_32 = BX_CPU_THIS_PTR dr1;
      break;
    case 2: // DR2
      val_32 = BX_CPU_THIS_PTR dr2;
      break;
    case 3: // DR3
      val_32 = BX_CPU_THIS_PTR dr3;
      break;
    case 4: // DR4
      val_32 = BX_CPU_THIS_PTR dr4;
      break;
    case 5: // DR5
      val_32 = BX_CPU_THIS_PTR dr5;
      break;
    case 6: // DR6
      val_32 = BX_CPU_THIS_PTR dr6;
      break;
    case 7: // DR7
      val_32 = BX_CPU_THIS_PTR dr7;
      break;
    default:
      bx_panic("MOV_RdDd: control register index out of range\n");
      val_32 = 0;
    }
  BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, val_32);
#endif
}


 INLINE void
bx_cpu_c::LMSW_Ew()
{
#if BX_CPU < 2
  bx_panic("LMSW_Ew(): not supported on 8086!\n");
#else
  Bit16u msw;
  Boolean pe, mp, em, ts;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  if ( protected_mode() ) {
    if ( CPL != 0 ) {
      bx_printf("LMSW: CPL != 0, CPL=%u\n", (unsigned) CPL);
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }


  if (bx_dbg._protected) bx_printf("lmsw():\n");

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    msw = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &msw);
    }

  /* LMSW does not affect PG,CD,NW,AM,WP,NE,ET bits, and cannot clear PE */
  pe = msw & 0x01; msw >>= 1;
  mp = msw & 0x01; msw >>= 1;
  em = msw & 0x01; msw >>= 1;
  ts = msw & 0x01;
  if (bx_dbg._protected)
    bx_printf("LMSW_Ew: pe=%x,mp=%x,em=%x,ts=%x\n",
      (unsigned) pe, (unsigned) mp, (unsigned) em, (unsigned) ts);

  if (mp || ts) {
    bx_panic("unsupported setting of MP, EM, TS bits in LMSW instructions\n");
    }
  BX_CPU_THIS_PTR cr0.em = em;

  if (pe) { /* switching to protected mode */
    if (protected_mode()) { /* already there */
      if (bx_dbg._protected)
        bx_printf("LMSW: already in protected mode\n");
      }
    else {
      enter_protected_mode();
      }
    }

  else { /* set bit to real mode */
    if ( real_mode() ) {
      /* stay in real mode, no transition */
      return;
      }
    else {
      bx_panic("LMSW_Ew: transition to real mode not supported!\n");
      }
    }
#endif /* BX_CPU < 2 */
}

 INLINE void
bx_cpu_c::SMSW_Ew()
{
#if BX_CPU < 2
  bx_panic("SMSW_Ew: not supported yet!\n");
#else
  Bit16u msw;

  if (bx_dbg._protected)
    bx_printf("SMSW_Ew:!\n");

#if BX_CPU == 2
  msw = 0xfff0; /* 80286 init value */
  msw |= (BX_CPU_THIS_PTR cr0.ts << 3) |
         (BX_CPU_THIS_PTR cr0.em << 2) |
         (BX_CPU_THIS_PTR cr0.mp << 1) |
         BX_CPU_THIS_PTR cr0.pe;
#else /* 386+ */
  /* reserved bits 0 ??? */
  /* should NE bit be included here ??? */
  // should ET bit be included here (AW)
  msw =  (BX_CPU_THIS_PTR cr0.ts << 3) |
         (BX_CPU_THIS_PTR cr0.em << 2) |
         (BX_CPU_THIS_PTR cr0.mp << 1) |
         BX_CPU_THIS_PTR cr0.pe;
#endif


  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, msw);  // zeros out high 16bits
      }
    else {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, msw);
      }
    }
  else {
    write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &msw);
    }

#endif
}


  INLINE void
bx_cpu_c::MOV_CdRd()
{
  /* mov general register data to control register */
#if BX_CPU < 3
  bx_panic("MOV_CdRd: not supported on < 386\n");
#else
  Bit32u val_32;
  Boolean prev_pe, prev_pg;


  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   *   Protected mode: #GP(0) if attempt to write a 1 to any reserved bit
   *     of CR4
   */

  decode_exgx();
  if (BX_CPU_THIS_PTR rm_type != BX_REGISTER_REF) {
    bx_panic("MOV_CdRd(): rm field not a register!\n");
    }

  invalidate_prefetch_q();

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_CdRd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  val_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);

  switch (BX_CPU_THIS_PTR reg_index) {
    case 0: /* CR0 (MSW) */
      /*bx_printf("MOV_CdRd:CR0: R32 = %08x\n", (unsigned) val_32);*/
      prev_pe = BX_CPU_THIS_PTR cr0.pe;
      prev_pg = BX_CPU_THIS_PTR cr0.pg;

      BX_CPU_THIS_PTR cr0.pe = val_32 & 0x01;
      BX_CPU_THIS_PTR cr0.mp = (val_32 >> 1) & 0x01;
      BX_CPU_THIS_PTR cr0.em = (val_32 >> 2) & 0x01;
      BX_CPU_THIS_PTR cr0.ts = (val_32 >> 3) & 0x01;
      /* BX_CPU_THIS_PTR cr0.et ??? */
#if BX_CPU >= 4
      BX_CPU_THIS_PTR cr0.ne = (val_32 >> 5)  & 0x01;
      BX_CPU_THIS_PTR cr0.wp = (val_32 >> 16) & 0x01;
      BX_CPU_THIS_PTR cr0.am = (val_32 >> 18) & 0x01;
      BX_CPU_THIS_PTR cr0.nw = (val_32 >> 29) & 0x01;
      BX_CPU_THIS_PTR cr0.cd = (val_32 >> 30) & 0x01;
#endif
      BX_CPU_THIS_PTR cr0.pg = (val_32 >> 31) & 0x01;
      if (BX_CPU_THIS_PTR cr0.ts)
        bx_panic("MOV_CdRd:CR0: R32 = %08x\n", (unsigned) val_32);

      if (prev_pe==0 && BX_CPU_THIS_PTR cr0.pe) {
        enter_protected_mode();
        }
      else if (prev_pe==1 && BX_CPU_THIS_PTR cr0.pe==0) {
        enter_real_mode();
        }

      if (prev_pg==0 && BX_CPU_THIS_PTR cr0.pg)
        enable_paging();
      else if (prev_pg==1 && BX_CPU_THIS_PTR cr0.pg==0)
        disable_paging();
      break;
    case 1: /* CR1 */
      bx_panic("MOV_CdRd: CR1 not implemented yet\n");
      break;
    case 2: /* CR2 */
      bx_printf("MOV_CdRd: CR2 not implemented yet\n");
      if (bx_dbg.creg)
        bx_printf("MOV_CdRd: CR2 = reg\n");
      BX_CPU_THIS_PTR cr2 = val_32;
      break;
    case 3: /* CR3 */
      if (bx_dbg.creg)
        bx_printf("MOV_CdRd:(%08x)\n", (unsigned) val_32);
      CR3_change(val_32);
      BX_INSTR_TLB_CNTRL(BX_INSTR_MOV_CR3, val_32);
      break;
    case 4: /* CR4 */
      bx_panic("MOV_CdRd: CR4 not implemented yet\n");
      break;
    default:
      bx_panic("MOV_CdRd: control register index out of range\n");
      break;
    }
#endif
}

  INLINE void
bx_cpu_c::MOV_RdCd()
{
  /* mov control register data to register */
#if BX_CPU < 3
  bx_panic("MOV_RdCd: not supported on < 386\n");
#else
  Bit32u val_32;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  /* NOTES:
   *   32bit operands always used
   *   r/m field specifies general register
   *   mod field should always be 11 binary
   *   reg field specifies which special register
   */

  decode_exgx();
  if (BX_CPU_THIS_PTR rm_type != BX_REGISTER_REF) {
    bx_panic("MOV_RdCd(): rm field not a register!\n");
    }

  if (protected_mode() && CPL!=0) {
    bx_panic("MOV_RdCd: CPL!=0\n");
    /* #GP(0) if CPL is not 0 */
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  switch (BX_CPU_THIS_PTR reg_index) {
    case 0: /* CR0 (MSW) */
      val_32 = (BX_CPU_THIS_PTR cr0.pe) |
               (BX_CPU_THIS_PTR cr0.mp << 1) |
               (BX_CPU_THIS_PTR cr0.em << 2) |
               (BX_CPU_THIS_PTR cr0.ts << 3) |
               /* ???(BX_CPU_THIS_PTR cr0.et << 4) |*/
#if BX_CPU >= 4
               (BX_CPU_THIS_PTR cr0.ne << 5) |
               (BX_CPU_THIS_PTR cr0.wp << 16) |
               (BX_CPU_THIS_PTR cr0.am << 18) |
               (BX_CPU_THIS_PTR cr0.nw << 29) |
               (BX_CPU_THIS_PTR cr0.cd << 30) |
#endif
               (BX_CPU_THIS_PTR cr0.pg << 31);
      break;
    case 1: /* CR1 */
      bx_panic("MOV_RdCd: CR1 not implemented yet\n");
      val_32 = 0;
      break;
    case 2: /* CR2 */
      if (bx_dbg.creg)
        bx_printf("MOV_RdCd: CR2\n");
      val_32 = BX_CPU_THIS_PTR cr2;
      break;
    case 3: /* CR3 */
      if (bx_dbg.creg)
        bx_printf("MOV_RdCd: reading CR3\n");
      val_32 = BX_CPU_THIS_PTR cr3;
      break;
    case 4: /* CR4 */
      bx_panic("MOV_RdCd: CR4 not implemented yet\n");
      val_32 = 0;
      break;
    default:
      bx_panic("MOV_RdCd: control register index out of range\n");
      val_32 = 0;
    }
  BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, val_32);
#endif
}

INLINE void bx_cpu_c::MOV_TdRd() {bx_panic("MOV_TdRd:\n");}
INLINE void bx_cpu_c::MOV_RdTd() {bx_panic("MOV_RdTd:\n");}

  INLINE void
bx_cpu_c::WAIT()
{
#if BX_CPU < 3
  bx_panic("WAIT: not implemented for < 386\n");
#elif BX_CPU > 3
  bx_panic("WAIT: not implemented for > 386\n");
#else // BX_CPU == 3

  BX_CPU_THIS_PTR single_step_event = 0; // ??? WAIT doesn't generate single steps

  if ( BX_CPU_THIS_PTR cr0.ts && BX_CPU_THIS_PTR cr0.mp ) {
    exception(BX_NM_EXCEPTION, 0, 0); // no error
    }
#endif
}

  INLINE void
bx_cpu_c::LOADALL()
{
#if BX_CPU < 2
  bx_panic("undocumented LOADALL instruction not supported on 8086\n");
#else
  Bit16u msw, tr, flags, ip, ldtr;
  Bit16u ds_raw, ss_raw, cs_raw, es_raw;
  Bit16u di, si, bp, sp, bx, dx, cx, ax;
  Bit16u base_15_0, limit;
  Bit8u  base_23_16, access;

  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

#if BX_CPU > 2
  bx_panic("loadall: not implemented for 386\n");
  /* ??? need to set G and other bits, and compute .limit_scaled also */
  /* for all segments CS,DS,SS,... */
#endif

  if (protected_mode()) {
    bx_panic(
      "LOADALL not yet supported for protected mode\n");
    }

  if (bx_dbg._protected) bx_printf("loadall():\n");

  /* MSW */
  bx_mem.read_physical(0x806, 2, &msw);
  BX_CPU_THIS_PTR cr0.pe = (msw & 0x01); msw >>= 1;
  BX_CPU_THIS_PTR cr0.mp = (msw & 0x01); msw >>= 1;
  BX_CPU_THIS_PTR cr0.em = (msw & 0x01); msw >>= 1;
  BX_CPU_THIS_PTR cr0.ts = (msw & 0x01);
  if (bx_dbg._protected)
    bx_printf("LOADALL: pe=%u, mp=%u, em=%u, ts=%u\n",
      (unsigned) BX_CPU_THIS_PTR cr0.pe, (unsigned) BX_CPU_THIS_PTR cr0.mp,
      (unsigned) BX_CPU_THIS_PTR cr0.em, (unsigned) BX_CPU_THIS_PTR cr0.ts);
  if (BX_CPU_THIS_PTR cr0.pe || BX_CPU_THIS_PTR cr0.mp || BX_CPU_THIS_PTR cr0.em || BX_CPU_THIS_PTR cr0.ts)
    bx_panic("LOADALL set PE, MP, EM or TS bits in MSW!\n");

  /* TR */
  bx_mem.read_physical(0x816, 2, &tr);
  BX_CPU_THIS_PTR tr.selector.value = tr;
  BX_CPU_THIS_PTR tr.selector.rpl   = (tr & 0x03);  tr >>= 2;
  BX_CPU_THIS_PTR tr.selector.ti    = (tr & 0x01);  tr >>= 1;
  BX_CPU_THIS_PTR tr.selector.index = tr;
  bx_mem.read_physical(0x860, 2, &base_15_0);
  bx_mem.read_physical(0x862, 1, &base_23_16);
  bx_mem.read_physical(0x863, 1, &access);
  bx_mem.read_physical(0x864, 2, &limit);

  if (bx_dbg._protected) {
    bx_printf("loadall: tr.AR byte: %02x\n", (unsigned) access);
    bx_printf("  tr.base : %06x\n", (unsigned) BX_CPU_THIS_PTR tr.cache.u.tss286.base);
    bx_printf("  tr.limit: %04x\n", (unsigned) BX_CPU_THIS_PTR tr.cache.u.tss286.limit);
    }

  BX_CPU_THIS_PTR tr.cache.valid =
  BX_CPU_THIS_PTR tr.cache.p           = (access & 0x80) >> 7;
  BX_CPU_THIS_PTR tr.cache.dpl         = (access & 0x60) >> 5;
  BX_CPU_THIS_PTR tr.cache.segment     = (access & 0x10) >> 4;
  BX_CPU_THIS_PTR tr.cache.type        = (access & 0x0f);
  BX_CPU_THIS_PTR tr.cache.u.tss286.base  = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR tr.cache.u.tss286.limit = limit;

  if ( (BX_CPU_THIS_PTR tr.selector.value & 0xfffc) == 0 ) {
    if (bx_dbg._protected)
      bx_printf("loadall: tr.value null\n");
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if ( BX_CPU_THIS_PTR tr.cache.valid == 0 ) {
    if (bx_dbg._protected)
      bx_printf("LOADALL: TR.valid = 0\n");
    }
  if ( BX_CPU_THIS_PTR tr.cache.u.tss286.limit < 43 ) {
    if (bx_dbg._protected) {
      bx_printf("  tr.segment: %u\n", (unsigned) BX_CPU_THIS_PTR tr.cache.segment);
      bx_printf("  tr.base : %06x\n", (unsigned) BX_CPU_THIS_PTR tr.cache.u.tss286.base);
      bx_printf("  tr.limit: %04x\n", (unsigned) BX_CPU_THIS_PTR tr.cache.u.tss286.limit);
      bx_printf("  AR byte: %02x\n", (unsigned) access);
      bx_printf("loadall: tr.limit<43\n");
      }
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if ( BX_CPU_THIS_PTR tr.cache.type != 1 ) {
    if (bx_dbg._protected)
      bx_printf("LOADALL: TR.type(%u) != 1\n", (unsigned) BX_CPU_THIS_PTR tr.cache.type);
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if ( BX_CPU_THIS_PTR tr.cache.segment ) {
    if (bx_dbg._protected)
      bx_printf("LOADALL: TR.segment = 1\n");
    BX_CPU_THIS_PTR tr.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR tr.cache.valid==0) {
    BX_CPU_THIS_PTR tr.cache.u.tss286.base   = 0;
    BX_CPU_THIS_PTR tr.cache.u.tss286.limit  = 0;
    BX_CPU_THIS_PTR tr.cache.p            = 0;
    BX_CPU_THIS_PTR tr.selector.value     = 0;
    BX_CPU_THIS_PTR tr.selector.index     = 0;
    BX_CPU_THIS_PTR tr.selector.ti        = 0;
    BX_CPU_THIS_PTR tr.selector.rpl       = 0;
    }


  /* FLAGS */
  bx_mem.read_physical(0x818, 2, &flags);
  write_flags(flags, 1, 1);

  /* IP */
  bx_mem.read_physical(0x81a, 2, &ip);
  IP = ip;

  /* LDTR */
  bx_mem.read_physical(0x81c, 2, &ldtr);
  BX_CPU_THIS_PTR ldtr.selector.value = ldtr;
  BX_CPU_THIS_PTR ldtr.selector.rpl   = (ldtr & 0x03);  ldtr >>= 2;
  BX_CPU_THIS_PTR ldtr.selector.ti    = (ldtr & 0x01);  ldtr >>= 1;
  BX_CPU_THIS_PTR ldtr.selector.index = ldtr;
  if ( (BX_CPU_THIS_PTR ldtr.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR ldtr.cache.valid   = 0;
    BX_CPU_THIS_PTR ldtr.cache.p       = 0;
    BX_CPU_THIS_PTR ldtr.cache.segment = 0;
    BX_CPU_THIS_PTR ldtr.cache.type    = 0;
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.base = 0;
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.limit = 0;
    BX_CPU_THIS_PTR ldtr.selector.value = 0;
    BX_CPU_THIS_PTR ldtr.selector.index = 0;
    BX_CPU_THIS_PTR ldtr.selector.ti    = 0;
    if (bx_dbg._protected)
      bx_printf("loadall: invalidating LDTR, null selector loaded.\n");
    }
  else {
    bx_mem.read_physical(0x854, 2, &base_15_0);
    bx_mem.read_physical(0x856, 1, &base_23_16);
    bx_mem.read_physical(0x857, 1, &access);
    bx_mem.read_physical(0x858, 2, &limit);
    BX_CPU_THIS_PTR ldtr.cache.valid      =
    BX_CPU_THIS_PTR ldtr.cache.p          = access >> 7;
    BX_CPU_THIS_PTR ldtr.cache.dpl        = (access >> 5) & 0x03;
    BX_CPU_THIS_PTR ldtr.cache.segment    = (access >> 4) & 0x01;
    BX_CPU_THIS_PTR ldtr.cache.type       = (access & 0x0f);
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.base = (base_23_16 << 16) | base_15_0;
    BX_CPU_THIS_PTR ldtr.cache.u.ldt.limit = limit;

    if (access == 0) {
      bx_panic("loadall: LDTR case access byte=0.\n");
      }
    if ( BX_CPU_THIS_PTR ldtr.cache.valid==0 ) {
      bx_panic("loadall: ldtr.valid=0\n");
      }
    if (BX_CPU_THIS_PTR ldtr.cache.segment) { /* not a system segment */
      bx_printf("         AR byte = %02x\n", (unsigned) access);
      bx_panic("loadall: LDTR descriptor cache loaded with non system segment\n");
      }
    if ( BX_CPU_THIS_PTR ldtr.cache.type != 2 ) {
      bx_panic("loadall: LDTR.type(%u) != 2\n", (unsigned) (access & 0x0f));
      }
    if (bx_dbg._protected) {
      bx_printf("loadall: ldtr.limit = %04x\n",
        (unsigned) BX_CPU_THIS_PTR ldtr.cache.u.ldt.limit);
      bx_printf("loadall: ldtr.base = %04x\n",
        (unsigned) BX_CPU_THIS_PTR ldtr.cache.u.ldt.base);
      }
    }

  /* DS */
  bx_mem.read_physical(0x81e, 2, &ds_raw);
  BX_CPU_THIS_PTR ds.selector.value = ds_raw;
  BX_CPU_THIS_PTR ds.selector.rpl   = (ds_raw & 0x03);  ds_raw >>= 2;
  BX_CPU_THIS_PTR ds.selector.ti    = (ds_raw & 0x01);  ds_raw >>= 1;
  BX_CPU_THIS_PTR ds.selector.index = ds_raw;
  bx_mem.read_physical(0x848, 2, &base_15_0);
  bx_mem.read_physical(0x84a, 1, &base_23_16);
  bx_mem.read_physical(0x84b, 1, &access);
  bx_mem.read_physical(0x84c, 2, &limit);
  BX_CPU_THIS_PTR ds.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR ds.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ds.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR ds.cache.valid      =
  BX_CPU_THIS_PTR ds.cache.p          = (access & 0x01);

  if ( (BX_CPU_THIS_PTR ds.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR ds.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR ds.cache.valid==0  ||
      BX_CPU_THIS_PTR ds.cache.segment==0) {
    bx_panic("loadall: DS invalid\n");
    }

  /* SS */
  bx_mem.read_physical(0x820, 2, &ss_raw);
  BX_CPU_THIS_PTR ss.selector.value = ss_raw;
  BX_CPU_THIS_PTR ss.selector.rpl   = (ss_raw & 0x03); ss_raw >>= 2;
  BX_CPU_THIS_PTR ss.selector.ti    = (ss_raw & 0x01); ss_raw >>= 1;
  BX_CPU_THIS_PTR ss.selector.index = ss_raw;
  bx_mem.read_physical(0x842, 2, &base_15_0);
  bx_mem.read_physical(0x844, 1, &base_23_16);
  bx_mem.read_physical(0x845, 1, &access);
  bx_mem.read_physical(0x846, 2, &limit);
  BX_CPU_THIS_PTR ss.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR ss.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR ss.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR ss.cache.p          = (access & 0x01);

  if ( (BX_CPU_THIS_PTR ss.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR ss.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR ss.cache.valid==0  ||
      BX_CPU_THIS_PTR ss.cache.segment==0) {
    bx_panic("loadall: SS invalid\n");
    }


  /* CS */
  bx_mem.read_physical(0x822, 2, &cs_raw);
  BX_CPU_THIS_PTR cs.selector.value = cs_raw;
  BX_CPU_THIS_PTR cs.selector.rpl   = (cs_raw & 0x03); cs_raw >>= 2;

  if (bx_dbg._protected)
    bx_printf("LOADALL: setting cs.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR cs.selector.rpl);

  BX_CPU_THIS_PTR cs.selector.ti    = (cs_raw & 0x01); cs_raw >>= 1;
  BX_CPU_THIS_PTR cs.selector.index = cs_raw;
  bx_mem.read_physical(0x83c, 2, &base_15_0);
  bx_mem.read_physical(0x83e, 1, &base_23_16);
  bx_mem.read_physical(0x83f, 1, &access);
  bx_mem.read_physical(0x840, 2, &limit);
  BX_CPU_THIS_PTR cs.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR cs.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR cs.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR cs.cache.p          = (access & 0x01);

  if ( (BX_CPU_THIS_PTR cs.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR cs.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR cs.cache.valid==0  ||
      BX_CPU_THIS_PTR cs.cache.segment==0) {
    bx_panic("loadall: CS invalid\n");
    }

  /* ES */
  bx_mem.read_physical(0x824, 2, &es_raw);
  BX_CPU_THIS_PTR es.selector.value = es_raw;
  BX_CPU_THIS_PTR es.selector.rpl   = (es_raw & 0x03); es_raw >>= 2;
  BX_CPU_THIS_PTR es.selector.ti    = (es_raw & 0x01); es_raw >>= 1;
  BX_CPU_THIS_PTR es.selector.index = es_raw;
  bx_mem.read_physical(0x836, 2, &base_15_0);
  bx_mem.read_physical(0x838, 1, &base_23_16);
  bx_mem.read_physical(0x839, 1, &access);
  bx_mem.read_physical(0x83a, 2, &limit);
  BX_CPU_THIS_PTR es.cache.u.segment.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR es.cache.u.segment.limit = limit;
  BX_CPU_THIS_PTR es.cache.u.segment.a          = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.u.segment.r_w        = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.u.segment.c_ed       = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.u.segment.executable = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.segment    = (access & 0x01); access >>= 1;
  BX_CPU_THIS_PTR es.cache.dpl        = (access & 0x03); access >>= 2;
  BX_CPU_THIS_PTR es.cache.p          = (access & 0x01);

  if (bx_dbg._protected) {
    bx_printf("cs.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR cs.cache.dpl);
    bx_printf("ss.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR ss.cache.dpl);
    bx_printf("BX_CPU_THIS_PTR ds.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR ds.cache.dpl);
    bx_printf("BX_CPU_THIS_PTR es.dpl = %02x\n", (unsigned) BX_CPU_THIS_PTR es.cache.dpl);
    bx_printf("LOADALL: setting cs.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR cs.selector.rpl);
    bx_printf("LOADALL: setting ss.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR ss.selector.rpl);
    bx_printf("LOADALL: setting ds.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR ds.selector.rpl);
    bx_printf("LOADALL: setting es.selector.rpl to %u\n",
      (unsigned) BX_CPU_THIS_PTR es.selector.rpl);
    }

  if ( (BX_CPU_THIS_PTR es.selector.value & 0xfffc) == 0 ) {
    BX_CPU_THIS_PTR es.cache.valid = 0;
    }
  if (BX_CPU_THIS_PTR es.cache.valid==0  ||
      BX_CPU_THIS_PTR es.cache.segment==0) {
    bx_panic("loadall: ES invalid\n");
    }

  /* DI */
  bx_mem.read_physical(0x826, 2, &di);
  DI = di;

  /* SI */
  bx_mem.read_physical(0x828, 2, &si);
  SI = si;

  /* BP */
  bx_mem.read_physical(0x82a, 2, &bp);
  BP = bp;

  /* SP */
  bx_mem.read_physical(0x82c, 2, &sp);
  SP = sp;

  /* BX */
  bx_mem.read_physical(0x82e, 2, &bx);
  BX = bx;

  /* DX */
  bx_mem.read_physical(0x830, 2, &dx);
  DX = dx;

  /* CX */
  bx_mem.read_physical(0x832, 2, &cx);
  CX = cx;

  /* AX */
  bx_mem.read_physical(0x834, 2, &ax);
  AX = ax;

  /* GDTR */
  bx_mem.read_physical(0x84e, 2, &base_15_0);
  bx_mem.read_physical(0x850, 1, &base_23_16);
  bx_mem.read_physical(0x851, 1, &access);
  bx_mem.read_physical(0x852, 2, &limit);
  BX_CPU_THIS_PTR gdtr.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR gdtr.limit = limit;
  if (access)
    if (bx_dbg._protected)
      bx_printf("LOADALL: GDTR access bits not 0 (%02x).\n",
        (unsigned) access);

  /* IDTR */
  bx_mem.read_physical(0x85a, 2, &base_15_0);
  bx_mem.read_physical(0x85c, 1, &base_23_16);
  bx_mem.read_physical(0x85d, 1, &access);
  bx_mem.read_physical(0x85e, 2, &limit);
  BX_CPU_THIS_PTR idtr.base = (base_23_16 << 16) | base_15_0;
  BX_CPU_THIS_PTR idtr.limit = limit;
  if (bx_dbg._protected) {
    bx_printf("LOADALL: idtr.limit = %04x\n", (unsigned) BX_CPU_THIS_PTR idtr.limit);
    bx_printf("LOADALL: idtr.base  = %06x\n", (unsigned) BX_CPU_THIS_PTR idtr.base);
    if (access)
      bx_printf("LOADALL: IDTR access bits not 0 (%02x).\n",
        (unsigned) access);
    }
#endif
}


  INLINE void
bx_cpu_c::CPUID()
{
  if (v8086_mode()) bx_panic("proc_ctrl: v8086 mode unsupported\n");

  bx_panic("CPUID: not implemented\n");
  invalidate_prefetch_q();
}
