// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com







#include "bochs.h"




 INLINE void
bx_cpu_c::ARPL_EwGw()
{
#if BX_CPU < 2
  bx_panic("ARPL_EwRw: not supported on 8086!\n");
#else /* 286+ */

  Bit16u op2_16, op1_16;

  decode_exgx();


  if (protected_mode()) {
    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    if ( (op1_16 & 0x03) < (op2_16 & 0x03) ) {
      op1_16 = (op1_16 & 0xfffc) | (op2_16 & 0x03);
      /* now write back to destination */
      if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
        if (BX_CPU_THIS_PTR is_32bit_opsize) {
          // if 32bit opsize, then 0xff3f is or'd into
          // upper 16bits of register
          Bit32u op1_32;

          op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
          op1_32 = (op1_32 & 0xffff0000) | op1_16;
          op1_32 |= 0xff3f0000;
          BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, op1_32);
          }
        else {
          BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, op1_16);
          }
        }
      else {
        write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
        }
      set_ZF(1);
      }
    else {
      set_ZF(0);
      }
    }
  else {
    // bx_printf("ARPL_EwRw: not recognized in real mode\n");
    exception(6, 0, 0);
    return;
    }
#endif
}

 INLINE void
bx_cpu_c::LAR_GvEw()
{
  /* for 16 bit operand size mode */
  Bit16u raw_selector;
  bx_descriptor_t descriptor;
  bx_selector_t   selector;
  Bit32u dword1, dword2;


  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  if (real_mode()) {
    bx_panic("LAR_GvEw: not recognized in real mode\n");
    exception(6, 0, 0);
    return;
    }

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    raw_selector = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &raw_selector);
    }

  /* if selector null, clear ZF and done */
  if ( (raw_selector & 0xfffc) == 0 ) {
    set_ZF(0);
    return;
    }

  parse_selector(raw_selector, &selector);

  if ( !fetch_raw_descriptor2(&selector, &dword1, &dword2) ) {
    /* not within descriptor table */
    set_ZF(0);
    return;
    }

  parse_descriptor(dword1, dword2, &descriptor);

  if (descriptor.valid==0) {
    set_ZF(0);
    //bx_printf("lar(): descriptor valid bit cleared\n");
    return;
    }

  /* if source selector is visible at CPL & RPL,
   * within the descriptor table, and of type accepted by LAR instruction,
   * then load register with segment limit and set ZF
   */

  if ( descriptor.segment ) { /* normal segment */
    if ( descriptor.u.segment.executable && descriptor.u.segment.c_ed ) {
      /* ignore DPL for conforming segments */
      }
    else {
      if ( (descriptor.dpl<CPL) || (descriptor.dpl<selector.rpl) ) {
        set_ZF(0);
        return;
        }
      }
    set_ZF(1);
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      /* masked by 00FxFF00, where x is undefined */
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, dword2 & 0x00ffff00);
      }
    else {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, dword2 & 0xff00);
      }
    return;
    }
  else { /* system or gate segment */
    switch ( descriptor.type ) {
      case 1: /* available TSS */
      case 2: /* LDT */
      case 3: /* busy TSS */
      case 4: /* 286 call gate */
      case 5: /* task gate */
#if BX_CPU >= 3
      case 9:  /* available 32bit TSS */
      case 11: /* busy 32bit TSS */
      case 12: /* 32bit call gate */
#endif
        break;
      default: /* rest not accepted types to LAR */
        set_ZF(0);
        bx_printf("lar(): not accepted type\n");
        return;
        break;
      }

    if ( (descriptor.dpl<CPL) || (descriptor.dpl<selector.rpl) ) {
      set_ZF(0);
      return;
      }
    set_ZF(1);
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      /* masked by 00FxFF00, where x is undefined ??? */
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, dword2 & 0x00ffff00);
      }
    else {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, dword2 & 0xff00);
      }
    return;
    }
}

 INLINE void
bx_cpu_c::LSL_GvEw()
{
  /* for 16 bit operand size mode */
  Bit16u raw_selector;
  Bit32u limit32;
  //bx_descriptor_t descriptor;
  bx_selector_t   selector;
  Bit32u dword1, dword2;
  Bit32u descriptor_dpl;

  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  decode_exgx();

  if (real_mode()) {
    bx_panic("LSL_GvEw: not recognized in real mode\n");
    exception(6, 0, 0);
    return;
    }

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    raw_selector = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &raw_selector);
    }


  /* if selector null, clear ZF and done */
  if ( (raw_selector & 0xfffc) == 0 ) {
    set_ZF(0);
    return;
    }

  parse_selector(raw_selector, &selector);

  if ( !fetch_raw_descriptor2(&selector, &dword1, &dword2) ) {
    /* not within descriptor table */
    set_ZF(0);
    return;
    }

  //parse_descriptor(dword1, dword2, &descriptor);

  descriptor_dpl = (dword2 >> 13) & 0x03;

  if ( (dword2 & 0x00001000) == 0 ) { // system segment
    Bit32u type;

    type = (dword2 >> 8) & 0x0000000f;
    switch (type) {
      case 1: // 16bit TSS
      case 3: // 16bit TSS
      case 2: // LDT
      case 9: // 32bit TSS    G00A
      case 11:// 32bit TSS    G00A
        limit32 = (dword1 & 0x0000ffff) | (dword2 & 0x000f0000);
        if ( dword2 & 0x00800000 )
          limit32 = (limit32 << 12) | 0x00000fff;
        if ( (descriptor_dpl<CPL) || (descriptor_dpl<selector.rpl) ) {
          set_ZF(0);
          return;
          }
        goto lsl_ok;
        break;
      default:
        set_ZF(0);
        return;
      }
    }
  else { // data & code segment
    limit32 = (dword1 & 0x0000ffff) | (dword2 & 0x000f0000);
    if ( dword2 & 0x00800000 )
      limit32 = (limit32 << 12) | 0x00000fff;
    if ( (dword2 & 0x00000c00) == 0x00000c00 ) {
      // conforming code segment, no check done
      goto lsl_ok;
      }

    if ( (descriptor_dpl<CPL) || (descriptor_dpl<selector.rpl) ) {
      set_ZF(0);
      return;
      }
    goto lsl_ok;
    }

lsl_ok:
  /* all checks pass, limit32 is now byte granular, write to op1 */
  set_ZF(1);

  if (BX_CPU_THIS_PTR is_32bit_opsize)
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, limit32)
  else
    // chop off upper 16 bits
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, (Bit16u) limit32)
}

 INLINE void
bx_cpu_c::SLDT_Ew()
{
#if BX_CPU < 2
  bx_panic("SLDT_Ew: not supported on 8086!\n");
#else
  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  if (real_mode()) {
    /* not recognized in real address mode */
    bx_panic("SLDT_Ew: encountered in real mode.\n");
    exception(6, 0, 0);
    }
  else {
    Bit16u val16; 

    val16 = BX_CPU_THIS_PTR ldtr.selector.value;
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, val16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &val16);
      }
    }
#endif
}

 INLINE void
bx_cpu_c::STR_Ew()
{
  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  if (real_mode()) {
    // not recognized in real address mode
    bx_panic("STR_Ew: encountered in real mode.\n");
    exception(6, 0, 0);
    }
  else {
    Bit16u val16;

    val16 = BX_CPU_THIS_PTR tr.selector.value;
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, val16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &val16);
      }
    }
}

 INLINE void
bx_cpu_c::LLDT_Ew()
{
#if BX_CPU < 2
  bx_panic("LLDT_Ew: not supported on 8086!\n");
#else
  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  invalidate_prefetch_q();

  if (real_mode()) {
    bx_panic("lldt: not recognized in real mode\n");
    exception(6, 0, 0); /* not recognized in real mode */
    return;
    }
  else { /* protected mode */
    bx_descriptor_t  descriptor;
    bx_selector_t    selector;
    Bit16u raw_selector;
    Bit32u dword1, dword2;


    /* #GP(0) if the current privilege level is not 0 */
    if (CPL != 0) {
      bx_panic("LLDT: CPL != 0\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      raw_selector = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &raw_selector);
      }

    /* if selector is NULL, invalidate and done */
    if ((raw_selector & 0xfffc) == 0) {
      bx_printf("lldt: case selector is null, OK.\n");
      BX_CPU_THIS_PTR ldtr.selector.value = raw_selector;
      BX_CPU_THIS_PTR ldtr.cache.valid = 0;
      return;
      }

    /* parse fields in selector */
    parse_selector(raw_selector, &selector);

    // #GP(selector) if the selector operand does not point into GDT
    if (selector.ti != 0) {
      bx_printf("LLDT: selector.ti != 0\n");
      exception(BX_GP_EXCEPTION, raw_selector & 0xfffc, 0);
      }

    if ((selector.index*8 + 7) > BX_CPU_THIS_PTR gdtr.limit) {
      bx_panic("lldt: GDT: index > limit\n");
      exception(BX_GP_EXCEPTION, raw_selector & 0xfffc, 0);
      return;
      }

    access_linear(BX_CPU_THIS_PTR gdtr.base + selector.index*8,     4, 0,
      BX_READ, &dword1);
    access_linear(BX_CPU_THIS_PTR gdtr.base + selector.index*8 + 4, 4, 0,
      BX_READ, &dword2);

    parse_descriptor(dword1, dword2, &descriptor);

    /* if selector doesn't point to an LDT descriptor #GP(selector) */
    if ( (descriptor.valid==0) ||
        descriptor.segment  ||
        (descriptor.type!=2) ) {
      bx_printf("lldt: doesn't point to an LDT descriptor!\n");
      exception(BX_GP_EXCEPTION, raw_selector & 0xfffc, 0);
      }

    /* #NP(selector) if LDT descriptor is not present */
    if (descriptor.p==0) {
      bx_printf("lldt: LDT descriptor not present!\n");
      exception(BX_NP_EXCEPTION, raw_selector & 0xfffc, 0);
      }

    if (descriptor.u.ldt.limit < 7) {
      bx_printf("lldt: ldtr.limit < 7\n");
      }

    BX_CPU_THIS_PTR ldtr.selector = selector;
    BX_CPU_THIS_PTR ldtr.cache = descriptor;
    BX_CPU_THIS_PTR ldtr.cache.valid = 1;

    return;
    }
#endif
}

 INLINE void
bx_cpu_c::LTR_Ew()
{
#if BX_CPU < 2
  bx_panic("LTR_Ew: not supported on 8086!\n");
#else
  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");


  invalidate_prefetch_q();

  if (protected_mode()) {
    bx_descriptor_t  descriptor;
    bx_selector_t    selector;
    Bit16u raw_selector;
    Bit32u dword1, dword2;


    /* #GP(0) if the current privilege level is not 0 */
    if (CPL != 0) {
      bx_panic("LTR: CPL != 0\n");
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      raw_selector = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &raw_selector);
      }

    /* if selector is NULL, invalidate and done */
    if ((raw_selector & 0xfffc) == 0) {
      bx_panic("ltr: loading with NULL selector!\n");
      /* if this is OK, then invalidate and load selector & descriptor cache */
      /* load here */
      BX_CPU_THIS_PTR tr.selector.value = raw_selector;
      BX_CPU_THIS_PTR tr.cache.valid = 0;
      return;
      }

    /* parse fields in selector, then check for null selector */
    parse_selector(raw_selector, &selector);

    if (selector.ti) {
      bx_panic("ltr: selector.ti != 0\n");
      return;
      }

    /* fetch 2 dwords of descriptor; call handles out of limits checks */
    fetch_raw_descriptor(&selector, &dword1, &dword2, BX_GP_EXCEPTION);

    parse_descriptor(dword1, dword2, &descriptor);

    /* #GP(selector) if object is not a TSS or is already busy */
    if ( (descriptor.valid==0) || descriptor.segment  ||
         (descriptor.type!=1 && descriptor.type!=9) ) {
      bx_panic("ltr: doesn't point to an available TSS descriptor!\n");
      exception(BX_GP_EXCEPTION, raw_selector & 0xfffc, 0); /* 0 ??? */
      return;
      }

    /* #NP(selector) if TSS descriptor is not present */
    if (descriptor.p==0) {
      bx_panic("ltr: LDT descriptor not present!\n");
      exception(BX_NP_EXCEPTION, raw_selector & 0xfffc, 0); /* 0 ??? */
      return;
      }

    if (descriptor.type==1 && descriptor.u.tss286.limit<43) {
      bx_panic("ltr:286TSS: loading tr.limit < 43\n");
      }
    else if (descriptor.type==9 && descriptor.u.tss386.limit<103) {
      bx_panic("ltr:386TSS: loading tr.limit < 103\n");
      }

    BX_CPU_THIS_PTR tr.selector = selector;
    BX_CPU_THIS_PTR tr.cache    = descriptor;
    BX_CPU_THIS_PTR tr.cache.valid = 1;

    /* mark as busy */
    dword2 |= 0x00000200; /* set busy bit */
    access_linear(BX_CPU_THIS_PTR gdtr.base + selector.index*8 + 4, 4, 0,
      BX_WRITE, &dword2);

    return;
    }
  else {
    bx_panic("ltr_ew: not recognized in real-mode!\n");
    exception(6, 0, 0);
    return;
    }
#endif
}

 INLINE void
bx_cpu_c::VERR_Ew()
{
  /* for 16 bit operand size mode */
  Bit16u raw_selector;
  bx_descriptor_t descriptor;
  bx_selector_t   selector;
  Bit32u dword1, dword2;

  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");


  if (real_mode()) {
    bx_panic("VERR_Ew: not recognized in real mode\n");
    exception(6, 0, 0);
    return;
    }

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    raw_selector = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &raw_selector);
    }

  /* if selector null, clear ZF and done */
  if ( (raw_selector & 0xfffc) == 0 ) {
    set_ZF(0);
    bx_printf("VERR: null selector\n");
    return;
    }

  /* if source selector is visible at CPL & RPL,
   * within the descriptor table, and of type accepted by VERR instruction,
   * then load register with segment limit and set ZF */
  parse_selector(raw_selector, &selector);

  if ( !fetch_raw_descriptor2(&selector, &dword1, &dword2) ) {
    /* not within descriptor table */
    set_ZF(0);
    bx_printf("VERR: not in table\n");
    return;
    }

  parse_descriptor(dword1, dword2, &descriptor);

  if ( descriptor.segment==0 ) { /* system or gate descriptor */
    set_ZF(0); /* inaccessible */
    bx_printf("VERR: system descriptor\n");
    return;
    }

  if ( descriptor.valid==0 ) {
    set_ZF(0);
    bx_printf("VERR: valid bit cleared\n");
    return;
    }

/* ??? */
#if 0
  if ( descriptor.p==0 ) { /* if not present */
    set_ZF(0); /* inaccessible */
    bx_printf("VERR: not present\n");
    return;
    }
#endif

  /* normal data/code segment */
  if ( descriptor.u.segment.executable ) { /* code segment */
    /* ignore DPL for readable conforming segments */
    if ( descriptor.u.segment.c_ed &&
         descriptor.u.segment.r_w) {
      set_ZF(1); /* accessible */
      bx_printf("VERR: conforming code, OK\n");
      return;
      }
    if ( descriptor.u.segment.r_w==0 ) {
      set_ZF(0); /* inaccessible */
      bx_printf("VERR: code not readable\n");
      return;
      }
    /* readable, non-conforming code segment */
    if ( (descriptor.dpl<CPL) || (descriptor.dpl<selector.rpl) ) {
      set_ZF(0); /* inaccessible */
      bx_printf("VERR: non-coforming code not withing priv level\n");
      return;
      }
    set_ZF(1); /* accessible */
    bx_printf("VERR: code seg readable\n");
    return;
    }
  else { /* data segment */
    if ( (descriptor.dpl<CPL) || (descriptor.dpl<selector.rpl) ) {
      set_ZF(0); /* not accessible */
      bx_printf("VERR: data seg not withing priv level\n");
      return;
      }
    set_ZF(1); /* accessible */
    bx_printf("VERR: data segment OK\n");
    return;
    }
}

 INLINE void
bx_cpu_c::VERW_Ew()
{
  /* for 16 bit operand size mode */
  Bit16u raw_selector;
  bx_descriptor_t descriptor;
  bx_selector_t   selector;
  Bit32u dword1, dword2;

  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");


  if (real_mode()) {
    bx_panic("VERW_Ew: not recognized in real mode\n");
    exception(6, 0, 0);
    return;
    }

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    raw_selector = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &raw_selector);
    }

  /* if selector null, clear ZF and done */
  if ( (raw_selector & 0xfffc) == 0 ) {
    set_ZF(0);
    bx_printf("VERW: null selector\n");
    return;
    }

  /* if source selector is visible at CPL & RPL,
   * within the descriptor table, and of type accepted by VERW instruction,
   * then load register with segment limit and set ZF */
  parse_selector(raw_selector, &selector);

  if ( !fetch_raw_descriptor2(&selector, &dword1, &dword2) ) {
    /* not within descriptor table */
    set_ZF(0);
    bx_printf("VERW: not in table\n");
    return;
    }

  parse_descriptor(dword1, dword2, &descriptor);

  /* rule out system segments & code segments */
  if ( descriptor.segment==0 || descriptor.u.segment.executable ) {
    set_ZF(0);
    bx_printf("VERW: system seg or code\n");
    return;
    }

  if ( descriptor.valid==0 ) {
    set_ZF(0);
    bx_printf("VERW: valid bit cleared\n");
    return;
    }

/* ??? */
#if 0
  if ( descriptor.p==0 ) { /* if not present */
    set_ZF(0); /* inaccessible */
    bx_printf("VERW: not present\n");
    return;
    }
#endif

  /* data segment */
  if ( descriptor.u.segment.r_w ) { /* writable */
    if ( (descriptor.dpl<CPL) || (descriptor.dpl<selector.rpl) ) {
      set_ZF(0); /* not accessible */
      bx_printf("VERW: writable data seg not within priv level\n");
      return;
      }
    set_ZF(1); /* accessible */
    bx_printf("VERW: data seg writable\n");
    return;
    }

  set_ZF(0); /* not accessible */
  bx_printf("VERW: data seg not writable\n");
  return;
}

 INLINE void
bx_cpu_c::SGDT_Ms()
{
#if BX_CPU < 2
  bx_panic("SGDT_Ms: not supported on 8086!\n");
#else
  Bit16u limit_16;
  Bit32u base_32;

  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");


  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    /* undefined opcode exception */
    bx_panic("SGDT_Ms: use of register is undefined opcode.\n");
    exception(6, 0, 0);
    return;
    }

  limit_16 = BX_CPU_THIS_PTR gdtr.limit;
  base_32  = BX_CPU_THIS_PTR gdtr.base;
#if BX_CPU == 2
  base_32 |= 0xff000000; /* ??? */
#else /* 386+ */
  /* 32bit processors always write 32bits of base */
#endif
  write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &limit_16);

  write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+2, &base_32);

#endif
}

 INLINE void
bx_cpu_c::SIDT_Ms()
{
#if BX_CPU < 2
  bx_panic("SIDT_Ms: not supported on 8086!\n");
#else
  Bit16u limit_16;
  Bit32u base_32;

  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    /* undefined opcode exception */
    bx_panic("SIDT: use of register is undefined opcode.\n");
    exception(6, 0, 0);
    return;
    }

  limit_16 = BX_CPU_THIS_PTR idtr.limit;
  base_32  = BX_CPU_THIS_PTR idtr.base;

#if BX_CPU == 2
  base_32 |= 0xff000000;
#else /* 386+ */
  /* ??? regardless of operand size, all 32bits of base are stored */
#endif

  write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &limit_16);

  write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+2, &base_32);

#endif
}

 INLINE void
bx_cpu_c::LGDT_Ms()
{
#if BX_CPU < 2
  bx_panic("LGDT_Ms: not supported on 8086!\n");
#else

  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  invalidate_prefetch_q();

  if (protected_mode() && (CPL!=0)) {
    bx_panic("LGDT: protected mode: CPL!=0\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    bx_panic("LGDT generating exception 6\n");
    exception(6, 0, 0);
    return;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit16u limit_16;
    Bit32u base0_31;

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &limit_16);

    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &base0_31);

    BX_CPU_THIS_PTR gdtr.limit = limit_16;
    BX_CPU_THIS_PTR gdtr.base = base0_31;
    }
  else
#endif
    {
    Bit16u limit_16, base0_15;
    Bit8u base16_23;

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &limit_16);

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &base0_15);

    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 4, &base16_23);

    /* ignore high 8 bits */

    BX_CPU_THIS_PTR gdtr.limit = limit_16;
    BX_CPU_THIS_PTR gdtr.base = (base16_23 << 16) | base0_15;
    }
#endif
}

 INLINE void
bx_cpu_c::LIDT_Ms()
{
#if BX_CPU < 2
  bx_panic("LIDT_Ms: not supported on 8086!\n");
#else
  Bit16u limit_16;
  Bit32u base_32;


  if (v8086_mode()) bx_panic("protect_ctrl: v8086 mode unsupported\n");

  invalidate_prefetch_q();

  if (protected_mode()) {
    if (CPL != 0) {
      bx_panic("LIDT(): CPL(%u) != 0\n", (unsigned) CPL);
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    /* undefined opcode exception */
    bx_panic("LIDT generating exception 6\n");
    exception(6, 0, 0);
    return;
    }
  
  read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &limit_16);

  read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &base_32);

  BX_CPU_THIS_PTR idtr.limit = limit_16;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize)
    BX_CPU_THIS_PTR idtr.base = base_32;
  else
#endif
    BX_CPU_THIS_PTR idtr.base = base_32 & 0x00ffffff; /* ignore upper 8 bits */

#endif
}
