/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/








#include "bochs.h"






  INLINE void
bx_cpu_c::LES_GvMp()
{
  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    bx_panic("invalid use of LES, must use memory reference!\n");
    exception(6, 0, 0);
    return;
    }
  
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit16u es;
    Bit32u reg_32;

    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 4, &es);

    load_seg_reg(&BX_CPU_THIS_PTR es, es);

    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, reg_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit mode */
    Bit16u reg_16, es;

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &es);

    load_seg_reg(&BX_CPU_THIS_PTR es, es);

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, reg_16);
    }
}

  INLINE void
bx_cpu_c::LDS_GvMp()
{
  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    bx_panic("invalid use of LDS, must use memory reference!\n");
    exception(6, 0, 0);
    return;
    }
  
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit16u ds;
    Bit32u reg_32;

    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 4, &ds);

    load_seg_reg(&BX_CPU_THIS_PTR ds, ds);

    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, reg_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit mode */
    Bit16u reg_16, ds;

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &ds);

    load_seg_reg(&BX_CPU_THIS_PTR ds, ds);

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, reg_16);
    }
}

  void
bx_cpu_c::LFS_GvMp()
{
#if BX_CPU < 3
  bx_panic("lfs_gvmp: not supported on 8086\n");
#else /* 386+ */

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    bx_panic("invalid use of LFS, must use memory reference!\n");
    exception(6, 0, 0);
    return;
    }
  
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u reg_32;
    Bit16u fs;

    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 4, &fs);

    load_seg_reg(&BX_CPU_THIS_PTR fs, fs);

    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, reg_32);
    }
  else { /* 16 bit operand size */
    Bit16u reg_16;
    Bit16u fs;

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &fs);

    load_seg_reg(&BX_CPU_THIS_PTR fs, fs);

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, reg_16);
    }
#endif
}

  void
bx_cpu_c::LGS_GvMp()
{
#if BX_CPU < 3
  bx_panic("lgs_gvmp: not supported on 8086\n");
#else /* 386+ */

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    bx_panic("invalid use of LGS, must use memory reference!\n");
    exception(6, 0, 0);
    return;
    }
  
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u reg_32;
    Bit16u gs;

    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 4, &gs);

    load_seg_reg(&BX_CPU_THIS_PTR gs, gs);

    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, reg_32);
    }
  else { /* 16 bit operand size */
    Bit16u reg_16;
    Bit16u gs;

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &gs);

    load_seg_reg(&BX_CPU_THIS_PTR gs, gs);

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, reg_16);
    }
#endif
}

  void
bx_cpu_c::LSS_GvMp()
{
#if BX_CPU < 3
  bx_panic("lss_gvmp: not supported on 8086\n");
#else /* 386+ */

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    bx_panic("invalid use of LSS, must use memory reference!\n");
    exception(6, 0, 0);
    return;
    }
  
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u reg_32;
    Bit16u ss_raw;

    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_32);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 4, &ss_raw);

    load_seg_reg(&BX_CPU_THIS_PTR ss, ss_raw);

    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, reg_32);
    }
  else { /* 16 bit operand size */
    Bit16u reg_16;
    Bit16u ss_raw;

    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &reg_16);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr + 2, &ss_raw);

    load_seg_reg(&BX_CPU_THIS_PTR ss, ss_raw);

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, reg_16);
    }
#endif
}
