// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com






#include "bochs.h"



  INLINE void
bx_cpu_c::SHLD_EvGv(unsigned count)
{
  /* op1:op2 << count.  result stored in op1 */

  count &= 0x1f; // use only 5 LSB's


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, op2_32, result_32;

    if (!count) return; /* NOP */

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    result_32 = (op1_32 << count) | (op2_32 >> (32 - count));

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
      }
    else {
      write_RMW_virtual_dword(result_32);
      }

    /* set eflags:
     * SHLD count affects the following flags: S,Z,P,C,O
     */
    set_CF((op1_32 >> (32 - count)) & 0x01);
    set_OF(((op1_32 ^ result_32) & 0x80000000) > 0);
    set_ZF(result_32 == 0);
    set_PF_base(result_32);
    set_SF(result_32 >> 31);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, op2_16, result_16;
    Bit32u temp_32, result_32;

    if (!count) return; /* NOP */
    // count is 1..31

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    temp_32 = (op1_16 << 16) | (op2_16); // double formed by op1:op2
    result_32 = temp_32 << count;
    if (count > 16) {
      // hack to act like x86 SHLD when count > 16
      // actually shifting op1:op2:op2 << count
      result_32 |= (op2_16 << (count - 16));
      }
    result_16 = result_32 >> 16;

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
      }
    else {
      write_RMW_virtual_word(result_16);
      }

    /* set eflags:
     * SHLD count affects the following flags: S,Z,P,C,O
     */
    set_CF( (temp_32 >> (32 - count)) & 0x01 );
    set_OF(((op1_16 ^ result_16) & 0x8000) > 0);
    set_ZF(result_16 == 0);
    set_SF(result_16 >> 15);
    set_PF_base((Bit8u) result_16);
    }
}


  INLINE void
bx_cpu_c::SHRD_EvGv(unsigned count)
{
#if BX_CPU < 3
  bx_panic("shrd_evgvib: not supported on < 386\n");
#else
  count &= 0x1F; /* use only 5 LSB's */

  if (!count) return; /* NOP */

  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, op2_32, result_32;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    result_32 = (op2_32 << (32 - count)) | (op1_32 >> count);

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
      }
    else {
      write_RMW_virtual_dword(result_32);
      }

    /* set eflags:
     * SHRD count affects the following flags: S,Z,P,C,O
     */

    set_CF((op1_32 >> (count - 1)) & 0x01);
    set_ZF(result_32 == 0);
    set_SF(result_32 >> 31);
    /* for shift of 1, OF set if sign change occurred. */
    set_OF(((op1_32 ^ result_32) & 0x80000000) > 0);
    set_PF_base(result_32);
    }
  else { /* 16 bit opsize */
    Bit16u op1_16, op2_16, result_16;
    Bit32u temp_32, result_32;

    // count is 1..31

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    temp_32 = (op2_16 << 16) | op1_16; // double formed by op2:op1
    result_32 = temp_32 >> count;
    if (count > 16) {
      // hack to act like x86 SHLD when count > 16
      // actually shifting op2:op2:op1 >> count
      result_32 |= (op2_16 << (32 - count));
      }
    result_16 = result_32;

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
      }
    else {
      write_RMW_virtual_word(result_16);
      }

    /* set eflags:
     * SHRD count affects the following flags: S,Z,P,C,O
     */

    set_CF((temp_32 >> (count - 1)) & 0x01);
    set_ZF(result_16 == 0);
    set_SF(result_16 >> 15);
    /* for shift of 1, OF set if sign change occurred. */
    set_OF(((op1_16 ^ result_16) & 0x8000) > 0);
    set_PF_base((Bit8u) result_16);
    }
#endif /* BX_CPU >= 3 */
}


  INLINE void
bx_cpu_c::ROL_Eb(unsigned count)
{
  Bit8u op1_8, result_8;

  count &= 0x07; // use only lowest 3 bits

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_RMW_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  if (count) {
    result_8 = (op1_8 << count) | (op1_8 >> (8 - count));

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, result_8);
      }
    else {
      write_RMW_virtual_byte(result_8);
      }

    /* set eflags:
     * ROL count affects the following flags: C
     */

    set_CF(result_8 & 0x01);
    set_OF(((op1_8 ^ result_8) & 0x80) > 0);
    }
}



  INLINE void
bx_cpu_c::ROL_Ev(unsigned count)
{

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, result_32;

    count &= 0x1f; /* only use bottom 5 bits */

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    if (count) {
      result_32 = (op1_32 << count) | (op1_32 >> (32 - count));

      /* now write result back to destination */
      if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
        BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
        }
      else {
        write_RMW_virtual_dword(result_32);
        }

      /* set eflags:
       * ROL count affects the following flags: C
       */

      set_CF(result_32 & 0x01);
      set_OF(((op1_32 ^ result_32) & 0x80000000) > 0);
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, result_16;

    count &= 0x0f; // only use bottom 4 bits

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    if (count) {
      result_16 = (op1_16 << count) | (op1_16 >> (16 - count));

      /* now write result back to destination */
      if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
        BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
        }
      else {
        write_RMW_virtual_word(result_16);
        }

      /* set eflags:
       * ROL count affects the following flags: C
       */

      set_CF(result_16 & 0x01);
      set_OF(((op1_16 ^ result_16) & 0x8000) > 0);
      }
    }
}



  INLINE void
bx_cpu_c::ROR_Eb(unsigned count)
{
  Bit8u op1_8, result_8;
  Bit8u result_b7;


  count &= 0x07; /* use only bottom 3 bits */

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_RMW_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  if (count) {
    result_8 = (op1_8 >> count) | (op1_8 << (8 - count));

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, result_8);
      }
    else {
      write_RMW_virtual_byte(result_8);
      }

    /* set eflags:
     * ROR count affects the following flags: C
     */
    result_b7 = result_8 & 0x80;

    set_CF(result_b7 != 0);
    set_OF(((op1_8 ^ result_8) & 0x80) > 0);
    }
}



  INLINE void
bx_cpu_c::ROR_Ev(unsigned count)
{

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, result_32, result_b31;

    count &= 0x1f;  // use only 5 LSB's

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    if (count) {
      result_32 = (op1_32 >> count) | (op1_32 << (32 - count));

      /* now write result back to destination */
      if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
        BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
        }
      else {
        write_RMW_virtual_dword(result_32);
        }

      /* set eflags:
       * ROR count affects the following flags: C
       */
      result_b31 = result_32 & 0x80000000;

      set_CF(result_b31 != 0);
      set_OF(((op1_32 ^ result_32) & 0x80000000) > 0);
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, result_16, result_b15;

    count &= 0x0f;  // use only 4 LSB's

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    if (count) {
      result_16 = (op1_16 >> count) | (op1_16 << (16 - count));

      /* now write result back to destination */
      if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
        BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
        }
      else {
        write_RMW_virtual_word(result_16);
        }

      /* set eflags:
       * ROR count affects the following flags: C
       */
      result_b15 = result_16 & 0x8000;

      set_CF(result_b15 != 0);
      set_OF(((op1_16 ^ result_16) & 0x8000) > 0);
      }
    }
}



  INLINE void
bx_cpu_c::RCL_Eb(unsigned count)
{
  Bit8u op1_8, result_8;

  count = (count & 0x1F) % 9;


  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_RMW_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  if (count) {
    result_8 = (op1_8 << count) |
             (get_CF() << (count - 1)) |
             (op1_8 >> (9 - count));

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, result_8);
      }
    else {
      write_RMW_virtual_byte(result_8);
      }

    /* set eflags:
     * RCL count affects the following flags: C
     */
    set_OF(((op1_8 ^ result_8) & 0x80) > 0);
    set_CF((op1_8 >> (8 - count)) & 0x01);
    }
}


  INLINE void
bx_cpu_c::RCL_Ev(unsigned count)
{
  count &= 0x1F;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, result_32;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    if (!count) return;

    if (count==1) {
      result_32 = (op1_32 << 1) | get_CF();
      }
    else {
      result_32 = (op1_32 << count) |
                (get_CF() << (count - 1)) |
                (op1_32 >> (33 - count));
      }

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
      }
    else {
      write_RMW_virtual_dword(result_32);
      }

    /* set eflags:
     * RCL count affects the following flags: C
     */
    set_OF(((op1_32 ^ result_32) & 0x80000000) > 0);
    set_CF((op1_32 >> (32 - count)) & 0x01);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, result_16;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    count %= 17;

    if (!count) return;

    if (count==1) {
      result_16 = (op1_16 << 1) | get_CF();
      }
    else if (count==16) {
      result_16 = (get_CF() << 15) |
                  (op1_16 >> 1);
      }
    else { // 2..15
      result_16 = (op1_16 << count) |
                  (get_CF() << (count - 1)) |
                  (op1_16 >> (17 - count));
      }

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
      }
    else {
      write_RMW_virtual_word(result_16);
      }

    /* set eflags:
     * RCL count affects the following flags: C
     */

    set_OF(((op1_16 ^ result_16) & 0x8000) > 0);
    set_CF((op1_16 >> (16 - count)) & 0x01);
    }
}



  INLINE void
bx_cpu_c::RCR_Eb(unsigned count)
{
  Bit8u op1_8, result_8;

  count = ( count & 0x1F ) % 9;

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_RMW_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  if (count) {
    result_8 = (op1_8 >> count) |
             (get_CF() << (8 - count)) |
             (op1_8 << (9 - count));

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, result_8);
      }
    else {
      write_RMW_virtual_byte(result_8);
      }

    /* set eflags:
     * RCR count affects the following flags: C
     */

    set_CF((op1_8 >> (count - 1)) & 0x01);
    set_OF(((op1_8 ^ result_8) & 0x80) > 0);
    }
}



  INLINE void
bx_cpu_c::RCR_Ev(unsigned count)
{
  count = count & 0x1F;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, result_32;


    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    if (!count) return;

    if (count==1) {
      result_32 = (op1_32 >> 1) | (get_CF() << 31);
      }
    else {
      result_32 = (op1_32 >> count) |
                (get_CF() << (32 - count)) |
                (op1_32 << (33 - count));
      }

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
      }
    else {
      write_RMW_virtual_dword(result_32);
      }

    /* set eflags:
     * RCR count affects the following flags: C
     */

    set_CF((op1_32 >> (count - 1)) & 0x01);
    set_OF(((op1_32 ^ result_32) & 0x80000000) > 0);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, result_16;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    count %= 17;
    if (count) {
      result_16 = (op1_16 >> count) |
		(get_CF() << (16 - count)) |
		(op1_16 << (17 - count));

      /* now write result back to destination */
      if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
	BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
	}
      else {
	write_RMW_virtual_word(result_16);
        }

      /* set eflags:
       * RCR count affects the following flags: C
       */

      set_CF((op1_16 >> (count - 1)) & 0x01);
      set_OF(((op1_16 ^ result_16) & 0x8000) > 0);
      }
    }
}



  INLINE void
bx_cpu_c::SHL_Eb(unsigned count)
{
  Bit8u op1_8, result_8;

  count &= 0x1F;

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_RMW_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  if (!count) return;

  result_8 = (op1_8 << count);

  /* now write result back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, result_8);
    }
  else {
    write_RMW_virtual_byte(result_8);
    }

  SET_FLAGS_OSZAPC_8(op1_8, count, result_8, BX_INSTR_SHL8);
}





  INLINE void
bx_cpu_c::SHL_Ev(unsigned count)
{
  count &= 0x1F; /* use only 5 LSB's */

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, result_32;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    if (!count) return;

    result_32 = (op1_32 << count);

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
      }
    else {
      write_RMW_virtual_dword(result_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, count, result_32, BX_INSTR_SHL32);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, result_16;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    if (!count) return;

    result_16 = (op1_16 << count);

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
      }
    else {
      write_RMW_virtual_word(result_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, count, result_16, BX_INSTR_SHL16);
    }
}




  INLINE void
bx_cpu_c::SHR_Eb(unsigned count)
{
  Bit8u op1_8, result_8;

  count &= 0x1F;

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_RMW_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  if (!count) return;

  result_8 = (op1_8 >> count);

  /* now write result back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, result_8);
    }
  else {
    write_RMW_virtual_byte(result_8);
    }

  SET_FLAGS_OSZAPC_8(op1_8, count, result_8, BX_INSTR_SHR8);
}


  INLINE void
bx_cpu_c::SHR_Ev(unsigned count)
{
  count &= 0x1F; /* use only 5 LSB's */

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, result_32;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    if (!count) return;

    result_32 = (op1_32 >> count);

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
      }
    else {
      write_RMW_virtual_dword(result_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, count, result_32, BX_INSTR_SHR32);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, result_16;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    if (!count) return;

    result_16 = (op1_16 >> count);


    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
      }
    else {
      write_RMW_virtual_word(result_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, count, result_16, BX_INSTR_SHR16);
    }
}



  INLINE void
bx_cpu_c::SAR_Eb(unsigned count)
{
  Bit8u op1_8, result_8;


  count &= 0x1F;

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_RMW_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  if (!count) return;

  if (count < 8) {
    if (op1_8 & 0x80) {
      result_8 = (op1_8 >> count) | (0xff << (8 - count));
      }
    else {
      result_8 = (op1_8 >> count);
      }
    }
  else {
    if (op1_8 & 0x80) {
      result_8 = 0xff;
      }
    else {
      result_8 = 0;
      }
    }

  /* now write result back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, result_8);
    }
  else {
    write_RMW_virtual_byte(result_8);
    }

  /* set eflags:
   * SAR count affects the following flags: S,Z,P,C
   */

  if (count < 8) {
    set_CF((op1_8 >> (count - 1)) & 0x01);
    }
  else {
    if (op1_8 & 0x80) {
      set_CF(1);
      }
    else {
      set_CF(0);
      }
    }

  set_ZF(result_8 == 0);
  set_SF(result_8 >> 7);
  set_OF(0);
  set_PF_base(result_8);
}



  INLINE void
bx_cpu_c::SAR_Ev(unsigned count)
{
  count &= 0x1F;  /* use only 5 LSB's */

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32, result_32;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    if (!count) return;

    /* count < 32, since only lower 5 bits used */
    if (op1_32 & 0x80000000) {
      result_32 = (op1_32 >> count) | (0xffffffff << (32 - count));
      }
    else {
      result_32 = (op1_32 >> count);
      }

    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, result_32);
      }
    else {
      write_RMW_virtual_dword(result_32);
      }

    /* set eflags:
     * SAR count affects the following flags: S,Z,P,C
     */

    set_CF((op1_32 >> (count - 1)) & 0x01);
    set_ZF(result_32 == 0);
    set_SF(result_32 >> 31);
    set_OF(0);
    set_PF_base(result_32);
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u op1_16, result_16;

    /* op1 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_RMW_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    if (!count) return;

    if (count < 16) {
      if (op1_16 & 0x8000) {
	result_16 = (op1_16 >> count) | (0xffff << (16 - count));
	}
      else {
	result_16 = (op1_16 >> count);
	}
      }
    else {
      if (op1_16 & 0x8000) {
	result_16 = 0xffff;
	}
      else {
	result_16 = 0;
	}
      }



    /* now write result back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, result_16);
      }
    else {
      write_RMW_virtual_word(result_16);
      }

    /* set eflags:
     * SAR count affects the following flags: S,Z,P,C
     */
    if (count < 16) {
      set_CF((op1_16 >> (count - 1)) & 0x01);
      }
    else {
      if (op1_16 & 0x8000) {
	set_CF(1);
	}
      else {
	set_CF(0);
	}
      }

    set_ZF(result_16 == 0);
    set_SF(result_16 >> 15);
    set_OF(0);
    set_PF_base((Bit8u) result_16);
    }
}
