// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com








#include "bochs.h"





  INLINE void
bx_cpu_c::BOUND_GvMa()
{
#if BX_CPU < 2
  bx_panic("BOUND_GvMa: not supported on 8086!\n");
#else

  bx_printf("BOUND_GvMa: beta!\n");

  decode_exgx();

  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    /* undefined opcode exception */
    bx_panic("bound: op2 must be mem ref\n");
    exception(6, 0, 0);
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32s bound_min, bound_max;
    Bit32s op1_32;

    op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);
  
    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, (Bit32u *) &bound_min);
    read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+4, (Bit32u *) &bound_max);
  
    /* ??? */
    if ( (op1_32 < bound_min) || (op1_32 > bound_max) ) {
      bx_printf("BOUND: fails bounds test\n");
      exception(5, 0, 0);
      }
    }
  else {
    Bit16s bound_min, bound_max;
    Bit16s op1_16;

    op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);
  
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, (Bit16u *) &bound_min);
    read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr+2, (Bit16u *) &bound_max);
  
    /* ??? */
    if ( (op1_16 < bound_min) || (op1_16 > bound_max) ) {
      bx_printf("BOUND: fails bounds test\n");
      exception(5, 0, 0);
      }
    }

#endif
}

  INLINE void
bx_cpu_c::INT3()
{
  /* INT 3 is not IOPL sensitive */

//bx_panic("INT3: bailing\n");
  interrupt(3, 1, 0, 0);
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
}

  INLINE void
bx_cpu_c::INT_Ib()
{
  Bit8u imm8;

  imm8 = fetch_next_byte();

  if (v8086_mode() && (IOPL<3)) {
    //bx_printf("int_ib: v8086: IOPL<3\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    }

#if 0
if ( (imm8 == 0x21) && (AH == 0x4c) ) {
  fprintf(stderr, "# INT 21/4C called AL=%02x, BX=%04x\n",
          (unsigned) AL, (unsigned) BX);
  bx_printf("INT 21/4C called AL=%02x, BX=%04x\n", (unsigned) AL, (unsigned) BX);
  }
#endif

  interrupt(imm8, 1, 0, 0);
  BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
}

  INLINE void
bx_cpu_c::INTO()
{
  /* ??? is this IOPL sensitive ? */
  if (v8086_mode()) bx_panic("soft_int: v8086 mode unsupported\n");

  if (get_OF()) {
    interrupt(4, 1, 0, 0);
    BX_INSTR_FAR_BRANCH(BX_CPU_THIS_PTR cs.selector.value, BX_CPU_THIS_PTR eip);
    }
}
