/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/







#include "bochs.h"



  INLINE void
bx_cpu_c::PUSH_ES()
{
#if BX_CPU >= 3
  /* push_16() & push_32() will use StackAddrSize to determine
   * use of either SP or ESP
   */
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* OperandSize = 32 */
    push_32(BX_CPU_THIS_PTR es.selector.value);
    }
  else
#endif
    { /* OperandSize = 16 */
    push_16(BX_CPU_THIS_PTR es.selector.value);
    }
}

  INLINE void
bx_cpu_c::PUSH_CS()
{
#if BX_CPU >= 3
  /* push_16() & push_32() will use StackAddrSize to determine
   * use of either SP or ESP
   */
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* OperandSize = 32 */
    push_32(BX_CPU_THIS_PTR cs.selector.value);
    }
  else
#endif
    { /* OperandSize = 16 */
    push_16(BX_CPU_THIS_PTR cs.selector.value);
    }
}

  INLINE void
bx_cpu_c::PUSH_SS()
{
#if BX_CPU >= 3
  /* push_16() & push_32() will use StackAddrSize to determine
   * use of either SP or ESP
   */
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* OperandSize = 32 */
    push_32(BX_CPU_THIS_PTR ss.selector.value);
    }
  else
#endif
    { /* OperandSize = 16 */
    push_16(BX_CPU_THIS_PTR ss.selector.value);
    }
}

  INLINE void
bx_cpu_c::PUSH_DS()
{
#if BX_CPU >= 3
  /* push_16() & push_32() will use StackAddrSize to determine
   * use of either SP or ESP
   */
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* OperandSize = 32 */
    push_32(BX_CPU_THIS_PTR ds.selector.value);
    }
  else
#endif
    { /* OperandSize = 16 */
    push_16(BX_CPU_THIS_PTR ds.selector.value);
    }
}

  INLINE void
bx_cpu_c::PUSH_FS()
{
#if BX_CPU >= 3
  /* push_16() & push_32() will use StackAddrSize to determine
   * use of either SP or ESP
   */
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* OperandSize = 32 */
    push_32(BX_CPU_THIS_PTR fs.selector.value);
    }
  else
#endif
    { /* OperandSize = 16 */
    push_16(BX_CPU_THIS_PTR fs.selector.value);
    }
}

  INLINE void
bx_cpu_c::PUSH_GS()
{
#if BX_CPU >= 3
  /* push_16() & push_32() will use StackAddrSize to determine
   * use of either SP or ESP
   */
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* OperandSize = 32 */
    push_32(BX_CPU_THIS_PTR gs.selector.value);
    }
  else
#endif
    { /* OperandSize = 16 */
    push_16(BX_CPU_THIS_PTR gs.selector.value);
    }
}

  INLINE void
bx_cpu_c::POP_ES()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u es;

    pop_32(&es);

    load_seg_reg(&BX_CPU_THIS_PTR es, (Bit16u) es);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit opsize */
    Bit16u es;

    pop_16(&es);

    load_seg_reg(&BX_CPU_THIS_PTR es, es);
    }
}


  INLINE void
bx_cpu_c::POP_SS()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u ss_raw;

    pop_32(&ss_raw);

    load_seg_reg(&BX_CPU_THIS_PTR ss, (Bit16u) ss_raw);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit opsize */
    Bit16u ss_raw;

    pop_16(&ss_raw);

    load_seg_reg(&BX_CPU_THIS_PTR ss, ss_raw);
    }
}

  INLINE void
bx_cpu_c::POP_DS()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u ds_raw;

    pop_32(&ds_raw);

    load_seg_reg(&BX_CPU_THIS_PTR ds, (Bit16u) ds_raw);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit opsize */
    Bit16u ds_raw;

    pop_16(&ds_raw);

    load_seg_reg(&BX_CPU_THIS_PTR ds, ds_raw);
    }
}

  INLINE void
bx_cpu_c::POP_FS()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u fs_raw;

    pop_32(&fs_raw);

    load_seg_reg(&BX_CPU_THIS_PTR fs, (Bit16u) fs_raw);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit opsize */
    Bit16u fs_raw;

    pop_16(&fs_raw);

    load_seg_reg(&BX_CPU_THIS_PTR fs, fs_raw);
    }
}

  INLINE void
bx_cpu_c::POP_GS()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u gs_raw;

    pop_32(&gs_raw);

    load_seg_reg(&BX_CPU_THIS_PTR gs, (Bit16u) gs_raw);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit opsize */
    Bit16u gs_raw;

    pop_16(&gs_raw);

    load_seg_reg(&BX_CPU_THIS_PTR gs, gs_raw);
    }
}

  INLINE void
bx_cpu_c::PUSHAD()
{
#if BX_CPU < 2
  bx_panic("PUSHAD: not supported on an 8086\n");
#else
  Bit32u temp_ESP;

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    temp_ESP = ESP;
  else
    temp_ESP = SP;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u esp;

    if (protected_mode()) {
      if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, 32) ) {
        bx_panic("PUSHAD(): stack doesn't have enough room!\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }
    else {
      if (temp_ESP < 32)
        bx_panic("pushad: eSP < 32\n");
      }

    esp = ESP;

    /* ??? optimize this by using virtual write, all checks passed */
    push_32(EAX);
    push_32(ECX);
    push_32(EDX);
    push_32(EBX);
    push_32(esp);
    push_32(EBP);
    push_32(ESI);
    push_32(EDI);
    }
  else
#endif
    {
    Bit16u sp;

#if BX_CPU >= 2
    if (protected_mode()) {
      if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, 16) ) {
        bx_panic("PUSHA(): stack doesn't have enough room!\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }
    else
#endif
      {
      if (temp_ESP < 16)
        bx_panic("pushad: eSP < 16\n");
      }

    sp = SP;

    /* ??? optimize this by using virtual write, all checks passed */
    push_16(AX);
    push_16(CX);
    push_16(DX);
    push_16(BX);
    push_16(sp);
    push_16(BP);
    push_16(SI);
    push_16(DI);
    }
  if (bx_dbg._protected)
    bx_printf("PUSHAD\n");
#endif
}

  INLINE void
bx_cpu_c::POPAD()
{
#if BX_CPU < 2
  bx_panic("POPAD not supported on an 8086\n");
#else /* 286+ */

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u edi, esi, ebp, etmp, ebx, edx, ecx, eax;

    if (protected_mode()) {
      if ( !can_pop(32) ) {
        bx_panic("pop_ad: not enough bytes on stack\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }

    if (bx_dbg._protected)
      bx_printf("POPAD:\n");
    /* ??? optimize this */
    pop_32(&edi);
    pop_32(&esi);
    pop_32(&ebp);
    pop_32(&etmp); /* value for ESP discarded */
    pop_32(&ebx);
    pop_32(&edx);
    pop_32(&ecx);
    pop_32(&eax);

    EDI = edi;
    ESI = esi;
    EBP = ebp;
    EBX = ebx;
    EDX = edx;
    ECX = ecx;
    EAX = eax;
    }
  else
#endif
    {
    Bit16u di, si, bp, tmp, bx, dx, cx, ax;

    if (protected_mode()) {
      if ( !can_pop(16) ) {
        bx_panic("pop_a: not enough bytes on stack\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }

    if (bx_dbg._protected)
      bx_printf("POPA:\n");
    /* ??? optimize this */
    pop_16(&di);
    pop_16(&si);
    pop_16(&bp);
    pop_16(&tmp); /* value for SP discarded */
    pop_16(&bx);
    pop_16(&dx);
    pop_16(&cx);
    pop_16(&ax);

    DI = di;
    SI = si;
    BP = bp;
    BX = bx;
    DX = dx;
    CX = cx;
    AX = ax;
    }
#endif
}

  INLINE void
bx_cpu_c::PUSH_Iv()
{
#if BX_CPU < 2
  bx_panic("PUSH_Iv: not supported on 8086!\n");
#else

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u imm32;

    imm32 = fetch_next_dword();

    push_32(imm32);
    /* will return after error */
    }
  else
    {
    Bit16u imm16;

    imm16 = fetch_next_word();

    push_16(imm16);
    /* will return after error */
    }
#endif
}

  INLINE void
bx_cpu_c::PUSH_Ib()
{
#if BX_CPU < 2
  bx_panic("PUSH_Ib: not supported on 8086!\n");
#else
  Bit8s  imm8;


  imm8 = fetch_next_byte();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32s imm32;

    imm32 = imm8;
    push_32(imm32);
    /* will return after error */
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit operand size */
    Bit16s imm16;

    imm16 = imm8;
    push_16(imm16);
    /* will return after error */
    }
#endif /* BX_CPU < 2*/
}

  INLINE void
bx_cpu_c::POP_Ev()
{

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u val32;

    // decode_exgx needs to be called after the pop_32(), since eSP
    // changes during the pop, and decode_exgx returns a value
    // which may use eSP in the calculation
    pop_32(&val32);
    decode_exgx();

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, val32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &val32);
      }
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u val16;

    // decode_exgx needs to be called after the pop_16(), since eSP
    // changes during the pop, and decode_exgx returns a value
    pop_16(&val16);
    decode_exgx();

    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, val16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &val16);
      }
    }
}

  INLINE void
bx_cpu_c::PUSH_Ev()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    push_32(op1_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    push_16(op1_16);
    }
}


  INLINE void
bx_cpu_c::ENTER_IwIb()
{
#if BX_CPU < 2
  bx_panic("ENTER_IwIb: not supported by 8086!\n");
#else
  Bit16u dynamic_size;
  Bit8u  level;
  Bit32u frame_ptr32;
  Bit16u frame_ptr16;

  dynamic_size = fetch_next_word();

  level = fetch_next_byte();

  invalidate_prefetch_q();

  level %= 32;
/* ??? */
if (level) bx_panic("enter(): level > 0\n");
//if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b && BX_CPU_THIS_PTR is_32bit_opsize==0) {
//  bx_printf("enter(): stacksize!=opsize: I'm unsure of the code for this\n");
//  bx_panic("         The Intel manuals are a mess on this one!\n");
//  }

  if ( protected_mode() ) {
    Bit32u bytes_to_push, temp_ESP;

    if (level == 0) {
      if (BX_CPU_THIS_PTR is_32bit_opsize)
        bytes_to_push = 4 + dynamic_size;
      else
        bytes_to_push = 2 + dynamic_size;
      }
    else { /* level > 0 */
      if (BX_CPU_THIS_PTR is_32bit_opsize)
        bytes_to_push = 4 + (level-1)*4 + 4 + dynamic_size;
      else
        bytes_to_push = 2 + (level-1)*2 + 2 + dynamic_size;
      }
    if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
      temp_ESP = ESP;
    else
      temp_ESP = SP;
    if ( !can_push(&BX_CPU_THIS_PTR ss.cache, temp_ESP, bytes_to_push) ) {
      bx_panic("ENTER: not enough room on stack!\n");
      exception(BX_SS_EXCEPTION, 0, 0);
      }
    if (bx_dbg._protected) bx_printf("ENTER(protected mode): succeeding\n");
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize)
    push_32(EBP);
  else
    push_16(BP);

  // can just do frame_ptr32 = ESP for either case ???
  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    frame_ptr32 = ESP;
  else
    frame_ptr32 = SP;

  if (level > 0) {
    /* do level-1 times */
    while (--level) {
      if (BX_CPU_THIS_PTR is_32bit_opsize) {
        Bit32u temp32;

        if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
          EBP -= 4;
          read_virtual_dword(&BX_CPU_THIS_PTR ss, EBP, &temp32);
          ESP -= 4;
          write_virtual_dword(&BX_CPU_THIS_PTR ss, ESP, &temp32);
          }
        else { /* 16bit stacksize */
          BP -= 4;
          read_virtual_dword(&BX_CPU_THIS_PTR ss, BP, &temp32);
          SP -= 4;
          write_virtual_dword(&BX_CPU_THIS_PTR ss, SP, &temp32);
          }
        }
      else { /* 16bit opsize */
        Bit16u temp16;

        if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
          EBP -= 2;
          read_virtual_word(&BX_CPU_THIS_PTR ss, EBP, &temp16);
          ESP -= 2;
          write_virtual_word(&BX_CPU_THIS_PTR ss, ESP, &temp16);
          }
        else { /* 16bit stacksize */
          BP -= 2;
          read_virtual_word(&BX_CPU_THIS_PTR ss, BP, &temp16);
          SP -= 2;
          write_virtual_word(&BX_CPU_THIS_PTR ss, SP, &temp16);
          }
        }
      } /* while (--level) */

    /* push(frame pointer) */
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
        ESP -= 4;
        write_virtual_dword(&BX_CPU_THIS_PTR ss, ESP, &frame_ptr32);
        }
      else {
        SP -= 4;
        write_virtual_dword(&BX_CPU_THIS_PTR ss, SP, &frame_ptr32);
        }
      }
    else { /* 16bit opsize */
      if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
        frame_ptr16 = frame_ptr32;
        ESP -= 2;
        write_virtual_word(&BX_CPU_THIS_PTR ss, ESP, &frame_ptr16);
        }
      else {
        frame_ptr16 = frame_ptr32;
        SP -= 2;
        write_virtual_word(&BX_CPU_THIS_PTR ss, SP, &frame_ptr16);
        }
      }
    } /* if (level > 0) ... */

  if (BX_CPU_THIS_PTR is_32bit_opsize)
    EBP = frame_ptr32;
  else
    BP = frame_ptr32;

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b) { /* 32bit stacksize */
    ESP = ESP - dynamic_size;
    }
  else { /* 16bit stack */
    SP = SP - dynamic_size;
    }
#endif
}

  INLINE void
bx_cpu_c::LEAVE()
{
#if BX_CPU < 2
  bx_panic("LEAVE: not supported by 8086!\n");
#else
  Bit32u temp_EBP;


  invalidate_prefetch_q();

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    temp_EBP = EBP;
  else
#endif
    temp_EBP = BP;

  if ( protected_mode() ) {
    if (BX_CPU_THIS_PTR ss.cache.u.segment.c_ed) { /* expand up */
      if (temp_EBP <= BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled) {
        bx_panic("LEAVE: BP > BX_CPU_THIS_PTR ss.limit\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }
    else { /* normal */
      if (temp_EBP > BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled) {
        bx_panic("LEAVE: BP > BX_CPU_THIS_PTR ss.limit\n");
        exception(BX_SS_EXCEPTION, 0, 0);
        return;
        }
      }
    if (bx_dbg._protected) bx_printf("LEAVE(protected mode): succeeding\n");
    }


  // delete frame
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    ESP = EBP;
  else
#endif
    SP = BP;

  // restore frame pointer
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u temp32;

    pop_32(&temp32);
    EBP = temp32;
    }
  else
#endif
    {
    Bit16u temp16;

    pop_16(&temp16);
    BP = temp16;
    }
#endif
}
