/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/






#include "bochs.h"


/* MOVSB ES:[EDI], DS:[ESI]   DS may be overridden
 *   mov string from DS:[ESI] into ES:[EDI]
 */

  INLINE void
bx_cpu_c::MOVSB_XbYb()
{
  bx_segment_reg_t *seg;
  Bit8u temp8;

  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u esi, edi;

    esi = ESI;
    edi = EDI;

    read_virtual_byte(seg, esi, &temp8);

    write_virtual_byte(&BX_CPU_THIS_PTR es, edi, &temp8);

    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement ESI, EDI */
      esi--;
      edi--;
      }
    else {
      /* increment ESI, EDI */
      esi++;
      edi++;
      }

    ESI = esi;
    EDI = edi;
    }

  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit address mode */
    Bit16u si, di;

    si = SI;
    di = DI;

    read_virtual_byte(seg, si, &temp8);

    write_virtual_byte(&BX_CPU_THIS_PTR es, di, &temp8);

    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement SI, DI */
      si--;
      di--;
      }
    else {
      /* increment SI, DI */
      si++;
      di++;
      }

    SI = si;
    DI = di;
    }
}

  INLINE void
bx_cpu_c::MOVSW_XvYv()
{
  bx_segment_reg_t *seg;


  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u  temp32;

    Bit32u esi, edi;

    esi = ESI;
    edi = EDI;

    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      read_virtual_dword(seg, esi, &temp32);

      write_virtual_dword(&BX_CPU_THIS_PTR es, edi, &temp32);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        esi -= 4;
        edi -= 4;
        }
      else {
        /* increment ESI */
        esi += 4;
        edi += 4;
        }
      } /* if (BX_CPU_THIS_PTR is_32bit_opsize) ... */
    else { /* 16 bit opsize mode */
      Bit16u  temp16;

      read_virtual_word(seg, esi, &temp16);

      write_virtual_word(&BX_CPU_THIS_PTR es, edi, &temp16);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        esi -= 2;
        edi -= 2;
        }
      else {
        /* increment ESI */
        esi += 2;
        edi += 2;
        }
      }

    ESI = esi;
    EDI = edi;
    }

  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address mode */
    Bit16u si, di;

    si = SI;
    di = DI;

#if BX_CPU >= 3
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      Bit32u temp32;

      read_virtual_dword(seg, si, &temp32);

      write_virtual_dword(&BX_CPU_THIS_PTR es, di, &temp32);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        si -= 4;
        di -= 4;
        }
      else {
        /* increment ESI */
        si += 4;
        di += 4;
        }
      } /* if (BX_CPU_THIS_PTR is_32bit_opsize) ... */
    else
#endif /* BX_CPU >= 3 */
      { /* 16 bit opsize mode */
      Bit16u  temp16;

      read_virtual_word(seg, si, &temp16);

      write_virtual_word(&BX_CPU_THIS_PTR es, di, &temp16);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement SI, DI */
        si -= 2;
        di -= 2;
        }
      else {
        /* increment SI, DI */
        si += 2;
        di += 2;
        }
      }

    SI = si;
    DI = di;
    }
}

  INLINE void
bx_cpu_c::CMPSB_XbYb()
{
  bx_segment_reg_t *seg;
  Bit8u op1_8, op2_8, diff_8;


  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u esi, edi;

    esi = ESI;
    edi = EDI;

    read_virtual_byte(seg, esi, &op1_8);

    read_virtual_byte(&BX_CPU_THIS_PTR es, edi, &op2_8);

    diff_8 = op1_8 - op2_8;

    SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_CMPS8);

    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement ESI */
      esi--;
      edi--;
      }
    else {
      /* increment ESI */
      esi++;
      edi++;
      }

    EDI = edi;
    ESI = esi;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address mode */
    Bit16u si, di;

    si = SI;
    di = DI;

    read_virtual_byte(seg, si, &op1_8);

    read_virtual_byte(&BX_CPU_THIS_PTR es, di, &op2_8);

    diff_8 = op1_8 - op2_8;

    SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_CMPS8);

    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement ESI */
      si--;
      di--;
      }
    else {
      /* increment ESI */
      si++;
      di++;
      }

    DI = di;
    SI = si;
    }
}

  INLINE void
bx_cpu_c::CMPSW_XvYv()
{
  bx_segment_reg_t *seg;


  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u op1_32, op2_32, diff_32;
    Bit32u esi, edi;

    esi = ESI;
    edi = EDI;


    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      read_virtual_dword(seg, esi, &op1_32);

      read_virtual_dword(&BX_CPU_THIS_PTR es, edi, &op2_32);

      diff_32 = op1_32 - op2_32;

      SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_CMPS32);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        esi -= 4;
        edi -= 4;
        }
      else {
        /* increment ESI */
        esi += 4;
        edi += 4;
        }
      }
    else { /* 16 bit opsize */
      Bit16u op1_16, op2_16, diff_16;

      read_virtual_word(seg, esi, &op1_16);

      read_virtual_word(&BX_CPU_THIS_PTR es, edi, &op2_16);

      diff_16 = op1_16 - op2_16;

      SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_CMPS16);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        esi -= 2;
        edi -= 2;
        }
      else {
        /* increment ESI */
        esi += 2;
        edi += 2;
        }
      }


    EDI = edi;
    ESI = esi;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit address mode */
    Bit16u si, di;

    si = SI;
    di = DI;

#if BX_CPU >= 3
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      Bit32u op1_32, op2_32, diff_32;

      read_virtual_dword(seg, si, &op1_32);

      read_virtual_dword(&BX_CPU_THIS_PTR es, di, &op2_32);

      diff_32 = op1_32 - op2_32;

      SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_CMPS32);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        si -= 4;
        di -= 4;
        }
      else {
        /* increment ESI */
        si += 4;
        di += 4;
        }
      }
    else
#endif /* BX_CPU >= 3 */
      { /* 16 bit opsize */
      Bit16u op1_16, op2_16, diff_16;

      read_virtual_word(seg, si, &op1_16);

      read_virtual_word(&BX_CPU_THIS_PTR es, di, &op2_16);

      diff_16 = op1_16 - op2_16;

      SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_CMPS16);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        si -= 2;
        di -= 2;
        }
      else {
        /* increment ESI */
        si += 2;
        di += 2;
        }
      }


    DI = di;
    SI = si;
    }
}

  INLINE void
bx_cpu_c::SCASB_ALXb()
{
  Bit8u op1_8, op2_8, diff_8;


#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u edi;

    edi = EDI;

    op1_8 = AL;

    read_virtual_byte(&BX_CPU_THIS_PTR es, edi, &op2_8);

    diff_8 = op1_8 - op2_8;

    SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_SCAS8);


    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement ESI */
      edi--;
      }
    else {
      /* increment ESI */
      edi++;
      }

    EDI = edi;
    }

  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address mode */
    Bit16u di;

    di = DI;

    op1_8 = AL;

    read_virtual_byte(&BX_CPU_THIS_PTR es, di, &op2_8);

    diff_8 = op1_8 - op2_8;

    SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_SCAS8);

    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement ESI */
      di--;
      }
    else {
      /* increment ESI */
      di++;
      }

    DI = di;
    }
}

  INLINE void
bx_cpu_c::SCASW_eAXXv()
{
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u edi;

    edi = EDI;

    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      Bit32u op1_32, op2_32, diff_32;

      op1_32 = EAX;
      read_virtual_dword(&BX_CPU_THIS_PTR es, edi, &op2_32);

      diff_32 = op1_32 - op2_32;

      SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_SCAS32);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        edi -= 4;
        }
      else {
        /* increment ESI */
        edi += 4;
        }
      }
    else { /* 16 bit opsize */
      Bit16u op1_16, op2_16, diff_16;

      op1_16 = AX;
      read_virtual_word(&BX_CPU_THIS_PTR es, edi, &op2_16);

      diff_16 = op1_16 - op2_16;

      SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_SCAS16);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        edi -= 2;
        }
      else {
        /* increment ESI */
        edi += 2;
        }
      }

    EDI = edi;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address mode */
    Bit16u di;

    di = DI;

#if BX_CPU >= 3
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      Bit32u op1_32, op2_32, diff_32;

      op1_32 = EAX;
      read_virtual_dword(&BX_CPU_THIS_PTR es, di, &op2_32);

      diff_32 = op1_32 - op2_32;

      SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_SCAS32);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        di -= 4;
        }
      else {
        /* increment ESI */
        di += 4;
        }
      }
    else
#endif /* BX_CPU >= 3 */
      { /* 16 bit opsize */
      Bit16u op1_16, op2_16, diff_16;

      op1_16 = AX;
      read_virtual_word(&BX_CPU_THIS_PTR es, di, &op2_16);

      diff_16 = op1_16 - op2_16;

      SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_SCAS16);

      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        di -= 2;
        }
      else {
        /* increment ESI */
        di += 2;
        }
      }

    DI = di;
    }
}

  INLINE void
bx_cpu_c::STOSB_YbAL()
{
  Bit8u al;

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u edi;

    edi = EDI;

    al = AL;
    write_virtual_byte(&BX_CPU_THIS_PTR es, edi, &al);

    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement EDI */
      edi--;
      }
    else {
      /* increment EDI */
      edi++;
      }

    EDI = edi;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address size */
    Bit16u di;

    di = DI;

    al = AL;
    write_virtual_byte(&BX_CPU_THIS_PTR es, di, &al);

    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement EDI */
      di--;
      }
    else {
      /* increment EDI */
      di++;
      }

    DI = di;
    }
}

  INLINE void
bx_cpu_c::STOSW_YveAX()
{
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u edi;

    edi = EDI;

    if (BX_CPU_THIS_PTR is_32bit_opsize) {
        Bit32u eax;

        eax = EAX;
        write_virtual_dword(&BX_CPU_THIS_PTR es, edi, &eax);

        if (BX_CPU_THIS_PTR eflags.df) {
          /* decrement EDI */
          edi -= 4;
          }
        else {
          /* increment EDI */
          edi += 4;
          }
      } /* if (BX_CPU_THIS_PTR is_32bit_opsize) ... */
    else { /* 16 bit opsize mode */
        Bit16u ax;

        ax = AX;
        write_virtual_word(&BX_CPU_THIS_PTR es, edi, &ax);

        if (BX_CPU_THIS_PTR eflags.df) {
          /* decrement EDI */
          edi -= 2;
          }
        else {
          /* increment EDI */
          edi += 2;
          }
      }

    EDI = edi;
    }

  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address size */
    Bit16u di;

    di = DI;

#if BX_CPU >= 3
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
        Bit32u eax;

        eax = EAX;
        write_virtual_dword(&BX_CPU_THIS_PTR es, di, &eax);

        if (BX_CPU_THIS_PTR eflags.df) {
          /* decrement EDI */
          di -= 4;
          }
        else {
          /* increment EDI */
          di += 4;
          }
      } /* if (BX_CPU_THIS_PTR is_32bit_opsize) ... */
    else
#endif /* BX_CPU >= 3 */
      { /* 16 bit opsize mode */
        Bit16u ax;

        ax = AX;
        write_virtual_word(&BX_CPU_THIS_PTR es, di, &ax);

        if (BX_CPU_THIS_PTR eflags.df) {
          /* decrement EDI */
          di -= 2;
          }
        else {
          /* increment EDI */
          di += 2;
          }
      }

    DI = di;
    }
}


  INLINE void
bx_cpu_c::LODSB_ALXb()
{
  bx_segment_reg_t *seg;
  Bit8u al;

  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u esi;

    esi = ESI;

    read_virtual_byte(seg, esi, &al);

    AL = al;
    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement ESI */
      esi--;
      }
    else {
      /* increment ESI */
      esi++;
      }

    ESI = esi;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address mode */
    Bit16u si;

    si = SI;

    read_virtual_byte(seg, si, &al);

    AL = al;
    if (BX_CPU_THIS_PTR eflags.df) {
      /* decrement ESI */
      si--;
      }
    else {
      /* increment ESI */
      si++;
      }

    SI = si;
    }
}

  INLINE void
bx_cpu_c::LODSW_eAXXv()
{
  bx_segment_reg_t *seg;

  if (BX_CPU_THIS_PTR segment_override) {
    seg = BX_CPU_THIS_PTR segment_override;
    }
  else {
    seg = &BX_CPU_THIS_PTR ds;
    }

#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR is_32bit_addrsize) {
    Bit32u esi;

    esi = ESI;

    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      Bit32u eax;

      read_virtual_dword(seg, esi, &eax);

      EAX = eax;
      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        esi -= 4;
        }
      else {
        /* increment ESI */
        esi += 4;
        }
      } /* if (BX_CPU_THIS_PTR is_32bit_opsize) ... */
    else { /* 16 bit opsize mode */
      Bit16u ax;
      read_virtual_word(seg, esi, &ax);

      AX = ax;
      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        esi -= 2;
        }
      else {
        /* increment ESI */
        esi += 2;
        }
      }

    ESI = esi;
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16bit address mode */
    Bit16u si;

    si = SI;

#if BX_CPU >= 3
    if (BX_CPU_THIS_PTR is_32bit_opsize) {
      Bit32u eax;

      read_virtual_dword(seg, si, &eax);

      EAX = eax;
      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        si -= 4;
        }
      else {
        /* increment ESI */
        si += 4;
        }
      } /* if (BX_CPU_THIS_PTR is_32bit_opsize) ... */
    else
#endif /* BX_CPU >= 3 */
      { /* 16 bit opsize mode */
      Bit16u ax;

      read_virtual_word(seg, si, &ax);

      AX = ax;
      if (BX_CPU_THIS_PTR eflags.df) {
        /* decrement ESI */
        si -= 2;
        }
      else {
        /* increment ESI */
        si += 2;
        }
      }

    SI = si;
    }
}
