// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"





#if BX_SUPPORT_TASKING

#if BX_CPU >= 2

// Notes:
// ======
// Step 2: TSS descriptor is not busy TS (for IRET); GP (for JMP, CALL, INT)
//   returns error code (Task's backlink TSS)???

// *   TSS selector must map to GDT
// *   TSS is stored in linear address space
// * what to do with I/O Map Base
// * what to do with T flag
// * where to set CR3 and flush paging cache
// * what happens when fault occurs, with some seg regs having valid bit cleared?
// * should check validity of current TR(TSS) before writing into it
// 

  void
bx_cpu_c::task_switch(bx_selector_t *tss_selector,
                 bx_descriptor_t *tss_descriptor, unsigned source,
                 Bit32u dword1, Bit32u dword2)
{
  Bit32u obase32; // base address of old TSS
  Bit32u nbase32; // base address of new TSS
  Bit32u temp32, temp_cr3;
  Bit16u raw_cs_selector, raw_ss_selector, raw_ds_selector, raw_es_selector,
         raw_fs_selector, raw_gs_selector, raw_ldt_selector;
  Bit16u temp16, trap_word;
  bx_selector_t cs_selector, ss_selector, ds_selector, es_selector,
                fs_selector, gs_selector, ldt_selector, old_tr_selector;
  bx_descriptor_t cs_descriptor, ss_descriptor, ds_descriptor, es_descriptor,
                  fs_descriptor, gs_descriptor, ldt_descriptor;

//bx_printf("XXXXX task switch in XXXXX\n");
//fprintf(stderr, "# XXXXX task switch in XXXXX\n");

  invalidate_prefetch_q();


  // ======================
  // 286 Task State Segment
  // ======================
  // dynamic item                      | hex  dec  offset
  // 0       task LDT selector         | 2a   42
  // 1       DS selector               | 28   40
  // 1       SS selector               | 26   38
  // 1       CS selector               | 24   36
  // 1       ES selector               | 22   34
  // 1       DI                        | 20   32
  // 1       SI                        | 1e   30
  // 1       BP                        | 1c   28
  // 1       SP                        | 1a   26
  // 1       BX                        | 18   24
  // 1       DX                        | 16   22
  // 1       CX                        | 14   20
  // 1       AX                        | 12   18
  // 1       flag word                 | 10   16
  // 1       IP (entry point)          | 0e   14
  // 0       SS for CPL 2              | 0c   12
  // 0       SP for CPL 2              | 0a   10
  // 0       SS for CPL 1              | 08   08
  // 0       SP for CPL 1              | 06   06
  // 0       SS for CPL 0              | 04   04
  // 0       SP for CPL 0              | 02   02
  //         back link selector to TSS | 00   00


  // ======================
  // 386 Task State Segment
  // ======================
  // |31            16|15                    0|
  // |I/O Map Base    |000000000000000000000|T| 64  static
  // |0000000000000000| LDT                   | 60  static
  // |0000000000000000| GS selector           | 5c  dynamic
  // |0000000000000000| FS selector           | 58  dynamic
  // |0000000000000000| DS selector           | 54  dynamic
  // |0000000000000000| SS selector           | 50  dynamic
  // |0000000000000000| CS selector           | 4c  dynamic
  // |0000000000000000| ES selector           | 48  dynamic
  // |                EDI                     | 44  dynamic
  // |                ESI                     | 40  dynamic
  // |                EBP                     | 3c  dynamic
  // |                ESP                     | 38  dynamic
  // |                EBX                     | 34  dynamic
  // |                EDX                     | 30  dynamic
  // |                ECX                     | 2c  dynamic
  // |                EAX                     | 28  dynamic
  // |                EFLAGS                  | 24  dynamic
  // |                EIP (entry point)       | 20  dynamic
  // |           CR3 (PDPR)                   | 1c  static
  // |000000000000000 | SS for CPL 2          | 18  static
  // |           ESP for CPL 2                | 14  static
  // |000000000000000 | SS for CPL 1          | 10  static
  // |           ESP for CPL 1                | 0c  static
  // |000000000000000 | SS for CPL 0          | 08  static
  // |           ESP for CPL 0                | 04  static
  // |000000000000000 | back link to prev TSS | 00  dynamic (updated only when return expected)


  // The following checks are made before calling task_switch(), for
  // JMP & CALL only.  These checks are NOT made for exceptions, interrupts, & IRET
  //
  //   1) TSS DPL must be >= CPL
  //   2) TSS DPL must be >= TSS selector RPL
  //   3) TSS descriptor is not busy.  TS(for IRET); GP(for JMP, CALL, INT)

// ??? hack
if (tss_descriptor->type == 3) tss_descriptor->type = 1;
else if (tss_descriptor->type == 11) tss_descriptor->type = 9;

  // Task State Seg must be present, else #NP(TSS selector)
  if (tss_descriptor->p==0) {
    bx_printf("task_switch: TSS.p == 0\n");
    exception(BX_NP_EXCEPTION, tss_selector->value & 0xfffc, 0);
    return;
    }

  if (tss_descriptor->type == 1) {
    nbase32 = tss_descriptor->u.tss286.base; // new TSS.base
    // TSS segment limit >= 0x67, else #TS(new task tss)
    if (tss_selector->ti ||
        tss_descriptor->valid==0 ||
        tss_descriptor->u.tss286.limit < 43) {
      bx_panic("task_switch(): TSS not valid\n");
      exception(BX_TS_EXCEPTION, tss_selector->value & 0xfffc, 0);
      return;
      }
    }
  else if (tss_descriptor->type == 9) {
    nbase32 = tss_descriptor->u.tss386.base; // new TSS.base
    if (tss_selector->ti ||
        tss_descriptor->valid==0 ||
        tss_descriptor->u.tss386.limit < 0x67) {
      bx_panic("task_switch(): TR not valid\n");
      exception(BX_TS_EXCEPTION, tss_selector->value & 0xfffc, 0);
      return;
      }
    }
  else {
    bx_panic("error in task_switch TSS type\n");
    nbase32 = 0; // keep compiler happy
    }

  // See if current TR is valid, may be 32 or 16bit
  if (BX_CPU_THIS_PTR tr.cache.type == 1) {
    if (BX_CPU_THIS_PTR tr.cache.valid==0 ||
        BX_CPU_THIS_PTR tr.cache.u.tss286.limit < 43) {
      bx_panic("task_switch(): TR16 not valid.\n");
      //exception(BX_TS_EXCEPTION, BX_CPU_THIS_PTR tr.selector.value & 0xfffc, 0);
      }
//bx_printf("tasking: saving state to 16bit TSS\n");

    // save machine state (dynamic part) in current 16bit TSS
    obase32 = BX_CPU_THIS_PTR tr.cache.u.tss286.base; // current TSS.base
    temp16 = IP; access_linear(obase32 + 14, 2, 0, BX_WRITE, &temp16);
    temp16 = read_flags();
    if (source == BX_TASK_FROM_IRET) {
      // NT flags in old task is cleared with an IRET
      temp16 &= ~0x4000;
      }
                 access_linear(obase32 + 16, 2, 0, BX_WRITE, &temp16);
    temp16 = AX; access_linear(obase32 + 18, 2, 0, BX_WRITE, &temp16);
    temp16 = CX; access_linear(obase32 + 20, 2, 0, BX_WRITE, &temp16);
    temp16 = DX; access_linear(obase32 + 22, 2, 0, BX_WRITE, &temp16);
    temp16 = BX; access_linear(obase32 + 24, 2, 0, BX_WRITE, &temp16);
    temp16 = SP; access_linear(obase32 + 26, 2, 0, BX_WRITE, &temp16);
    temp16 = BP; access_linear(obase32 + 28, 2, 0, BX_WRITE, &temp16);
    temp16 = SI; access_linear(obase32 + 30, 2, 0, BX_WRITE, &temp16);
    temp16 = DI; access_linear(obase32 + 32, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR es.selector.value;
                 access_linear(obase32 + 34, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR cs.selector.value;
                 access_linear(obase32 + 36, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR ss.selector.value;
                 access_linear(obase32 + 38, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR ds.selector.value;
                 access_linear(obase32 + 40, 2, 0, BX_WRITE, &temp16);
    }
  else if (BX_CPU_THIS_PTR tr.cache.type == 9) {
    if (BX_CPU_THIS_PTR tr.cache.valid==0 ||
        BX_CPU_THIS_PTR tr.cache.u.tss386.limit < 0x67) {
      bx_panic("task_switch(): TR32 not valid.\n");
      //exception(BX_TS_EXCEPTION, BX_CPU_THIS_PTR tr.selector.value & 0xfffc, 0);
      }
//bx_printf("tasking: saving state to 32bit TSS\n");

    obase32 = BX_CPU_THIS_PTR tr.cache.u.tss386.base; // current TSS.base
    temp32 = EIP; access_linear(obase32 + 0x20, 4, 0, BX_WRITE, &temp32);
    temp32 = read_eflags();
    if (source == BX_TASK_FROM_IRET) {
      // NT flags in old task is cleared with an IRET
      temp32 &= ~0x00004000;
      }
                  access_linear(obase32 + 0x24, 4, 0, BX_WRITE, &temp32);
    temp32 = EAX; access_linear(obase32 + 0x28, 4, 0, BX_WRITE, &temp32);
    temp32 = ECX; access_linear(obase32 + 0x2c, 4, 0, BX_WRITE, &temp32);
    temp32 = EDX; access_linear(obase32 + 0x30, 4, 0, BX_WRITE, &temp32);
    temp32 = EBX; access_linear(obase32 + 0x34, 4, 0, BX_WRITE, &temp32);
    temp32 = ESP; access_linear(obase32 + 0x38, 4, 0, BX_WRITE, &temp32);
    temp32 = EBP; access_linear(obase32 + 0x3c, 4, 0, BX_WRITE, &temp32);
    temp32 = ESI; access_linear(obase32 + 0x40, 4, 0, BX_WRITE, &temp32);
    temp32 = EDI; access_linear(obase32 + 0x44, 4, 0, BX_WRITE, &temp32);
    temp16 = BX_CPU_THIS_PTR es.selector.value;
                  access_linear(obase32 + 0x48, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR cs.selector.value;
                  access_linear(obase32 + 0x4c, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR ss.selector.value;
                  access_linear(obase32 + 0x50, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR ds.selector.value;
                  access_linear(obase32 + 0x54, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR fs.selector.value;
                  access_linear(obase32 + 0x58, 2, 0, BX_WRITE, &temp16);
    temp16 = BX_CPU_THIS_PTR gs.selector.value;
                  access_linear(obase32 + 0x5c, 2, 0, BX_WRITE, &temp16);
    }
  else {
    bx_panic("error in task_switch()\n");
    obase32 = 0;
    }

//fprintf(stderr, "# obase=%08x, nbase=%08x\n", obase32, nbase32);

  // load TR register with selector to the new task's TSS descriptor
  old_tr_selector = BX_CPU_THIS_PTR tr.selector; // save for later
  BX_CPU_THIS_PTR tr.selector = *tss_selector;
  BX_CPU_THIS_PTR tr.cache    = *tss_descriptor;


  // ==================================================
  // Effect of task switch on Busy, NT, and Link Fields
  // ==================================================

  // Field         jump        call/interrupt          iret
  // -----------------------------------------------------------------
  // new busy bit  Bit set     Bit set                 No change
  // old busy bit  Bit cleared No change               Bit is cleared
  // new NT flag   No change   Flag set                No change
  // old NT flag   No change   No change               Flag is cleared
  // new link      No change   old task's TSS selector No change
  // old link      No change   No change               No change

  // effect on Busy bit of new task
  if ( (source==BX_TASK_FROM_JUMP) || (source==BX_TASK_FROM_CALL_OR_INT) ) {
    // set the new task's busy bit
    dword2 |= 0x00000200;
    access_linear(BX_CPU_THIS_PTR gdtr.base + tss_selector->index*8 + 4, 4, 0,
      BX_WRITE, &dword2);
    }

  // effect on Busy bit of old task
  if ( (source==BX_TASK_FROM_JUMP) || (source==BX_TASK_FROM_IRET) ) {
    // Bit is cleared
    access_linear(BX_CPU_THIS_PTR gdtr.base + old_tr_selector.index*8 + 4,
      4, 0, BX_READ, &temp32);
    temp32 &= 0xfffffdff;
    access_linear(BX_CPU_THIS_PTR gdtr.base + old_tr_selector.index*8 + 4,
      4, 0, BX_WRITE, &temp32);
    }

  // effect on NT flag of new task
  if ( source==BX_TASK_FROM_CALL_OR_INT ) {
    // flag is set
    BX_CPU_THIS_PTR eflags.nt = 1;
    // ??? does this also change the value in the TSS?
    }

  // effect on NT flag of old task
  if ( source==BX_TASK_FROM_IRET ) {
    // flag is cleared
    // (handled above while writing CPU state to old TSS)
    }

  // effect on link field of new task
  if ( source==BX_TASK_FROM_CALL_OR_INT ) {
    // set to selector of old task's TSS
    temp16 = old_tr_selector.value;
    access_linear(nbase32 + 0, 2, 0, BX_WRITE, &temp16);
    }

  // effect on link field of old task
  // (none)


  // set TS bit in CR0 register
  BX_CPU_THIS_PTR cr0.ts = 1;

  // load new task's state from its TSS and continue execution
  // LDTR, CR3, EFLAGS, EIP, EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI
  // ES, CS, SS, DS, FS, GS
  // Errors which are detected in this step occur in context of the new task

  BX_CPU_THIS_PTR es.cache.valid = 0;
  BX_CPU_THIS_PTR cs.cache.valid = 0;
  BX_CPU_THIS_PTR ss.cache.valid = 0;
  BX_CPU_THIS_PTR ds.cache.valid = 0;
  BX_CPU_THIS_PTR fs.cache.valid = 0;
  BX_CPU_THIS_PTR gs.cache.valid = 0;
  BX_CPU_THIS_PTR ldtr.cache.valid = 0;

  if (tss_descriptor->type == 1) {
//bx_printf("tasking: loading state from 16bit TSS\n");
    access_linear(nbase32 + 14, 2, 0, BX_READ, &temp16); EIP = temp16;
    access_linear(nbase32 + 16, 2, 0, BX_READ, &temp16);
      write_eflags(temp16, 1, 1, 1, 1); // zero out upper word
    // incoming TSS is 16bit:
    //   - upper word of general registers is set to 0xFFFF
    //   - upper word of eflags is zero'd
    //   - FS, GS are zero'd
    //   - upper word of eIP is zero'd
    access_linear(nbase32 + 18, 2, 0, BX_READ, &temp16); EAX = 0xffff0000 | temp16;
    access_linear(nbase32 + 20, 2, 0, BX_READ, &temp16); ECX = 0xffff0000 | temp16;
    access_linear(nbase32 + 22, 2, 0, BX_READ, &temp16); EDX = 0xffff0000 | temp16;
    access_linear(nbase32 + 24, 2, 0, BX_READ, &temp16); EBX = 0xffff0000 | temp16;
    access_linear(nbase32 + 26, 2, 0, BX_READ, &temp16); ESP = 0xffff0000 | temp16;
    access_linear(nbase32 + 28, 2, 0, BX_READ, &temp16); EBP = 0xffff0000 | temp16;
    access_linear(nbase32 + 30, 2, 0, BX_READ, &temp16); ESI = 0xffff0000 | temp16;
    access_linear(nbase32 + 32, 2, 0, BX_READ, &temp16); EDI = 0xffff0000 | temp16;

    access_linear(nbase32 + 34, 2, 0, BX_READ, &raw_es_selector);
    access_linear(nbase32 + 36, 2, 0, BX_READ, &raw_cs_selector);
    access_linear(nbase32 + 38, 2, 0, BX_READ, &raw_ss_selector);
    access_linear(nbase32 + 40, 2, 0, BX_READ, &raw_ds_selector);
    access_linear(nbase32 + 42, 2, 0, BX_READ, &raw_ldt_selector);

    raw_fs_selector = 0; // ??? use a NULL selector
    raw_gs_selector = 0; // ??? use a NULL selector
    // No CR3 change for 286 task switch
    }
  else {
//bx_printf("tasking: loading state from 32bit TSS\n");
    access_linear(nbase32 + 0x1c, 4, 0, BX_READ, &temp32); temp_cr3 = temp32;
    access_linear(nbase32 + 0x20, 4, 0, BX_READ, &temp32); EIP = temp32;
    access_linear(nbase32 + 0x24, 4, 0, BX_READ, &temp32); write_eflags(temp32, 1,1,1,1);
    access_linear(nbase32 + 0x28, 4, 0, BX_READ, &temp32); EAX = temp32;
    access_linear(nbase32 + 0x2c, 4, 0, BX_READ, &temp32); ECX = temp32;
    access_linear(nbase32 + 0x30, 4, 0, BX_READ, &temp32); EDX = temp32;
    access_linear(nbase32 + 0x34, 4, 0, BX_READ, &temp32); EBX = temp32;
    access_linear(nbase32 + 0x38, 4, 0, BX_READ, &temp32); ESP = temp32;
    access_linear(nbase32 + 0x3c, 4, 0, BX_READ, &temp32); EBP = temp32;
    access_linear(nbase32 + 0x40, 4, 0, BX_READ, &temp32); ESI = temp32;
    access_linear(nbase32 + 0x44, 4, 0, BX_READ, &temp32); EDI = temp32;
    access_linear(nbase32 + 0x48, 2, 0, BX_READ, &raw_es_selector);
    access_linear(nbase32 + 0x4c, 2, 0, BX_READ, &raw_cs_selector);
    access_linear(nbase32 + 0x50, 2, 0, BX_READ, &raw_ss_selector);
    access_linear(nbase32 + 0x54, 2, 0, BX_READ, &raw_ds_selector);
    access_linear(nbase32 + 0x58, 2, 0, BX_READ, &raw_fs_selector);
    access_linear(nbase32 + 0x5c, 2, 0, BX_READ, &raw_gs_selector);
    access_linear(nbase32 + 0x60, 2, 0, BX_READ, &raw_ldt_selector);
    access_linear(nbase32 + 0x64, 2, 0, BX_READ, &trap_word);

    if (trap_word & 0x0001) {
      bx_panic("task_switch: T bit set in new TSS: not implemented yet.\n");
      }

    // Tell paging unit about new cr3 value
    CR3_change(temp_cr3);
    BX_INSTR_TLB_CNTRL(BX_INSTR_TASKSWITCH, temp_cr3);
    }


  //
  // start by parsing selectors for each segment register
  // so that if an error occurs, these values are filled in.
  //

  // LDTR
  parse_selector(raw_ldt_selector, &ldt_selector);
  BX_CPU_THIS_PTR ldtr.selector = ldt_selector;

  // CS, SS
  parse_selector(raw_cs_selector, &cs_selector); BX_CPU_THIS_PTR cs.selector = cs_selector;
  parse_selector(raw_ss_selector, &ss_selector); BX_CPU_THIS_PTR ss.selector = ss_selector;

  // DS, ES, FS, GS
  parse_selector(raw_ds_selector, &ds_selector); BX_CPU_THIS_PTR ds.selector = ds_selector;
  parse_selector(raw_es_selector, &es_selector); BX_CPU_THIS_PTR es.selector = es_selector;
  parse_selector(raw_fs_selector, &fs_selector); BX_CPU_THIS_PTR fs.selector = fs_selector;
  parse_selector(raw_gs_selector, &gs_selector); BX_CPU_THIS_PTR gs.selector = gs_selector;

  //
  // LDT selector of new task is valid, else #TS(new task's LDT)
  //

  if (ldt_selector.ti) {
    bx_panic("task_switch: ldt.ti=1\n");
    //exception(BX_TS_EXCEPTION, raw_ldt_selector & 0xfffc, 0);
    return;
    }
  if ( (raw_ldt_selector & 0xfffc) != 0 ) { // not NULL selector
    fetch_raw_descriptor(&ldt_selector, &dword1, &dword2, 
      BX_TS_EXCEPTION);
    parse_descriptor(dword1, dword2, &ldt_descriptor);

    // AR byte from LDT descriptor must specify LDT segment else #TS(LDT)
    if (ldt_descriptor.valid==0 ||
        ldt_descriptor.type!=2 ||
        ldt_descriptor.segment ||
        ldt_descriptor.u.ldt.limit<7) {
      bx_panic("task_switch: bad LDT segment\n");
      //exception(BX_TS_EXCEPTION, raw_ldt_selector & 0xfffc, 0);
      return;
      }

    // load LDT cache with new LDT descriptor and set valid bit
    BX_CPU_THIS_PTR ldtr.cache    = ldt_descriptor;
    }

  if (v8086_mode()) {
    // load seg regs as 8086 registers

    load_seg_reg(&BX_CPU_THIS_PTR cs, raw_cs_selector);
    load_seg_reg(&BX_CPU_THIS_PTR ss, raw_ss_selector);
    load_seg_reg(&BX_CPU_THIS_PTR ds, raw_ds_selector);
    load_seg_reg(&BX_CPU_THIS_PTR es, raw_es_selector);
    load_seg_reg(&BX_CPU_THIS_PTR fs, raw_fs_selector);
    load_seg_reg(&BX_CPU_THIS_PTR gs, raw_gs_selector);
//bx_printf("XXXXX task switch out XXXXX\n");
    return;

#if 0
    BX_CPU_THIS_PTR cs.selector.value = raw_cs_selector;
    BX_CPU_THIS_PTR cs.selector.rpl = 3;
    BX_CPU_THIS_PTR cs.cache.valid = 1;
    BX_CPU_THIS_PTR cs.cache.p = 1;
    BX_CPU_THIS_PTR cs.cache.dpl = 3;
    BX_CPU_THIS_PTR cs.cache.segment = 1; // regular segment
    BX_CPU_THIS_PTR cs.cache.u.segment.executable = 1; // code segment
    BX_CPU_THIS_PTR cs.cache.u.segment.c_ed = 0; // expand up
    BX_CPU_THIS_PTR cs.cache.u.segment.r_w = 1; // writeable
    BX_CPU_THIS_PTR cs.cache.u.segment.a = 1; // accessed
    BX_CPU_THIS_PTR cs.cache.u.segment.base = raw_cs_selector << 4;
    BX_CPU_THIS_PTR cs.cache.u.segment.limit        = 0xffff;
    BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled = 0xffff;
    BX_CPU_THIS_PTR cs.cache.u.segment.g     = 0; // byte granular
    BX_CPU_THIS_PTR cs.cache.u.segment.d_b   = 0; // default 16bit size
    BX_CPU_THIS_PTR cs.cache.u.segment.avl   = 0;
 
    BX_CPU_THIS_PTR ss.selector.value = raw_ss_selector;
    BX_CPU_THIS_PTR ss.selector.rpl = 3;
    BX_CPU_THIS_PTR ss.cache.valid = 1;
    BX_CPU_THIS_PTR ss.cache.p = 1;
    BX_CPU_THIS_PTR ss.cache.dpl = 3;
    BX_CPU_THIS_PTR ss.cache.segment = 1; // regular segment
    BX_CPU_THIS_PTR ss.cache.u.segment.executable = 0; // data segment
    BX_CPU_THIS_PTR ss.cache.u.segment.c_ed = 0; // expand up
    BX_CPU_THIS_PTR ss.cache.u.segment.r_w = 1; // writeable
    BX_CPU_THIS_PTR ss.cache.u.segment.a = 1; // accessed
    BX_CPU_THIS_PTR ss.cache.u.segment.base = raw_ss_selector << 4;
    BX_CPU_THIS_PTR ss.cache.u.segment.limit        = 0xffff;
    BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled = 0xffff;
    BX_CPU_THIS_PTR ss.cache.u.segment.g     = 0; // byte granular
    BX_CPU_THIS_PTR ss.cache.u.segment.d_b   = 0; // default 16bit size
    BX_CPU_THIS_PTR ss.cache.u.segment.avl   = 0;
#endif
    }

  //
  // Code segment DPL matches selector RPL, else #TS(new code segment)
  //

  // if new cs selector is null #TS(CS)
  if ( (raw_cs_selector & 0xfffc)==0 ) {
    bx_panic("task_switch: CS NULL\n");
    //exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  fetch_raw_descriptor(&cs_selector, &dword1, &dword2, 
    BX_TS_EXCEPTION);
  parse_descriptor(dword1, dword2, &cs_descriptor);
  
  if (cs_descriptor.dpl != cs_selector.rpl) {
    bx_printf("task_switch: CS.dpl != CS.rpl\n");
    exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }


  //
  // SS selector is valid, else #TS(new stack segment)
  //

  // if new stack selector is null #TS(SS)
  if ( (raw_ss_selector & 0xfffc)==0 ) {
    bx_panic("task_switch: SS NULL\n");
    //exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  // SS selector must be within its descriptor table limits else #TS(SS)
  fetch_raw_descriptor(&ss_selector, &dword1, &dword2, 
    BX_TS_EXCEPTION);
  parse_descriptor(dword1, dword2, &ss_descriptor);


  // SS descriptor AR byte must must indicate writable data segment,
  // else #TS(SS)
  if (ss_descriptor.valid==0 ||
      ss_descriptor.segment==0 ||
      ss_descriptor.u.segment.executable ||
      ss_descriptor.u.segment.r_w==0) {
    bx_printf("task_switch: SS not valid\n");
    exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }


  //
  // Stack segment is present in memory, else #SF(new stack segment)
  //
  if (ss_descriptor.p==0) {
    bx_panic("task_switch: SS not present\n");
    //exception(BX_SS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }
  
  //
  // Stack segment DPL matches CPL, else #TS(new stack segment)
  //

  if (ss_descriptor.dpl != CPL) {
    bx_panic("task_switch: SS.rpl != CPL\n");
    //exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }
  BX_CPU_THIS_PTR ss.cache    = ss_descriptor;

  
  //
  // LDT of new task is present in memory, else #TS(new tasks's LDT)
  //
  if (ldt_descriptor.p==0) {
    bx_panic("task_switch: LDT.p==0\n");
    //exception(BX_TS_EXCEPTION, raw_ldt_selector & 0xfffc, 0);
    return;
    }

  //
  // CS selector is valid, else #TS(new code segment)
  //

  // CS descriptor AR byte must indicate code segment else #TS(CS)
  if (cs_descriptor.valid==0 || cs_descriptor.segment==0 ||
      cs_descriptor.u.segment.executable==0) {
    bx_panic("task_switch: CS not valid executable seg\n");
    //exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

//???
  // if non-conforming then DPL must equal CPL else #TS(CS)
  if (cs_descriptor.u.segment.c_ed==0 &&
      cs_descriptor.dpl!=CPL) {
    bx_panic("task_switch: non-conforming: CS.dpl!=CPL\n");
    exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

//???
  // if conforming then DPL must be <= CPL else #TS(CS)
  if (cs_descriptor.u.segment.c_ed &&
      cs_descriptor.dpl>CPL) {
    bx_panic("task_switch: conforming: CS.dpl>CPL\n");
    exception(BX_TS_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }

  BX_CPU_THIS_PTR cs.cache    = cs_descriptor;

  //
  // Code segment is present in memory, else #NP(new code segment)
  //
  if (cs_descriptor.p==0) {
    bx_panic("task_switch: CS.p==0\n");
    //exception(BX_NP_EXCEPTION, raw_cs_selector & 0xfffc, 0);
    return;
    }


  //
  // Stack segment DPL matches selector RPL, else #TS(new stack segment)
  //
  if (ss_descriptor.dpl != ss_selector.rpl) {
    bx_panic("task_switch: SS.dpl != SS.rpl\n");
    //exception(BX_TS_EXCEPTION, raw_ss_selector & 0xfffc, 0);
    return;
    }

  //
  // DS, ES, FS, GS selectors are valid, else #TS(new data segment)
  // DS, ES, FS, GS segments are readable, else #TS(new data segment)
  // DS, ES, FS, GS segments are present in memory, else #NP(new data segment)
  // DS, ES, FS, GS segment DPL >= CPL (unless conforming segments), else #TS(new data segment)
  //

  //   if new selector is not null then perform following checks:
  //     index must be within its descriptor table limits else #TS(selector)
  //     AR byte must indicate data or readable code else #TS(selector)
  //     if data or non-conforming code then:
  //       DPL must be >= CPL else #TS(selector)
  //       DPL must be >= RPL else #TS(selector)
  //     AR byte must indicate PRESENT else #NP(selector)
  //     load cache with new segment descriptor and set valid bit

  // DS
  if ( (raw_ds_selector & 0xfffc) != 0 ) { // not NULL
    fetch_raw_descriptor(&ds_selector, &dword1, &dword2, 
      BX_TS_EXCEPTION);
    parse_descriptor(dword1, dword2, &ds_descriptor);
    BX_CPU_THIS_PTR ds.cache    = ds_descriptor;
    if (ds_descriptor.valid==0 || ds_descriptor.segment==0 ||
        (ds_descriptor.u.segment.executable &&
         ds_descriptor.u.segment.r_w==0)) {
      bx_panic("task_switch: DS not valid\n");
      exception(BX_TS_EXCEPTION, raw_ds_selector & 0xfffc, 0);
      return;
      }
    // if data or non-conforming code
    if (ds_descriptor.type<12 &&
        (ds_descriptor.dpl<CPL || ds_descriptor.dpl<ds_selector.rpl)) {
      bx_panic("task_switch: DS.dpl not valid\n");
      exception(BX_TS_EXCEPTION, raw_ds_selector & 0xfffc, 0);
      return;
      }
    if (ds_descriptor.p==0) {
      bx_panic("task_switch: DS.p==0\n");
      //exception(BX_NP_EXCEPTION, raw_ds_selector & 0xfffc, 0);
      return;
      }
    }

  // ES
  if ( (raw_es_selector & 0xfffc) != 0 ) { // not NULL
    fetch_raw_descriptor(&es_selector, &dword1, &dword2, 
      BX_TS_EXCEPTION);
    parse_descriptor(dword1, dword2, &es_descriptor);
    BX_CPU_THIS_PTR es.cache    = es_descriptor;
    if (es_descriptor.valid==0 || es_descriptor.segment==0 ||
        (es_descriptor.u.segment.executable &&
         es_descriptor.u.segment.r_w==0)) {
      bx_panic("task_switch: ES not valid\n");
      exception(BX_TS_EXCEPTION, raw_es_selector & 0xfffc, 0);
      return;
      }
    // if data or non-conforming code
    if (es_descriptor.type<12 &&
        (es_descriptor.dpl<CPL || es_descriptor.dpl<es_selector.rpl)) {
      bx_panic("task_switch: ES.dpl not valid\n");
      exception(BX_TS_EXCEPTION, raw_es_selector & 0xfffc, 0);
      return;
      }
    if (es_descriptor.p==0) {
      bx_panic("task_switch: ES.p==0\n");
      //exception(BX_NP_EXCEPTION, raw_es_selector & 0xfffc, 0);
      return;
      }
    }


  // FS
  if ( (raw_fs_selector & 0xfffc) != 0 ) { // not NULL
    fetch_raw_descriptor(&fs_selector, &dword1, &dword2, 
      BX_TS_EXCEPTION);
    parse_descriptor(dword1, dword2, &fs_descriptor);
    BX_CPU_THIS_PTR fs.cache    = fs_descriptor;
    if (fs_descriptor.valid==0 || fs_descriptor.segment==0 ||
        (fs_descriptor.u.segment.executable &&
         fs_descriptor.u.segment.r_w==0)) {
      bx_panic("task_switch: FS not valid\n");
      exception(BX_TS_EXCEPTION, raw_fs_selector & 0xfffc, 0);
      return;
      }
    // if data or non-conforming code
    if (fs_descriptor.type<12 &&
        (fs_descriptor.dpl<CPL || fs_descriptor.dpl<fs_selector.rpl)) {
      bx_panic("task_switch: FS.dpl not valid\n");
      exception(BX_TS_EXCEPTION, raw_fs_selector & 0xfffc, 0);
      return;
      }
    if (fs_descriptor.p==0) {
      bx_panic("task_switch: FS.p==0\n");
      //exception(BX_NP_EXCEPTION, raw_fs_selector & 0xfffc, 0);
      return;
      }
    }

  // GS
  if ( (raw_gs_selector & 0xfffc) != 0 ) { // not NULL
    fetch_raw_descriptor(&gs_selector, &dword1, &dword2, 
      BX_TS_EXCEPTION);
    parse_descriptor(dword1, dword2, &gs_descriptor);
    BX_CPU_THIS_PTR gs.cache    = gs_descriptor;
    if (gs_descriptor.valid==0 || gs_descriptor.segment==0 ||
        (gs_descriptor.u.segment.executable &&
         gs_descriptor.u.segment.r_w==0)) {
      bx_panic("task_switch: GS not valid\n");
      exception(BX_TS_EXCEPTION, raw_gs_selector & 0xfffc, 0);
      return;
      }
    // if data or non-conforming code
    if (gs_descriptor.type<12 &&
        (gs_descriptor.dpl<CPL || gs_descriptor.dpl<gs_selector.rpl)) {
      bx_panic("task_switch: GS.dpl not valid\n");
      exception(BX_TS_EXCEPTION, raw_gs_selector & 0xfffc, 0);
      return;
      }
    if (gs_descriptor.p==0) {
      bx_panic("task_switch: GS.p==0\n");
      //exception(BX_NP_EXCEPTION, raw_gs_selector & 0xfffc, 0);
      return;
      }
    }

  // check out paging considerations.  Is the whole TSS supposed to
  // be in 1 page?  If that's the case, then only 1 page fault should
  // occur in the beginning? */
//bx_printf("XXXXX task switch out XXXXX\n");
}
#endif



#else  // BX_SUPPORT_TASKING


// for non-support of hardware tasking

#if BX_CPU >= 2
  /* corresponds to SWITCH_TASKS algorithm in Intel documentation */
  void
bx_cpu_c::task_switch(bx_selector_t *selector,
                 bx_descriptor_t *descriptor, unsigned source,
                 Bit32u dword1, Bit32u dword2)
{
  UNUSED(selector);
  UNUSED(descriptor);
  UNUSED(source);
  UNUSED(dword1);
  UNUSED(dword2);

  bx_printf("task_switch(): not complete\n");
}
#endif

#endif // BX_SUPPORT_TASKING
