/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.

BOCHS is commercial software.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/

/* NOTE: Some comments may closely follow those from Intel 386 through
 *       Pentium Processor Family Developer's Manual, Volume 3.
 *       I've done this in areas where Bochs source code closely models
 *       Intel's published pseudo-code, as a way to reference between
 *       the two.
 */


#include "bochs.h"



#if BX_SUPPORT_V8086_MODE


#if BX_CPU >= 3


  void
bx_cpu_c::stack_return_to_v86(Bit32u new_eip, Bit32u raw_cs_selector, Bit32u flags32)
{
  Bit32u temp_ESP, new_esp;
  Bit16u raw_es_selector, raw_ds_selector, raw_fs_selector,
         raw_gs_selector, raw_ss_selector;


  /* Must be 32bit effective opsize, VM is in upper 16bits of eFLAGS */
  /* CPL = 0 to get here */

  /*    unused (ESP from TSS) ???
   * ----------------
   * |     | OLD GS | eSP+32
   * |     | OLD FS | eSP+28
   * |     | OLD DS | eSP+24
   * |     | OLD ES | eSP+20
   * |     | OLD SS | eSP+16
   * |  OLD ESP     | eSP+12
   * | OLD EFLAGS   | eSP+8
   * |     | OLD CS | eSP+4
   * |  OLD EIP     | eSP+0
   * ----------------
   */

  if (BX_CPU_THIS_PTR ss.cache.u.segment.d_b)
    temp_ESP = ESP;
  else
    temp_ESP = SP;

  /* top 36 bytes of stack must be within stack limits, else #GP(0) */
  if ( !can_pop(36) ) {
    bx_panic("iret: VM: top 36 bytes not within limits\n");
    exception(BX_SS_EXCEPTION, 0, 0);
    return;
    }

  /* if new IP not within code segment limits, #GP(0) */
  if ( new_eip > 0xffff ) {
    bx_printf("stack_return_to_v86: IP > 65535 limit\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    }

  /* EFLAGS = SS:[ESP + 8] , sets VM */
  write_eflags(flags32, /*change IOPL*/ 1, /*change IF*/ 1,
                  /*change VM*/ 1, /*change RM*/ 0);

  /* load CS:EIP from stack; already read and passed as args */
  BX_CPU_THIS_PTR cs.selector.value = raw_cs_selector;
  EIP = new_eip;

  /* load SS:ESP from stack */
  access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 12,
    4, 0, BX_READ, &new_esp);

  access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 16,
    2, 0, BX_READ, &raw_ss_selector);

  /* load ES,DS,FS,GS from stack */
  access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 20,
    2, 0, BX_READ, &raw_es_selector);

  access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 24,
    2, 0, BX_READ, &raw_ds_selector);

  access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 28,
    2, 0, BX_READ, &raw_fs_selector);

  access_linear(BX_CPU_THIS_PTR ss.cache.u.segment.base + temp_ESP + 32,
    2, 0, BX_READ, &raw_gs_selector);

  BX_CPU_THIS_PTR es.selector.value = raw_es_selector;
  BX_CPU_THIS_PTR ds.selector.value = raw_ds_selector;
  BX_CPU_THIS_PTR fs.selector.value = raw_fs_selector;
  BX_CPU_THIS_PTR gs.selector.value = raw_gs_selector;

  BX_CPU_THIS_PTR ss.selector.value = raw_ss_selector;
  ESP = new_esp;

  init_v8086_mode();
  if (bx_dbg.v8086) bx_printf("stack_return_to_v86: VM bit set on stack!\n");
}


  void
bx_cpu_c::stack_return_from_v86(void)
{
  //bx_printf("stack_return_from_v86:\n");
  exception(BX_GP_EXCEPTION, 0, 0);

#if 0
  if (IOPL != 3) {
    // trap to virtual 8086 monitor
    bx_printf("stack_return_from_v86: IOPL != 3\n");
    exception(BX_GP_EXCEPTION, 0, 0);
    }

  if (BX_CPU_THIS_PTR is_32bit_opsize) {
    Bit32u eip, ecs_raw, eflags;

// ??? should be some stack checks here
    pop_32(&eip);
    pop_32(&ecs_raw);
    pop_32(&eflags);

    load_seg_reg(&BX_CPU_THIS_PTR cs, (Bit16u) ecs_raw);
    BX_CPU_THIS_PTR eip = eip;
    write_eflags(eflags, /*IOPL*/ CPL==0, /*IF*/ 1, /*VM*/ 0, /*RF*/ 1);
    }
  else {
    Bit16u ip, cs_raw, flags;

// ??? should be some stack checks here
    pop_16(&ip);
    pop_16(&cs_raw);
    pop_16(&flags);

    load_seg_reg(&BX_CPU_THIS_PTR cs, cs_raw);
    BX_CPU_THIS_PTR eip = (Bit32u) ip;
    write_flags(flags, /*IOPL*/ CPL==0, /*IF*/ 1);
    }
#endif
}


  void
bx_cpu_c::init_v8086_mode(void)
{
  BX_CPU_THIS_PTR cs.cache.valid                  = 1;
  BX_CPU_THIS_PTR cs.cache.p                      = 1;
  BX_CPU_THIS_PTR cs.cache.dpl                    = 3;
  BX_CPU_THIS_PTR cs.cache.segment                = 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.executable   = 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.c_ed         = 0;
  BX_CPU_THIS_PTR cs.cache.u.segment.r_w          = 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.a            = 1;
  BX_CPU_THIS_PTR cs.cache.u.segment.base         = BX_CPU_THIS_PTR cs.selector.value << 4;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit        = 0xffff;
  BX_CPU_THIS_PTR cs.cache.u.segment.limit_scaled = 0xffff;
  BX_CPU_THIS_PTR cs.cache.u.segment.g            = 0;
  BX_CPU_THIS_PTR cs.cache.u.segment.d_b          = 0;
  BX_CPU_THIS_PTR cs.cache.u.segment.avl          = 0;
  BX_CPU_THIS_PTR cs.selector.rpl                 = 3;

  BX_CPU_THIS_PTR ss.cache.valid                  = 1;
  BX_CPU_THIS_PTR ss.cache.p                      = 1;
  BX_CPU_THIS_PTR ss.cache.dpl                    = 3;
  BX_CPU_THIS_PTR ss.cache.segment                = 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.executable   = 0;
  BX_CPU_THIS_PTR ss.cache.u.segment.c_ed         = 0;
  BX_CPU_THIS_PTR ss.cache.u.segment.r_w          = 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.a            = 1;
  BX_CPU_THIS_PTR ss.cache.u.segment.base         = BX_CPU_THIS_PTR ss.selector.value << 4;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit        = 0xffff;
  BX_CPU_THIS_PTR ss.cache.u.segment.limit_scaled = 0xffff;
  BX_CPU_THIS_PTR ss.cache.u.segment.g            = 0;
  BX_CPU_THIS_PTR ss.cache.u.segment.d_b          = 0;
  BX_CPU_THIS_PTR ss.cache.u.segment.avl          = 0;
  BX_CPU_THIS_PTR ss.selector.rpl                 = 3;

  BX_CPU_THIS_PTR es.cache.valid                  = 1;
  BX_CPU_THIS_PTR es.cache.p                      = 1;
  BX_CPU_THIS_PTR es.cache.dpl                    = 3;
  BX_CPU_THIS_PTR es.cache.segment                = 1;
  BX_CPU_THIS_PTR es.cache.u.segment.executable   = 0;
  BX_CPU_THIS_PTR es.cache.u.segment.c_ed         = 0;
  BX_CPU_THIS_PTR es.cache.u.segment.r_w          = 1;
  BX_CPU_THIS_PTR es.cache.u.segment.a            = 1;
  BX_CPU_THIS_PTR es.cache.u.segment.base         = BX_CPU_THIS_PTR es.selector.value << 4;
  BX_CPU_THIS_PTR es.cache.u.segment.limit        = 0xffff;
  BX_CPU_THIS_PTR es.cache.u.segment.limit_scaled = 0xffff;
  BX_CPU_THIS_PTR es.cache.u.segment.g            = 0;
  BX_CPU_THIS_PTR es.cache.u.segment.d_b          = 0;
  BX_CPU_THIS_PTR es.cache.u.segment.avl          = 0;
  BX_CPU_THIS_PTR es.selector.rpl                 = 3;

  BX_CPU_THIS_PTR ds.cache.valid                  = 1;
  BX_CPU_THIS_PTR ds.cache.p                      = 1;
  BX_CPU_THIS_PTR ds.cache.dpl                    = 3;
  BX_CPU_THIS_PTR ds.cache.segment                = 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.executable   = 0;
  BX_CPU_THIS_PTR ds.cache.u.segment.c_ed         = 0;
  BX_CPU_THIS_PTR ds.cache.u.segment.r_w          = 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.a            = 1;
  BX_CPU_THIS_PTR ds.cache.u.segment.base         = BX_CPU_THIS_PTR ds.selector.value << 4;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit        = 0xffff;
  BX_CPU_THIS_PTR ds.cache.u.segment.limit_scaled = 0xffff;
  BX_CPU_THIS_PTR ds.cache.u.segment.g            = 0;
  BX_CPU_THIS_PTR ds.cache.u.segment.d_b          = 0;
  BX_CPU_THIS_PTR ds.cache.u.segment.avl          = 0;
  BX_CPU_THIS_PTR ds.selector.rpl                 = 3;

  BX_CPU_THIS_PTR fs.cache.valid                  = 1;
  BX_CPU_THIS_PTR fs.cache.p                      = 1;
  BX_CPU_THIS_PTR fs.cache.dpl                    = 3;
  BX_CPU_THIS_PTR fs.cache.segment                = 1;
  BX_CPU_THIS_PTR fs.cache.u.segment.executable   = 0;
  BX_CPU_THIS_PTR fs.cache.u.segment.c_ed         = 0;
  BX_CPU_THIS_PTR fs.cache.u.segment.r_w          = 1;
  BX_CPU_THIS_PTR fs.cache.u.segment.a            = 1;
  BX_CPU_THIS_PTR fs.cache.u.segment.base         = BX_CPU_THIS_PTR fs.selector.value << 4;
  BX_CPU_THIS_PTR fs.cache.u.segment.limit        = 0xffff;
  BX_CPU_THIS_PTR fs.cache.u.segment.limit_scaled = 0xffff;
  BX_CPU_THIS_PTR fs.cache.u.segment.g            = 0;
  BX_CPU_THIS_PTR fs.cache.u.segment.d_b          = 0;
  BX_CPU_THIS_PTR fs.cache.u.segment.avl          = 0;
  BX_CPU_THIS_PTR fs.selector.rpl                 = 3;

  BX_CPU_THIS_PTR gs.cache.valid                  = 1;
  BX_CPU_THIS_PTR gs.cache.p                      = 1;
  BX_CPU_THIS_PTR gs.cache.dpl                    = 3;
  BX_CPU_THIS_PTR gs.cache.segment                = 1;
  BX_CPU_THIS_PTR gs.cache.u.segment.executable   = 0;
  BX_CPU_THIS_PTR gs.cache.u.segment.c_ed         = 0;
  BX_CPU_THIS_PTR gs.cache.u.segment.r_w          = 1;
  BX_CPU_THIS_PTR gs.cache.u.segment.a            = 1;
  BX_CPU_THIS_PTR gs.cache.u.segment.base         = BX_CPU_THIS_PTR gs.selector.value << 4;
  BX_CPU_THIS_PTR gs.cache.u.segment.limit        = 0xffff;
  BX_CPU_THIS_PTR gs.cache.u.segment.limit_scaled = 0xffff;
  BX_CPU_THIS_PTR gs.cache.u.segment.g            = 0;
  BX_CPU_THIS_PTR gs.cache.u.segment.d_b          = 0;
  BX_CPU_THIS_PTR gs.cache.u.segment.avl          = 0;
  BX_CPU_THIS_PTR gs.selector.rpl                 = 3;
}


#endif /* BX_CPU >= 3 */





#else  // BX_SUPPORT_V8086_MODE

// non-support of v8086 mode

  void
bx_cpu_c::stack_return_to_v86(Bit32u new_eip, Bit32u raw_cs_selector, Bit32u flags32)
{
  bx_printf("stack_return_to_v86: VM bit set in EFLAGS stack image\n");
  v8086_message();
}

  void
bx_cpu_c::stack_return_from_v86(void)
{
  bx_printf("stack_return_from_v86:\n");
  v8086_message();
}

  void
bx_cpu_c::v8086_message(void)
{
  bx_printf("Program compiled with BX_SUPPORT_V8086_MODE = 0\n");
  bx_printf("You need to rerun the configure script and recompile\n");
  bx_printf("  to use virtual-8086 mode features.\n");
  bx_panic("Bummer!\n");
}
#endif // BX_SUPPORT_V8086_MODE
