// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com

#ifndef  __cplusplus
// strict C programs at least need Bit8u, Bit16u, Bit32u, Boolean typedefs
#include "config.h"
#endif

#if (BX_DEBUGGER==0) && (BX_DISASM==1)

// If only using disassembly class, just need a few declarations.
// No external stuff, so can declare as C++

Boolean bx_dbg_fetch_mem(Bit32u addr, unsigned len, Bit8u *buf);
void dbg_xlate_linear2phy(Bit32u linear, Bit32u *phy, Boolean *valid);
#endif

unsigned long crc32(unsigned char *buf, int len);

#if BX_DEBUGGER

#ifdef  __cplusplus
extern "C" {
#endif

#define BX_DBG_REG_EAX          10
#define BX_DBG_REG_ECX          11
#define BX_DBG_REG_EDX          12
#define BX_DBG_REG_EBX          13
#define BX_DBG_REG_ESP          14
#define BX_DBG_REG_EBP          15
#define BX_DBG_REG_ESI          16
#define BX_DBG_REG_EDI          17
#define BX_DBG_REG_EIP          18
#define BX_DBG_REG_EFLAGS       19
#define BX_DBG_REG_CS           20
#define BX_DBG_REG_SS           21
#define BX_DBG_REG_DS           22
#define BX_DBG_REG_ES           23
#define BX_DBG_REG_FS           24
#define BX_DBG_REG_GS           25
 
Bit32u bx_dbg_get_register(unsigned which_reg);
 


#define BX_MAX_PATH     256
 

// Flex defs
extern int yylex(void);
extern char *yytext;  // Using the pointer option rather than array
extern int yywrap(void);

// Yacc defs
extern int yyparse(void);
extern void yyerror(char *s);


void bx_dbg_continue(void);
void bx_dbg_single_step(Bit32u count);
void bx_dbg_set(char *p1, char *p2, char *p3);
void bx_dbg_breakpoint(char *p1, char *p2);
void bx_dbg_del_breakpoint(unsigned handle);
void bx_dbg_exit(int code);
void bx_dbg_vbreakpoint(Boolean specific, Bit32u cs, Bit32u eip);
void bx_dbg_lbreakpoint(Boolean specific, Bit32u laddress);
void bx_dbg_pbreakpoint(Boolean specific, Bit32u paddress);
void bx_dbg_info_bpoints(void);
void bx_dbg_quit(void);
void bx_dbg_info_program(void);
void bx_dbg_info_registers(void);
void bx_dbg_info_dirty(void);
void bx_dbg_examine(char *command, char *format, Boolean format_passed,
                    Bit32u addr, Boolean addr_passed);
void bx_dbg_setpmem(Bit32u addr, unsigned len, Bit32u val);
Boolean bx_dbg_fetch_mem(Bit32u addr, unsigned len, Bit8u *buf);
Boolean bx_dbg_set_mem(Bit32u addr, unsigned len, Bit8u *buf);
void bx_dbg_set_symbol(char *symbol, Bit32u val);
Boolean bx_dbg_set_reg(unsigned reg, Bit32u val);
void bx_dbg_query(char *);
void bx_dbg_take(char *, unsigned n);
void bx_dbg_dump_cpu(void);
void bx_dbg_set_cpu_command(void);
void bx_dbg_disassemble_command(Bit32u addr1, Bit32u addr2);
void bx_dbg_instrument_command(char *);
void bx_dbg_loader_command(char *);
void bx_dbg_doit_command(void);
void bx_dbg_crc_command(Bit32u addr1, Bit32u addr2);

Boolean bx_dbg_cpu_icount(Bit32u cs, Bit32u eip,
                       Bit32u laddr, Bit32u paddr, Boolean paddr_valid,
                       Boolean intr_pending, Boolean is_32bit_code);
void dbg_xlate_linear2phy(Bit32u linear, Bit32u *phy, Boolean *valid);

int bx_dbg_extensions(char *command);
void bx_add_lex_input(char *buf);


typedef struct {
  Bit16u  cs;
  Bit32u  eip;
  Bit32u  linear_addr;
  Bit32u  phy_addr;
  Boolean phy_addr_valid;
  } bx_instr_addr_t;
 
typedef struct {
  Boolean          event_occurred; // any break condition will trip this

  struct {
    Boolean          intr_pending;
    bx_instr_addr_t  iaddr;
    Boolean          is_32bit_code;
    } before;

  struct {
    Boolean          intr_pending;
    bx_instr_addr_t  iaddr;
    Boolean          is_32bit_code;
    } after;
 
  Boolean          is_instr_count_bpoint;
  Bit32u              instr_count;

  Boolean          is_vir_iaddr_bpoint;
  unsigned            vir_iaddr_handle;
 
  Boolean          is_lin_iaddr_bpoint;
  unsigned            lin_iaddr_handle;
 
  Boolean          is_phy_iaddr_bpoint;
  unsigned            phy_iaddr_handle;

  Boolean          is_io_bpoint;
  unsigned            io_handle;
  unsigned            io_addr;
 
  Boolean          is_soft_int_bpoint;
  Boolean             soft_int_handle;
  Boolean             soft_int_num;
 
  Boolean          is_hard_int_bpoint;
  Boolean             hard_int_handle;
  Boolean             hard_int_num;
 
  Boolean          is_memory_bpoint; // do later
 
  Boolean          is_opcode_bpoint; // do later
 
  Boolean          is_exception_bpoint; // do later
 
  // DMA stuff // do later
 
  Boolean          is_user_interrupt;
  } bx_bpoint_state_t;



enum bx_disposition_e { bx_disp_disable=0, bx_disp_keep=1, bx_disp_delete=2 };

typedef struct {
  enum bx_disposition_e disposition;
  Boolean               enabled;
  unsigned long         ignorecount;
  } bx_bpoint_attr_t;


#define BX_MAX_IADDR_BPOINTS 10

#define BX_BPOINT_TYPE_NONE     0
#define BX_BPOINT_TYPE_VIRTUAL  1
#define BX_BPOINT_TYPE_LINEAR   2
#define BX_BPOINT_TYPE_PHYSICAL 3

typedef struct {
  Bit32u instr_count_max;
  Boolean take_dma;
  Boolean take_irq;

  unsigned num_iaddr_bpoints;

  unsigned bpoint_type[BX_MAX_IADDR_BPOINTS];
  bx_bpoint_attr_t attr[BX_MAX_IADDR_BPOINTS];
  union {
    struct {
      Bit32u           cs; // really only use 16 bits
      Bit32u           eip;
      } vir;

    struct {
      Bit32u           addr;
      } lin;

    struct {
      Bit32u           addr;
      } phy;
    } iaddr[BX_MAX_IADDR_BPOINTS];

  } bx_bpoint_requests_t;



extern volatile Boolean bx_dbg_interrupt_requested;
extern bx_bpoint_requests_t bx_dbg_bpoint_requests;

extern int bx_dbg_main(int argc, char *argv[]);

#ifdef  __cplusplus
}
#endif

void bx_dbg_loader(char *path);

typedef struct {
    unsigned long eax;
    unsigned long ebx;
    unsigned long ecx;
    unsigned long edx;
    unsigned long ebp;
    unsigned long esi;
    unsigned long edi;
    unsigned long esp;
    unsigned long eflags;
    unsigned long eip;
    struct { unsigned long sel, des_l, des_h, valid; } cs;
    struct { unsigned long sel, des_l, des_h, valid; } ss;
    struct { unsigned long sel, des_l, des_h, valid; } ds;
    struct { unsigned long sel, des_l, des_h, valid; } es;
    struct { unsigned long sel, des_l, des_h, valid; } fs;
    struct { unsigned long sel, des_l, des_h, valid; } gs;
    struct { unsigned long sel, des_l, des_h, valid; } ldtr;
    struct { unsigned long sel, des_l, des_h, valid; } tr;
    struct { unsigned long base, limit; } gdtr;
    struct { unsigned long base, limit; } idtr;
    unsigned long dr0, dr1, dr2, dr3, dr4, dr5, dr6, dr7;
    unsigned long tr3, tr4, tr5, tr6, tr7;
    unsigned long cr0, cr1, cr2, cr3, cr4;
    unsigned inhibit_int;
    } bx_dbg_cpu_t;

void bx_dbg_query_pending(void);
extern Boolean bx_dbg_take_dma;
extern Boolean bx_dbg_take_irq;
extern Boolean bx_dbg_event_reports;

void bx_dbg_dma_report(Bit32u addr, unsigned len, unsigned what, Bit32u val);
void bx_dbg_irq_report(unsigned irq);
void bx_dbg_a20_report(unsigned val);
void bx_dbg_io_report(Bit32u addr, unsigned size, unsigned op, Bit32u val);
void bx_dbg_ucmem_report(Bit32u addr, unsigned size, unsigned op, Bit32u val);
Bit32u bx_dbg_get_eflags(void);
Boolean bx_dbg_set_cpu(bx_dbg_cpu_t *cpu);

#define BX_DBG_DMA_REPORT(addr, len, what, val) \
        bx_dbg_dma_report(addr, len, what, val)
#define BX_DBG_IRQ_REPORT(vector) \
        bx_dbg_irq_report(vector)
#define BX_DBG_A20_REPORT(val) \
        bx_dbg_a20_report(val)
#define BX_DBG_IO_REPORT(addr, size, op, val) \
        bx_dbg_io_report(addr, size, op, val)
#define BX_DBG_UCMEM_REPORT(addr, size, op, val) \
        bx_dbg_ucmem_report(addr, size, op, val)

extern unsigned char bx_dbg_dirty_pages[];

#define BX_DBG_DIRTY_PAGE(page) bx_dbg_dirty_pages[page] = 1;

#else  // #if BX_DEBUGGER
  // debugger not compiled in, use empty stubs
#define BX_CPU_ICOUNT()
#define BX_DBG_DMA_REPORT(addr, len, what, val)
#define BX_DBG_IRQ_REPORT(vector)
#define BX_DBG_A20_REPORT(val)
#define BX_DBG_IO_REPORT(addr, size, op, val)
#define BX_DBG_UCMEM_REPORT(addr, size, op, val)
#define BX_DBG_DIRTY_PAGE(page)

#endif // #if BX_DEBUGGER
