%{
#include <string.h>
#include "debug.h"
#include "parser.h"

int bx_yyinput(char *buf, int max_size);
#undef YY_INPUT
#define YY_INPUT(buf, ret, max_size) (ret = bx_yyinput(buf, max_size))

static char    *lex_input_ptr = NULL;
static unsigned lex_input_size = 0;
%}


%%
[ \t]+          ; // eat up whitespace
set             { yylval.sval = strdup(yytext); return(BX_TOKEN_SET); }
dis             { yylval.sval = strdup(yytext); return(BX_TOKEN_DIS); }
on              { yylval.sval = strdup(yytext); return(BX_TOKEN_ON); }
off             { yylval.sval = strdup(yytext); return(BX_TOKEN_OFF); }
crc             { yylval.sval = strdup(yytext); return(BX_TOKEN_CRC); }
c               { yylval.sval = strdup(yytext); return(BX_TOKEN_CONTINUE); }
stepi           |
step            |
si              |
s               { yylval.sval = strdup(yytext); return(BX_TOKEN_SINGLE_STEP); }
vb              { yylval.sval = strdup(yytext); return(BX_TOKEN_VBREAKPOINT); }
vbreak          { yylval.sval = strdup(yytext); return(BX_TOKEN_VBREAKPOINT); }
lb              { yylval.sval = strdup(yytext); return(BX_TOKEN_LBREAKPOINT); }
lbreak          { yylval.sval = strdup(yytext); return(BX_TOKEN_LBREAKPOINT); }
pb              { yylval.sval = strdup(yytext); return(BX_TOKEN_PBREAKPOINT); }
pbreak          { yylval.sval = strdup(yytext); return(BX_TOKEN_PBREAKPOINT); }
info            { yylval.sval = strdup(yytext); return(BX_TOKEN_INFO); }
break           { yylval.sval = strdup(yytext); return(BX_TOKEN_PBREAKPOINT); }
b               { yylval.sval = strdup(yytext); return(BX_TOKEN_PBREAKPOINT); }
program         { yylval.sval = strdup(yytext); return(BX_TOKEN_PROGRAM); }
registers       { yylval.sval = strdup(yytext); return(BX_TOKEN_REGISTERS); }
dirty           { yylval.sval = strdup(yytext); return(BX_TOKEN_DIRTY); }
dump_cpu        { yylval.sval = strdup(yytext); return(BX_TOKEN_DUMP_CPU); }
delete          { yylval.sval = strdup(yytext); return(BX_TOKEN_DEL_BREAKPOINT); }
del             { yylval.sval = strdup(yytext); return(BX_TOKEN_DEL_BREAKPOINT); }
d               { yylval.sval = strdup(yytext); return(BX_TOKEN_DEL_BREAKPOINT); }
quit            { yylval.sval = strdup(yytext); return(BX_TOKEN_QUIT); }
q               { yylval.sval = strdup(yytext); return(BX_TOKEN_QUIT); }
x               { yylval.sval = strdup(yytext); return(BX_TOKEN_EXAMINE); }
xp              { yylval.sval = strdup(yytext); return(BX_TOKEN_EXAMINE); }
setpmem         { yylval.sval = strdup(yytext); return(BX_TOKEN_SETPMEM); }
query           { yylval.sval = strdup(yytext); return(BX_TOKEN_QUERY); }
pending         { yylval.sval = strdup(yytext); return(BX_TOKEN_PENDING); }
take            { yylval.sval = strdup(yytext); return(BX_TOKEN_TAKE); }
dma             { yylval.sval = strdup(yytext); return(BX_TOKEN_DMA); }
irq             { yylval.sval = strdup(yytext); return(BX_TOKEN_IRQ); }
set_cpu         { yylval.sval = strdup(yytext); return(BX_TOKEN_SET_CPU); }
disas           |
disassemble     { yylval.sval = strdup(yytext); return(BX_TOKEN_DISASSEMBLE); }
instrument      { yylval.sval = strdup(yytext); return(BX_TOKEN_INSTRUMENT); }
start           { yylval.sval = strdup(yytext); return(BX_TOKEN_START); }
stop            { yylval.sval = strdup(yytext); return(BX_TOKEN_STOP); }
reset           { yylval.sval = strdup(yytext); return(BX_TOKEN_RESET); }
print           { yylval.sval = strdup(yytext); return(BX_TOKEN_PRINT); }
loader          { yylval.sval = strdup(yytext); return(BX_TOKEN_LOADER); }
doit            { yylval.sval = strdup(yytext); return(BX_TOKEN_DOIT); }
\"[^\"\n]*\"    { yylval.sval = strdup(yytext); return(BX_TOKEN_STRING); }
\/[0-9]*[xduotsibhwg][xduotsibhwg] { yylval.sval = strdup(yytext); return(BX_TOKEN_XFORMAT); }
\/[0-9]*[xduotsibhwg]              { yylval.sval = strdup(yytext); return(BX_TOKEN_XFORMAT); }
\/[0-9]+                           { yylval.sval = strdup(yytext); return(BX_TOKEN_XFORMAT); }
0x[0-9a-fA-F]+  { yylval.uval = strtoul(yytext+2, NULL, 16); return(BX_TOKEN_NUMERIC); }
0[0-7]+         { yylval.uval = strtoul(yytext+1, NULL, 8); return(BX_TOKEN_NUMERIC); }
[0-9]+          { yylval.uval = strtoul(yytext, NULL, 10); return(BX_TOKEN_NUMERIC); }
$[a-zA-Z_][a-zA-Z0-9_]* { yylval.sval = strdup(yytext); return(BX_TOKEN_SYMBOLNAME); }
\n              { return('\n'); }
[#][^\n]*    ; // eat up comments '//'
.               { return(yytext[0]); }
%%

// [A-Za-z_][A-Za-z0-9_]*   { yylval.sval = strdup(yytext); return(BX_TOKEN_GENERIC); }

  int
bx_yyinput(char *buf, int max_size)
{
  int len;

  if (lex_input_size == 0) {
    fprintf(stderr, "lex: no characters in string input buffer.\n");
    exit(1);
    }

  len = strlen(lex_input_ptr);
  if (len > max_size)
    max_size = len;

  memcpy(buf, lex_input_ptr, max_size);
  lex_input_size -= max_size;

  return(max_size);
}


  void
bx_add_lex_input(char *buf)
{
  lex_input_ptr  = buf;
  lex_input_size = strlen(buf);

  // Since we're parsing from strings, flush out
  // all current buffer state, so the next read
  // requests from yyinput

  yy_flush_buffer( YY_CURRENT_BUFFER );
}
