// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
//
// BOCHS is commercial software.
//
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
//
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
//
//     EMail:    bochs@world.std.com



// possible types passed to BX_INSTR_TLB_CNTRL()
#define BX_INSTR_MOV_CR3     10
#define BX_INSTR_INVLPG      11
#define BX_INSTR_TASKSWITCH  12

// possible types passed to BX_INSTR_CACHE_CNTRL()
#define BX_INSTR_INVD      20
#define BX_INSTR_WBINVD    21


// called from the CPU core
void bx_instr_cnear_branch(Bit32u new_eip);
void bx_instr_ucnear_branch(Bit32u new_eip);
void bx_instr_far_branch(Bit32u new_cs, Bit32u new_eip);
void bx_instr_opcode_byte1(Bit8u);
void bx_instr_opcode_byte2(Bit8u);
void bx_instr_opcode_g1ebib(unsigned nnn);
void bx_instr_opcode_g1eviv(unsigned nnn);
void bx_instr_opcode_g1evib(unsigned nnn);
void bx_instr_opcode_g2ebib(unsigned nnn);
void bx_instr_opcode_g2evib(unsigned nnn);
void bx_instr_opcode_g2eb1(unsigned nnn);
void bx_instr_opcode_g2ev1(unsigned nnn);
void bx_instr_opcode_g2ebcl(unsigned nnn);
void bx_instr_opcode_g2evcl(unsigned nnn);
void bx_instr_opcode_g3eb(unsigned nnn);
void bx_instr_opcode_g3ev(unsigned nnn);
void bx_instr_opcode_g4(unsigned nnn);
void bx_instr_opcode_g5(unsigned nnn);
void bx_instr_opcode_g6(unsigned nnn);
void bx_instr_opcode_g7(unsigned nnn);
void bx_instr_opcode_g8evib(unsigned nnn);
void bx_instr_mem_code(Bit32u linear, unsigned size);
void bx_instr_mem_data(Bit32u linear, unsigned size, unsigned rw);
void bx_instr_opcode_begin(void);
void bx_instr_opcode_end(void);
void bx_instr_exception(unsigned vector);
void bx_instr_tlb_cntrl(unsigned what, Bit32u newval);
void bx_instr_cache_cntrl(unsigned what);
void bx_instr_hwinterrupt(unsigned vector, Bit32u cs, Bit32u eip);
void bx_instr_init(void);
void bx_instr_shutdown(void);
void bx_instr_opcode_repeating(void);

// called from the debug prompt
void bx_instr_start(void);
void bx_instr_stop(void);
void bx_instr_reset(void);
void bx_instr_print(void);

#if BX_INSTRUMENTATION
#  define BX_INSTR_INIT()                      bx_instr_init()
#  define BX_INSTR_SHUTDOWN()                  bx_instr_shutdown()
#  define BX_INSTR_CNEAR_BRANCH(new_eip)       bx_instr_cnear_branch(new_eip)
#  define BX_INSTR_UCNEAR_BRANCH(new_eip)      bx_instr_ucnear_branch(new_eip)
#  define BX_INSTR_FAR_BRANCH(new_cs, new_eip) bx_instr_far_branch(new_cs, new_eip)
#  define BX_INSTR_OPCODE_BEGIN()              bx_instr_opcode_begin()
#  define BX_INSTR_OPCODE_END()                bx_instr_opcode_end()
#  define BX_INSTR_OPCODE_BYTE1(b)             bx_instr_opcode_byte1(b)
#  define BX_INSTR_OPCODE_BYTE2(b)             bx_instr_opcode_byte2(b)
#  define BX_INSTR_OPCODE_G1EbIb(nnn)          bx_instr_opcode_g1ebib(nnn)
#  define BX_INSTR_OPCODE_G1EvIv(nnn)          bx_instr_opcode_g1eviv(nnn)
#  define BX_INSTR_OPCODE_G1EvIb(nnn)          bx_instr_opcode_g1evib(nnn)
#  define BX_INSTR_OPCODE_G2EbIb(nnn)          bx_instr_opcode_g2ebib(nnn)
#  define BX_INSTR_OPCODE_G2EvIb(nnn)          bx_instr_opcode_g2evib(nnn)
#  define BX_INSTR_OPCODE_G2Eb1(nnn)           bx_instr_opcode_g2eb1(nnn)
#  define BX_INSTR_OPCODE_G2Ev1(nnn)           bx_instr_opcode_g2ev1(nnn)
#  define BX_INSTR_OPCODE_G2EbCL(nnn)          bx_instr_opcode_g2ebcl(nnn)
#  define BX_INSTR_OPCODE_G2EvCL(nnn)          bx_instr_opcode_g2evcl(nnn)
#  define BX_INSTR_OPCODE_G3Eb(nnn)            bx_instr_opcode_g3eb(nnn)
#  define BX_INSTR_OPCODE_G3Ev(nnn)            bx_instr_opcode_g3ev(nnn)
#  define BX_INSTR_OPCODE_G4(nnn)              bx_instr_opcode_g4(nnn)
#  define BX_INSTR_OPCODE_G5(nnn)              bx_instr_opcode_g5(nnn)
#  define BX_INSTR_OPCODE_G6(nnn)              bx_instr_opcode_g6(nnn)
#  define BX_INSTR_OPCODE_G7(nnn)              bx_instr_opcode_g7(nnn)
#  define BX_INSTR_OPCODE_G8EvIb(nnn)          bx_instr_opcode_g8evib(nnn)
#  define BX_INSTR_MEM_CODE(linear, size)      bx_instr_mem_code(linear, size)
#  define BX_INSTR_MEM_DATA(linear, size, rw)  bx_instr_mem_data(linear, size, rw)
#  define BX_INSTR_EXCEPTION(vector)           bx_instr_exception(vector)
#  define BX_INSTR_TLB_CNTRL(what, newval)     bx_instr_tlb_cntrl(what, newval)
#  define BX_INSTR_CACHE_CNTRL(what)           bx_instr_cache_cntrl(what)
#  define BX_INSTR_HWINTERRUPT(vector, cs, eip) bx_instr_hwinterrupt(vector, cs, eip)
#  define BX_INSTR_OPCODE_REPEATING()           bx_instr_opcode_repeating()
#else
#  define BX_INSTR_INIT()
#  define BX_INSTR_SHUTDOWN()
#  define BX_INSTR_CNEAR_BRANCH(new_eip)
#  define BX_INSTR_UCNEAR_BRANCH(new_eip)
#  define BX_INSTR_FAR_BRANCH(new_cs, new_eip)
#  define BX_INSTR_OPCODE_BEGIN()
#  define BX_INSTR_OPCODE_END()
#  define BX_INSTR_OPCODE_BYTE1(b)
#  define BX_INSTR_OPCODE_BYTE2(b)
#  define BX_INSTR_OPCODE_G1EbIb(nnn)
#  define BX_INSTR_OPCODE_G1EvIv(nnn)
#  define BX_INSTR_OPCODE_G1EvIb(nnn)
#  define BX_INSTR_OPCODE_G2EbIb(nnn)
#  define BX_INSTR_OPCODE_G2EvIb(nnn)
#  define BX_INSTR_OPCODE_G2Eb1(nnn)
#  define BX_INSTR_OPCODE_G2Ev1(nnn)
#  define BX_INSTR_OPCODE_G2EbCL(nnn)
#  define BX_INSTR_OPCODE_G2EvCL(nnn)
#  define BX_INSTR_OPCODE_G3Eb(nnn)
#  define BX_INSTR_OPCODE_G3Ev(nnn)
#  define BX_INSTR_OPCODE_G4(nnn)
#  define BX_INSTR_OPCODE_G5(nnn)
#  define BX_INSTR_OPCODE_G6(nnn)
#  define BX_INSTR_OPCODE_G7(nnn)
#  define BX_INSTR_OPCODE_G8EvIb(nnn)
#  define BX_INSTR_MEM_CODE(linear, size)
#  define BX_INSTR_MEM_DATA(linear, size, rw)
#  define BX_INSTR_EXCEPTION(vector)
#  define BX_INSTR_TLB_CNTRL(what, newval)
#  define BX_INSTR_CACHE_CNTRL(what)
#  define BX_INSTR_HWINTERRUPT(vector, cs, eip)
#  define BX_INSTR_OPCODE_REPEATING()
#endif
