// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com





#include "bochs.h"


#if BX_USE_CMOS_SMF
bx_cmos_c bx_cmos;
#define this NULL
#endif


bx_cmos_c::bx_cmos_c(void)
{
  BX_CMOS_THIS s.cmos_mem_address = 0;
}
 
bx_cmos_c::~bx_cmos_c(void)
{
  // nothing for now
}


  void
bx_cmos_c::init(bx_devices_c *d)
{
  unsigned i;

  // CMOS RAM & RTC

  BX_CMOS_THIS devices = d;

  BX_CMOS_THIS devices->register_io_read_handler(this,
       read_handler, 0x0070,
       "CMOS RAM");
  BX_CMOS_THIS devices->register_io_read_handler(this,
       read_handler,
       0x0071,
       "CMOS RAM");
  BX_CMOS_THIS devices->register_io_write_handler(this,
       write_handler,
       0x0070, "CMOS RAM");
  BX_CMOS_THIS devices->register_io_write_handler(this,
       write_handler,
       0x0071, "CMOS RAM");
  BX_CMOS_THIS devices->register_irq(8, "CMOS RTC");

  BX_CMOS_THIS s.timer_index =
    bx_pc_system.register_timer( this, timer_handler,
      1000000, 1,0); // continuous, not-active

  for (i=0; i<BX_NUM_CMOS_REGS; i++) {
    BX_CMOS_THIS s.reg[i] = 0;
    }

  BX_CMOS_THIS s.timeval = time(NULL);
  update_clock();
  BX_CMOS_THIS s.reg[0x0a] = 0x26;
  BX_CMOS_THIS s.reg[0x0b] = 0x02;
  BX_CMOS_THIS s.reg[0x0c] = 0x00;
  BX_CMOS_THIS s.reg[0x0d] = 0x01;
  if ( (BX_CMOS_THIS s.reg[0x0b] & 0x80) == 0 )
    bx_pc_system.activate_timer(BX_CMOS_THIS s.timer_index, 1000000, 1);
}


  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_cmos_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_CMOS_SMF
  bx_cmos_c *class_ptr = (bx_cmos_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}

  Bit32u
bx_cmos_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif
  Bit8u ret8;

  if (io_len > 1)
    bx_panic("cmos: io read from address %08x len=%u\n",
             (unsigned) address, (unsigned) io_len);

  if (bx_dbg.cmos)
    bx_printf("CMOS read of CMOS register 0x%x\n",
      (unsigned) BX_CMOS_THIS s.cmos_mem_address);


  switch (address) {
    case 0x0071:
      if (BX_CMOS_THIS s.cmos_mem_address >= BX_NUM_CMOS_REGS) {
        bx_panic("unsupported cmos io read, register(0x%02x)!\n",
            (unsigned) BX_CMOS_THIS s.cmos_mem_address);
        }

      ret8 = BX_CMOS_THIS s.reg[BX_CMOS_THIS s.cmos_mem_address];
      // all bits of Register C are cleared after a read occurs.
      if (BX_CMOS_THIS s.cmos_mem_address == 0x0c)
        BX_CMOS_THIS s.reg[0x0c] = 0x00;
      return(ret8);
      break;
    default:
      bx_panic("unsupported cmos read, address=%0x%x!\n",
        (unsigned) address);
      return(0);
      break;
    }
}


  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_cmos_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_CMOS_SMF
  bx_cmos_c *class_ptr = (bx_cmos_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_cmos_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_CMOS_SMF

  if (io_len > 1)
    bx_panic("cmos: io write to address %08x len=%u\n",
             (unsigned) address, (unsigned) io_len);

  if (bx_dbg.cmos)
    bx_printf("CMOS write to address: 0x%x = 0x%x\n",
      (unsigned) address, (unsigned) value);


  /* CMOS RAM */
  switch (address) {
    case 0x0070:
      BX_CMOS_THIS s.cmos_mem_address = value & 0x3F;
      break;

    case 0x0071:
      if (BX_CMOS_THIS s.cmos_mem_address >= BX_NUM_CMOS_REGS) {
        bx_panic("unsupported cmos io write, register(0x%02x)=%02x!\n",
            (unsigned) BX_CMOS_THIS s.cmos_mem_address, (unsigned) value);
        return;
        }
      switch (BX_CMOS_THIS s.cmos_mem_address) {
        case 0x00: // seconds
        case 0x01: // seconds alarm
        case 0x02: // minutes
        case 0x03: // minutes alarm
        case 0x04: // hours
        case 0x05: // hours alarm
        case 0x06: // day of the week
        case 0x07: // day of the month
        case 0x08: // month
        case 0x09: // year
          //bx_printf("CMOS: write reg %02xh: value = %02xh\n",
          //    (unsigned) BX_CMOS_THIS s.cmos_mem_address, (unsigned) value);
          break;

        case 0x0a: // status register A
          if (value != 0x26)
            bx_panic("unsupported cmos io write (0x71), case 0x0a = %02x!\n",
              (unsigned) value);
          break;
        case 0x0b: // status register B
          // setting bit 7 (halt clock), clears bit 4
          if (value & 0x80)
            value &= ~0x10;
          if ( (value & 0x80) && ((BX_CMOS_THIS s.reg[0x0b] & 0x80)==0) ) {
            // bit7: transition from 0 to 1, clock halts
            }
          if ( ((value & 0x80)==0) && (BX_CMOS_THIS s.reg[0x0b] & 0x80) ) {
            // bit7: transition from 1 to 0, clock resumes
            bx_pc_system.activate_timer(BX_CMOS_THIS s.timer_index, 1000000, 1);
            }
          if (value & 0x40)
            bx_panic("cmos: write status reg B, periodic interrupt enabled.\n");
          if (value & 0x20)
            bx_panic("cmos: write status reg B, alarm interrupt enabled.\n");
          if (value & 0x08)
            bx_panic("cmos: write status reg B, square wave enabled.\n");
          if (value & 0x04)
            bx_panic("cmos: write status reg B, binary format enabled.\n");
          if ( !(value & 0x02) )
            bx_panic("cmos: write status reg B, 12 hour mode enabled.\n");
          if (value & 0x01)
            bx_panic("cmos: write status reg B, daylight savings enabled.\n");
          break;
        case 0x0c: // status register C
          bx_panic("cmos: status register C write (read-only register)\n");
          break;
        case 0x0d: // status register D
          bx_panic("cmos: status register D write\n");
          break;

        case 0x0e: // diagnostic status
          bx_printf("CMOS: write register 0Eh: %02x\n", (unsigned) value);
          break;

	case 0x0f: /* shutdown status */
          switch (value) {
            case 0: /* proceed with normal POST (soft reset) */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0F set to 0: shutdown action = normal POST\n");
              break;
            case 0x02: /* shutdown after memory test */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0Fh: request to change shutdown action"
                             " to shutdown after memory test\n");
              break;
            case 4: /* jump to disk bootstrap routine */
              bx_printf("CMOS: Reg 0Fh: request to change shutdown action "
                             "to jump to disk bootstrap routine.\n");
              break;
            case 9: /* return to BIOS extended memory block move
                       (interrupt 15h, func 87h was in progress) */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0Fh: request to change shutdown action "
                             "to return to BIOS extended memory block move.\n");
              break;
            case 0x0a: /* jump to DWORD pointer at 40:67 */
              if (bx_dbg.reset)
                bx_printf("CMOS: Reg 0Fh: request to change shutdown action"
                             " to jump to DWORD at 40:67\n");
              break;
            default:
              bx_panic("unsupported cmos io write to reg F, case %x!\n",
                (unsigned) value);
              break;
            }
          break;

        default:
          bx_printf("CMOS: write reg %02xh: value = %02xh\n",
            (unsigned) BX_CMOS_THIS s.cmos_mem_address, (unsigned) value);
          break;
        }

      BX_CMOS_THIS s.reg[BX_CMOS_THIS s.cmos_mem_address] = value;
      break;
    }
}


  void
bx_cmos_c::checksum_cmos(void)
{
  unsigned i;
  Bit16u sum;

  sum = 0;
  for (i=0x10; i<=0x2d; i++) {
    sum += BX_CMOS_THIS s.reg[i];
    }
  BX_CMOS_THIS s.reg[0x2e] = (sum >> 8) & 0xff; /* checksum high */
  BX_CMOS_THIS s.reg[0x2f] = (sum & 0xff);      /* checksum low */

bx_printf("cmos: cmos[0x14] = %02x\n", (unsigned) BX_CMOS_THIS s.reg[0x14]);
}

  void
bx_cmos_c::timer_handler(void *this_ptr)
{
  bx_cmos_c *class_ptr = (bx_cmos_c *) this_ptr;

  class_ptr->timer();
}

  void
bx_cmos_c::timer()
{
  // if clock is halted, nothing to do
  if (BX_CMOS_THIS s.reg[0x0b] & 0x80)
    return;

  // clock is active
  // assume 1 more second has elapsed

  BX_CMOS_THIS s.timeval++;
  update_clock();

  // if update interrupts are enabled, trip IRQ 8, and
  // update status register C
  if (BX_CMOS_THIS s.reg[0x0b] & 0x10) { // bit 4 is set
    BX_CMOS_THIS s.reg[0x0c] &= 0x90; // IRQ=8 | US=1
    BX_CMOS_THIS devices->pic->trigger_irq(8);
    }
}


  void
bx_cmos_c::update_clock()
{
  struct tm *time_calendar;
  unsigned year, month, day;
  Bit8u val_bcd;

  time_calendar = localtime(& BX_CMOS_THIS s.timeval);

  // update seconds
  val_bcd =
     ((time_calendar->tm_sec  / 10) << 4) |
     (time_calendar->tm_sec % 10);
  BX_CMOS_THIS s.reg[0x00] = val_bcd;

  // update minutes
  val_bcd =
     ((time_calendar->tm_min  / 10) << 4) |
     (time_calendar->tm_min % 10);
  BX_CMOS_THIS s.reg[0x02] = val_bcd;

  // update hours
  val_bcd =
     ((time_calendar->tm_hour  / 10) << 4) |
     (time_calendar->tm_hour % 10);
  BX_CMOS_THIS s.reg[0x04] = val_bcd;

  // update day of the week
  day = time_calendar->tm_wday + 1; // 0..6 to 1..7
  BX_CMOS_THIS s.reg[0x06] = ((day / 10) << 4) | (day % 10);

  // update day of the month
  day = time_calendar->tm_mday;
  BX_CMOS_THIS s.reg[0x07] = ((day / 10) << 4) | (day % 10);

  // update month
  month   = time_calendar->tm_mon + 1;
  BX_CMOS_THIS s.reg[0x08] = ((month / 10) << 4) | (month % 10);

  // update year
  year = time_calendar->tm_year % 100;
  BX_CMOS_THIS s.reg[0x09] = ((year  / 10) << 4) | (year % 10);
}
