// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com



#include "bochs.h"

extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
}



#define INDEX_PULSE_CYCLE 10

#if BX_USE_HD_SMF
bx_hard_drive_c bx_hard_drive;
#define this NULL
#endif


//static unsigned im_here = 0;

bx_hard_drive_c::bx_hard_drive_c(void)
{
  // nothing for now
}

bx_hard_drive_c::~bx_hard_drive_c(void)
{
  // nothing for now
}



  void
bx_hard_drive_c::init(bx_devices_c *d, bx_cmos_c *cmos)
{
  struct stat stat_buf;
  int ret;

  BX_HD_THIS devices = d;

  /* HARD DRIVE 0 */

  BX_HD_THIS devices->register_irq(14, "Hard Drive 0");
  for (unsigned addr=0x01F0; addr<=0x01F7; addr++) {
    BX_HD_THIS devices->register_io_read_handler(this, read_handler,
                                        addr, "Hard Drive 0");
    BX_HD_THIS devices->register_io_write_handler(this, write_handler,
                                        addr, "Hard Drive 0");
    }

  /* open hard drive image file */
  BX_HD_THIS s.hard_drive.fd = open(bx_options.diskc.path, O_RDWR
#ifdef O_BINARY
                                    | O_BINARY
#endif
                                   );

  if (BX_HD_THIS s.hard_drive.fd < 0) {
    bx_panic("could not open hard drive image file '%s'!\n",
      bx_options.diskc.path);
    }

  /* look at size of image file to calculate disk geometry */
  ret = fstat(BX_HD_THIS s.hard_drive.fd, &stat_buf);
  if (ret) {
    perror("fstat'ing hard drive image file");
    bx_panic("fstat() returns error!\n");
    }

  cmos->s.reg[0x12] = 0x00; /* start out with: no drive 0, no drive 1 */

  // Flag drive type as Fh, use extended CMOS location as real type
  cmos->s.reg[0x12] = (cmos->s.reg[0x12] & 0x0f) | 0xf0;
  cmos->s.reg[0x19] = 47; // user definable type
  cmos->s.reg[0x1b] = (bx_options.diskc.cylinders & 0x00ff);
  cmos->s.reg[0x1c] = (bx_options.diskc.cylinders & 0xff00) >> 8;
  cmos->s.reg[0x1d] = (bx_options.diskc.heads);
  cmos->s.reg[0x1e] = 0xff; // -1
  cmos->s.reg[0x1f] = 0xff; // -1
  cmos->s.reg[0x20] = 0xc0 | ((bx_options.diskc.heads > 8) << 3);
  cmos->s.reg[0x21] = cmos->s.reg[0x1b];
  cmos->s.reg[0x22] = cmos->s.reg[0x1c];
  cmos->s.reg[0x23] = bx_options.diskc.spt;

  if ( bx_options.bootdrive[0] == 'c' ) {
    // system boot sequence C:, A:
    cmos->s.reg[0x2d] &= 0xdf;
    }
  else { // 'a'
    // system boot sequence A:, C:
    cmos->s.reg[0x2d] |= 0x20;
    }

  BX_HD_THIS s.hard_drive.cylinders = bx_options.diskc.cylinders;
  BX_HD_THIS s.hard_drive.heads     = bx_options.diskc.heads;
  BX_HD_THIS s.hard_drive.sectors   = bx_options.diskc.spt;

  //switch (stat_buf.st_size) {
  //  }

  BX_HD_THIS s.controller.status.busy           = 0;
  BX_HD_THIS s.controller.status.drive_ready    = 1;
  BX_HD_THIS s.controller.status.write_fault    = 0;
  BX_HD_THIS s.controller.status.seek_complete  = 1;
  BX_HD_THIS s.controller.status.drq            = 0;
  BX_HD_THIS s.controller.status.corrected_data = 0;
  BX_HD_THIS s.controller.status.index_pulse    = 0;
  BX_HD_THIS s.controller.status.index_pulse_count = 0;
  BX_HD_THIS s.controller.status.err            = 0;

  BX_HD_THIS s.controller.error_register = 0;
  BX_HD_THIS s.controller.diag_mode      = 0;
  BX_HD_THIS s.controller.diag_byte      = 0;
  BX_HD_THIS s.controller.drive_select   = 0;
  BX_HD_THIS s.controller.head_no        = 0;
  BX_HD_THIS s.controller.sector_count   = 0;
  BX_HD_THIS s.controller.sector_no      = 0;
  BX_HD_THIS s.controller.cylinder_no    = 0;
  BX_HD_THIS s.controller.current_command = 0x00;
  BX_HD_THIS s.controller.buffer_index = 0;

  BX_HD_THIS s.controller.control.reserved    = 1;
  BX_HD_THIS s.controller.control.reset       = 0;
  BX_HD_THIS s.controller.control.disable_irq = 0;
}



  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_hard_drive_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_HD_SMF
  bx_hard_drive_c *class_ptr = (bx_hard_drive_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}


  Bit32u
bx_hard_drive_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_HD_SMF
  Bit8u value8;
  Bit16u value16;

  if (io_len==2 && address!=0x1f0) {
    bx_panic("disk: non-byte IO read to %04x\n", (unsigned) address);
    }

  if (bx_dbg.disk)
    bx_printf("disk: IO read from %04x\n", (unsigned) address);

  switch (address) {
    case 0x1f0: /* hard disk data (16bit) */
      if (io_len != 2) {
        bx_panic("disk: non-word IO read from %04x\n", (unsigned) address);
        }
      switch (BX_HD_THIS s.controller.current_command) {
        case 0x20: /* read sectors */
          if (BX_HD_THIS s.controller.buffer_index >= 512)
            bx_panic("disk: IO read(1f0): buffer_index >= 512\n");
          value16  = BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index];
          value16 |= (BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index+1] << 8);
          BX_HD_THIS s.controller.buffer_index += 2;
          
          /* if buffer completely read */
          if (BX_HD_THIS s.controller.buffer_index >= 512) {
            /* update sector count, sector number, cylinder,
             * drive, head, status
             * if there are more sectors, read next one in...
             */
            BX_HD_THIS s.controller.buffer_index = 0;
            /* sector count of 0 means 256, it'll wrap */
            BX_HD_THIS s.controller.sector_count--;
            BX_HD_THIS s.controller.sector_no++;
            if (BX_HD_THIS s.controller.sector_no > BX_HD_THIS s.hard_drive.sectors) {
              BX_HD_THIS s.controller.sector_no = 1;
              BX_HD_THIS s.controller.head_no++;
              if (BX_HD_THIS s.controller.head_no >= BX_HD_THIS s.hard_drive.heads) {
                BX_HD_THIS s.controller.head_no = 0;
                BX_HD_THIS s.controller.cylinder_no++;
                if (BX_HD_THIS s.controller.cylinder_no >= BX_HD_THIS s.hard_drive.cylinders)
                  BX_HD_THIS s.controller.cylinder_no = BX_HD_THIS s.hard_drive.cylinders - 1;
                }
              }

            BX_HD_THIS s.controller.status.busy = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.write_fault = 0;
            BX_HD_THIS s.controller.status.seek_complete = 0;
            BX_HD_THIS s.controller.status.corrected_data = 0;
            BX_HD_THIS s.controller.status.err = 0;

            if (BX_HD_THIS s.controller.sector_count==0) {
              BX_HD_THIS s.controller.status.drq = 0;
              }
            else { /* read next one into controller buffer */
              unsigned long logical_sector;
              int ret;

              BX_HD_THIS s.controller.status.drq = 1;
              BX_HD_THIS s.controller.status.seek_complete = 1;

              logical_sector = (BX_HD_THIS s.controller.cylinder_no * BX_HD_THIS s.hard_drive.heads *
                                BX_HD_THIS s.hard_drive.sectors) +
                               (BX_HD_THIS s.controller.head_no * BX_HD_THIS s.hard_drive.sectors) +
                               (BX_HD_THIS s.controller.sector_no - 1);

              if (logical_sector >= (BX_HD_THIS s.hard_drive.cylinders *
                  BX_HD_THIS s.hard_drive.heads * BX_HD_THIS s.hard_drive.sectors))
                bx_panic("disk: read sectors: out of bounds\n");
              ret = lseek(BX_HD_THIS s.hard_drive.fd, logical_sector * 512, SEEK_SET);

              if (ret < 0)
                bx_panic("disk: could lseek() hard drive image file\n");
              ret = ::read(BX_HD_THIS s.hard_drive.fd,
                           (void *) BX_HD_THIS s.controller.buffer, 512);
              if (ret < 512) {
                bx_printf("logical sector was %u\n", (unsigned) logical_sector);
                bx_panic("disk: could not read() hard drive image file\n");
                }

              BX_HD_THIS s.controller.buffer_index = 0;
              if (!BX_HD_THIS s.controller.control.disable_irq)
                BX_HD_THIS devices->pic->trigger_irq(14);
              }
            }
          return(value16);
          break;
        default:
          bx_panic("disk: IO read(1f0h): current command is %02xh\n",
            (unsigned) BX_HD_THIS s.controller.current_command);
        }
      break;

    case 0x1f1: /* hard disk error register */
      if (BX_HD_THIS s.controller.diag_mode) {
        value8 = BX_HD_THIS s.controller.diag_byte;
        return(value8);
        BX_HD_THIS s.controller.diag_mode = 0;
        }
      value8 = BX_HD_THIS s.controller.error_register;
      return(value8);
      break;

    case 0x1f2: /* hard disk sector count */
      if (BX_HD_THIS s.controller.current_command==0x20 ||
          BX_HD_THIS s.controller.current_command==0x30) {
        value8 = BX_HD_THIS s.controller.sector_count;
        return(value8);
        }
      bx_panic("disk: IO read(0x1f2): current command not read/write\n");
      break;

    case 0x1f3: // sector number
      value8 = BX_HD_THIS s.controller.sector_no;
      return(value8);
    case 0x1f4: // cylinder low
      value8 = (BX_HD_THIS s.controller.cylinder_no & 0x00ff);
      return(value8);
    case 0x1f5: // cylinder high
      value8 = ((BX_HD_THIS s.controller.cylinder_no & 0x0300) >> 8);
      return(value8);

    case 0x1f6: // hard disk drive and head register
      value8 = (1 << 7) | // extended data field for ECC
               (0 << 7) | // 1=LBA mode, 0=CHSmode
               (1 << 5) | // 01b = 512 sector size
               (BX_HD_THIS s.controller.drive_select << 4) |
               (BX_HD_THIS s.controller.head_no << 0);
      return(value8);
      break;

    case 0x1f7: // Hard Disk Status
    case 0x3f6: // Hard Disk Alternate Status
      value8 = (
        (BX_HD_THIS s.controller.status.busy << 7) |
        (BX_HD_THIS s.controller.status.drive_ready << 6) |
        (BX_HD_THIS s.controller.status.write_fault << 5) |
        (BX_HD_THIS s.controller.status.seek_complete << 4) |
        (BX_HD_THIS s.controller.status.drq << 3) |
        (BX_HD_THIS s.controller.status.corrected_data << 2) |
        (BX_HD_THIS s.controller.status.index_pulse << 1) |
        (BX_HD_THIS s.controller.status.err) );
      BX_HD_THIS s.controller.status.index_pulse_count++;
      BX_HD_THIS s.controller.status.index_pulse = 0;
      if (BX_HD_THIS s.controller.status.index_pulse_count >= INDEX_PULSE_CYCLE) {
        BX_HD_THIS s.controller.status.index_pulse = 1;
        BX_HD_THIS s.controller.status.index_pulse_count = 0;
        }
      return(value8);
      break;

    case 0x3f7: // Hard Disk Address Register
      //bx_printf("hard drive: address register unsupported\n");
      return(0);
      break;

    default:
      bx_panic("hard drive: io read to address %x unsupported\n",
        (unsigned) address);
    }

  bx_panic("hard drive: shouldnt get here!\n");
  return(0);
}


  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_hard_drive_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_HD_SMF
  bx_hard_drive_c *class_ptr = (bx_hard_drive_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_hard_drive_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_HD_SMF
  unsigned long logical_sector;
  int ret;
  Boolean prev_control_reset;

  if (io_len==2 && address!=0x1f0) {
    bx_panic("disk: non-byte IO write to %04x\n", (unsigned) address);
    }

  if (bx_dbg.disk)
    bx_printf("disk: IO write to %04x = %02x\n",
      (unsigned) address, (unsigned) value);

  switch (address) {
    case 0x1f0: 
      if (io_len != 2) {
        bx_panic("disk: non-word IO read from %04x\n", (unsigned) address);
        }
      switch (BX_HD_THIS s.controller.current_command) {
        case 0x30:
          if (BX_HD_THIS s.controller.buffer_index >= 512)
            bx_panic("disk: IO write(1f0): buffer_index >= 512\n");
          BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index] = value;
          BX_HD_THIS s.controller.buffer[BX_HD_THIS s.controller.buffer_index+1] = (value >> 8);
          BX_HD_THIS s.controller.buffer_index += 2;

          /* if buffer completely writtten */
          if (BX_HD_THIS s.controller.buffer_index >= 512) {
            unsigned long logical_sector;
            int ret;

            logical_sector = (BX_HD_THIS s.controller.cylinder_no * BX_HD_THIS s.hard_drive.heads *
                              BX_HD_THIS s.hard_drive.sectors) +
                             (BX_HD_THIS s.controller.head_no * BX_HD_THIS s.hard_drive.sectors) +
                             (BX_HD_THIS s.controller.sector_no - 1);

            if (logical_sector >= (BX_HD_THIS s.hard_drive.cylinders *
                BX_HD_THIS s.hard_drive.heads * BX_HD_THIS s.hard_drive.sectors))
              bx_panic("disk: write sectors: out of bounds\n");
            ret = lseek(BX_HD_THIS s.hard_drive.fd, logical_sector * 512, SEEK_SET);
            if (ret < 0)
              bx_panic("disk: could lseek() hard drive image file\n");

            ret = ::write(BX_HD_THIS s.hard_drive.fd,
                          (void *) BX_HD_THIS s.controller.buffer, 512);
            if (ret < 512)
              bx_panic("disk: could not write() hard drive image file\n");

            BX_HD_THIS s.controller.buffer_index = 0;

            /* update sector count, sector number, cylinder,
             * drive, head, status
             * if there are more sectors, read next one in...
             */
            /* sector count of 0 means 256, it'll wrap */
            BX_HD_THIS s.controller.sector_count--;
            BX_HD_THIS s.controller.sector_no++;
            if (BX_HD_THIS s.controller.sector_no > BX_HD_THIS s.hard_drive.sectors) {
              BX_HD_THIS s.controller.sector_no = 1;
              BX_HD_THIS s.controller.head_no++;
              if (BX_HD_THIS s.controller.head_no >= BX_HD_THIS s.hard_drive.heads) {
                BX_HD_THIS s.controller.head_no = 0;
                BX_HD_THIS s.controller.cylinder_no++;
                if (BX_HD_THIS s.controller.cylinder_no >= BX_HD_THIS s.hard_drive.cylinders)
                  BX_HD_THIS s.controller.cylinder_no = BX_HD_THIS s.hard_drive.cylinders - 1;
                }
              }

            /* When the write is complete, controller clears the DRQ bit and
             * sets the BSY bit.
             * If at least one more sector is to be written, controller sets DRQ bit,
             * clears BSY bit, and issues IRQ 14
             */

            if (BX_HD_THIS s.controller.sector_count!=0) {
              BX_HD_THIS s.controller.status.busy = 0;
              BX_HD_THIS s.controller.status.drive_ready = 1;
              BX_HD_THIS s.controller.status.drq = 1;
              BX_HD_THIS s.controller.status.corrected_data = 0;
              BX_HD_THIS s.controller.status.err = 0;
              }
            else { /* no more sectors to write */
              BX_HD_THIS s.controller.status.busy = 0;
              BX_HD_THIS s.controller.status.drive_ready = 1;
              BX_HD_THIS s.controller.status.drq = 0;
              BX_HD_THIS s.controller.status.err = 0;
              BX_HD_THIS s.controller.status.corrected_data = 0;
              }
            if (!BX_HD_THIS s.controller.control.disable_irq)
              BX_HD_THIS devices->pic->trigger_irq(14);
            }
          break;

        default:
          bx_panic("disk: IO write(1f0h): current command is %02xh\n",
            (unsigned) BX_HD_THIS s.controller.current_command);
        }
      break;

    case 0x1f1: /* hard disk write precompensation */
      if (bx_dbg.disk) {
        if (value == 0xff)
          bx_printf("disk: no precompensation\n");
        else
          bx_printf("disk: precompensation value %02x\n", (unsigned) value);
        }
      break;

    case 0x1f2: /* hard disk sector count */
      BX_HD_THIS s.controller.sector_count = value;
      if (bx_dbg.disk)
        bx_printf("disk: sector count = %u\n", (unsigned) value);
      break;

    case 0x1f3: /* hard disk sector number */
      BX_HD_THIS s.controller.sector_no = value;
      if (bx_dbg.disk)
        bx_printf("disk: sector number = %u\n", (unsigned) value);
      break;

    case 0x1f4: /* hard disk cylinder low */
      BX_HD_THIS s.controller.cylinder_no = (BX_HD_THIS s.controller.cylinder_no & 0x0300) | value;
      if (bx_dbg.disk)
        bx_printf("disk: cylinder low = %02xh\n", (unsigned) value);
      break;

    case 0x1f5: /* hard disk cylinder high */
      value &= 0x03; /* only use 2 bits for a total of 10bits */
      BX_HD_THIS s.controller.cylinder_no = (BX_HD_THIS s.controller.cylinder_no & 0x00ff) |
                               (value << 8);
      if (bx_dbg.disk)
        bx_printf("disk: cylinder high = %02xh\n", (unsigned) value);
      break;

    case 0x1f6: /* hard disk drive and head register */
      // b7 0
      // b6 1=LBA mode, 0=CHS mode (LBA not supported)
      // b5 0
      // b4: DRV
      // b3..0 HD3..HD0
      if ( (value & 0xe0) != 0xa0 ) // 101xxxxx
        bx_printf("disk: IO write 1f6 (%02x): not 101xxxxxb\n", (unsigned) value);
      BX_HD_THIS s.controller.drive_select = (value >> 4) & 0x01;
      BX_HD_THIS s.controller.head_no = (value & 0x0f);
      break;

    case 0x1f7: /* hard disk command */
      if (BX_HD_THIS s.controller.status.busy)
        bx_panic("hard disk: command sent, controller BUSY\n");
      if ( (value & 0xf0) == 0x10 )
        value = 0x10;
      switch (value) {

        case 0x10: /* calibrate drive */
          if (BX_HD_THIS s.controller.drive_select != 0)
            bx_panic("disk: calibrate drive != 0\n");
          /* move head to cylinder 0, issue IRQ 14 */
          BX_HD_THIS s.controller.error_register = 0;
          BX_HD_THIS s.controller.cylinder_no = 0;
          BX_HD_THIS s.controller.status.busy = 0;
          BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.seek_complete = 1;
          BX_HD_THIS s.controller.status.drq = 0;
          BX_HD_THIS s.controller.status.err = 0;
          if (!BX_HD_THIS s.controller.control.disable_irq)
            BX_HD_THIS devices->pic->trigger_irq(14);
          break;

        case 0x20: /* read multiple sectors, with retries */
          /* update sector_no, always points to current sector
           * after each sector is read to buffer, DRQ bit set and issue IRQ 14
           * if interrupt handler transfers all data words into main memory,
           * and more sectors to read, then set BSY bit again, clear DRQ and
           * read next sector into buffer
           * sector count of 0 means 256 sectors
           */

          BX_HD_THIS s.controller.current_command = value;

          logical_sector = (BX_HD_THIS s.controller.cylinder_no * BX_HD_THIS s.hard_drive.heads *
                            BX_HD_THIS s.hard_drive.sectors) +
                           (BX_HD_THIS s.controller.head_no * BX_HD_THIS s.hard_drive.sectors) +
                           (BX_HD_THIS s.controller.sector_no - 1);

          if (logical_sector >=
              (BX_HD_THIS s.hard_drive.cylinders * BX_HD_THIS s.hard_drive.heads * BX_HD_THIS s.hard_drive.sectors)) {
            bx_panic("disk: read sectors: out of bounds\n");
            }
          ret = lseek(BX_HD_THIS s.hard_drive.fd, logical_sector * 512, SEEK_SET);

          if (ret < 0) {
            bx_panic("disk: could not lseek() hard drive image file\n");
            }
          ret = ::read(BX_HD_THIS s.hard_drive.fd,
                       (void *) BX_HD_THIS s.controller.buffer, 512);
          if (ret < 512) {
            bx_printf("logical sector was %u\n", (unsigned) logical_sector);
            bx_panic("disk: could not read() hard drive image file\n");
            }

          BX_HD_THIS s.controller.error_register = 0;
          BX_HD_THIS s.controller.status.busy  = 0;
          BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.seek_complete = 1;
          BX_HD_THIS s.controller.status.drq   = 1;
          BX_HD_THIS s.controller.status.corrected_data = 0;
          BX_HD_THIS s.controller.status.err   = 0;
          BX_HD_THIS s.controller.buffer_index = 0;
          if (!BX_HD_THIS s.controller.control.disable_irq)
            BX_HD_THIS devices->pic->trigger_irq(14);
          break;

        case 0x30: /* write multiple sectors, with retries */
          /* update sector_no, always points to current sector
           * after each sector is read to buffer, DRQ bit set and issue IRQ 14
           * if interrupt handler transfers all data words into main memory,
           * and more sectors to read, then set BSY bit again, clear DRQ and
           * read next sector into buffer
           * sector count of 0 means 256 sectors
           */

          if (BX_HD_THIS s.controller.status.busy) {
            bx_panic("disk: write command: BSY bit set\n");
            }
          BX_HD_THIS s.controller.current_command = value;

          // implicit seek done :^)
          BX_HD_THIS s.controller.error_register = 0;
          BX_HD_THIS s.controller.status.busy = 0;
          // BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.seek_complete = 1;
          BX_HD_THIS s.controller.status.drq = 1;
          BX_HD_THIS s.controller.status.err   = 0;
          BX_HD_THIS s.controller.buffer_index = 0;
          break;

        case 0x90: // Drive Diagnostic
          if (BX_HD_THIS s.controller.status.busy) {
            bx_panic("disk: diagnostic command: BSY bit set\n");
            }
          BX_HD_THIS s.controller.diag_byte = 0x81; // Drive 1 failed, no error on drive 0
          BX_HD_THIS s.controller.diag_mode = 1;
          // BX_HD_THIS s.controller.status.busy = 0; // not needed
          BX_HD_THIS s.controller.status.drq = 0;
          BX_HD_THIS s.controller.status.err = 0;
          break;

        case 0x91: // initialize drive parameters
          if (BX_HD_THIS s.controller.status.busy) {
            bx_panic("disk: init drive parameters command: BSY bit set\n");
            }
          // sets logical geometry of specified drive
          bx_printf("initialize drive params\n");
          bx_printf("  sector count = %u\n",
            (unsigned) BX_HD_THIS s.controller.sector_count);
          bx_printf("  drive select = %u\n",
            (unsigned) BX_HD_THIS s.controller.drive_select);
          bx_printf("  head number = %u\n",
            (unsigned) BX_HD_THIS s.controller.head_no);
          if (BX_HD_THIS s.controller.drive_select != 0) {
            bx_panic("disk: init drive params: drive != 0\n");
            //BX_HD_THIS s.controller.error_register = 0x12;
            BX_HD_THIS s.controller.status.busy = 0;
            BX_HD_THIS s.controller.status.drive_ready = 1;
            BX_HD_THIS s.controller.status.drq = 0;
            BX_HD_THIS s.controller.status.err = 0;
            if (!BX_HD_THIS s.controller.control.disable_irq)
              BX_HD_THIS devices->pic->trigger_irq(14);
            break;
            }
          if (BX_HD_THIS s.controller.sector_count != BX_HD_THIS s.hard_drive.sectors)
            bx_panic("disk: init drive params: sector count doesnt match\n");
          if ( BX_HD_THIS s.controller.head_no != (BX_HD_THIS s.hard_drive.heads-1) )
            bx_panic("disk: init drive params: head number doesn't match\n");
          BX_HD_THIS s.controller.status.busy = 0;
          BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.drq = 0;
          BX_HD_THIS s.controller.status.err = 0;
          if (!BX_HD_THIS s.controller.control.disable_irq)
            BX_HD_THIS devices->pic->trigger_irq(14);
          break;

        case 0xec: // Get Drive Info (unsupported)
        case 0xf0: // ???
        case 0xa1: // ???
fprintf(stderr, "# *** unsupported drive command %02x\n", (unsigned) value);
          BX_HD_THIS s.controller.status.busy = 0;
          BX_HD_THIS s.controller.status.drive_ready = 1;
          BX_HD_THIS s.controller.status.err = 1;
          BX_HD_THIS s.controller.error_register = 0x04; // command ABORTED
          BX_HD_THIS s.controller.status.drq = 0;
          if (!BX_HD_THIS s.controller.control.disable_irq)
            BX_HD_THIS devices->pic->trigger_irq(14);
          break;

        default:
          bx_panic("IO write(1f7h): command %02x\n", (unsigned) value);
        }
      break;

    case 0x3f6: // hard disk adapter control
      prev_control_reset = BX_HD_THIS s.controller.control.reset;
      BX_HD_THIS s.controller.control.reserved       = value & 0x08;
      BX_HD_THIS s.controller.control.reset          = value & 0x04;
      BX_HD_THIS s.controller.control.disable_irq    = value & 0x02;
      //fprintf(stderr, "# hard drive: adpater control reg: reserved = %d\n",
      //  (unsigned) (BX_HD_THIS s.controller.control.reserved) ? 1 : 0);
      //fprintf(stderr, "# hard drive: adpater control reg: reset controller = %d\n",
      //  (unsigned) (BX_HD_THIS s.controller.control.reset) ? 1 : 0);
      //fprintf(stderr, "# hard drive: adpater control reg: disable_irq(14) = %d\n",
      //  (unsigned) (BX_HD_THIS s.controller.control.disable_irq) ? 1 : 0);
      if (prev_control_reset && !BX_HD_THIS s.controller.control.reset) {
        // transition from 1 to 0: complete controller reset
        bx_printf("hard drive: RESET\n");
        BX_HD_THIS s.controller.status.busy           = 0;
        BX_HD_THIS s.controller.status.drive_ready    = 1;
        BX_HD_THIS s.controller.status.write_fault    = 0;
        BX_HD_THIS s.controller.status.seek_complete  = 0;
        BX_HD_THIS s.controller.status.drq            = 0;
        BX_HD_THIS s.controller.status.corrected_data = 0;
        BX_HD_THIS s.controller.status.err            = 0;

        BX_HD_THIS s.controller.error_register = 0;
        BX_HD_THIS s.controller.diag_mode      = 0;
        BX_HD_THIS s.controller.diag_byte      = 0;
        BX_HD_THIS s.controller.drive_select   = 0;
        BX_HD_THIS s.controller.head_no        = 0;
        BX_HD_THIS s.controller.sector_count   = 0;
        BX_HD_THIS s.controller.sector_no      = 0;
        BX_HD_THIS s.controller.cylinder_no    = 0;
        BX_HD_THIS s.controller.current_command = 0x00;
        BX_HD_THIS s.controller.buffer_index = 0;

        BX_HD_THIS s.controller.control.reserved    = 1;
        BX_HD_THIS s.controller.control.reset       = 0;
        BX_HD_THIS s.controller.control.disable_irq = 0;
        }
      break;

    default:
      bx_panic("hard drive: io write to address %x = %02x\n",
        (unsigned) address, (unsigned) value);
    }
}

  void
bx_hard_drive_c::close_harddrive(void)
{
  if (BX_HD_THIS s.hard_drive.fd > -1) {
    close(BX_HD_THIS s.hard_drive.fd);
    }
}
