// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com




#include "bochs.h"


#if BX_USE_UM_SMF
bx_unmapped_c bx_unmapped;
#define this NULL
#endif


bx_unmapped_c::bx_unmapped_c(void)
{
  s.test_port16=0;
  s.test_port32=0;

  s.port80 = 0x00;
  s.port8e = 0x00;

  s.bios_message_i = 0;
}

bx_unmapped_c::~bx_unmapped_c(void)
{
  // nothing for now
}

  void
bx_unmapped_c::init(bx_devices_c *d)
{
  BX_UM_THIS devices = d;

  for (Bit32u addr=0; addr<0x10000; addr++) {
    BX_UM_THIS devices->register_io_read_handler(this, read_handler,
                                      addr, "Unmapped");
    BX_UM_THIS devices->register_io_write_handler(this, write_handler,
                                      addr, "Unmapped");
    }
}


  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_unmapped_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_UM_SMF
  bx_unmapped_c *class_ptr = (bx_unmapped_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}

  Bit32u
bx_unmapped_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_UM_SMF
  UNUSED(io_len);

/* ??? put in for external environment testing */
if (address == 0xfffc)
  return(BX_UM_THIS s.test_port32);
if (address == 0xfffe)
  return(BX_UM_THIS s.test_port16);

  /* This function gets called for access to any IO ports which
     are not mapped to any device handler.  Reads return 0 */

if (address >= 0x02e0 && address <= 0x02ef) return(0);
  switch (address) {
    case 0x80:
      return(BX_UM_THIS s.port80);
      break;
    case 0x8e:
      return(BX_UM_THIS s.port8e);
      break;
    case 0x03df:
      /* bx_printf("unsupported IO read from port %04x (CGA)\n", address);*/
      break;
    case 0x023a:
    case 0x02f8: /* UART */
    case 0x02f9: /* UART */
    case 0x02fb: /* UART */
    case 0x02fc: /* UART */
    case 0x02fd: /* UART */
    case 0x02ea:
    case 0x02eb:
    case 0x03e8:
    case 0x03e9:
    case 0x03ea:
    case 0x03eb:
    case 0x03ec:
    case 0x03ed:
    case 0x03f8: /* UART */
    case 0x03f9: /* UART */
    case 0x03fb: /* UART */
    case 0x03fc: /* UART */
    case 0x03fd: /* UART */
    case 0x17c6:
      /* bx_printf("unsupported IO read from port %04x\n", address); */
      break;
    default:
      if (bx_dbg.unsupported_io)
        bx_printf("unsupported IO read from port %04x\n", address);
    }
  return(0);
}


  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_unmapped_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_UM_SMF
  bx_unmapped_c *class_ptr = (bx_unmapped_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_unmapped_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_UM_SMF
  UNUSED(io_len);

/* ??? put in for external environment testing */
if (address == 0xfffc)
  BX_UM_THIS s.test_port32 = value;
if (address == 0xfffe)
  BX_UM_THIS s.test_port16 = value;

if (address == 0xfff0) {
  BX_UM_THIS s.bios_message[BX_UM_THIS s.bios_message_i] = (Bit8u) value; 
  BX_UM_THIS s.bios_message_i ++;
  if ( BX_UM_THIS s.bios_message_i >= BX_BIOS_MESSAGE_SIZE ) {
    BX_UM_THIS s.bios_message[ BX_BIOS_MESSAGE_SIZE - 1] = 0;
    BX_UM_THIS s.bios_message_i = 0;
    bx_printf("BIOS message: %s\n", BX_UM_THIS s.bios_message);
    }
  else if ((value & 0xff) == '\n') {
    BX_UM_THIS s.bios_message[ BX_UM_THIS s.bios_message_i ] = 0;
    BX_UM_THIS s.bios_message_i = 0;
    bx_printf("BIOS message: %s\n", BX_UM_THIS s.bios_message);
    }
  return;
  }


/* ??? */
if (address==0xee || address==0xed) return;

  /* This function gets called for access to any IO ports which
     are not mapped to any device handler.  Writes to an unmapped
     IO port are ignored. */

if (address >= 0x02e0 && address <= 0x02ef) return;
  switch (address) {
    case 0x80: /* diagnostic test port to display progress of POST */
      /*bx_printf("Diagnostic port 80h: write = %02xh\n", (unsigned) value);*/
      BX_UM_THIS s.port80 = value;
      return;
      break;

    case 0x8e: /* ??? */
      BX_UM_THIS s.port8e = value;
      return;

    case 0x2f2:
    case 0x2f3:
    case 0x2f4:
    case 0x2f5:
    case 0x2f6:
    case 0x2f7:
    case 0x3e8:
    case 0x3e9:
    case 0x3eb:
    case 0x3ec:
    case 0x3ed:
      bx_printf("unsupported IO write to port %04x of %02x\n",
        address, value);
      break;
    case 0xfedc:
      bx_dbg.debugger = (value > 0);
fprintf(stderr, "# DEBUGGER = %u\n", (unsigned) bx_dbg.debugger);
      break;
    default:
      if (bx_dbg.unsupported_io)
        bx_printf("unsupported IO write to port %04x of %02x\n",
          address, value);
    }
}
