// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com







#include "bochs.h"
extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
}




#if BX_PROVIDE_CPU_MEMORY

  void
bx_memory_c::write_physical(Bit32u addr, unsigned len, void *data)
{
  Bit8u *data_ptr;
  Bit32u a20addr;

  a20addr = A20ADDR(addr);

#ifdef BX_LITTLE_ENDIAN
  data_ptr = (Bit8u *) data;
#else // BX_BIG_ENDIAN
  data_ptr = (Bit8u *) data + (len - 1);
#endif




  if ( (a20addr + len) <= BX_MEM_THIS len ) {

write_one:
    if ( (a20addr & 0xfff80000) != 0x00080000 ) {
      // addr *not* in range 00080000 .. 000FFFFF
      bx_mem.vector[a20addr] = *data_ptr;
      BX_DBG_DIRTY_PAGE(a20addr >> 12);
inc_one:
      if (len == 1) return;
      len--;
      addr++;
      a20addr = A20ADDR(addr);
#ifdef BX_LITTLE_ENDIAN
      data_ptr++;
#else // BX_BIG_ENDIAN
      data_ptr--;
#endif
      goto write_one;
      }
  
    // addr in range 00080000 .. 000FFFFF

    if (a20addr <= 0x0009ffff) {
      // regular memory 80000 .. 9FFFF
      bx_mem.vector[a20addr] = *data_ptr;
      BX_DBG_DIRTY_PAGE(a20addr >> 12);
      goto inc_one;
      }
    if (a20addr <= 0x000bffff) {
      // VGA memory A0000 .. BFFFF
      bx_devices.vga->mem_write(a20addr, *data_ptr);
      BX_DBG_UCMEM_REPORT(a20addr, 1, BX_WRITE, *data_ptr);
      goto inc_one;
      }
    // adapter ROM     C0000 .. DFFFF
    // ROM BIOS memory E0000 .. FFFFF
    // (ignore write)
    //bx_printf("ROM lock %08x: len=%u\n",
    //  (unsigned) a20addr, (unsigned) len);
    goto inc_one;
    }
  else {
    unsigned i;
    for (i = 0; i < len; i++) {
      if (a20addr < BX_MEM_THIS len) {
        bx_mem.vector[a20addr] = *data_ptr;
        BX_DBG_DIRTY_PAGE(a20addr >> 12);
        }
      // otherwise ignore byte, since it overruns memory
      addr++;
      a20addr = A20ADDR(addr);
#ifdef BX_LITTLE_ENDIAN
      data_ptr++;
#else // BX_BIG_ENDIAN
      data_ptr--;
#endif
      }
    return;
    }
}



  void
bx_memory_c::read_physical(Bit32u addr, unsigned len, void *data)
{
  Bit8u *data_ptr;
  Bit32u a20addr;

  a20addr = A20ADDR(addr);

#ifdef BX_LITTLE_ENDIAN
  data_ptr = (Bit8u *) data;
#else // BX_BIG_ENDIAN
  data_ptr = (Bit8u *) data + (len - 1);
#endif
 

 
  if ( (a20addr + len) <= BX_MEM_THIS len ) {
read_one:
    if ( (a20addr & 0xfff80000) != 0x00080000 ) {
      // addr *not* in range 00080000 .. 000FFFFF
      *data_ptr = bx_mem.vector[a20addr];
inc_one:
      if (len == 1) return;
      len--;
      addr++;
      a20addr = A20ADDR(addr);
#ifdef BX_LITTLE_ENDIAN
      data_ptr++;
#else // BX_BIG_ENDIAN
      data_ptr--;
#endif
      goto read_one;
      }
 
    // addr in range 00080000 .. 000FFFFF
    if ((a20addr <= 0x0009ffff) || (a20addr >= 0x000c0000) ) {
      // regular memory 80000 .. 9FFFF, C0000 .. F0000
      *data_ptr = bx_mem.vector[a20addr];
      goto inc_one;
      }
    // VGA memory A0000 .. BFFFF
    *data_ptr = bx_devices.vga->mem_read(a20addr);
    BX_DBG_UCMEM_REPORT(a20addr, 1, BX_READ, *data_ptr);
    goto inc_one;
    }
  else {
    unsigned i;
    for (i = 0; i < len; i++) {
      if (a20addr < BX_MEM_THIS len)
        *data_ptr = bx_mem.vector[a20addr];
      else
        *data_ptr = 0xff;
      addr++;
      a20addr = A20ADDR(addr);
#ifdef BX_LITTLE_ENDIAN
      data_ptr++;
#else // BX_BIG_ENDIAN
      data_ptr--;
#endif
      }
    return;
    }
}

#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
  Bit32u
bx_memory_c::get_memory_in_k(void)
{
fprintf(stderr, "# get_memory_in_k() = %u\n", (unsigned)
  (BX_MEM_THIS megabytes * 1024));

  return(BX_MEM_THIS megabytes * 1024);
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
  // bx_memory_c constructor
bx_memory_c::bx_memory_c(void)
{
  bx_mem.vector = NULL;
  bx_mem.len    = 0;
  bx_mem.megabytes = 0;
}
#endif // #if BX_PROVIDE_CPU_MEMORY



#if BX_PROVIDE_CPU_MEMORY
  // bx_memory_c constructor
bx_memory_c::bx_memory_c(size_t memsize)
{
  fprintf(stderr, "# bx_memory_c::bx_memory_c(size_t) called\n");
  fprintf(stderr, "#   memsize = %u\n", (unsigned) memsize);
  vector = new Bit8u[memsize];
  len    = memsize;
  megabytes = len / (1024*1024);
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
// bx_memory_c destructor
bx_memory_c::~bx_memory_c(void)
{
  if (this-> vector != NULL) {
    delete this->vector;
    }
  else {
    fprintf(stderr, "#   memory not freed as it wasn't allocated!\n");
    }
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
  void
bx_memory_c::init_memory(int memsize)
{
  // you can pass 0 if memory has been allocated already through
  // the constructor, or the desired size of memory if it hasn't

  if (BX_MEM_THIS vector == NULL) {
    // memory not already allocated, do now...
    fprintf(stderr, "# bx_memory_c::init_memory(int): allocating memory.\n");
    fprintf(stderr, "#   memsize = %u\n", (unsigned) memsize);
    BX_MEM_THIS vector = new Bit8u[memsize];
    BX_MEM_THIS len    = memsize;
    BX_MEM_THIS megabytes = memsize / (1024*1024);
    }
  // initialize all memory to 0x00
  memset(BX_MEM_THIS vector, 0x00, BX_MEM_THIS len);

  // initialize ROM area (0xc0000 .. 0xfffff) to 0xff
  memset(BX_MEM_THIS vector + 0xc0000, 0xff, 0x40000);
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
  void
bx_memory_c::load_ROM(const char *path, Bit32u romaddress)
{
  struct stat stat_buf;
  int fd, ret;
  unsigned long size, offset;

  // read in ROM BIOS image file
  fd = open(path, O_RDONLY
#ifdef O_BINARY
            | O_BINARY
#endif
           );
  if (fd < 0) {
    fprintf(stderr, "load_ROM: couldn't open ROM image file '%s'.\n", path);
    exit(1);
    }
  ret = fstat(fd, &stat_buf);
  if (ret) {
    fprintf(stderr, "load_ROM: couldn't stat ROM image file '%s'.\n", path);
    exit(1);
    }

  size = stat_buf.st_size;

  if ( (romaddress + size) > BX_MEM_THIS len ) {
    fprintf(stderr, "load_ROM: ROM address range > physical memsize!\n");
    exit(1);
    }

  offset = 0;
  while (size > 0) {
    ret = read(fd, (void *) &BX_MEM_THIS vector[romaddress + offset], size);
    if (ret <= 0) {
      fprintf(stderr, "load_ROM: read failed on ROM BIOS image\n");
      exit(1);
      }
    size -= ret;
    offset += ret;
    }
  close(fd);
  fprintf(stderr, "# load_ROM: ROM BIOS '%s', size=%u read into memory at %08x\n",
          path, (unsigned) stat_buf.st_size, (unsigned) romaddress);
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if ( BX_DEBUGGER || BX_DISASM )
  Boolean
bx_dbg_fetch_mem(Bit32u addr, unsigned len, Bit8u *buf)
{
  if ( (addr + len) > bx_mem.len ) {
    return(0); // error, beyond limits of memory
    }
  for (; len>0; len--) {
#if BX_SUPPORT_VGA
    if ( (addr & 0xfffe0000) == 0x000a0000 ) {
      *buf = bx_devices.vga->mem_read(addr);
      }
    else
#endif
      *buf = bx_mem.vector[addr];
    buf++;
    addr++;
    }
  return(1);
}
#endif

#if BX_DEBUGGER
  Boolean
bx_dbg_set_mem(Bit32u addr, unsigned len, Bit8u *buf)
{
  if ( (addr + len) > bx_mem.len ) {
    return(0); // error, beyond limits of memory
    }
  for (; len>0; len--) {
#if BX_SUPPORT_VGA
    if ( (addr & 0xfffe0000) == 0x000a0000 ) {
      bx_devices.vga->mem_write(addr, *buf);
      }
    else
#endif
      bx_mem.vector[addr] = *buf;
    buf++;
    addr++;
    }
  return(1);
}
#endif
